(* General Soundness/Safeness Proofs *)
(* Copyright (c) 2005, Rodrigo Ferreira *)

Set Implicit Arguments.

Section SoundSafe.

Variable prog : Type.
Variable Step: prog -> prog -> Prop.

Inductive Step_n: nat -> prog -> prog -> Prop :=
  | Step_O : forall P,
             Step_n O P P
  | Step_S : forall n P P' P'',
             Step P P' ->
             Step_n n P' P'' ->
             Step_n (S n) P P''.

Variable spec : Type.
Variable Infer: spec -> prog -> Prop.

(* Soundness *)
Section Sound.

Variable Progress : forall S P, Infer S P -> exists P', Step P P'.
Variable Preservation : forall S P P', Infer S P -> Step P P' -> exists S', Infer S' P'.

Theorem Soundness : forall n S P, Infer S P -> exists P', Step_n n P P'.
Proof.
intro n.
induction n.
intros S P Hinf.
exists P.
apply Step_O.
intros S P Hinf.
elim (Progress Hinf).
intros P' Hstep.
elim (Preservation Hinf Hstep).
intros S' Hinf'.
elim (IHn S' P' Hinf').
intros P'' Hstep_n.
exists P''.
apply Step_S with P'.
exact Hstep.
exact Hstep_n.
Qed.

End Sound.

(* Safeness *)
Section Safe.

Variable S : spec.
Variable P : prog.

Variable Soundness : forall n S P, Infer S P -> exists P', Step_n n P P'.
Variable WellFormed : Infer S P.

Theorem Safeness : forall n, exists P', Step_n n P P'.
Proof.
intro n.
apply Soundness with S.
exact WellFormed.
Qed.

End Safe.

End SoundSafe.
