
(** ** Utilities for manipulating & reasoning natural numbers *)
(** By Nadeem A. Hamid *)

Require Import Div2.
Require Import EqNat.
Require Import Lt.
Require Import Le.
Require Import Arith.

Ltac FAuto := auto with fsl.

(** blt_nat and related properties *)
Fixpoint blt_nat (n m : nat) {struct n} : bool :=
  match n, m with
  | O, O => false
  | O, S _ => true
  | S _, O => false
  | S n1, S m1 => blt_nat n1 m1
  end.


Lemma blt_S_false : forall a : nat, blt_nat (S a) a = false.
Proof.
simple induction a.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in |- *.
auto.
Qed.
Hint Resolve blt_S_false: fsl.

Lemma blt_S_eq :
 forall (a b : nat) (t : bool), blt_nat a b = t -> blt_nat (S a) (S b) = t.
Proof.
intros.
unfold blt_nat in |- *.
fold blt_nat in |- *.
auto.
Qed.
Hint Resolve blt_S_eq: fsl.

Lemma blt_lt_true : forall a b : nat, a < b -> blt_nat a b = true.
Proof.
simple induction a.
intros.
inversion H.
unfold blt_nat in |- *.
auto.
unfold blt_nat in |- *.
auto.
intros.
inversion H0.
apply blt_S_eq.
apply H.
apply lt_n_Sn.
apply blt_S_eq.
apply H.
rewrite <- H2 in H0.
apply lt_S_n.
auto.
Qed.
Hint Resolve blt_lt_true: fsl.


Lemma blt_Sm_false :
 forall a b : nat, blt_nat a b = false -> blt_nat (S a) b = false.
Proof.
double induction a b.
intros.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in H0.
absurd (true = false).
discriminate.
auto.
intros.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in H1.
fold blt_nat in H1.
cut (blt_nat (S n0) n = false).
intros.
apply blt_S_eq.
auto.
apply H0.
auto.
Qed.
Hint Resolve blt_Sm_false: fsl.


Lemma blt_false : forall a b : nat, a < b -> blt_nat b a = false.
Proof.
intros.
induction H as [| m H HrecH].
apply blt_S_false.
apply blt_Sm_false.
auto.
Qed.
Hint Resolve blt_false: fsl.


Lemma blt_eq_false : forall a : nat, blt_nat a a = false.
Proof.
simple induction a.
unfold blt_nat in |- *.
auto.
intros.
unfold blt_nat in |- *.
auto.
Qed.
Hint Resolve blt_eq_false: fsl.


Lemma le_blt_false : forall a b : nat, a <= b -> blt_nat b a = false.
Proof.
intros.
induction H as [| m H HrecH].
apply blt_eq_false.
apply blt_false.
unfold lt in |- *.
apply le_n_S.
auto.
Qed.
Hint Resolve le_blt_false: fsl.



(** Property of natural numbers *)
Lemma beq_neq_false : forall a b : nat, a <> b -> false = beq_nat a b.
Proof.
double induction a b.
intros.
absurd (0 <> 0); auto.
intros.
unfold beq_nat in |- *.
auto.
intros.
unfold beq_nat in |- *.
auto.
intros.
unfold beq_nat in |- *.
fold beq_nat in |- *.
apply H0.
auto.
Qed.
Hint Resolve beq_neq_false: fsl.

Lemma eq_neq_Sl : forall a b : nat, a = b -> S a <> b.
Proof.
double induction a b.
intros.
auto.
intros.
absurd (0 = S n).
discriminate.
auto.
intros.
absurd (0 = S n).
discriminate.
auto.
intros.
apply not_eq_S.
auto.
Qed.
Hint Resolve eq_neq_Sl: fsl.


Theorem eqorneq_nat : forall a a' : nat, a = a' \/ a <> a'.
Proof.
intros.
compare a a'.
auto.
auto.
Qed.
Hint Resolve eqorneq_nat: fsl.

Theorem le_or_gt : forall m n : nat, n <= m \/ n > m.
Proof.
double induction m n.
left.
auto.
intros.
right.
red in |- *.
red in |- *.
apply le_n_S.
apply le_O_n.
intros.
left.
apply le_O_n.
intros.
cut (n <= n1 \/ n > n1).
intros.
inversion_clear H1.
left.
apply le_n_S.
auto.
right.
unfold gt in |- *.
unfold gt in H2.
apply lt_n_S.
auto.
apply H0.
Qed.
Hint Resolve le_or_gt: fsl.

Lemma gt_lt : forall p q : nat, p > q -> q < p.
Proof.
intros. 
induction p.
inversion H.
apply le_lt_n_Sm.
apply gt_S_le.
assumption.
Qed.
Hint Resolve gt_lt : fsl.

Lemma lt_neq : forall m n : nat, n < m -> n <> m.
Proof.
simple induction m.
intros.
inversion H.
intros.
induction n0 as [| n0 Hrecn0].
auto.
cut (n0 < n).
intros.
cut (n0 <> n).
intros.
auto.
apply H.
auto.
apply lt_S_n.
auto.
Qed.
Hint Resolve lt_neq: fsl.

Lemma lt_S_neq : forall n x : nat, n < S x -> x <> n -> n < x.
Proof.
double induction n x.
intros.
absurd (0 <> 0).
auto.
auto.
intros.
apply lt_O_Sn.
intros.
unfold lt in H0.
cut (S n0 <= 0).
intros.
absurd (S n0 <= 0).
apply le_Sn_O.
auto.
apply le_S_n.
auto.
intros.
apply lt_n_S.
cut (n1 = n0 \/ n1 <> n0).
intros.
inversion_clear H3.
absurd (S n1 <> S n1).
auto.
rewrite <- H4 in H2.
auto.
cut (n1 < S n0).
intros.
apply (H0 n0 H3).
auto.
apply lt_S_n.
auto.
apply eqorneq_nat.
Qed.
Hint Resolve lt_S_neq: fsl.


Lemma lt_S_eq_or_lt : forall n x : nat, n < S x -> x = n \/ n < x.
Proof.
intros.
cut (x = n \/ x <> n).
intros.
inversion_clear H0.
left; auto.
right; apply lt_S_neq.
auto.
auto.
apply eqorneq_nat.
Qed.
Hint Resolve lt_S_eq_or_lt: fsl.


Lemma plus_n_O_n : forall n : nat, n + 0 = n.
Proof.
simple induction n.
unfold plus in |- *.
auto.
intros.
unfold plus in |- *.
fold plus in |- *.
rewrite H.
auto.
Qed.
Hint Resolve plus_n_O_n: fsl.


Lemma plus_S_neq : forall n m : nat, n <> n + S m.
Proof.
simple induction n.
intros.
unfold plus in |- *.
auto.
intros.
unfold plus in |- *.
fold plus in |- *.
auto.
Qed.
Hint Resolve plus_S_neq: fsl.


Lemma neq_Sn_n : forall n : nat, S n <> n.
Proof.
simple induction n; [ discriminate | intros; apply not_eq_S; assumption ].
Qed.
Hint Resolve neq_Sn_n: fsl.


Lemma beq_nat_sym : forall n m : nat, beq_nat n m = beq_nat m n.
Proof.
double induction n m; auto; intros; compute in |- *; auto.
Qed.
Hint Resolve beq_nat_sym: fsl.

Lemma not_beq_Sn_n : forall n : nat, beq_nat (S n) n = false.
Proof.
intros; apply sym_eq; FAuto.
Qed.
Hint Resolve not_beq_Sn_n: fsl.


Lemma not_beq_nat_big_small : forall m n : nat, n < m -> beq_nat m n = false.
Proof.
intros; apply sym_eq; apply beq_neq_false; apply sym_not_eq; FAuto.
Qed.
Hint Resolve not_beq_nat_big_small: fsl.

Lemma not_beq_nat_small_big : forall m n : nat, n < m -> beq_nat n m = false.
Proof.
intros; replace (beq_nat n m) with (beq_nat m n); FAuto.
Qed.
Hint Resolve not_beq_nat_small_big: fsl.
