(* The Readers and Writers program *)
(* Copyright (c) 2005, FLINT Group, Yale University *)

Require Import Classical.

Require Import Util.
(* Require Import HiMach. *)
Require Import SoundSafe.
Require Import Omega.
Require Import EqNat.
Require Import ThMach.
Require Import natutil.

Set Implicit Arguments.

Ltac SplitAllHyps :=
  repeat
   match goal with (* splits all conjunctions in hypotheses *)
   | H:(_ /\ _) |- _ => inversion_clear H
   end.

(* code heap labels *)
Notation L_reader := 0 (only parsing).
Notation L_cont1 := 1 (only parsing).
Notation L_getw := 2 (only parsing).
Notation L_inc_cnt := 3 (only parsing).
Notation L_cont2 := 4 (only parsing).
Notation L_cont3 := 5 (only parsing).
Notation L_relw := 6 (only parsing).
Notation L_cont4 := 7 (only parsing).
Notation L_writer := 8 (only parsing).
Notation L_cont5 := 9 (only parsing).


(* variables *)
Notation V_length := 100 (only parsing).
Notation rf := 0 (only parsing).
Notation wf := V_length (only parsing).
Notation cnt := 200 (only parsing).
Notation writ := 201 (only parsing).
Notation l := 202 (only parsing).
Notation v := 203 (only parsing).

(* definition of code heap *)

Definition facq : label -> label :=
  fun (f : label) =>
      if (beq_w f L_cont1) 
      then L_reader
      else if (beq_w f L_cont3)
           then L_cont2
	   else 100.

Definition fpwrit :  label -> label :=
  fun (f : label) =>
      if (beq_w f L_inc_cnt) 
      then L_getw
      else if (beq_w f L_cont5)
           then L_writer
           else 100.


Definition C_acq : label -> block :=
  fun (f : label) =>
    seq (yield)
   (seq (movi r0 0)
   (seq (movi r1 l)
   (seq (ld r2 r1 0)
   (seq (bgt r2 r0 (facq f))
   (seq (st r1 0 tn)
        (jd f)
   ))))).

Definition C_facq : label -> label -> block :=
  fun (f0 f1 : label) =>
    seq yield
   (seq (movi r0 0)
   (seq (movi r1 l)
   (seq (ld r2 r1 0)
   (seq (bgt r2 r0 f0)
   (seq (st r1 0 tn)
        (jd f1)
   ))))).

Definition C_rel : label -> block :=
  fun (f : label) =>
    seq (yield)
   (seq (movi r0 0)
   (seq (st r0 l r0)
        (jd f)
   )).

Definition C_pwrit : label -> block :=
  fun (f : label) =>
    seq (yield)
   (seq (movi r0 0)
   (seq (movi r1 writ)
   (seq (ld r2 r1 0)
   (seq (beq r2 r0 (fpwrit f))
   (seq (st r1 0 r0)
        (jd f)
   ))))).

Definition C_fpwrit : label -> label -> block :=
  fun (f0 f1 : label) =>
    seq yield
   (seq (movi r0 0)
   (seq (movi r1 writ)
   (seq (ld r2 r1 0)
   (seq (beq r2 r0 f0)
   (seq (st r1 0 r0)
        (jd f1)
   ))))).

Definition C_vwrit : label -> block :=
  fun (f : label) =>
   seq (movi r1 1)
   (seq (movi r2 writ)
   (seq (st r2 0 r1)
        (jd f)
   )).

Definition C_reader : block :=
  C_acq L_cont1.

Definition C_cont1 : block :=
    seq (yield)
   (seq (movi r0 0)
   (seq (movi r1 cnt)
   (seq (ld r2 r1 0)
   (seq (yield)
   (seq (beq r2 r0 L_getw)
        (jd L_inc_cnt)
   ))))).

Definition C_getw : block :=
  C_pwrit(L_inc_cnt).

Definition C_inc_cnt : block :=
    seq (yield)
   (seq (movi r1 cnt)
   (seq (ld r2 r1 0)
   (seq (yield)
   (seq (movi r3 1)
   (seq (add r2 r2 r3)
   (seq (st r1 0 r2)
   (seq (yield)
   (seq (movi r1 1)
   (seq (st ta rf r1)
        (C_rel L_cont2)
   ))))))))).

Definition C_cont2 : block :=
  C_acq L_cont3.

Definition C_cont3 : block :=
    seq (yield)
   (seq (movi r0 0)
   (seq (st ta rf r0)
   (seq (yield)
   (*(seq (movi r0 0) *)
   (seq (movi r1 cnt)
   (seq (ld r2 r1 0)
   (seq (yield)
   (seq (movi r3 1)
   (seq (sub r2 r2 r3)
   (seq (st r1 0 r2)
   (seq (yield)
   (seq (beq r2 r0 L_relw)
        (C_rel L_reader)
   ))))))))))).

Definition C_relw : block :=
  seq (yield)
      (C_vwrit L_cont4).

Definition C_cont4 : block :=
  C_rel L_reader.

Definition C_writer : block :=
  C_pwrit L_cont5.

Definition C_cont5 : block :=
    seq (movi r1 1)
   (seq (st ta wf r1)
   (seq (yield)
   (seq (movi r2 v)
   (seq (st r2 0 r1)
   (seq (yield)
   (seq (movi r0 0)
   (seq (st ta wf r0)
        (C_vwrit L_writer)
   ))))))).

Definition CdHeap : code :=
  fun f : label =>
    if (beq_w f L_reader)
    then SomeT C_reader
    else if (beq_w f L_cont1)
         then SomeT C_cont1
	 else if (beq_w f L_getw)
	      then SomeT C_getw
	      else if (beq_w f L_inc_cnt)
		   then SomeT C_inc_cnt
		   else if (beq_w f L_cont2)
		        then SomeT C_cont2
			else if (beq_w f L_cont3)
			     then SomeT C_cont3
			     else if (beq_w f L_relw)
			     then SomeT C_relw
			     else if (beq_w f L_cont4)
			          then SomeT C_cont4
			          else if (beq_w f L_writer)
				       then SomeT C_writer
				       else if (beq_w f L_cont5)
					    then SomeT C_cont5
					    else NoneT.

Fixpoint mem_sigma (M : heap) (base lth : nat) {struct lth} : nat :=
    match lth with
        O => O
      | S lth' => M(base + lth) + (mem_sigma M base lth')
    end.

Definition inv0 : INV :=
  fun M : heap => M cnt >= mem_sigma M rf V_length.

Definition inv1 : INV :=
  fun M : heap => M l = 0 -> mem_sigma M rf V_length = M cnt.


Definition inv2 : INV :=
  fun M : heap => mem_sigma M wf V_length <= 1.

Definition inv3 : INV :=
  fun M : heap => mem_sigma M wf V_length = 1 -> (M cnt = 0 /\ M writ = 0).

Definition inv4 : INV :=
  fun M : heap => M writ = 1 -> M cnt = 0.

Definition inv5 : INV :=
  fun M : heap => M writ = 0 \/ M writ = 1.

Definition Inv : INV :=
  fun M : heap => inv0 M /\ inv1 M /\ inv2 M /\ inv3 M /\ inv4 M /\ inv5 M.

Definition T_reader_A : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in
        M (rf + R ta) = M' (rf + R ta) /\ (M (rf + R ta) = 1 -> M v = M' v) 
	/\ (M l = R tn -> M l = M' l /\ M cnt = M' cnt 
	                  /\ (forall i : nat, i >= 0 -> i < V_length -> 
                                                    M (rf + i) = M' (rf + i))
	                  /\ (M writ = 0 -> mem_sigma M wf V_length = 0 -> 
                                M' writ = 0 /\ mem_sigma M' wf V_length = 0))
(*
		/\ (M cnt > 0 -> M' writ = 0 /\ mem_sigma M' wf V_length = 0)
*)

	/\ (M l <> R tn -> M' l <> R tn).

Definition T_reader_G : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in
        (forall i : nat, i >= 0 -> i < V_length -> M (wf + i) = M' (wf + i)) /\ 
	(forall i : nat, i >= 0 -> i < V_length -> i <> R ta -> M (rf + i) = M' (rf + i)) /\ 
	M v = M' v /\ 
	(M l <> R tn -> M l <> 0 -> M l = M' l /\ M cnt = M' cnt /\ 
	                                  M (rf + R ta) = M' (rf + R ta) /\ M writ = M' writ) /\
	((M l = R tn \/ M l = 0) -> (M' l = R tn \/ M' l = 0)).

Definition T_writer_A : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in
        M (wf + R ta) = M' (wf + R ta) /\ 
	(M (wf + R ta) = 1 -> M v = M' v /\ M writ = M' writ /\ M cnt = M' cnt).

Definition T_writer_G : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in
        (forall i : nat, i >= 0 -> i < V_length -> M (rf + i) = M' (rf + i)) /\ 
	(forall i : nat, i >= 0 -> i < V_length -> i <> R ta -> M (wf + i) = M' (wf + i)) /\ 
	(M writ = 0 -> M (wf + R ta) = 0 -> M v = M' v /\ M' writ = 0 /\  M' (wf + R ta) = 0) /\
	(*(M cnt > 0 -> M v = M' v /\ M' writ = 0 /\  M' (wf + R ta) = 0) /\*)
	(M l = M' l) /\ (M cnt = M' cnt).

Definition T_G_id : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  M = M'.

Definition TnTa : precond :=
  fun s : state =>
    let (R,M) := s
    in  R tn > 0 /\ R ta < V_length /\ R ta >= 0.

Definition T_reader_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\ M (rf + R ta) = 0 /\ M l <> R tn.

Definition T_cont1_p : precond :=
  fun s : state =>
    let (R,M) := s
    in  
        TnTa (R,M) /\ M (rf + R ta) = 0 /\ M l = R tn /\ mem_sigma M rf V_length = M cnt.

Definition T_getw_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        T_cont1_p (R,M) /\ M cnt = 0.

Definition T_inc_cnt_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        T_cont1_p (R,M) /\ M writ = 0 /\ mem_sigma M wf V_length = 0.

(* p7 *)
Definition T_cont2_p : precond :=
  fun s : state =>
    let (R,M) := s
    in  
        TnTa (R,M) /\ M (rf + R ta) = 1 /\ M l <> R tn.

(* p8 *)
Definition T_cont3_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\ M (rf + R ta) = 1 /\ M l = R tn /\ mem_sigma M rf V_length = M cnt.

(* p10 *)
Definition T_relw_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\  M l = R tn /\ M cnt = 0 /\ M writ = 0 /\ M (rf + R ta) = 0 
        /\ mem_sigma M rf V_length = M cnt
        /\ mem_sigma M wf V_length = 0.

(* M (rf + R ta) = 0 can be infered *)

(* p11 *)
Definition T_cont4_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        (* TnTa (R,M) /\  M l = R tn /\ M writ = 1 /\ mem_sigma M rf V_length = M cnt.*)
	TnTa (R,M) /\  M l = R tn /\ M (rf + R ta) = 0 /\ mem_sigma M rf V_length = M cnt.

(* p12 *)
Definition T_writer_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\  M (wf + R ta) = 0.

(* p13 *)
Definition T_cont5_p : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\  M (wf + R ta) = 0  /\ M writ = 0 /\ M cnt = 0 /\ mem_sigma M wf V_length = 0.

Definition T_cont5_G : guarantee :=
  fun (s : state) (M' : heap) =>
    let (R,M) := s
    in  ext_w M (wf + R ta) 1 = M'.

Definition Psi : cspec :=
  fun f : label =>
    if (beq_w f L_reader)
    then SomeT (T_reader_p, T_G_id, T_reader_A, T_reader_G)
    else if (beq_w f L_cont1)
         then SomeT (T_cont1_p, T_G_id, T_reader_A, T_reader_G)
	 else if (beq_w f L_getw)
	      then SomeT (T_getw_p, T_G_id, T_reader_A, T_reader_G)
	      else if (beq_w f L_inc_cnt)
		   then SomeT (T_inc_cnt_p, T_G_id, T_reader_A, T_reader_G)
		   else if (beq_w f L_cont2)
		        then SomeT (T_cont2_p, T_G_id, T_reader_A, T_reader_G)
			else if (beq_w f L_cont3)
			     then SomeT (T_cont3_p, T_G_id, T_reader_A, T_reader_G)
			     else if (beq_w f L_relw)
			     then SomeT (T_relw_p, T_G_id, T_reader_A, T_reader_G)
			     else if (beq_w f L_cont4)
			          then SomeT (T_cont4_p, T_G_id, T_reader_A, T_reader_G)
			          else if (beq_w f L_writer)
				       then SomeT (T_writer_p, T_G_id, T_writer_A, T_writer_G)
				       else if (beq_w f L_cont5)
					    then SomeT (T_cont5_p, T_cont5_G, T_writer_A, T_writer_G)
					    else NoneT.


(* The following Hypothesis is trivial to prove at meta level *)
(* I just omit the proof and treat them as hypothesis.        *)
(* Their correctness is human checkable                       *)

Hypothesis TnTa_p1 : forall R M M', TnTa (R,M) -> TnTa (R,M').
Hypothesis TnTa_p2 : forall R M M' r n, TnTa (R,M) -> r <> tn -> r<> ta -> TnTa (ext_r R r n, M').
Hypothesis TnTa_p3 : forall R M base, TnTa (R,M) -> base + R ta < base + 100.


Hypothesis mem_p1 : forall M M', (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) 
                               -> mem_sigma M 0 100 = mem_sigma M' 0 100.
Hypothesis mem_p2 : forall loc base length n M, (loc >= (base + length) \/ loc < base) -> 
                                 mem_sigma (ext_w M loc n) base length = mem_sigma M base length.
Hypothesis mem_p3 : forall M M', (forall i : nat, i >= 0 -> i < 100 -> M (rf + i) = M' (rf + i))
                                    -> mem_sigma M rf V_length = mem_sigma M' rf V_length.
Hypothesis mem_p4 : forall M R, TnTa(R,M) -> M (rf + R ta) = 0 -> 
                    (mem_sigma (ext_w M (rf + R ta) 1) rf V_length = mem_sigma M rf V_length + 1).
Hypothesis mem_p5 : forall M R, TnTa(R,M) -> M (rf + R ta) = 1 -> 
                    (mem_sigma (ext_w M (rf + R ta) 0) rf V_length + 1 = mem_sigma M rf V_length).

Lemma Gr_p1 : forall R M, T_reader_G (R,M) M.
intros.
unfold T_reader_G.
tauto.
Qed.
Hypothesis Gr_p2 : forall R M, M l = 0 -> TnTa (R,M) -> T_reader_G (R,M) (ext_w M l (R tn)).

Hypothesis Gr_p3 : forall R M n, M l = R tn -> T_reader_G (R,M) (ext_w M writ n).

Lemma Gr_p4 : forall R M M' r n, r <> ta -> r <> tn -> T_reader_G (ext_r R r n, M) M' -> T_reader_G (R,M) M'.
intros.
assert (beq_r ta r = false).
apply trv8; auto.
assert (beq_r tn r = false).
apply trv8; auto.
unfold T_reader_G in H1.
unfold ext_r in H1.
unfold ext in H1.
rewrite H2 in H1.
rewrite H3 in H1.
unfold T_reader_G.
auto.
Qed.
Hypothesis Gr_p5 : forall R M, M l = R tn -> T_reader_G (R, M) (ext_w M l 0).
Hypothesis Gr_p6 : forall R M n, M l = R tn -> T_reader_G (R, M) (ext_w M cnt n).
Hypothesis Gr_p7 : forall R M n, M l = R tn -> T_reader_G (R,M) (ext_w M (rf + R ta) n).

Lemma Ar_p1 : forall R M M' r n, r <> ta -> r <> tn -> T_reader_A (R,M) M' -> T_reader_A (ext_r R r n, M) M'.
intros.
assert (beq_r ta r = false).
apply trv8; auto.
assert (beq_r tn r = false).
apply trv8; auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
rewrite H2.
rewrite H3.
unfold T_reader_G in H1.
auto.
Qed.

Hypothesis Inv_p1 : forall M loc n, Inv M -> M l = 0 -> 
                         loc > 199 -> loc <> cnt -> loc <> writ -> Inv (ext_w M loc n).

Hypothesis Inv_p2 : forall M, Inv M -> M cnt <> 0 -> (M writ = 0 /\ mem_sigma M wf V_length = 0).

Hypothesis Inv_p3 : forall M, Inv M -> M writ <> 0 -> Inv (ext_w M writ 0).

Hypothesis Inv_p4 : forall M n, Inv M -> M l = 0 -> mem_sigma (ext_w M l n) 0 100 = ext_w M l n cnt.

Hypothesis Inv_p5 : forall M, Inv M -> mem_sigma M rf V_length = M cnt -> Inv (ext_w M l 0).

Hypothesis Inv_p6 : forall M, Inv M -> M writ <> 0 -> mem_sigma M wf V_length = 0.

Hypothesis Inv_p7 : forall R M n, TnTa (R,M) -> M l = R tn -> M writ = 0 
                             -> n >= mem_sigma M rf V_length -> mem_sigma M wf V_length = 0
                                       -> Inv M -> Inv (ext_w M cnt n).

Hypothesis Inv_p8 : forall R M, Inv M -> M l = R tn -> M cnt > mem_sigma M rf V_length 
                                       -> Inv (ext_w M (R ta + rf) 1).

Hypothesis Inv_p8' : forall R M, Inv M -> M l = R tn -> Inv (ext_w M (R ta + rf) 0).

Hypothesis Inv_p9 : forall R M, R r2 = mem_sigma M 0 100 -> 
                                     Inv M -> Inv (ext_w M cnt (R r2)).
Hypothesis Inv_p10 :
  forall M, Inv M -> M cnt = 0 -> mem_sigma M wf V_length = 0 -> Inv (ext_w M writ 1).

Lemma T_cont1_p_p1 :
   forall (R : regs) (M M' : heap),
       Inv M -> T_cont1_p (R, M) -> T_reader_A (R, M) M' -> T_cont1_p (R, M').
unfold T_cont1_p.
unfold T_reader_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
split.
rewrite <- H2; auto.
assert (M 202 = M' 202 /\ M 200 = M' 200 /\
            (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
            (M 201 = 0 -> mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H3; auto.
SplitAllHyps.
split.
rewrite <- H9.
auto.
rewrite <- H7.
assert (mem_sigma M 0 100 = mem_sigma M' 0 100).
apply mem_p1; auto.
rewrite <- H11; auto.
Qed.

Lemma T_cont1_p_p2 :
   forall (R : regs) (M : heap) r n,
       T_cont1_p (R, M) -> (r = r0 \/ r = r1 \/ r=r2 \/ r = r3) -> T_cont1_p (ext_r R r n, M).
unfold T_cont1_p.
intros.
SplitAllHyps.
inversion_clear H0.
rewrite H3.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
inversion_clear H3.
rewrite H0.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
inversion_clear H0.
rewrite H3.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
rewrite H3.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
Qed.

Lemma wf_frel : 
  forall (f:label) (p1 p2 :precond), 
    get Psi f (p2, T_G_id, T_reader_A, T_reader_G) -> 
    (forall R M M', p1 (R,M) -> T_reader_A (R,M) M' -> p1 (R,M')) ->
    (forall R M, p1 (R,M) -> M l = R tn /\ mem_sigma M rf V_length = M cnt) ->
    (forall R M n, p1 (R,M) -> p1 (ext_r R r0 n, M)) ->
    (forall R M, p1 (R,M) -> p2 (R, ext_w M l 0)) ->
    I_infer Inv Psi (p1, T_G_id, T_reader_A, T_reader_G) (C_rel f).
intros.
unfold C_rel.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.
intros.
apply H0 with M; auto.

(* movi r0 0 *)
apply I_infer_movi with
(fun s : state => let (R,M) := s in p1 (R,M) /\ R r0 = 0) T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
split; auto.
intros.

apply Gr_p4 with r0 0; auto.
discriminate.
discriminate.
intros.

apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r0 0; auto.
discriminate.
discriminate.

(* st r0 202 r0 *)
apply I_infer_st with p2 T_G_id T_reader_A T_reader_G.
intros.
inversion_clear H5.
rewrite H7.
assert (0 + l = l); auto.
rewrite H5.
split; auto.
assert (M 202 = R tn /\ mem_sigma M 0 100 = M 200).
apply H1; auto.
inversion_clear H8.
apply Inv_p5; auto.

intros.
unfold T_G_id in H6.
rewrite <- H6.
inversion_clear H5.
rewrite H8.
assert (0 + l = l); auto.
rewrite H5.
assert (M 202 = R tn /\ mem_sigma M 0 100 = M 200).
apply H1; auto.
inversion_clear H9.
apply Gr_p5; auto.
auto.
auto.

(* jd f *)
apply I_infer_jd with p2 T_G_id T_reader_A T_reader_G; auto.
Qed.


Lemma wf_inc_cnt : I_infer Inv Psi (T_inc_cnt_p, T_G_id, T_reader_A, T_reader_G) C_inc_cnt.
unfold C_inc_cnt.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.

Lemma T_inc_cnt_p_p1 : 
   forall (R : regs) (M M' : heap),
   Inv M -> T_inc_cnt_p (R, M) -> T_reader_A (R, M) M' -> T_inc_cnt_p (R, M').
unfold T_inc_cnt_p.
intros.
SplitAllHyps.
split.
apply T_cont1_p_p1 with M; auto.
unfold T_reader_A in H1.
unfold T_cont1_p in H2.
SplitAllHyps.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H8; auto.
SplitAllHyps.
apply H14; auto.
Qed.
apply T_inc_cnt_p_p1.

(* mivi *)
apply I_infer_movi with 
(fun s:state => let (R,M) := s in T_inc_cnt_p (R,M) /\ R r1 = cnt)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_inc_cnt_p.
intros.
SplitAllHyps.
split.
split.
apply T_cont1_p_p2; auto.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* ld r2 r1 0 *)
apply I_infer_ld with
(fun s : state => let (R, M) := s in T_inc_cnt_p (R, M) /\ R r1 = cnt /\ R r2 = M cnt)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_inc_cnt_p.
intros.
SplitAllHyps.
split.
split.
apply T_cont1_p_p2; auto.
split; auto.
rewrite H2.
auto.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* yield *)
apply I_infer_yield.
intros.
apply Gr_p1.
intros.
SplitAllHyps.
split.
apply T_inc_cnt_p_p1 with M; auto.
split; auto.
rewrite H4.
unfold T_reader_A in H1.
unfold T_inc_cnt_p in H2.
unfold T_cont1_p in H2.
SplitAllHyps.
assert (M 202 = M' 202 /\
        M 200 = M' 200 /\
        (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
        (M 201 = 0 ->
         mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H10; auto.
SplitAllHyps; auto.

(* movi *)
apply I_infer_movi with 
(fun s : state =>
      let (R, M) := s in T_inc_cnt_p (R, M) /\ R r1 = 200 /\ R r2 = M 200 /\ R r3 = 1)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_inc_cnt_p.
intros.
SplitAllHyps.
split.
split.
apply T_cont1_p_p2; auto.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.


intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* add r2 r2 r3 *)
apply I_infer_add with 
(fun s : state =>
      let (R, M) := s in
      T_inc_cnt_p (R, M) /\ R r1 = 200 /\ R r2 = M 200 + 1)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_inc_cnt_p.
intros.
SplitAllHyps.
split.
split.
apply T_cont1_p_p2; auto.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* st r1 0 r2 *)
apply I_infer_st with 
(fun s : state => let (R,M) := s in 
    TnTa (R,M) /\ M (rf + R ta) = 0 /\ M l = R tn /\ mem_sigma M rf V_length + 1 = M cnt)
        (* /\ M writ = 0 /\ mem_sigma M wf V_length = 0)*)
T_G_id T_reader_A T_reader_G.
intros.
unfold T_inc_cnt_p in H0.
unfold T_cont1_p in H0.
SplitAllHyps.
rewrite H0.
assert (cnt + 0 = cnt).
auto.
rewrite H7.
split.

apply Inv_p7 with R; auto.
rewrite H8.
rewrite H3.
omega.

split.
apply TnTa_p1 with M; auto.
split.
unfold TnTa in H4.
SplitAllHyps.
assert ((0 + R ta) = R ta).
auto.
rewrite H10.
assert (R ta < 200).
omega.
assert (R ta <> cnt).
apply lt_neq; auto.
assert (false = beq_nat (R ta) cnt).
apply beq_neq_false; auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H14.
rewrite H10 in H2; auto.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
assert (mem_sigma (ext_w M cnt (R r2)) rf V_length = mem_sigma M rf V_length).
apply mem_p2; auto.
omega.
rewrite H9.
rewrite H8.
rewrite <- H3.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
(*split.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
assert (mem_sigma (ext_w M cnt (R r2)) wf V_length = mem_sigma M wf V_length).
apply mem_p2; auto.
rewrite H9; auto.*)
unfold T_G_id.
intros.
rewrite <- H1.
unfold T_inc_cnt_p in H0.
unfold T_cont1_p in H0.
SplitAllHyps.
rewrite H0.
assert (cnt + 0 = cnt); auto.


apply Gr_p6; auto.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id.
auto.
unfold T_reader_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M;auto.
split.
rewrite <- H2; auto.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H3; auto.
SplitAllHyps.
split.
(* rewrite <- H11; auto.*)
rewrite <- H9; auto.
(* split. *)
assert (mem_sigma M rf V_length = mem_sigma M' rf V_length).

apply mem_p3.
auto.
(* rewrite <- H13.
rewrite <- H9. *)
rewrite <- H11.
rewrite <- H7.
auto.
(* apply H14; auto.*)

(*movi r1 1 *)
apply I_infer_movi with
     (fun s : state =>
      let (R, M) := s in
      TnTa (R, M) /\
      M (0 + R ta) = 0 /\ M 202 = R tn /\ mem_sigma M 0 100 + 1 = M 200 /\ R r1 = 1)
      (* /\ M 201 = 0 /\ mem_sigma M 100 100 = 0) *)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
tauto.

intros.
apply Gr_p4 with r1 1; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 1; auto.
discriminate.
discriminate.

(* st ta 0 r1 *)
apply I_infer_st with
(fun s : state =>
      let (R, M) := s in
      TnTa (R, M) /\
      M (rf + R ta) = 1 /\
      M l = R tn /\ mem_sigma M 0 100 = M cnt)
      (* /\ M 201 = 0 /\ mem_sigma M 100 100 = 0) *)
T_G_id T_reader_A T_reader_G.
intros.
SplitAllHyps.
split.
rewrite H5.
apply Inv_p8; auto.
omega.
split.
apply TnTa_p1 with M.
auto.
assert ((ext_w M (R ta + 0) (R r1) (0 + R ta) = 1 /\
   ext_w M (R ta + 0) (R r1) 202 = R tn) /\
   mem_sigma (ext_w M (R ta + 0) (R r1)) 0 100 =
   ext_w M (R ta + 0) (R r1) 200).
split.
assert (R ta + 0 = 0 + R ta); auto.
(* rewrite H6. *)
rewrite H4.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (true = beq_nat (0 + R ta) (0 + R ta)).
apply beq_nat_refl.
(* rewrite <- H8. *)
rewrite <- H6.
auto.
unfold TnTa in H1.
SplitAllHyps.
assert (0 + R ta = R ta); auto.
(* rewrite H9. *)
rewrite H7.
assert (R ta < 202).
apply lt_trans with 100; omega.
assert (R ta <> l).
apply lt_neq; auto.
assert (false = beq_nat l (R ta)).
apply beq_neq_false.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H11; auto.


assert (0 + R ta = R ta + 0).
auto.
rewrite <- H4.
assert (mem_sigma (ext_w M (rf + R ta) (R r1)) rf V_length = mem_sigma M rf V_length + 1).
rewrite H5.
apply mem_p4; auto.
rewrite H6.
assert (0 + R ta = R ta); auto.
rewrite H7.
assert (R ta < cnt).
unfold TnTa in H1.
SplitAllHyps.
apply lt_trans with 100; omega.
assert (R ta <> cnt).
apply lt_neq; auto.
assert (false = beq_nat cnt (R ta)).
apply beq_neq_false.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H10; auto.
tauto.
(*split. tauto.
split. tauto.
split. tauto.
unfold TnTa in H1.
SplitAllHyps.
assert (R ta < writ).
apply lt_trans with 100; omega.
assert (R ta <> writ).
apply lt_neq; auto.
assert (false = beq_nat writ (R ta)).
apply beq_neq_false.
auto.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (R ta + 0 = R ta); auto.
rewrite H15.
rewrite <- H14; auto.
assert (mem_sigma (ext_w M (R ta + 0) (R r1)) wf V_length = mem_sigma M wf V_length).
apply mem_p2; auto.
omega.
rewrite H15; auto.*)

unfold T_G_id.
intros.
rewrite <- H1.
SplitAllHyps.
assert (rf + R ta = R ta + rf); auto.
rewrite <- H5.
apply Gr_p7; auto.
auto.
auto.

apply wf_frel with T_cont2_p; auto.
unfold Psi.
unfold get.
auto.
unfold T_reader_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H2; auto.
SplitAllHyps.
split.
rewrite <- H1; auto.
split.
rewrite <- H8; auto.
assert (mem_sigma M rf V_length = mem_sigma M' rf V_length).
apply mem_p1; auto.
rewrite <- H10.
rewrite <- H6; auto.
tauto.

unfold T_cont2_p.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M.
auto.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
assert ((rf + R ta) < l).
apply lt_trans with (0 + 100).

apply TnTa_p3 with M; auto.
omega.
assert (rf + R ta <> l).
apply lt_neq; auto.
assert (false = beq_nat (rf + R ta) l).
apply beq_neq_false; auto.
rewrite <- H5.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
unfold TnTa in H0.
inversion_clear H0.
apply lt_neq; auto.
(*assert (mem_sigma (ext_w M l 0) wf V_length = mem_sigma M wf V_length).
apply mem_p2.
omega.
rewrite H4; auto.*)

Qed.


Lemma wf_cont2 : I_infer Inv Psi (T_cont2_p, T_G_id, T_reader_A, T_reader_G) (C_acq L_cont3).
unfold C_acq.
unfold facq.
unfold beq_w.
unfold beq_nat.


(* yield *)
apply I_infer_yield.
unfold T_G_id.
auto.

Lemma T_cont2_p_p1 :
   forall (R : regs) (M M' : heap),
   Inv M -> T_cont2_p (R, M) -> T_reader_A (R, M) M' -> T_cont2_p (R, M').

unfold T_cont2_p.
unfold T_reader_A.

intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
split.
rewrite <- H2; auto.
apply H5; auto.

Qed.
apply T_cont2_p_p1.

Hypothesis T_cont2_p_p2 : forall R M r n, r <> tn -> r <> ta -> T_cont2_p (R,M) ->
                                                  T_cont2_p (ext_r R r n, M).
(* movi r0 0 *)
apply I_infer_movi with 
  (fun s : state => let (R,M):=s in T_cont2_p (R,M) /\ R r0 = 0) T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
split.
apply T_cont2_p_p2; auto.
discriminate.
discriminate.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* movi r1 l *)
apply I_infer_movi with 
  (fun s : state => let (R,M):=s in T_cont2_p (R,M) /\ R r0 = 0 /\ R r1 = l) 
  T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
inversion_clear H0.
split.
apply T_cont2_p_p2; auto.
discriminate.
discriminate.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* ld r2 r1 0 *)
apply I_infer_ld with 
  (fun s : state => let (R,M):=s in T_cont2_p (R,M) /\ R r0 = 0 /\ R r1 = l /\ R r2 = M l) 
  T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
split.
apply T_cont2_p_p2; auto.
discriminate.
discriminate.
rewrite H3.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* bgt r2 r0 0 *)
apply I_infer_bgt with 
(fun s : state =>
      let (R, M) := s in
      T_cont2_p (R, M) /\ R r0 = 0 /\ R r1 = l /\ R r2 = M l /\ M l = 0)
T_cont2_p T_reader_G T_G_id T_reader_A T_reader_A T_reader_G T_reader_G.
tauto.
unfold T_G_id.
intros.
rewrite <- H2.


apply Gr_p1.
auto.
auto.
unfold Psi.
unfold get.
auto.
intros.
SplitAllHyps.
split; [auto | split ; [auto | split; [auto | split; auto]]].
rewrite H5 in H1.
rewrite H0 in H1.
auto with arith.
auto.
auto.
auto.

(* st r1 0 tn *)
apply I_infer_st with
(*(fun s : state =>
      let (R, M) := s in
      TnTa (R,M) /\ M (rf + R ta) = 0 /\ R r0 = 0 /\ R r1 = l /\ R r2 = 0 /\ M l = R tn)*)
T_cont3_p T_G_id T_reader_A T_reader_G.
intros.
SplitAllHyps.
split.
rewrite H2.
assert (202 + 0 = 202).
auto.
rewrite H4.
apply Inv_p1; auto with arith.
discriminate.
unfold T_cont2_p in H1.
SplitAllHyps.
unfold T_cont3_p.
split.
apply TnTa_p1 with M.
auto.
assert ((ext_w M (R r1 + 0) (R tn) (0 + R ta) = 1 /\
   ext_w M (R r1 + 0) (R tn) 202 = R tn) /\
   mem_sigma (ext_w M (R r1 + 0) (R tn)) 0 100 =
   ext_w M (R r1 + 0) (R tn) 200).
split.
unfold TnTa in H4.
SplitAllHyps.
rewrite H2.
assert (0 + R ta < 202 + 0).
assert (0 + R ta = R ta); auto.
rewrite H8.
apply lt_trans with 100; omega.
assert (0 + R ta <> 202 + 0).
apply lt_neq; auto.
assert (false = beq_nat (0 + R ta) (202 + 0)).
apply beq_neq_false.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H11.
auto.

rewrite H2.
assert (202 + 0 = 202); auto.
rewrite H6.
apply Inv_p4; auto.
tauto.


unfold T_G_id.
intros.
rewrite <- H1.
unfold T_cont2_p in H0.
SplitAllHyps.
rewrite H3.
assert ((202 + 0) = 202); auto.
rewrite H7.
apply Gr_p2; auto.
auto.
auto.

(* jd *)
apply I_infer_jd with T_cont3_p T_G_id T_reader_A T_reader_G; auto.

unfold Psi.
unfold get.
auto.
Qed.

Lemma wf_cont3: I_infer Inv Psi (T_cont3_p, T_G_id, T_reader_A, T_reader_G) C_cont3.
unfold C_cont3.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.

Lemma T_cont3_p_p1 : 
   forall (R : regs) (M M' : heap),
   Inv M -> T_cont3_p (R, M) -> T_reader_A (R, M) M' -> T_cont3_p (R, M').
unfold T_cont3_p.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.

unfold T_reader_A in H1.
unfold T_cont3_p in H2.
SplitAllHyps.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H6; auto.
SplitAllHyps.
split.
rewrite <- H4.
auto.
split.
rewrite <- H9; auto.
rewrite <- H7.
assert ( mem_sigma M 0 100 = mem_sigma M' 0 100 ).
apply mem_p3.
auto.
rewrite <- H11; auto.
Qed.
apply T_cont3_p_p1.

(* movi *)
apply I_infer_movi with 
(fun s:state => let (R,M) := s in T_cont3_p (R,M) /\ R r0 = 0)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_cont3_p.
intros.
SplitAllHyps.
split.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* st ta 0 r0 *)
apply I_infer_st with
(fun s : state =>
      let (R, M) := s in 
	R r0 = 0 /\TnTa (R,M) /\ M (rf + R ta) = 0 /\ M l = R tn 
	/\ mem_sigma M rf V_length + 1 = M cnt )
T_G_id T_reader_A T_reader_G.
intros.
unfold T_cont3_p in H0.
SplitAllHyps.
split.
rewrite H2.
apply Inv_p8'; auto.
split.
auto.
split.
apply TnTa_p1 with M.
auto.
assert ((ext_w M (R ta + 0) (R r0) (0 + R ta) = 0 /\
   ext_w M (R ta + 0) (R r0) 202 = R tn) /\
   mem_sigma (ext_w M (R ta + 0) (R r0)) 0 100 + 1 =
   ext_w M (R ta + 0) (R r0) 200).
split.
assert (R ta + 0 = 0 + R ta); auto.
rewrite H4.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (true = beq_nat (0 + R ta) (0 + R ta)).
apply beq_nat_refl.
rewrite <- H6.
auto.
unfold TnTa in H0.
SplitAllHyps.
assert (0 + R ta = R ta); auto.
rewrite H7.
assert (R ta < 202).
apply lt_trans with 100; omega.
assert (R ta <> l).
apply lt_neq; auto.
assert (false = beq_nat l (R ta)).
apply beq_neq_false.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H11; auto.


assert (0 + R ta = R ta + 0).
auto.
rewrite <- H4.
assert (mem_sigma (ext_w M (rf + R ta) (R r0)) rf V_length + 1 = mem_sigma M rf V_length).
rewrite H2.
apply mem_p5; auto.
rewrite H6.
assert (0 + R ta = R ta); auto.
rewrite H7.
assert (R ta < cnt).
unfold TnTa in H0.
SplitAllHyps.
apply lt_trans with 100; omega.
assert (R ta <> cnt).
apply lt_neq; auto.
assert (false = beq_nat cnt (R ta)).
apply beq_neq_false.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H10; auto.
tauto.
unfold T_G_id.
intros.
rewrite <- H1.
SplitAllHyps.
assert (rf + R ta = R ta + rf); auto.
rewrite <- H0.
unfold T_cont3_p in H2.
SplitAllHyps.
apply Gr_p7; auto.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id.
auto.
unfold T_reader_A.
intros.
SplitAllHyps.
split.
auto.
split.
apply TnTa_p1 with M;auto.
split.
rewrite <- H2; auto.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H3; auto.
SplitAllHyps.
split.
rewrite <- H10; auto.
assert (mem_sigma M rf V_length = mem_sigma M' rf V_length).

apply mem_p3.
auto.
rewrite <- H12.
rewrite <- H8.
auto.

(* movi r1 cnt *)
apply I_infer_movi with
(fun s : state =>
let (R, M) := s in
    R r0 = 0 /\ TnTa (R, M) /\ M (0 + R ta) = 0 /\ M 202 = R tn /\ mem_sigma M 0 100 + 1 = M 200 /\
    R r1 = cnt)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
tauto.
intros.
SplitAllHyps.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
SplitAllHyps.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
SplitAllHyps.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* ld r2 r1 0 *)
apply I_infer_ld with
(fun s : state => let (R, M) := s in 
R r0 = 0 /\ TnTa (R, M) /\ M (0 + R ta) = 0 /\ M 202 = R tn /\ mem_sigma M 0 100 + 1 = M 200
  /\ R r1 = cnt /\ R r2 = M cnt)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
split; auto.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
split.
tauto.
split.
tauto.
split; auto.
rewrite H6.
auto.

intros.
apply Gr_p4 with r2 (M (R r1 + 0)); auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r2 (M (R r1 + 0)); auto.
discriminate.
discriminate.

(* yield *)
apply I_infer_yield.
intros.
apply Gr_p1.
intros.
SplitAllHyps.
split.
auto.
split.
apply TnTa_p1 with M; auto.
unfold T_reader_A in H1.
SplitAllHyps.
assert (M 202 = M' 202 /\
        M 200 = M' 200 /\
        (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
        (M 201 = 0 ->
         mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H9; auto.
SplitAllHyps.
split.
rewrite <- H7; auto.
split.
rewrite <- H12; auto.
rewrite <- H10.
assert (mem_sigma M rf V_length = mem_sigma M' rf V_length).

apply mem_p3.
auto.
rewrite <- H14.
auto.

(* movi r3 1 *)
apply I_infer_movi with 
(fun s : state =>
      let (R, M) := s in
      R r0 = 0 /\
      TnTa (R, M) /\
      M (0 + R ta) = 0 /\
      M 202 = R tn /\
      mem_sigma M 0 100 + 1 = M 200 /\ M writ = 0 /\ mem_sigma M wf V_length = 0 
      /\ R r1 = 200 /\ R r2 = M 200 /\ R r3 = 1)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
split.
auto.
split.
apply TnTa_p2 with M; auto.

discriminate.
discriminate.
split; auto.
split; auto.
split; auto.
assert (M cnt > 0).
omega.
assert (0 <> M cnt).
apply lt_neq.
auto.
assert (M writ = 0 /\ mem_sigma M wf V_length = 0).
apply Inv_p2; auto.
inversion_clear H9.
tauto.
intros.
apply Gr_p4 with r0 1; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r0 1; auto.
discriminate.
discriminate.

(* sub r2 r2 r3 *)
apply I_infer_sub with 
(fun s : state =>
      let (R, M) := s in
      R r0 = 0 /\
      TnTa (R, M) /\
      M (0 + R ta) = 0 /\
      M l = R tn /\
      M writ = 0 /\
      mem_sigma M wf V_length = 0 /\
      R r1 = 200 /\ R r2 = mem_sigma M 0 100)
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.

intros.
SplitAllHyps.
split.
auto.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
split; auto.
split; auto.
split; auto.
rewrite H10.
rewrite H8.
split; auto.
Hypothesis tmp1 : forall M, mem_sigma M 0 100 + 1 = M 200 -> M 200 - 1 = mem_sigma M 0 100.
apply tmp1; auto.

intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 cnt; auto.
discriminate.
discriminate.

(* st r1 0 r2 *)
apply I_infer_st with 
(fun s : state => let (R,M) := s in 
    R r0 = 0 /\ TnTa (R,M) /\ M (rf + R ta) = 0 /\ M l = R tn /\ 
    M writ = 0 /\
    mem_sigma M wf V_length = 0 /\ 
    mem_sigma M rf V_length = M cnt /\
    R r2 = M cnt)
        (* /\ M writ = 0 /\ mem_sigma M wf V_length = 0) *)
T_G_id T_reader_A T_reader_G.
intros.

SplitAllHyps.
rewrite H6.
assert (cnt + 0 = cnt).
auto.
rewrite H7.
split.

apply Inv_p9; auto.

split.
auto.
split.
apply TnTa_p1 with M; auto.
split.
unfold TnTa in H0.
SplitAllHyps.
assert ((0 + R ta) = R ta).
auto.
rewrite H10.
assert (R ta < 200).
omega.
assert (R ta <> cnt).
apply lt_neq; auto.
assert (false = beq_nat (R ta) cnt).
apply beq_neq_false; auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H14.
rewrite H10 in H2; auto.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
split.
assert (mem_sigma (ext_w M cnt (R r2)) wf V_length = mem_sigma M wf V_length).
apply mem_p2; auto.
rewrite H9; auto.
assert (mem_sigma (ext_w M cnt (R r2)) rf V_length = mem_sigma M rf V_length).
apply mem_p2; auto.
omega.
rewrite H9.
split; auto.


unfold T_G_id.
intros.
rewrite <- H1.

SplitAllHyps.
rewrite H7.
assert (cnt + 0 = cnt); auto.
rewrite H8.

apply Gr_p6; auto.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.

intros.
SplitAllHyps.
split; auto.
split.
apply TnTa_p1 with M; auto.
unfold T_reader_A in H1.
SplitAllHyps.
assert (M 202 = M' 202 /\
        M 200 = M' 200 /\
        (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
        (M 201 = 0 ->
         mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H10; auto.
SplitAllHyps.
split.
rewrite <- H8; auto.
split.
rewrite <- H13; auto.
assert (M' 201 = 0 /\ mem_sigma M' 100 100 = 0).
apply H16; auto.
inversion_clear H15.
split; auto.
split; auto.
rewrite <- H11.
assert (mem_sigma M rf V_length = mem_sigma M' rf V_length).

apply mem_p3.
auto.
rewrite <- H15.
auto.

(* beq r1 r0 L_relw *)
apply I_infer_beq with
(*(fun s : state =>
      let (R, M) := s in
      TnTa (R, M) /\
      M (0 + R ta) = 0 /\ M 202 = R tn /\ mem_sigma M 0 100 = M 200 /\ M writ = 0)*)
T_cont4_p T_relw_p T_G_id T_G_id T_reader_A T_reader_A T_reader_G T_reader_G.
intros.
SplitAllHyps.
unfold T_relw_p.
split; auto.
split; auto.
split.
rewrite H2 in H1.
rewrite H1 in H9; auto.
auto.
unfold T_G_id.
intros.
rewrite <- H2.
apply Gr_p1.
auto.
auto.
unfold Psi.
unfold get.
auto.
intros.
SplitAllHyps.
unfold T_cont4_p.
auto.
unfold T_G_id.
intros.
rewrite <- H2.
apply Gr_p1.
auto.
auto.

Lemma wf_cont4 : I_infer Inv Psi (T_cont4_p, T_G_id, T_reader_A, T_reader_G) (C_rel L_reader).
apply wf_frel with T_reader_p; auto.
unfold Psi.
unfold get.
auto.

Lemma T_cont4_p_p1 : 
forall (R : regs) (M M' : heap), T_cont4_p (R, M) -> T_reader_A (R, M) M' -> T_cont4_p (R, M').
unfold T_cont4_p.
unfold T_reader_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H2; auto.
SplitAllHyps.
split.
rewrite <- H8.
auto.
split.
rewrite <- H1; auto.
rewrite <- H6.
assert ( mem_sigma M 0 100 = mem_sigma M' 0 100 ).
apply mem_p3.
auto.
rewrite <- H10.
auto.
Qed.
apply T_cont4_p_p1.
unfold T_cont4_p.
intros.
SplitAllHyps.
auto.
unfold T_reader_p.
unfold T_cont4_p.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
unfold TnTa in H0.
SplitAllHyps.
assert (rf + R ta = R ta); auto.
rewrite H4.
assert (R ta < l).
apply lt_trans with 100; omega.
assert (R ta <> l).
apply lt_neq; auto.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (false = beq_nat (R ta) l).
apply beq_neq_false; auto.
rewrite <- H8.
rewrite H4 in H1.
unfold beq_nat.
split; auto.
apply lt_neq; auto.
Qed.
apply wf_cont4.
Qed.

Lemma wf_fvwrit : 
  forall (f:label) (p1 p2 :precond), 
    get Psi f (p2, T_G_id, T_reader_A, T_reader_G) -> 
    (*(forall R M M', p1 (R,M) -> T_reader_A (R,M) M' -> p1 (R,M')) -> *)
    (forall R M, p1 (R,M) -> M l = R tn /\ M writ = 0 /\ M cnt = 0 /\ mem_sigma M wf V_length = 0) ->
    (forall R M r n, p1 (R,M) -> (r = r1 \/ r = r2) -> p1 (ext_r R r n, M)) ->
    (forall R M, p1 (R,M) -> p2 (R, ext_w M writ 1)) ->
    I_infer Inv Psi (p1, T_reader_G, T_reader_A, T_reader_G) (C_vwrit f).
intros.
unfold C_vwrit.

(* movi r1 1 *)
apply I_infer_movi with 
(fun s : state => let (R,M) := s in p1 (R,M) /\ R r1 = 1) T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
split.
apply H1; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
intros.
apply Gr_p4 with r1 1; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 1; auto.
discriminate.
discriminate.

(* movi r2 writ *)
apply I_infer_movi with 
(fun s : state => let (R,M) := s in p1 (R,M) /\ R r1 = 1 /\ R r2 = writ) 
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
inversion_clear H4.
split.
apply H1; auto.
auto.

intros.
apply Gr_p4 with r1 1; auto.
discriminate.
discriminate.
intros.
apply Ar_p1; auto.
discriminate.
discriminate.
intros.
apply Gr_p4 with r1 1; auto.
discriminate.
discriminate.

(* st r2 0 r1 *)
apply I_infer_st with p2 T_G_id T_reader_A T_reader_G.
intros.
SplitAllHyps.
split.
rewrite H7.
assert (writ + 0 = writ); auto.
rewrite H6.
rewrite H4.
apply Inv_p10; auto.
assert (M l = R tn /\ M writ = 0 /\ M cnt = 0 /\ mem_sigma M wf V_length = 0).
apply H0; auto.
SplitAllHyps.
auto.
assert (M l = R tn /\ M writ = 0 /\ M cnt = 0 /\ mem_sigma M wf V_length = 0).
apply H0; auto.
SplitAllHyps.
auto.
rewrite H4.
rewrite H7.
auto.
unfold T_G_id.
intros.
rewrite <- H5.
SplitAllHyps.
rewrite H8.
assert (writ + 0 = writ); auto.
apply Gr_p3.
assert (M 202 = R tn /\ M 201 = 0 /\ M 200 = 0 /\ mem_sigma M 100 100 = 0).
apply H0; auto.
SplitAllHyps.
auto.
auto.
auto.

(* jd f *)
apply I_infer_jd with p2 T_G_id T_reader_A T_reader_G; auto.
Qed.

Lemma wf_relw : I_infer Inv Psi (T_relw_p, T_G_id, T_reader_A, T_reader_G) C_relw.
unfold C_relw.
apply I_infer_yield.
unfold T_G_id.
intros.
auto.

Lemma T_relw_p_p1 : 
forall (R : regs) (M M' : heap),
   T_relw_p (R, M) -> T_reader_A (R, M) M' -> T_relw_p (R, M').
unfold T_relw_p.
unfold T_reader_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
assert (M 202 = M' 202 /\
       M 200 = M' 200 /\
       (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
       (M 201 = 0 ->
        mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H2; auto.
SplitAllHyps.
rewrite <- H11.
rewrite <- H9.
split; auto.
split; auto.
assert (M' 201 = 0 /\ mem_sigma M' 100 100 = 0).
apply H14; auto.
inversion_clear H13.
split; auto.
split.
rewrite <- H1; auto.
split; auto.
assert (mem_sigma M rf V_length = mem_sigma M' rf V_length).
apply mem_p1; auto.
rewrite <- H13.
auto.
Qed.

intros.
apply T_relw_p_p1 with M; auto.

apply wf_fvwrit with T_cont4_p; auto.
unfold Psi.
unfold get.
auto.

unfold T_relw_p.
intros.
SplitAllHyps.
tauto.
unfold T_relw_p.
unfold TnTa.
intros.
SplitAllHyps.
inversion_clear H0.
rewrite H8.
unfold ext_r.
unfold ext.
unfold beq_r.
tauto.
rewrite H8.
unfold ext_r.
unfold ext.
unfold beq_r.
tauto.
unfold T_relw_p.
unfold T_cont4_p.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
unfold TnTa in H0.
SplitAllHyps.
assert (rf + R ta = R ta); auto.
rewrite H7.
split.
assert (R ta < writ).
apply lt_trans with 100; auto.
omega.
assert (R ta <> 201).
apply lt_neq; auto.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (false = beq_nat (R ta) writ).
apply beq_neq_false; auto.
rewrite <- H11.
rewrite <- H7; auto.
assert (mem_sigma (ext_w M writ 1) rf V_length = mem_sigma M rf V_length).
apply mem_p2; auto.
omega.
rewrite H9.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
Qed.


Lemma wf_reader : I_infer Inv Psi (T_reader_p, T_G_id, T_reader_A, T_reader_G) C_reader.
unfold C_reader.
unfold C_acq.
unfold facq.
unfold beq_w.
unfold beq_nat.

(* yield *)
apply I_infer_yield.
unfold T_G_id.
auto.

Lemma T_reader_p_p1 :
   forall (R : regs) (M M' : heap),
   Inv M -> T_reader_p (R, M) -> T_reader_A (R, M) M' -> T_reader_p (R, M').

unfold T_reader_p.
unfold T_reader_A.

intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.

split.
rewrite <- H2; auto.
apply H5; auto.
Qed.
apply T_reader_p_p1.

(* movi r0 0 *)

apply I_infer_movi with 
  (fun s : state => let (R,M):=s in T_reader_p (R,M) /\ R r0 = 0) T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_reader_p.
intros.
SplitAllHyps.
split.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* movi r1 l *)
apply I_infer_movi with 
  (fun s : state => let (R,M):=s in T_reader_p (R,M) /\ R r0 = 0 /\ R r1 = l) 
  T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_reader_p.
intros.
SplitAllHyps.
split.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* ld r2 r1 0 *)
apply I_infer_ld with 
  (fun s : state => let (R,M):=s in T_reader_p (R,M) /\ R r0 = 0 /\ R r1 = l /\ R r2 = M l) 
  T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
unfold T_reader_p.
intros.
SplitAllHyps.
split.
split.
apply TnTa_p2 with M; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
split. auto.
split. auto.
assert (R r1 + 0 = R r1).
auto.
rewrite H4.
rewrite H3. auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* bgt r2 r0 0 *)
apply I_infer_bgt with 
(fun s : state =>
      let (R, M) := s in
      T_reader_p (R, M) /\ R r0 = 0 /\ R r1 = l /\ R r2 = M l /\ M l = 0)
T_reader_p T_reader_G T_G_id T_reader_A T_reader_A T_reader_G T_reader_G.
tauto.
unfold T_G_id.
intros.
rewrite <- H2.


apply Gr_p1.
auto.
auto.
unfold Psi.
unfold get.
auto.
intros.
SplitAllHyps.
split; [auto | split ; [auto | split; [auto | split; auto]]].
rewrite H5 in H1.
rewrite H0 in H1.
auto with arith.
auto.
auto.
auto.

(* st r1 0 tn *)
apply I_infer_st with
(*(fun s : state =>
      let (R, M) := s in
      TnTa (R,M) /\ M (rf + R ta) = 0 /\ R r0 = 0 /\ R r1 = l /\ R r2 = 0 /\ M l = R tn)*)
T_cont1_p T_G_id T_reader_A T_reader_G.
intros.
SplitAllHyps.
split.
(*Hypothesis Inv_p1 : forall M loc n, Inv M -> M l = 0 -> 
                         loc > 199 -> loc <> cnt -> loc <> writ -> Inv (ext_w M loc n).*)
rewrite H2.
assert (l + 0 = l).
auto.
rewrite H4.
apply Inv_p1; auto with arith.
discriminate.
unfold T_reader_p in H1.
SplitAllHyps.
unfold T_cont1_p.
split.
apply TnTa_p1 with M.
auto.
assert ((ext_w M (R r1 + 0) (R tn) (0 + R ta) = 0 /\
   ext_w M (R r1 + 0) (R tn) 202 = R tn) /\
   mem_sigma (ext_w M (R r1 + 0) (R tn)) 0 100 =
   ext_w M (R r1 + 0) (R tn) 200).
split.
unfold TnTa in H4.
SplitAllHyps.
rewrite H2.
assert (0 + R ta < 202 + 0).
assert (0 + R ta = R ta); auto.
rewrite H8.
apply lt_trans with 100; omega.
assert (0 + R ta <> 202 + 0).
apply lt_neq; auto.
assert (false = beq_nat (0 + R ta) (202 + 0)).
apply beq_neq_false.
auto.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H11.
assert (202 + 0 = 202); auto.
(* Hypothesis Inv_p4 : forall M n, Inv M -> M l = 0 -> mem_sigma (ext_w M l n) 0 100 = ext_w M l n cnt.*)
rewrite H2.
assert (202 + 0 = 202); auto.
rewrite H6.
apply Inv_p4; auto.
tauto.


unfold T_G_id.
intros.
rewrite <- H1.
unfold T_reader_p in H0.
SplitAllHyps.
rewrite H3.
assert ((l + 0) = l); auto.
rewrite H7.
(* Hypothesis Gr_p2 : forall R M, M l = 0 -> TnTa (R,M) -> M l = 0 -> T_reader_G (R,M) (ext_w M l (R tn)).*)
apply Gr_p2; auto.
auto.
auto.

(* jd *)
apply I_infer_jd with T_cont1_p T_G_id T_reader_A T_reader_G; auto.

unfold Psi.
unfold get.
auto.
Qed.

Lemma wf_cont1 : I_infer Inv Psi (T_cont1_p, T_G_id, T_reader_A, T_reader_G) C_cont1.
unfold C_cont1.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.

apply T_cont1_p_p1.

(* movi r0 0 *)
apply I_infer_movi with 
(fun s : state => let (R,M) := s in T_cont1_p (R,M) /\ R r0 = 0) T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
split.
apply T_cont1_p_p2; auto.

unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* movi r1 200 *)
apply I_infer_movi with 
(fun s : state => let (R,M) := s in T_cont1_p (R,M) /\ R r0 = 0 /\ R r1 = cnt) 
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
inversion_clear H0.
split.
apply T_cont1_p_p2; auto.

unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* ld r2 r1 0 *)
apply I_infer_ld with
(fun s : state => let (R,M) := s in T_cont1_p (R,M) /\ R r0 = 0 /\ R r1 = cnt /\ R r2 = M cnt) 
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
split.
apply T_cont1_p_p2; auto.

unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
split; auto.
rewrite H3.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* yield *)
apply I_infer_yield.
unfold T_reader_G; tauto.

intros.
SplitAllHyps.
split.
apply T_cont1_p_p1 with M; auto.

split; auto.
split; auto.
unfold T_cont1_p in H2.
unfold T_reader_A in H1.
SplitAllHyps.
assert (M 202 = M' 202 /\ M 200 = M' 200 /\
            (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
            (M 201 = 0 -> mem_sigma M 100 100 = 0 -> M' 201 = 0 /\ mem_sigma M' 100 100 = 0)).
apply H9; auto.
SplitAllHyps.
rewrite <- H10.
auto.

(* beq r2 r0 L_getw *)
apply I_infer_beq with T_inc_cnt_p T_getw_p T_G_id T_G_id T_reader_A T_reader_A T_reader_G T_reader_G.
unfold T_getw_p.
intros.
SplitAllHyps.
split; auto.
rewrite <- H5.
rewrite H0 in H1.
auto.

unfold T_G_id.
intros.
rewrite <- H2.
apply Gr_p1.
auto.
auto.
unfold Psi.
unfold get.
auto.
unfold T_inc_cnt_p.
intros.
SplitAllHyps.
split; auto.

apply Inv_p2; auto.
rewrite H0 in H1.
rewrite <- H5.
auto.

unfold T_G_id.
intros.
rewrite <- H2.
apply Gr_p1.
auto.
auto.

(* jd inc_cnt *)
apply I_infer_jd with T_inc_cnt_p T_G_id T_reader_A T_reader_G; auto.
unfold Psi.
unfold get.
auto.
Qed.

Lemma wf_getw : I_infer Inv Psi (T_getw_p, T_G_id, T_reader_A, T_reader_G) C_getw.
unfold C_getw.
unfold C_pwrit.
unfold fpwrit.
unfold beq_w.
unfold beq_nat.

(* yield *)
apply I_infer_yield.
intros.
unfold T_G_id.
auto.

unfold T_getw_p.
intros.
inversion_clear H0.
split.
apply T_cont1_p_p1 with M; auto.
unfold T_reader_A in H1.
unfold T_cont1_p in H2.
SplitAllHyps.
assert (M 202 = M' 202 /\ M 200 = M' 200 /\
           (forall i : nat, i >= 0 -> i < 100 -> M (0 + i) = M' (0 + i)) /\
           (M 201 = 0 -> mem_sigma M 100 100 = 0 -> M' 201 = 0  /\ mem_sigma M' 100 100 = 0)).
apply H7; auto.
SplitAllHyps.
rewrite <- H8; auto.
Lemma T_getw_p_p1 : 
  forall R M r n, T_getw_p (R,M) -> (r = r0 \/ r = r1 \/ r = r2 \/ r = r3) -> T_getw_p (ext_r R r n, M).
unfold T_getw_p.
intros.
inversion_clear H.
split.
apply T_cont1_p_p2; auto.
auto.
Qed.

(* movi r0 0 *)
apply I_infer_movi with
(fun s:state => let (R,M) := s in T_getw_p (R,M) /\ R r0 = 0) T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
split.
apply T_getw_p_p1; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* movi r1 201 *)
apply I_infer_movi with
(fun s:state => let (R,M) := s in T_getw_p (R,M) /\ R r0 = 0 /\ R r1 = writ) 
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
inversion_clear H0.
split.
apply T_getw_p_p1; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* ld r2 r1 0 *)
apply I_infer_ld with
(fun s:state => let (R,M) := s in T_getw_p (R,M) /\ R r0 = 0 /\ R r1 = writ /\ R r2 = M writ) 
T_reader_G T_reader_A T_reader_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
split.
apply T_getw_p_p1; auto.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
split; auto.
rewrite H3; auto.

unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_A.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
unfold T_reader_G.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

(* beq r2 r0 2 *)
apply I_infer_beq with 
(fun s:state => 
  let (R,M) := s in M writ <> 0 /\ T_getw_p (R,M) /\ R r0 = 0 /\ R r1 = writ /\ R r2 = M writ)
T_getw_p T_reader_G T_G_id T_reader_A T_reader_A T_reader_G T_reader_G.
tauto.
unfold T_G_id.
intros.
rewrite <- H2.
apply Gr_p1.
auto.
auto.
unfold Psi.
unfold get.
auto.
intros.
split.
SplitAllHyps.
rewrite <- H5.
rewrite <- H0.
auto.
auto.
auto.
auto.
auto.

(* st r1 0 r0 *)
apply I_infer_st with T_inc_cnt_p T_G_id T_reader_A T_reader_G.
intros.
SplitAllHyps.
split.
rewrite H2.
assert (R r1 + 0 = R r1); auto.
rewrite H4.
rewrite H3.
apply Inv_p3; auto.
unfold T_inc_cnt_p.
unfold T_getw_p in H0.
unfold T_cont1_p in H0.
unfold T_cont1_p.
SplitAllHyps.
split.
split.
apply TnTa_p1 with M; auto.
unfold TnTa in H0.
SplitAllHyps.
rewrite H3.
assert (0 + R ta < 201 + 0).
assert (0 + R ta = R ta); auto.
rewrite H10.
apply lt_trans with 100; omega.
assert (0 + R ta <> 201 + 0).
apply lt_neq; auto.
assert (false = beq_nat (0 + R ta) (201 + 0)).
apply beq_neq_false.
auto.
assert ((ext_w M (201 + 0) (R r0) (0 + R ta) = 0 /\
   ext_w M (201 + 0) (R r0) 202 = R tn) /\
   mem_sigma (ext_w M (201 + 0) (R r0)) 0 100 = ext_w M (201 + 0) (R r0) 200).
split.
unfold ext_w.
unfold ext.
unfold beq_w.
rewrite <- H13.
auto.
auto.

tauto.
rewrite H3.
rewrite H2.
assert (201 + 0 = 201).
auto.
rewrite H8.
split.
unfold ext_w.
unfold ext.
unfold beq_w.
unfold beq_nat.
auto.
assert (mem_sigma M wf V_length = 0).
apply Inv_p6; auto.
assert (mem_sigma (ext_w M writ 0) wf V_length = mem_sigma M wf V_length).
apply mem_p2; auto.
rewrite H11; auto.

unfold T_G_id.
intros.
rewrite <- H1.
(* Hypothesis Gr_p3 : forall R M n, T_reader_G (R,M) (ext_w M writ n).*)
SplitAllHyps.
rewrite H4.
assert (201 + 0 = 201).
auto.
rewrite H5.
apply Gr_p3.
unfold T_getw_p in H0.
unfold T_cont1_p in H0.
SplitAllHyps.
auto.
auto.
auto.

(* jd L_inc_cnt *)
apply I_infer_jd with T_inc_cnt_p T_G_id T_reader_A T_reader_G; auto.
unfold Psi.
unfold get.
auto.
Qed.


Lemma wf_writer : I_infer Inv Psi (T_writer_p, T_G_id, T_writer_A, T_writer_G) (C_writer).
unfold C_writer.
unfold C_pwrit.
unfold fpwrit.
unfold beq_w.
unfold beq_nat.

apply I_infer_yield.
unfold T_G_id.
intros.
auto.

Lemma T_writer_p_p1 :
   forall (R : regs) (M M' : heap),
   Inv M -> T_writer_p (R, M) -> T_writer_A (R, M) M' -> T_writer_p (R, M').
unfold T_writer_p.
unfold T_writer_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
rewrite <- H2; auto.
Qed.

apply T_writer_p_p1.

(* movi r0 0 *)

Lemma reg_update_p :
    forall (R:regs) (r1 r2:reg) (n: nat), r1 <> r2 -> (ext_r R r1 n) r2 = R r2.
intros.
unfold ext_r.
unfold ext.
assert (beq_r r2 r1 = false).
apply trv8; auto.
rewrite H0.
auto.
Qed.

Lemma T_writer_G_p1 :
    forall (R : regs) (r: reg) (M M' : heap) (n : nat),
        r <> ta -> T_writer_G (ext_r R r n, M) M' -> T_writer_G (R,M) M'.
unfold T_writer_G.
intros R r M M' n rneq.
assert (ext_r R r n ta = R ta).
apply reg_update_p; auto.
rewrite H.
auto.
Qed.

apply I_infer_movi with
(fun s:state => let (R,M) := s in T_writer_p (R,M) /\ R r0 = 0) T_writer_G T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
split.

Lemma T_writer_p_p2:
    forall (R: regs) (r: reg) (M : heap) (n : nat),
        r <> ta -> r <> tn -> T_writer_p (R, M) -> T_writer_p (ext_r R r n, M).
unfold T_writer_p.
intros.
inversion_clear H1.
split.
apply TnTa_p2 with M; auto.
assert (ext_r R r n ta = R ta).
apply reg_update_p; auto.
rewrite H1; auto.
Qed.
apply T_writer_p_p2; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

intros.
apply T_writer_G_p1 with r0 0; auto.
discriminate.

Lemma T_writer_A_p1 :
    forall (R : regs) (r: reg) (M M' : heap) (n : nat),
        r <> ta -> T_writer_A (R,M) M' -> T_writer_A (ext_r R r n, M) M'.
unfold T_writer_A.
intros R r M M' n rneq.
assert (ext_r R r n ta = R ta).
apply reg_update_p; auto.
rewrite H.
auto.
Qed.

intros.
apply T_writer_A_p1; auto.
discriminate.

intros.
apply T_writer_G_p1 with r0 0; auto.
discriminate.

(* movi r1 201 *)
apply I_infer_movi with
(fun s:state => let (R,M) := s in T_writer_p (R,M) /\ R r0 = 0 /\ R r1 = writ) 
T_writer_G T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
inversion_clear H0.
split.
apply T_writer_p_p2; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.

intros.
apply T_writer_G_p1 with r1 201; auto.
discriminate.
intros.
apply T_writer_A_p1; auto.
discriminate.
intros.
apply T_writer_G_p1 with r1 201; auto.
discriminate.

(* ld r2 r1 0 *)
apply I_infer_ld with
(fun s:state => let (R,M) := s in T_writer_p (R,M) /\ R r0 = 0 /\ R r1 = writ /\ R r2 = M writ) 
T_writer_G T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
SplitAllHyps.
split.
apply T_writer_p_p2; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
split; auto.
split; auto.
rewrite H3; auto.

intros.
apply T_writer_G_p1 with r2 (M (R r1 + 0)); auto.
discriminate.
intros.
apply T_writer_A_p1; auto.
discriminate.
intros.
apply T_writer_G_p1 with r2 (M (R r1 + 0)); auto.
discriminate.

(* beq r2 r0 2 *)
apply I_infer_beq with 
(fun s:state => 
  let (R,M) := s in M writ <> 0 /\ T_writer_p (R,M) /\ R r0 = 0 /\ R r1 = writ /\ R r2 = M writ)
T_writer_p T_writer_G T_G_id T_writer_A T_writer_A T_writer_G T_writer_G.
tauto.
unfold T_G_id.
intros.
rewrite <- H2.
Lemma T_writer_G_p2 : forall (R: regs) (M : heap), T_writer_G (R,M) M.
unfold T_writer_G.
tauto.
Qed.
apply T_writer_G_p2.
auto.
auto.
unfold Psi.
unfold get.
auto.
intros.
split.
SplitAllHyps.
rewrite <- H5.
rewrite <- H0.
auto.
auto.
auto.
auto.
auto.

(* st r1 0 r0 *)
apply I_infer_st with T_cont5_p T_cont5_G T_writer_A T_writer_G.
intros.
SplitAllHyps.
split.
rewrite H2.
assert (R r1 + 0 = R r1); auto.
rewrite H4.
rewrite H3.
apply Inv_p3; auto.
unfold T_cont5_p.
unfold T_writer_p in H0.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
unfold TnTa in H4.
SplitAllHyps.
rewrite H3.
assert(201 + 0 = 201).
auto.
rewrite H7.
Lemma mem_p6 : forall (m n:nat) (M : heap), ext_w M m n m = n.
intros.
unfold ext_w.
unfold ext.
assert (beq_w m m = true).
apply trv9; auto.
rewrite H.
auto.
Qed.
Lemma mem_p7 : forall (m n k k': nat) (M : heap), m <> n -> M n = k' -> ext_w M m k n = k'.
intros.
unfold ext_w.
unfold ext.
unfold beq_w.
assert (false = beq_nat n m).
apply beq_neq_false.
auto.
rewrite <- H1.
auto.
Qed.
split.
assert (100 + R ta < 201).
omega.
assert (100 + R ta <> 201).
apply lt_neq; auto.
apply mem_p7; auto.
split.
rewrite H2.
apply mem_p6; auto.
split.
assert (200 <> 201).
discriminate.
apply mem_p7; auto.

Hypothesis Inv_p11:
  forall M, Inv M -> M writ <> 0 -> M cnt = 0.

apply Inv_p11; auto.

assert (mem_sigma M wf V_length = 0).
apply Inv_p6; auto.
assert (mem_sigma (ext_w M writ 0) wf V_length = mem_sigma M wf V_length).
apply mem_p2; auto.
rewrite H2.
rewrite H10.
auto.

unfold T_cont5_G.
intros.
rewrite <- H1.
(* Hypothesis Gr_p3 : forall R M n, T_reader_G (R,M) (ext_w M writ n).*)
SplitAllHyps.
rewrite H4.
assert (201 + 0 = 201).
auto.
rewrite H5.
rewrite H3.
Hypothesis T_writer_G_p3 : forall R M, M writ <> 0 -> T_writer_G (R,M) (ext_w (ext_w M writ 0) (wf + R ta) 1).
apply T_writer_G_p3; auto.
auto.
auto.

(* jd L_inc_cnt *)
apply I_infer_jd with T_cont5_p T_cont5_G T_writer_A T_writer_G; auto.
unfold Psi.
unfold get.
auto.
Qed.


Lemma wf_cont5 : I_infer Inv Psi (T_cont5_p, T_cont5_G, T_writer_A, T_writer_G) (C_cont5).
unfold C_cont5.

(* movi r1 1 *)
apply I_infer_movi with 
(fun s : state => let (R,M) := s in T_cont5_p (R,M) /\ R r1 = 1) T_cont5_G T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
split.

unfold T_cont5_p in H0.
unfold T_cont5_p.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
assert (ext_r R r1 1 ta = R ta).
apply reg_update_p; auto.
discriminate.
rewrite H4.
auto.
tauto.

unfold T_cont5_G.
intros.
assert (ext_r R r1 1 ta = R ta).
apply reg_update_p; auto.
discriminate.
rewrite H2 in H1.
auto.

intros.
apply T_writer_A_p1; auto.
discriminate.

intros.
apply T_writer_G_p1 with r1 1; auto.
discriminate.

(* st ta 100 r1 *)
Definition T_cont5_p' : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\  M (wf + R ta) = 1  /\ M writ = 0 /\ M cnt = 0.

apply I_infer_st with
T_cont5_p' T_G_id T_writer_A T_writer_G.

unfold T_cont5_p.
intros.
SplitAllHyps.
split.

Hypothesis Inv_p12 : 
  forall (R:regs) (M : heap), TnTa (R,M) -> Inv M -> mem_sigma M wf V_length = 0 -> M writ = 0 
                      -> M cnt = 0 -> Inv (ext_w M (R ta + wf) 1).

rewrite H2.
apply Inv_p12; auto.

unfold T_cont5_p'.
split.
apply TnTa_p1 with M; auto.
rewrite H2.
split.
assert (R ta + 100 = 100 + R ta).
auto with arith.
rewrite <- H5.
apply mem_p6; auto.

Lemma TnTa_p4 : forall (R : regs) (M : heap), TnTa(R,M) -> R ta + wf < 200.
intros.
unfold TnTa in H.
SplitAllHyps.
omega.
Qed.

split.
apply mem_p7; auto.
assert (R ta + wf < 201).
apply lt_trans with 200.
apply TnTa_p4 with M; auto.
auto.
apply lt_neq; auto.
apply mem_p7; auto.
assert (R ta + wf < 200).
apply TnTa_p4 with M; auto.
apply lt_neq; auto.

unfold T_G_id.
intros.
rewrite <- H1.
unfold T_cont5_G.
inversion_clear H0.
rewrite H3.
assert (100 + R ta = R ta + 100).
auto with arith.
rewrite H0.
auto.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id.
auto.

Lemma T_cont5_p'_p1 :
   forall (R : regs) (M M' : heap),
   Inv M -> T_cont5_p' (R, M) -> T_writer_A (R, M) M' -> T_cont5_p' (R, M').
unfold T_cont5_p'.
unfold T_writer_A.
intros.
SplitAllHyps.
split.
apply TnTa_p1 with M; auto.
assert (M 203 = M' 203 /\ M 201 = M' 201 /\ M 200 = M' 200).
apply H3; auto.
SplitAllHyps.
split.
rewrite <- H2; auto.
split.
rewrite <- H5; auto.
rewrite <- H9; auto.
Qed.

apply T_cont5_p'_p1; auto.

(* movi r2 v *)

apply I_infer_movi with 
(fun s : state => let (R,M) := s in T_cont5_p' (R,M) /\ R r2 = v) T_writer_G T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
split.

Lemma T_cont5_p'_p2 : forall (R : regs) (M : heap) (r : reg) (n : nat),
                         r <> tn -> r <> ta -> T_cont5_p' (R, M) -> T_cont5_p' (ext_r R r n, M).
unfold T_cont5_p'.
intros.
SplitAllHyps.
split.
apply TnTa_p2 with M; auto.
assert (ext_r R r n ta = R ta).
apply reg_update_p; auto.
rewrite H4.
auto.
Qed.

apply T_cont5_p'_p2; auto.
discriminate.
discriminate.
unfold ext_r.
unfold ext.
unfold beq_r.
auto.
intros.
apply T_writer_G_p1 with r2 v; auto.
discriminate.
intros.
apply T_writer_A_p1; auto.
discriminate.
intros.
apply T_writer_G_p1 with r2 v; auto.
discriminate.

(* st r2 0 r1 *)
apply I_infer_st with T_cont5_p' T_G_id T_writer_A T_writer_G.
unfold T_cont5_p'.
intros.
SplitAllHyps.
split.
rewrite H2.
auto.

split.
apply TnTa_p1 with M; auto.
split.
apply mem_p7; auto.
assert (R ta + wf < 200).
apply TnTa_p4 with M; auto.
rewrite H2.
assert (v + 0 = v); auto.
rewrite H6.
assert (100 + R ta <> 203).
apply lt_neq; auto.
assert (wf + R ta = R ta + wf). 
auto with arith.
rewrite H7.
apply lt_trans with 200; auto.
auto.

split.
apply mem_p7; auto.
rewrite H2.
assert (v + 0 = v); auto. 
rewrite H4.
discriminate.
apply mem_p7; auto.
rewrite H2.
assert (v + 0 = v); auto. 
rewrite H4.
discriminate.
unfold T_G_id.
intros.
rewrite <- H1.
unfold T_cont5_p' in H0.
SplitAllHyps.

Hypothesis T_writer_G_pp : 
  forall (R : regs) (M : heap) (n : nat), T_writer_G (R, M) (ext_w M v n).

rewrite H3.
assert (v + 0 = v); auto.
rewrite H5.
apply T_writer_G_pp.
auto.
auto.

(* yield *)
apply I_infer_yield.
unfold T_G_id; auto.
apply T_cont5_p'_p1.

(* mivi r0 0 *)
apply I_infer_movi with
(fun s : state => let (R, M) := s in T_cont5_p' (R, M) /\ R r0 = 0) T_writer_G T_writer_A T_writer_G.
discriminate.
discriminate.
split.
apply T_cont5_p'_p2; auto.
discriminate.
discriminate.
auto.
intros.
apply T_writer_G_p1 with r0 0; auto.
discriminate.
intros.
apply T_writer_A_p1; auto.
discriminate.
intros.
apply T_writer_G_p1 with r0 0; auto.
discriminate.

unfold C_vwrit.

(* st ta wf r0 *)
Definition tmpG : guarantee :=
    fun (s : state) (M' : heap) =>
        let (R, M) := s
        in
            ext_w M writ 1 = M'.

Definition tmpPre : precond :=
  fun s : state =>
    let (R,M) := s
    in
        TnTa (R,M) /\  mem_sigma M wf V_length = 0  /\ M writ = 0 /\ M cnt = 0.

apply I_infer_st with tmpPre tmpG T_writer_A T_writer_G.
unfold T_cont5_p'.
unfold tmpPre.
intros.
SplitAllHyps.
rewrite H2.
split.

Hypothesis Inv_p13 : forall (R : regs) (M : heap), TnTa (R,M) -> Inv M -> Inv (ext_w M (R ta + wf) 0).
apply Inv_p13; auto.
split.
apply TnTa_p1 with M; auto.
Hypothesis mem_p8 : 
  forall (R: regs) (M : heap), Inv M -> M (wf + R ta) = 1 -> 
                              mem_sigma (ext_w M (R ta + wf) 0) wf V_length = 0.
split.
apply mem_p8; auto.
split.
apply mem_p7; auto.
assert (R ta + wf < 201).
apply lt_trans with 200.
apply TnTa_p4 with M; auto.
auto.
apply lt_neq; auto.
apply mem_p7; auto.
assert (R ta + wf < 200).
apply TnTa_p4 with M; auto.
apply lt_neq; auto.
unfold tmpG.
intros.
rewrite <- H1.
unfold T_cont5_p' in H0.
SplitAllHyps.
rewrite H3.
Hypothesis T_writer_G_p4 : 
    forall (R : regs) (M : heap), TnTa (R, M) -> M (wf + R ta) = 1 ->
                                    T_writer_G (R, M) (ext_w (ext_w M (R ta + wf) 0) writ 1).
apply T_writer_G_p4; auto.
auto.
auto.

(* movi r2 v *)

apply I_infer_movi with 
(fun s : state => let (R,M) := s in tmpPre (R,M) /\ R r1 = 1) tmpG T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
split.

Lemma tmpPre_p1 : forall (R : regs) (M : heap) (r : reg) (n : nat),
                         r <> tn -> r <> ta -> tmpPre (R, M) -> tmpPre (ext_r R r n, M).
unfold tmpPre.
intros.
SplitAllHyps.
split.
apply TnTa_p2 with M; auto.
auto.
Qed.


apply tmpPre_p1; auto.
discriminate.
discriminate.
auto.
auto.
intros.
apply T_writer_A_p1; auto.
discriminate.
intros.
apply T_writer_G_p1 with r1 1; auto.
discriminate.

(* movi r2 writ *)

apply I_infer_movi with 
(fun s : state => let (R,M) := s in tmpPre (R,M) /\ R r1 = 1 /\ R r2 = writ) 
         tmpG T_writer_A T_writer_G.
discriminate.
discriminate.
intros.
inversion_clear H0.
split.

apply tmpPre_p1; auto.
discriminate.
discriminate.
auto.
auto.
intros.
apply T_writer_A_p1; auto.
discriminate.
intros.
apply T_writer_G_p1 with r1 1; auto.
discriminate.

(* st r2 0 r1 *)
apply I_infer_st with
T_writer_p T_G_id T_writer_A T_writer_G.
unfold tmpPre.
unfold T_writer_p.
intros.
SplitAllHyps.
rewrite H3.
assert (writ + 0 = writ).
auto.
rewrite H5.
rewrite H0.
split.
apply Inv_p10; auto.
split.
apply TnTa_p1 with M; auto.
assert (M (wf + R ta) = 0).
Hypothesis mem_p9 : 
    forall (R : regs) (M : heap), TnTa (R, M) -> mem_sigma M 100 100 = 0 -> M (wf + R ta) = 0.

apply mem_p9; auto.
apply mem_p7; auto.
assert (wf + R ta <> 201).
apply lt_neq.
apply lt_trans with 200.
assert (wf + R ta = R ta + wf).
auto with arith.
rewrite H8.
apply TnTa_p4 with M; auto.
auto.
auto.

unfold T_G_id.
unfold tmpG.
intros.
rewrite <- H1.
SplitAllHyps.
rewrite H0.
rewrite H4.
auto.
auto.
auto.

(* jd writer *)
apply I_infer_jd with T_writer_p T_G_id T_writer_A T_writer_G; auto.
unfold Psi.
unfold get.
auto.
Qed.
