(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SCAP                                                        *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: ccap.v,v 1.1.1.1 2006/12/07 16:43:25 ssg05 Exp $ *)

Require Import ListB_MAPT.
Require Import machine.

Require Import extac.

Open Local Scope tm_scope.

Definition stpred := state -> Prop.

Definition acpred := state -> state -> Prop.

Definition assume := acpred.

Definition guarantee := acpred.

Definition imply (p p' : stpred) :=
  forall s, p s -> p' s.

Definition Imply (g g' : acpred) :=
  forall s s', g s s' -> g' s s'.

Notation " p ---> p' " := (imply p p')
  (at level 75, right associativity) : ccap_scope.

Notation " G ===> G' " := (Imply G G')
  (at level 75, right associativity) : ccap_scope.

Open Local Scope ccap_scope.

Inductive cdspec : Type :=
  | mk_cdspec : forall (p : stpred) (g : guarantee)
       (A : assume) (G : guarantee), cdspec.

Definition Aok (A : assume) :=
  forall s s' pc pc' w w',
    A s s' <-> A (rset (pcset s pc) ra w) (rset (pcset s' pc') ra w').

Definition Gok (G : guarantee) :=
  forall s s' pc pc' w w',
    G s s' <-> G (rset (pcset s pc) ra w) (rset (pcset s' pc') ra w').

Module CdSpec <: ONE_TYPE.
  Definition B := cdspec.
End CdSpec.

Module CHSpecDef := ListMapT CdSpec.
Module CHSpec := MapEx CHSpecDef.
(* psi : chspec *)
Definition chspec := CHSpec.map.

Notation YIELD := 10.
Parameter FORK : word.
Parameter EXIT : word.

(* Definition gimp (g g' : guarantee) :=  *)
(*   forall s s', g s s' -> g' s s'. *)

(* Definition pimp (p p' : stpred) := *)
(*   forall s, p s -> p s'. *)

(* Notation " A ===> B" := forall s  *)
Fixpoint wfis (psi : chspec) (p : stpred) (g : guarantee)
  (A : assume) (G : guarantee) (is : instrseq) {struct is} : Prop :=
  match is with 
    | ione (j f) =>
          (forall s, p s 
            -> exists p'', exists g'', exists A'', exists G'',
              CHSpec.get psi f = SomeT (mk_cdspec p'' g'' A'' G'')
              /\ (optp p'' (nextS (j f) s))
              /\ (forall s',
                optg g'' (nextS (j f) s) s' -> g s s')
              /\ (Aok A'')
              /\ (Gok G'')
              /\ (A ===> A'')
              /\ (G'' ===> G))
          
    | ione (jal f) =>
      match beq_nat f YIELD with
        | true =>
          (forall s, p s
            -> exists p', exists g', exists A', exists G',
              CHSpec.get psi ((pcget s) + 1) = SomeT (mk_cdspec p' g' A' G')
              /\ optp (g s) (nextS (jal YIELD) s)
              /\ optp p' (nextS (jal (pcget s + 1)) s)
              /\ (Aok A')
              /\ (Gok G')
              /\ (forall s s', p' s -> A' s s' -> p' s')
              /\ (g' ===> G')
              /\ (A ===> A')
              /\ (G' ===> G))
        | _ =>
          (forall s, p s 
            -> exists p'', exists g'', exists A'', exists G'',
              CHSpec.get psi f = SomeT (mk_cdspec p'' g'' A'' G'')
              /\ (optp p'' (nextS (jal f) s))
              /\ (forall s',
                optg g'' (nextS (jal f) s) s' -> g s s')
              /\ (Aok A'')
              /\ (Gok G'')
              /\ (A ===> A'')
              /\ (G'' ===> G))
      end
    | ione (jr r) =>
          (forall s, p s 
            -> exists p'', exists g'', exists A'', exists G'',
              CHSpec.get psi (rget s r) = SomeT (mk_cdspec p'' g'' A'' G'')
              /\ (optp p'' (nextS (jr r) s))
              /\ (forall s',
                optg g'' (nextS (jr r) s) s' -> g s s')
              /\ (Aok A'')
              /\ (Gok G'')
              /\ (A ===> A'')
              /\ (G'' ===> G))

    | ione _ => False

    | iseq (j f) is' => False
    | iseq (jal f) is' => False
    | iseq (jr r) is' => False

    | iseq (blt rs rt f) is' =>
          (forall s, p s -> 
            if blt_nat (S_gR s rs) (S_gR s rt)
              then 
                (exists p'', exists g'', exists A'', exists G'',
                  CHSpec.get psi f = SomeT (mk_cdspec p'' g'' A'' G'')
                  /\ (optp p'' (nextS (blt rs rt f) s))
                  /\ (forall s',
                    optg g'' (nextS (blt rs rt f) s) s' -> g s s')
                  /\ (Aok A'')
                  /\ (Gok G'')
                  /\ (A ===> A'')
                  /\ (G'' ===> G))
              else
                (exists p', exists g' : guarantee,
                  optp p' (nextS (blt rs rt f) s)
                  /\ (forall s', optg g' (nextS (blt rs rt f) s) s' -> g s s')
                  /\ wfis psi p' g' A G is'))

    | iseq (beq rs rt f) is' =>
          (forall s, p s -> 
            if beq_nat (S_gR s rs) (S_gR s rt)
              then 
                (exists p'', exists g'', exists A'', exists G'',
                  CHSpec.get psi f = SomeT (mk_cdspec p'' g'' A'' G'')
                  /\ (optp p'' (nextS (beq rs rt f) s))
                  /\ (forall s',
                    optg g'' (nextS (beq rs rt f) s) s' -> g s s')
                  /\ (Aok A'')
                  /\ (Gok G'')
                  /\ (A ===> A'')
                  /\ (G'' ===> G))
              else
                (exists p', exists g',
                  optp p' (nextS (beq rs rt f) s)
                  /\ (forall s', optg g' (nextS (beq rs rt f) s) s' -> g s s')
                  /\ wfis psi p' g' A G is'))

    | iseq (bgtz rs f) is' =>
          (forall s, p s -> 
            if blt_nat O (S_gR s rs)
              then 
                (exists p'', exists g'', exists A'', exists G'',
                  CHSpec.get psi f = SomeT (mk_cdspec p'' g'' A'' G'')
                  /\ (optp p'' (nextS (bgtz rs f) s))
                  /\ (forall s',
                    optg g'' (nextS (bgtz rs f) s) s' -> g s s')
                  /\ (Aok A'')
                  /\ (Gok G'')
                  /\ (A ===> A'')
                  /\ (G'' ===> G))
              else
                (exists p', exists g',
                  optp p' (nextS (bgtz rs f) s)
                  /\ (forall s', optg g' (nextS (bgtz rs f) s) s' -> g s s')
                  /\ wfis psi p' g' A G is'))
            
    | iseq i is' =>
          (forall s, p s
            -> exists p', exists g',
              optp p' (nextS i s)
              /\ (forall s',
                optg g' (nextS i s) s' -> g s s')
              /\ wfis psi p' g' A G is')
  end.

Definition wfch (psi psi' : chspec) (ch : codeheap) :=
  forall f p g A G,
    CHSpec.lookup psi' f (mk_cdspec p g A G)
    -> (exists is : _, codeheap.getis ch f = Some is
      /\ wfis psi p g A G is).

Lemma Imply_trans : forall g g' g'',
  g ===> g' 
  -> g' ===> g''
  -> g ===> g''.
Proof.
  intros g g' g'' H12 H23 s s' H1.
  apply H23.
  apply H12; trivial.
Qed.
Implicit Arguments Imply_trans [g g' g''].
(* *********************************************************** *)

Require Import ListA.

Module QueueElem <: LISTELEM.
  Definition A := word.
End QueueElem.

Module Queue := List QueueElem.

Definition queue := Queue.list.

(* *********************************************************** *)

Ltac injcdspec :=
  match goal with
    | H : CHSpec.lookup ?psi ?pc (mk_cdspec ?pi ?gi ?Ai ?Gi),
      H1 : CHSpec.lookup ?psi ?pc (mk_cdspec ?pj ?gj ?Aj ?Gj) 
      |- _ => let H := fresh "H" in
        (assert (H := CHSpec.lookup_unique H H1);
          injection H; intros; subst pj gj Aj Gj; clear H H1)
  end.

Fixpoint NIx' (Q : queue) (psi : chspec) 
  (A : assume) (G : guarantee) {struct Q} : Prop :=
  match Q with
    | Queue.nil => True
    | Queue.cons pc Q =>
      NIx' Q psi A G 
      /\ exists p', exists g', exists A', exists G',
        CHSpec.lookup psi pc (mk_cdspec p' g' A' G')
        /\ (G ===> A')
        /\ (G' ===> A)
  end.

Definition NIx (Q :queue) (psi : chspec) 
  (A : assume) (G : guarantee) : Prop :=
  forall i pc',
    Queue.get Q i = Some pc' 
    -> exists p', exists g', exists A', exists G',
      CHSpec.lookup psi pc' (mk_cdspec p' g' A' G')
      /\ (G ===> A')
      /\ (G' ===> A).

Lemma NIx'_NIx : forall Q psi A G,
  NIx' Q psi A G -> NIx Q psi A G.
Proof.
  induction Q as [ | pc' Q IHpc' ]; simpl; intros.
    unfold NIx; simpl; intros.
    destruct i; discriminate.
  destruct H as [HNIx' [p' [g' [A' [G' [Hpc' H]]]]]].
  intros i pc Hpc.
  simpl in Hpc.
  destruct i.
    injection Hpc; intro; subst pc'; clear Hpc.
    exists p'; exists g'; exists A'; exists G'.
    split_lr.
  apply ((IHpc' psi A G HNIx') i pc Hpc).
Qed.
Implicit Arguments NIx'_NIx [Q psi A G].

Lemma NIx_NIx' : forall Q psi A G,
  NIx Q psi A G -> NIx' Q psi A G.
Proof.
  induction Q as [ | pc' Q IHpc' ]; simpl; intros.
    trivial.
  split.
    apply IHpc'; trivial.
    unfold NIx in H |- * .
    intros i pc Hi.
    substH H with (H (S i) pc Hi).
    trivial.
  substH H with (H O pc' (refl_equal _)).
  trivial.
Qed.
Implicit Arguments NIx_NIx' [Q psi A G].

(* Lemma NIx_add_w_Q : forall Q f psi A G, *)
(*   NIx (Queue.add f Q) psi A G *)
(*   -> NIx Q psi A G. *)
(* Proof. *)
(*   intros Q f psi A G H. *)
(*   intros i pc Hpc. *)
(*   substH H with (H i pc). *)
(*   destruct (Queue.get_add_elim Hpc). *)
(*   assert (Hx := Queue.get_i_get_add_i f Hpc). *)
(*   apply (H i pc Hx). *)
(* Qed. *)

Lemma NIx_add_s_Q :
  forall Q psi f A G p' g' A' G',
    NIx Q psi A G
    -> CHSpec.lookup psi f (mk_cdspec p' g' A' G')
    -> G ===> A' /\ G' ===> A 
    -> NIx (Queue.add f Q) psi A G.
Proof.
  intros Q psi f A G p' g' A' G'.
  intros HNIx Hf H.
  unfold NIx.
  intros i pc' Hi.
  destruct (Queue.get_add_elim Hi) as [[Heq Hdel] | [i' Hi']].
  subst pc'.
    exists p'; exists g'; exists A'; exists G'.
    split_lr.
  destruct Hi' as [Hi' [l' [Hl' Hdel]]].
  apply (HNIx i' pc' Hi').
Qed.

Lemma NIx_del_w_Q : forall Q i Q' psi A G,
  Queue.del Q i = Some Q'
  -> NIx Q psi A G
  -> NIx Q' psi A G.
Proof.
  intros Q i Q' psi A G Hdel HNIx.
  intros i' pc' Hi'.
  destruct (Queue.del_get_get Hdel Hi') as [i'' Hi''].
  apply (HNIx i'' pc' Hi'').
Qed.

Lemma NIx_cons_w_Q : forall Q pc psi A G,
  NIx (Queue.cons pc Q) psi A G
  -> NIx Q psi A G.
Proof.
  intros Q pc psi A G H.
  substH H with (NIx_NIx' H).
  simpl in H.
  destruct H.
  apply NIx'_NIx; trivial.
Qed.

Lemma NIx_cons_s_Q : forall Q psi A G f pi gi Ai Gi,
  NIx Q psi A G
  -> CHSpec.lookup psi f (mk_cdspec pi gi Ai Gi)
  -> (G ===> Ai) /\ (Gi ===> A)
  -> NIx (Queue.cons f Q) psi A G.
Proof.
  intros Q psi A G f pi gi Ai Gi HNIx Hf HAG.
  intros i pci Hpci.
  simpl in Hpci.
  destruct i.
    inversion Hpci; subst pci; clear Hpci.
    exists pi; exists gi; exists Ai; exists Gi.
    split_lr.
  apply (HNIx i pci Hpci).
Qed.

Fixpoint NI' (Q : queue) (psi : chspec) 
  (A0 : assume) (G0 : guarantee) {struct Q}: Prop :=
  match Q with 
    | Queue.nil => True
    | Queue.cons pc Q =>
      NI' Q psi A0 G0
      /\ exists p, exists g, exists A, exists G,
        CHSpec.lookup psi pc (mk_cdspec p g A G)
        /\ NIx Q psi A G
        /\ (G0 ===> A)
        /\ (G ===> A0)
  end.

Definition NI (Q : queue) (psi : chspec) 
  (A0 : assume) (G0 : guarantee) : Prop :=
  forall i pc,
    Queue.get Q i = Some pc
    -> exists p, exists g, exists A, exists G,
      CHSpec.lookup psi pc (mk_cdspec p g A G)
      /\ (exists Q',
        Queue.del Q i = Some Q'
        /\ NIx Q' psi A G
        /\ (G0 ===> A)
        /\ (G ===> A0)).

Lemma NI'_NI : forall Q psi A G,
  NI' Q psi A G -> NI Q psi A G.
Proof.
  induction Q as [ | pc Q IHpc ]; simpl; intros.
    intros i pc Hget.
    rewrite Queue.get_nil in Hget; discriminate.
  intros i pci Hget.
  destruct i.
    simpl in Hget.
    injection Hget; intro; subst pci; clear Hget.
    destruct H as [HNI' [p' [g' [A' [G' Hpc]]]]].
    destruct Hpc as [Hpc [HNIx [HGA' HG'A]]].
    exists p'; exists g'; exists A'; exists G'.
    split_l.
    exists Q.
    simpl.
    split_l.
    split; trivial; split_lr.
  simpl in Hget.
  destruct H as [HNI' H].
(*   destruct H as [p' [g' [A' [G' [Hpc [HNIx HGAG'A']]]]]]. *)
  substH IHpc with (IHpc psi A G HNI' i pci Hget).
  destruct IHpc as [pi [gi [Ai [Gi [Hpci IHpc]]]]].
  destruct IHpc as [Q' [Hdel IHpc]].
  exists pi; exists gi; exists Ai; exists Gi; split_l.
  exists (Queue.cons pc Q').
  split.
    simpl.
    rewrite Hdel.
    trivial.
  destruct IHpc as [IHpc [HGAi HGiA]].
  split_r by (split_lr).
  apply NIx'_NIx.
  simpl.
  split.
    apply NIx_NIx'; trivial.
  destruct H as [p' [g' [A' [G' [Hpc H]]]]].
  destruct H as [HNIx H].
  exists p'; exists g'; exists A'; exists G'; split_l.
  substH HNIx with (HNIx i pci Hget).
  destruct HNIx as [p'' [g'' [A'' [G'' [Hpci' HNIx]]]]].
  assert (Hx := CHSpec.lookup_unique Hpci Hpci').
  injection Hx; intros; subst p'' g'' A'' G''.
  destruct HNIx; split_lr.
Qed.
Implicit Arguments NI'_NI [Q psi A G].

Lemma NI_NI' : forall Q psi A G,
  NI Q psi A G -> NI' Q psi A G.
Proof.
  induction Q as [ | pc Q IHpc ]; simpl; intros; trivial.
  split.
    apply IHpc.
    intros i pci Hget.
    substH H with (H (S i) pci Hget).
    destruct H as [pi [gi [Ai [Gi [Hpci H]]]]].
    exists pi; exists gi; exists Ai; exists Gi; split_l.
    destruct H as [Q' [H HNIx]].
    destruct (Queue.del_cons_del H) as [Q'' [Heq HQ'']].
    subst Q'.
    exists Q''; split_l.
    destruct HNIx as [HNIx HAG].
    split_r.
    eapply NIx_cons_w_Q; eauto.
  destruct (H O pc (refl_equal _)) as 
    [p' [g' [A' [G' [Hpc [Q' [HQ' [HNIx HAG]]]]]]]].
  exists p'; exists g'; exists A'; exists G'; split_l.
  simpl in HQ'.
  injection HQ'; intro; subst Q'; clear HQ'.
  split_lr.
Qed.
Implicit Arguments NI_NI' [Q psi A G].

Lemma NI_add_s_Q :
  forall Q psi f A G p' g' A' G',
    NI Q psi A G
    -> CHSpec.lookup psi f (mk_cdspec p' g' A' G')
    -> NIx Q psi A' G'
    -> G ===> A' /\ G' ===> A
    -> NI (Queue.add f Q) psi A G.
Proof.
  intros Q psi f A G p' g' A' G'.
  intros HNI Hf HNIx HG'A'.
  intros i pci Hpci.
  destruct (Queue.get_add_elim Hpci) as [[Heq Hdel] | [i' Hi']].
    subst pci.
    exists p'; exists g'; exists A'; exists G'; split_l.
    exists Q; split_l.
    split_lr.
  destruct Hi' as [Hpci' [Q' [HQ' Hdel]]].
  substH HNI with (HNI i' pci Hpci').
  destruct HNI as [pi [gi [Ai [Gi [Hini HNI]]]]].
  exists pi; exists gi; exists Ai; exists Gi; split_l.
  rewrite Hdel.
  destruct HNI as [Q'' [HQ'' [HNIx' HAG]]].
  rewrite HQ' in HQ''.
  injection HQ''; intro; subst Q''; clear HQ''.
  exists (Queue.add f Q').
  split_l.
  split_r.
  eapply NIx_add_s_Q; eauto.
  substH HNIx with (HNIx i' pci Hpci').
  destruct HNIx as [pj [gj [Aj [Gj [Hinj HGjAj]]]]].
  injcdspec.
  destruct HGjAj; split_lr.
Qed.

Lemma NI'_ex : 
  forall Q i psi A0 G0 f,
    NI' Q psi A0 G0
    -> Queue.get Q i = Some f
    -> exists pi, exists gi, exists Ai, exists Gi,
      CHSpec.lookup psi f (mk_cdspec pi gi Ai Gi).
Proof.
  induction Q as [ | pc Q IHpc ]; destruct i; simpl; 
    intros psi A0 G0 f.
    intros; discriminate.
    intros; discriminate.
    intros HNI Hi.
    destruct HNI as [HNI [p [g [A [G H]]]]].
    destruct H as [Hpc [H [HG0A HGA0]]].
    injection Hi; intro; subst f; clear Hi.
    exists p; exists g; exists A; exists G; trivial.
  intros HNI Hi.
  destruct HNI as [HNI [p [g [A [G H]]]]].
  destruct H as [Hpc [H [HG0A HGA0]]].
  eapply IHpc; eauto. 
Qed.

Lemma NI_G_Ai : 
  forall Q i psi A G f pi gi Ai Gi,
    NI Q psi A G
    -> Queue.get Q i = Some f
    -> CHSpec.lookup psi f (mk_cdspec pi gi Ai Gi)
    -> G ===> Ai.
Proof.
  intros Q i psi A G f pi gi Ai Gi.
  intros HNI Hget Hin.
  destruct (HNI i f Hget) as [p' [g' [A' [G' [Hpc' H]]]]].
  assert (Heq := CHSpec.lookup_unique Hin Hpc').
  injection Heq; intros; subst p' g' A' G'.
  destruct H as [Q' [HQ' [HNIx [HGAi HGiA]]]].
  trivial.
Qed.

Lemma NI_Gi_A : 
  forall Q i psi A G f pi gi Ai Gi,
    NI Q psi A G
    -> Queue.get Q i = Some f
    -> CHSpec.lookup psi f (mk_cdspec pi gi Ai Gi)
    -> Gi ===> A.
Proof.
  intros Q i psi A G f pi gi Ai Gi.
  intros HNI Hget Hin.
  destruct (HNI i f Hget) as [p' [g' [A' [G' [Hpc' H]]]]].
  assert (Heq := CHSpec.lookup_unique Hin Hpc').
  injection Heq; intros; subst p' g' A' G'.
  destruct H as [Q' [HQ' [HNIx [HGAi HGiA]]]].
  trivial.
Qed.

Lemma NI'_Gi_Ai : 
  forall Q psi A0 G0 i f pi gi Ai Gi i' f' pi' gi' Ai' Gi',
    NI' Q psi A0 G0
    -> i <> i'
    -> Queue.get Q i = Some f
    -> CHSpec.lookup psi f (mk_cdspec pi gi Ai Gi)
    -> Queue.get Q i' = Some f'
    -> CHSpec.lookup psi f' (mk_cdspec pi' gi' Ai' Gi')
    -> Gi ===> Ai'.
Proof.
  induction Q as [ | pc Q IHpc ]; simpl.
    intros psi A0 G0 i f pi gi Ai Gi i' f' pi' gi' Ai' Gi'.
    intros HNI Hii' Hi Hf Hi' Hf'.
    destruct i; try discriminate.
  intros psi A0 G0 i f pi gi Ai Gi i' f' pi' gi' Ai' Gi'.
  intros HNI Hii' Hi Hf Hi' Hf'.
  
  destruct i; destruct i'.
  (* case i=0, i'=0 *)
  destruct (Hii' (refl_equal _)).
  (* case i=0, i' <> 0 *)
  injection Hi; intros; subst; clear Hi.
  destruct HNI as [HNI [p [g [A [G [Hfx HNI']]]]]].
  destruct HNI' as [HNI' [HG0A HGA0]].
  assert (H := CHSpec.lookup_unique Hf Hfx).
  injection H; intros; subst; clear H Hfx.
  substH HNI' with (HNI' i' f' Hi').
  destruct HNI' as [p' [g' [A' [G' [Hf'x [HGA' HG'A]]]]]].
  assert (H := CHSpec.lookup_unique Hf' Hf'x).
  injection H; intros; subst; clear H Hf'x.
  trivial.
  (* case i <> 0, i' = 0 *)
  injection Hi'; intros; subst; clear Hi'.
  destruct HNI as [HNI [p [g [A [G [Hf'x HNI']]]]]].
  destruct HNI' as [HNI' [HG0A HGA0]].
  assert (H := CHSpec.lookup_unique Hf' Hf'x).
  injection H; intros; subst; clear H Hf'x.
  substH HNI' with (HNI' i f Hi).
  destruct HNI' as [p' [g' [A' [G' [Hfx [HGA' HG'A]]]]]].
  assert (H := CHSpec.lookup_unique Hf Hfx).
  injection H; intros; subst; clear H Hfx.
  trivial.
  (* case i <> 0, i' <> 0 *)
  destruct HNI as [HNI _].
  assert (Hii'x : i <> i').
  omega.
  eapply IHpc; eauto.
Qed.

Lemma NI_Gi_Ai : 
  forall Q psi A G i f pi gi Ai Gi i' f' pi' gi' Ai' Gi',
    NI Q psi A G
    -> i <> i'
    -> Queue.get Q i = Some f
    -> CHSpec.lookup psi f (mk_cdspec pi gi Ai Gi)
    -> Queue.get Q i' = Some f'
    -> CHSpec.lookup psi f' (mk_cdspec pi' gi' Ai' Gi')
    -> Gi ===> Ai'.
Proof.
  intros Q psi A G i f pi gi Ai Gi i' f' pi' gi' Ai' Gi'.
  intros H; intros.
  substH H with (NI_NI' H).
  eapply NI'_Gi_Ai; eauto.
Qed.

Lemma NI'_NIx : 
  forall Q psi A G,
  NI' Q psi A G
  -> NIx Q psi A G.
Proof.
  induction Q as [ | pc Q IHpc ]; simpl.
    intros.
    intros i pc Hx.
    simpl in Hx.
    destruct i; discriminate.
  intros psi A G [HNI H].
  destruct H as [p' [g' [A' [G' [Hpc [HNIx H]]]]]].
  eapply NIx_cons_s_Q; eauto.
Qed.

Lemma NI_NIx : forall Q psi A G,
  NI Q psi A G
  -> NIx Q psi A G.
Proof.
  intros Q psi A G HNI.
  substH HNI with (NI_NI' HNI).
  apply NI'_NIx; trivial.
Qed.

Lemma NI_weaken_A_G : 
  forall Q psi (A A' : assume) (G G' : guarantee),
    NI Q psi A G
    -> A ===> A'
    -> G' ===> G
    -> NI Q psi A' G'.
Proof.
  intros Q psi A A' G G' HNI HAA' HG'G.
  intros i pci Hpci.
  destruct (HNI i pci Hpci) as [pi [gi [Ai [Gi [Hini HNI']]]]].
  exists pi; exists gi; exists Ai; exists Gi; split_l.
  destruct HNI' as [Q' [HQ' [HNIx [HGAi HGiA]]]].
  exists Q'; split_l.
  split_l.
  split; eapply Imply_trans; eauto.
Qed.

Lemma NI_change_A_G :
  forall Q psi A0 G0 A G,
    NIx Q psi A G
    -> NI Q psi A0 G0
    -> NI Q psi A G.
Proof.
  intros Q psi A0 G0 A G.
  intros HNIx HNI.
  intros i pci Hpci.
  destruct (HNI i pci Hpci) as [pi [gi [Ai [Gi [Hini HNI']]]]].
  exists pi; exists gi; exists Ai; exists Gi; split_l.
  destruct HNI' as [Q' [HQ' [HNIx' [HGAi HGiA]]]].
  exists Q'; split_l.
  split_l.
  substH HNIx with (HNIx i pci Hpci).
  destruct HNIx as [p' [g' [A' [G' [Hin' HNIx]]]]].
  assert (Heq := CHSpec.lookup_unique Hini Hin').
  injection Heq; intros; subst p' g' A' G'; clear Heq.
  trivial.
Qed.

(*
Lemma NI_perm :
  forall Q psi f f' A G,
  NI (Queue.cons f (Queue.cons f' Q)) psi A G
  -> NI (Queue.cons f' (Queue.cons f Q)) psi A G.
Proof.
  simpl.
  intros Q psi f f' A G HNI.
  destruct HNI as [[H1 H2] H3].
  split.
    split_l.
    destruct H3 as [px [gx [Ax [Gx [Hf H3]]]]].
    exists px; exists gx; exists Ax; exists Gx.
    split_l.
    destruct H3 as [H3 H4].
    split_r.
    eapply NIx_weaken_Q; eauto.
  destruct H2 as [px [gx [Ax [Gx [Hf' H2]]]]].
  destruct H2 as [H2 H4].
  exists px; exists gx; exists Ax; exists Gx.
  split_l.
  split_r.
  destruct H3 as [p' [g' [A' [G' [Hf H3]]]]].
  destruct H3 as [H3 H5].
  unfold NIx in H3.
  substH H3 with (H3 O f' (refl_equal _)).
  destruct H3 as [p'' [g'' [A'' [G'' [Hf'x Hx]]]]].
  assert (Heq := CHSpec.lookup_unique Hf' Hf'x).
  injection Heq; intros; subst; clear Heq.
  eapply NIx_strengthen_Q; eauto.
  destruct Hx; split_lr.
Qed.
*)

Lemma NI_shift_out : 
  forall Q i Q' psi  A G pci pi gi Ai Gi,
    CHSpec.lookup psi pci (mk_cdspec pi gi Ai Gi)
    -> Queue.get Q i = Some pci
    -> Queue.del Q i = Some Q'
    -> NI Q psi A G
    -> NI Q' psi Ai Gi.
Proof.
  intros Q i Q1 psi Ax Gx pci pi gi Ai Gi.
  intros Hini Hpci Hdel HNI.
  intros i12 pci' Hget'.
  destruct (Queue.get_del_elim_2 Hpci Hdel Hget') 
    as [i2 [Hi2 [Q2 [HQ2 [i21 [Hdel' Hi21]]]]]].
  destruct (HNI i2 pci' Hi2) as [p' [g' [A' [G' [Hin' HNI']]]]].
  exists p'; exists g'; exists A'; exists G'; split_l.
  destruct HNI' as [Qx [HQx [HNIx HAxGx]]].
  rewrite HQ2 in HQx.
  injection HQx; intro; subst Qx; clear HQx.
  destruct (Queue.get_del Hget') as [Q12 HQ12].
  exists Q12; split_l.
  split.
    rewrite Hdel' in HQ12.
    eapply NIx_del_w_Q with (Q:=Q2); eauto.
  substH HNIx with (HNIx i21 pci Hi21).
  destruct HNIx as [p'' [g'' [A'' [G'' [Hin'' HAG'']]]]].
  injcdspec.
  destruct HAG''; split; trivial.
Qed.

Lemma NI_commu : 
  forall Q i Q' psi f p g A G pci pi gi Ai Gi,
    CHSpec.lookup psi f (mk_cdspec p g A G)
    -> Queue.get Q i = Some pci
    -> CHSpec.lookup psi pci (mk_cdspec pi gi Ai Gi)
    -> Queue.del Q i = Some Q'
    -> NI Q psi A G
    -> NI (Queue.add f Q') psi Ai Gi.
Proof.
  intros Q i Q' psi f p g A G pci pi gi Ai Gi.
  intros Hf Heq Hpci HQ' HNI.
  eapply NI_add_s_Q; eauto.
      eapply NI_shift_out; eauto.
    eapply NIx_del_w_Q; eauto.
    apply NI_NIx; trivial.
  assert (Hx : NIx Q psi A G).
    apply NI_NIx; trivial.
  substH Hx with (Hx i pci Heq).
  destruct Hx as [p' [g' [A' [G' [Hin' [HGA' HG'A]]]]]].
  injcdspec.
  split; trivial.
Qed.

(* *********************************************************** *)

Definition q_imply_TQ (Q : queue) (q : stpred) (psi : chspec) :=
  (forall i pci,
    Queue.get Q i = Some pci
    -> exists pi, exists gi, exists Ai, exists Gi,
      CHSpec.lookup psi pci (mk_cdspec pi gi Ai Gi)
      /\ (Aok Ai 
        /\ Gok Gi 
        /\ (forall s s', pi s -> Ai s s' -> pi s')
        /\ (gi ===> Gi))
      /\ (forall s , q s
        -> pi (rset (pcset s pci) ra pci))).

Lemma q_imply_TQ_add_w_Q :
  forall Q psi f q,
    q_imply_TQ (Queue.add f Q) q psi
    -> q_imply_TQ Q q psi.
Proof.
  intros Q psi f q Hq.
  intros i pci  Hpci.
  destruct (Queue.get_add f Hpci) as [i' Hi'].
  apply (Hq i' pci Hi').
Qed.

Lemma q_imply_TQ_add_s_Q :
  forall Q psi f q p g A G,
    q_imply_TQ Q q psi
    -> CHSpec.lookup psi f (mk_cdspec p g A G)
    -> ((Aok A
      /\ Gok G 
      /\ (forall s s', p s -> A s s' -> p s')
      /\ (g ===> G))
    /\ (forall s , q s
      -> p (rset (pcset s f) ra f)))
    -> q_imply_TQ (Queue.add f Q) q psi.
Proof.
  intros Q psi f q p g A G.
  intros HTQ Hf H.
  unfold q_imply_TQ.
  intros i pc Hi.
  simpl in Hi.
  destruct (Queue.get_add_elim Hi) as [Hx | Hx].
    destruct Hx as [Hpc Hdel].
    subst pc.
    exists p; exists g; exists A; exists G.
    split_lr.
  destruct Hx as [i' [Hpc [Q' [HQ' Hdel]]]].
  apply (HTQ i' pc Hpc).
Qed.

Lemma q_imply_TQ_cons_w_Q :
  forall Q psi f q,
    q_imply_TQ (Queue.cons f Q) q psi
    -> q_imply_TQ Q q psi.
Proof.
  induction Q as [ | pc Q IHpc]; simpl.
    intros. unfold q_imply_TQ. 
    simpl; intros. destruct i; discriminate.
  intros psi f q HTQ.
  intros i pci Hi.
  simpl in Hi.
  destruct i.
    inj_hyp Hi; subst pci.
    substH HTQ with (HTQ 1 pc (refl_equal _)).
    trivial.
  substH HTQ with (HTQ (S (S i)) pci Hi).
  trivial.
Qed.

Lemma q_imply_TQ_cons_s_Q :
  forall Q psi f q p g A G,
    q_imply_TQ Q q psi
    -> CHSpec.lookup psi f (mk_cdspec p g A G)
    -> ((Aok A
      /\ Gok G 
      /\ (forall s s', p s -> A s s' -> p s')
      /\ (g ===> G))
    /\ (forall s , q s
      -> p (rset (pcset s f) ra f)))
    -> q_imply_TQ (Queue.cons f Q) q psi.
Proof.
  intros Q psi f q p g A G.
  intros HTQ Hf H.
  unfold q_imply_TQ.
  intros i pc Hi.
  simpl in Hi.
  destruct i.
    injection Hi; intro; subst pc; clear Hi.
    exists p; exists g; exists A; exists G.
    split_lr.
  substH HTQ with (HTQ i pc Hi).
  trivial.
Qed.

Lemma q_imply_TQ_del_Q :
  forall Q i Q' psi q,
    q_imply_TQ Q q psi
    -> Queue.del Q i = Some Q'
    -> q_imply_TQ Q' q psi.
Proof.
  intros Q i Q' psi q Hq HQ'.
  intros i' pci' Hpci'.
  destruct (Queue.del_get_get HQ' Hpci') as [i'' Hi''].
  apply (Hq i'' pci' Hi'').
Qed.

(* *********************************************************** *)

Definition ccapWFTQ (Q : queue) (q : stpred) (psi : chspec)
  (A : assume) (G : guarantee) : Prop :=
  NI Q psi A G /\ q_imply_TQ Q q psi.

(*
Lemma ccapWFTQ_weaken_Q :
  forall Q0 Q q psi A G,
    Queue.sub Q0 Q
    -> ccapWFTQ Q q psi A G
    -> ccapWFTQ Q0 q psi A G.
Proof.
  intros Q0 Q q psi A G Hsub HWFTQ.
  destruct HWFTQ as [HNI HWFTQ].
  split_l by (eapply NI weaken_Q; eauto).
  intros i pci Hpci.
  destruct (Queue.sub_get Hsub Hpci) as [i' Hi'].
  apply (HWFTQ i' pci Hi').
Qed.
*)

Lemma ccapWFTQ_weaken_A_G :
  forall Q q Psi (A A' : assume) (G G' : guarantee),
    (A ===> A')
    -> (G' ===> G)
    -> ccapWFTQ Q q Psi A G
    -> ccapWFTQ Q q Psi A' G'.
Proof.
  intros Q q Psi A A' G G' HAA' HG'G.
  intros [HNI HWFTQ]. 
  split_r. 
  eapply NI_weaken_A_G; eauto.
Qed.

Lemma ccapWFTQ_strengthen_q :
  forall Q (q0 q : stpred) Psi A G,
    (forall s, q0 s -> q s)
    -> ccapWFTQ Q q Psi A G
    -> ccapWFTQ Q q0 Psi A G.
Proof.
  intros Q q0 q Psi A G Hq0q.
  intros [HNI HWFTQ].
  split_l.
  intros i pci Hpci.
  substH HWFTQ with (HWFTQ i pci Hpci).
  destruct HWFTQ as [pi [gi [Ai [Gi [Hin [HAGok Hqpi]]]]]].
  exists pi; exists gi; exists Ai; exists Gi.
  split_l.
  split_l.
  auto.
Qed.
