(**************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* CCAP over OCAP                                              *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: ccap_ocap.v,v 1.2 2006/12/28 02:28:01 ssg05 Exp $ *)

Require Import machine.
Require Import ccap.
Require Import ocap.
Require Import seplogic.
Require Import seplogic2.
Require Import ListA.

Require Import scap_ocap.

Open Local Scope seplogic_scope.
Open Local Scope tm_scope.
Open Local Scope ocap_scope.
Open Local Scope ccap_scope.

Require Import ListT.

Require Import extac.
Require Import deptac.

Module Type CCAPPARAM.

  Parameter rho_ccap : nat.
  Parameter rho_scap : nat.

  Axiom rho_ccap_neq_rho_scap : beq_nat rho_scap rho_ccap = false.

  Parameter p_ctxqueue : queue -> heap -> Prop.

End CCAPPARAM.

Module CcapOcap (Params : CCAPPARAM).

Export Params.

(* *********************************************************** *)

(* Inductive Lscap' : Type := *)
(*   | mk_Lscap' : forall (p : stpred) (g : guarantee), Lscap'. *)

Notation Lccap := ccap.cdspec.
Notation mk_Lccap := ccap.mk_cdspec.
Notation ccap_chspec := ccap.chspec.
Notation ccap_wfis := ccap.wfis.
Notation ocap_wfis := ocap.wfis.
Notation ccap_wfch := ccap.wfch.
Notation ocap_wfch := ocap.wfch.

Module CcapCHSpec := ccap.CHSpec.

(* *********************************************************** *)

Definition ccap_chspec_lift (psi : ccap_chspec) (OPsi : chspec) : Prop
:=
  (forall f p g A G,
    CcapCHSpec.lookup psi f (mk_Lccap p g A G) 
    -> OCHSpec.lookup OPsi f <* rho_ccap, (mk_Lccap p g A G) *>)
  /\ (forall f pi, 
    OCHSpec.lookup OPsi f pi
    -> exists p, exists g, exists A, exists G,
      pi = <* rho_ccap, (mk_Lccap p g A G) *>
      /\ CcapCHSpec.lookup psi f (mk_Lccap p g A G)).

Fixpoint lift_ccap_chspec (psi : ccap_chspec) : chspec :=
  match psi return chspec with 
    | ListT.nilT => ListT.nilT
    | ListT.consT (pairT f (mk_cdspec p g A G)) psi' =>
      ListT.consT (pairT f <* rho_ccap, mk_cdspec p g A G *>)
      (lift_ccap_chspec psi')
  end.

(*
Lemma lift_ccap_chspec_lift :
  forall psi : ccap_chspec,
    ccap_chspec_lift psi (lift_ccap_chspec psi).
Proof.
Admitted.
      
Lemma lookup_lift_lookup :
  forall OPsi psi f (theta : Lccap),
    CcapCHSpec.get psi f = SomeT theta
    -> ccap_chspec_lift psi OPsi
    -> OCHSpec.lookup OPsi f <* rho_ccap, theta *> .
Proof.

Admitted.
*)

Definition ccap_chspec_lift_with_s (psi : ccap_chspec) (OPsi : chspec) : Prop :=
  (forall f p g A G,
    CcapCHSpec.lookup psi f (mk_Lccap p g A G) 
    -> OCHSpec.lookup OPsi f <* rho_ccap, (mk_Lccap p g A G) *>)
  /\ (forall f pi, 
    OCHSpec.lookup OPsi f pi
    -> exists p, exists g, exists A, exists G,
      pi = <* rho_ccap, (mk_Lccap p g A G) *>
      /\ CcapCHSpec.lookup psi f (mk_Lccap p g A G)).

Definition ccap_chspec_sink (OPsi : chspec) (psi : ccap_chspec) : Prop :=
  forall f p g A G,
    (OCHSpec.lookup OPsi f <* rho_ccap, (mk_Lccap p g A G) *>
    -> CcapCHSpec.lookup psi f (mk_Lccap p g A G))
    /\ (CcapCHSpec.lookup psi f (mk_Lccap p g A G)
      -> OCHSpec.lookup OPsi f <* rho_ccap, (mk_Lccap p g A G) *>) .

Lemma lift_lookup : forall psi OPsi f p g A G,
  ccap_chspec_lift psi OPsi
  -> CcapCHSpec.lookup psi f (mk_Lccap p g A G)
  -> OCHSpec.lookup OPsi f <* rho_ccap, mk_Lccap p g A G *> .
Proof.
  intros psi OPsi f p g A G.
  unfold ccap_chspec_lift.
  intros [H1 H2] Hin.
  apply H1; trivial.
Qed.

Lemma sink_lookup : forall psi OPsi f p g A G,
  ccap_chspec_sink OPsi psi
  -> CcapCHSpec.lookup psi f (mk_Lccap p g A G)
  -> OCHSpec.lookup OPsi f <* rho_ccap, mk_Lccap p g A G *> .
Proof.
  intros psi OPsi f p g A G.
  unfold ccap_chspec_sink.
  intros H H1.
  substH H with (H f p g A G).
  destruct H.
  auto.
Qed.

Lemma lift_sub : forall psi psi' OPsi Psi f p g A G,
  ccap_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> CcapCHSpec.lookup psi' f (mk_Lccap p g A G)
  -> ccap_chspec_sink Psi psi'
  -> OCHSpec.lookup Psi f <*rho_ccap, mk_Lccap p g A G*>.
Proof.
  intros psi psi' OPsi Psi f p g A G.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f p g A G) as [Hin1 Hin2].
  auto.
Qed.

Lemma sink_sub : forall psi psi' OPsi Psi f p g A G,
  ccap_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> CcapCHSpec.lookup psi' f (mk_Lccap p g A G)
  -> ccap_chspec_sink Psi psi'
  -> OCHSpec.lookup Psi f <*rho_ccap, mk_Lccap p g A G *> .
Proof.
  intros psi psi' OPsi Psi f p g A G.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f p g A G) as [Hin1 Hin2].
  auto.
Qed.

Lemma sink_sub_2 : forall psi psi' OPsi Psi f p g A G,
  ccap_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> CcapCHSpec.lookup psi f (mk_Lccap p g A G)
  -> ccap_chspec_sink Psi psi'
  -> CcapCHSpec.lookup psi' f (mk_Lccap p g A G).
Proof.
  intros psi psi' OPsi Psi f p g A G.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f p g A G) as [Hin1 Hin2].
  destruct Hlift.
  auto.
Qed.


(** *********************************************************** *)

Definition itpccap (theta : Lccap) : assert :=
  match theta with
    | mk_cdspec p g A G =>
      fun (Psi : chspec) (s : state) =>
        match s with
          | mkst H R pc =>
              exists Q : queue, exists H1, exists H2,
                Heap.join H1 H2 H
                /\ p (mkst H1 R pc)
                /\ H2 ||- p_ctxqueue Q
                  /\ (exists psi, ccap_chspec_sink Psi psi
                    /\ ccapWFTQ Q (g (mkst H1 R pc)) psi A G)
        end
  end.

(* *********************************************************** *)

Definition p_yield (s : state) := 
  exists Q, s |||- top ** p_ctxqueue Q .

Definition g_yield (s s' : state) := 
  forall (Q : queue) A, 
    (s |||- A ** p_ctxqueue Q)
    -> exists i, exists pcx,
      Queue.get (Queue.add (rget s ra) Q) i = Some pcx
      /\ exists Q',
        (Queue.del (Queue.add (rget s ra) Q) i) = Some Q'
        /\ pcget s' = pcx
        /\ S_R (rset s ra pcx) = S_R s'
        /\ (s' |||- A ** p_ctxqueue Q').

Definition D_ccap0 := fun id =>
    if (beq_nat id rho_ccap) then SomeT <@ itpccap @> else
      NoneT.

Module ScapParam <: SCAPPARAM.

  Definition rho_scap := rho_scap.
  Definition D0 : idict := D_ccap0.

  Lemma rho_scap_not_indom_D0 : ~ DSpec.indom D0 rho_scap.
  Proof.
    unfold DSpec.indom, DSpec.get, DSpecDef.get, D0, D_ccap0; 
      intros [b H].
    unfold rho_scap in H.
    rewrite rho_ccap_neq_rho_scap in H.
    discriminate.
  Qed.

End ScapParam.

Module ScapOcap := ScapOcap ScapParam.

Export ScapOcap.

(* *********************************************************** *)

Definition OPsis : chspec := 
  OCHSpec.sig_map YIELD (<* rho_scap, mk_Lscap p_yield g_yield *>).

Definition D_ccap : idict := DSpec.merge D_ccap0 D_scap.

(* Definition D_ccap : idict := fun id => *)
(*   if (beq_nat id rho_ccap) then SomeT <@ itpccap @> else  *)
(*   if (beq_nat id rho_scap) then SomeT <@ itpscap @> else  *)
(*     NoneT. *)

(* *********************************************************** *)

Lemma ccap_yield : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee),
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> ccap_wfis psi p g A G (ione (jal YIELD))
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} 
    / (OCHSpec.merge OPsi OPsis) >>
    (ione (jal YIELD)).
Proof.
  intros D OPsi psi p g A G.
  unfold ccap_wfis; fold ccap_wfis.
  simplbnat.
  unfold wfis; fold wfis.
  intros HD HDscap Hpsi Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  simplst.
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [A' [G' Hwfis]]]].
  destruct Hwfis as [Hspc [Hg [Hp' Hwfis]]]. 
  destruct Hwfis as [HA' [HG' [Hp'A'p' [Hg'G' [HAA'' HG'G]]]]].
  exists <* rho_scap, mk_Lscap p_yield g_yield *> .
  split.
    eapply OCHSpec.lookup_sub_merge_r_lookup; eauto.
    unfold OPsis, OCHSpec.sig_map.
    OCHSpec.lookup_solve.
  unfold nextS.
  exists itpscap.
  split_l.
(* 1. proving (p, g, A, G) => (p_yield, g_yield) *)
  unfold itpscap, proj2_xcdspec.
  unfold p_yield, g_yield.
  split.
(* 1%1 proving p => p_yield *)
    assert (Hp_yield : H ||- top ** p_ctxqueue Q).
      sep_split into H1 H2; sep_trivial.
    exists Q; trivial.
(* 1%2. proving exists p', g_yield => p' *)
  (* after yield routine, the `state' (s' = <.H', R', pc'.>) 
    of the tm still satisfy the assertion. We can prove it 
    by the g of yield.  *)
  exists O; simpl.
  intros [H' R' pc'] Hg_yield.
  substH Hg_yield with (Hg_yield Q).
  assert (Hh : H ||- (fun h => h=H1) ** p_ctxqueue Q).
    apply sconj_intro with (ha:=H1) (hb:=H2); trivial.
    unfold hpred_true; trivial.
  substH Hg_yield with (Hg_yield (fun h => h = H1) Hh); clear Hh.
  destruct Hg_yield as [i [pci [Hpci [Q' [HQ' Hg_yield]]]]]. 
  destruct Hg_yield as [Hpc' [HReq Hgs]].
  subst R'.
  simplst.
  subst pc'.
  destruct HWFTQ as [HNI HWFTQ].
  (* after yield routine, there are two cases of tm. *)
  destruct (Queue.get_add_elim Hpci) as [Hi | Hi].
(* 1 % 2 % 1, case 1: the scheduler resume the former thread. *)
    destruct Hi as [Hx Hy].
    (* so pc' = pc + 1 *)
    subst pci.
    (* so the thread queue is unchanged *)
    rewrite HQ' in Hy; injection Hy; intro; subst Q'; clear Hy.
    (* so the current assertion is [|p', g', A', G'|] *)
    exists (<* rho_ccap, mk_Lccap p' g' A' G'*>).
    split.
      eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
      apply lift_lookup with psi; trivial.
    (* We will prove that the state satisfy the assertion *)
    unfold BigInterp; exists itpccap.
    split.
      unfold ScapParam.D0, D_ccap0.
      unfold DSpec.lookup, DSpec.get, DSpecDef.get.
      simplbnat; trivial.
    unfold itpccap.
    (* looking for H1' H2' *)
    destruct Hgs as [H1' [H2' [HJ' [HH1' HH2']]]].
    unfold hpred_true in HH1'; subst H1'.
    (* H1 is the heap of CCAP, so H1 is unchanged. H2 is described 
      by g_yield *)
    exists Q; exists H1; exists H2'.
    split_l.
    split_l.
    split_l.
    (* the TQ must be well-formed !! *)
    exists psi'; split_l.
    unfold ccapWFTQ.
    (* NI *)
    split.
      eapply NI_weaken_A_G; eauto.
    (* forall i, Q[i] is well-formed *)
    clear i Hpci HQ'.
    intros i pci Hpci.
    substH HWFTQ with (HWFTQ i pci Hpci).
    destruct HWFTQ as [pi [gi [Ai [Gi HWFTQ]]]].
    destruct HWFTQ as [Hpciinpsi [HWFTQ Hgpi]].
    destruct HWFTQ as [HAi [HGi [HpiAipi HgiGi]]].
    substH Hgpi with (Hgpi (rset (mkst H1 R YIELD) ra (pc+1))).
    simpl in Hgpi.
    simplst.
    substH Hgpi with (Hgpi Hg).
    exists pi; exists gi; exists Ai; exists Gi.
    split_l.
    split.
      split_l.
      split_l.
      split_lr.
    (* to prove g' s sx -> pi sx *)
    intros [Hx Rx pcx] Hg'.
    (* from pi = Ai => pi and (pi s) need prove (Ai s sx) *)
    apply (HpiAipi (mkst H1 (set R ra pci) pci)
      (mkst Hx (set Rx ra pci) pci)); trivial.
    (* from NI[...], we can have G => A_i *)
    assert (HGAi : G ===> Ai).
      apply (NI_G_Ai Q i psi' A G pci pi gi Ai Gi);
        trivial.
    (* need prove (G s sx) *)
    apply HGAi; clear HGAi.
    (* from G' => G, need prove (G' s sx) *)
    apply HG'G.
    (* because G' is ok and g' => G', the goal is complete *)
    assert (HG'G' := (HG' (mkst H1 (set R ra (pc + 1)) (pc+1))
      (mkst Hx Rx pcx) pci pci pci pci)).
    simplst.
    apply (proj1 HG'G').
    apply Hg'G'; trivial.
    (* the proof of 1st case is complete.*)
(* 1 % 2 % 2: the scheduler transfer control to others. *)
  destruct Hi as [i' [Hi' [Q'' [HQ'' Hdel]]]].
  rewrite HQ' in Hdel.
  injection Hdel; intro Heq; clear Hdel.
  generalize HWFTQ; intro HWFTQ'.
  substH HWFTQ with (HWFTQ i' pci Hi').
  destruct HWFTQ as [pi [gi [Ai [Gi HWFTQ]]]].
  destruct HWFTQ as [Hpciinpsi [HWFTQ Hgpi]].
  destruct HWFTQ as [HAi [HGi [HpiAipi HgiGi]]].
  substH Hgpi with (Hgpi (rset (mkst H1 R YIELD) ra (pc+1))).
  simpl in Hgpi.
  substH Hgpi with (Hgpi Hg).
  (* so the current assertion is [| pi, gi, Ai, Gi|] *)
  exists (<* rho_ccap, mk_Lccap pi gi Ai Gi *>).
  split.
    eapply sink_sub; eauto.
    eapply OCHSpec.sub_merge_elim_l; eauto.
  (* We will prove that the state satisfy the assertion *)
    unfold BigInterp; exists itpccap.
    split.
      unfold ScapParam.D0, D_ccap0.
      unfold DSpec.lookup, DSpec.get, DSpecDef.get.
      simplbnat; trivial.
  unfold itpccap.
  (* looking for H1' H2' *)
  destruct Hgs as [H1' [H2' [HJ' [HH1' HH2']]]].
  unfold hpred_true in HH1'; subst H1'.
  simplst.
  exists Q'; exists H1; exists H2'.
  split_l.
  split_l.
  split_l.
  (* the TQ must be well-formed !! *)
(*   apply ccapWFTQ_weaken_Q with (Queue.cons (pc + 1) Q). *)
(*     apply Queue.del_sub with (S i); trivial. *)
  exists psi'; split_l.
  unfold ccapWFTQ.
  split.
    rewrite Heq.
    apply (NI_commu Q i' Q'' psi' (pc+1) p' g' A' G'
      pci pi gi Ai Gi); eauto.
    eapply sink_sub_2; eauto.
    eapply OCHSpec.sub_merge_elim_l; eauto.
    eapply NI_weaken_A_G; eauto.
  (* forall i', Q[i'] is well-formed *)
  intros i0' pci0' Hpci0'.
  subst Q'.
  simpl in Hpci0'.
  destruct (Queue.get_add_elim Hpci0') as [Hi0' | Hi0'].
    (* 1st case : Q[0], pc + 1, should be well-formed *)
    destruct Hi0' as [Hx Hy].
    (* so pc' = pc + 1 *)
    subst pci0'.
    (* so the thread queue is unchanged *)
(*     rewrite HQ' in Hy; injection Hy; intro; subst Q'; clear Hy. *)
    exists p'; exists g'; exists A'; exists G'.
    split.
      eapply sink_sub_2; eauto.
      eapply OCHSpec.sub_merge_elim_l; eauto.
    split.
      split_l.
      split_l.
      split_lr.
    (* to prove g' s sx -> pi sx *)
    intros [Hx Rx pcx] Hgi.
    simplst.
    apply (Hp'A'p' (mkst H1 (set R ra (pc + 1)) (pc + 1))
      (mkst Hx (set Rx ra (pc + 1)) (pc + 1)) Hp').
    assert (HGiA' : Gi ===> A').
      apply Imply_trans with A; trivial.
      apply (NI_Gi_A Q i' psi' A G pci pi gi Ai Gi);
        trivial.
    apply HGiA'; clear HGiA'.
    assert (HGiGi := (HGi (mkst H1 (set R ra pci) pci)
      (mkst Hx Rx pcx) (pc+1) (pc+1) (pc+1) (pc+1))).
    simplst.
    apply (proj1 HGiGi).
    apply HgiGi; trivial.
  (* 2nd case : Q[i'], pci', should be well-formed *)
  destruct Hi0' as [i1' [Hi1' [Q1' [HQ1' Hdel']]]].
  simpl in Hpci0'.
  substH HWFTQ' with (q_imply_TQ_del_Q Q i' Q'' psi' 
    (g <. H1, R, pc .>) HWFTQ' HQ'').
  substH HWFTQ' with (HWFTQ' i1' pci0' Hi1').
  destruct HWFTQ' as [pi' [gi' [Ai' [Gi' HWFTQ']]]].
  destruct HWFTQ' as [Hpciinpsi' [HWFTQ' Hgpi']].
  destruct HWFTQ' as [HAi' [HGi' [Hpi'Ai'pi' Hgi'Gi']]].
  substH Hgpi' with (Hgpi' (rset (mkst H1 R YIELD) ra (pc+1))).
  simpl in Hgpi'.
  substH Hgpi' with (Hgpi' Hg).
  exists pi'; exists gi'; exists Ai'; exists Gi'.
  split_l.
  split.
    split_l.
    split_l.
    split_lr.
  (* to prove gi s sx -> pi' sx *)
  intros [Hx Rx pcx] Hgi'.
  simplst.
  apply (Hpi'Ai'pi' (mkst H1 (set R ra pci0') pci0')
    (mkst Hx (set Rx ra pci0') pci0') Hgpi').
  destruct (Queue.del_get_get_neq HQ'' Hi1')
    as [ix [Hix Hneq]].
  assert (HGiAi' : Gi ===> Ai').
    apply (NI_Gi_Ai Q psi' A G i' pci pi gi Ai Gi
       ix pci0' pi' gi' Ai' Gi');
      trivial.
  apply HGiAi'; clear HGiAi'.
  assert (HGiGi := (HGi (mkst H1 (set R ra pci) pci)
    (mkst Hx Rx pcx) pci0' pci0' pci0' pci0')).
  simplst.
  apply (proj1 HGiGi).
  apply HgiGi; trivial.
Qed.

Lemma ccap_j : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) (f : label),
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
(*   -> beq_nat f YIELD = false *)
  -> ccap_wfis psi p g A G (ione (j f))
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} 
  / (OCHSpec.merge OPsi OPsis) >>
  (ione (j f)).
Proof.
  intros D OPsi psi p g A G f.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros HD HDscap Hpsi Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p'' [g'' [A'' [G'' Hwfis]]]].
  destruct Hwfis as [Hfin [Hp'' [Hg'' [HA''ok [HG''ok [HA'' HG'']]]]]].
  exists <* rho_ccap, mk_Lccap p'' g'' A'' G'' *> .
  split.
    eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
    apply lift_lookup with psi; trivial.
  unfold nextS.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_weaken_A_G with A G; eauto.
  eapply ccapWFTQ_strengthen_q; eauto.
Qed.

Lemma ccap_jal : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) (f : label),
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> beq_nat f YIELD = false
  -> ccap_wfis psi p g A G (ione (jal f))
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} 
  / (OCHSpec.merge OPsi OPsis) >>
  (ione (jal f)).
Proof.
  intros D OPsi psi p g A G f.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros HD HDscap Hpsi Hf Hwfis.
  simplbnat.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p'' [g'' [A'' [G'' Hwfis]]]].
  destruct Hwfis as [Hfin [Hp'' [Hg'' [HAok'' [HGok'' [HA'' HG'']]]]]].
  exists <* rho_ccap, mk_Lccap p'' g'' A'' G'' *> .
  split.
    eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
    apply lift_lookup with psi; trivial.
  unfold nextS.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  eapply ccapWFTQ_weaken_A_G; eauto.
  eapply ccapWFTQ_strengthen_q; eauto.
Qed.

Lemma ccap_jr : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) (r : reg),
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> ccap_wfis psi p g A G (ione (jr r))
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} 
  / (OCHSpec.merge OPsi OPsis) >>
  (ione (jr r)).
Proof.
  intros D OPsi psi p g A G r.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros HD HDscap Hpsi Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p'' [g'' [A'' [G'' Hwfis]]]].
  destruct Hwfis as [Hrin [Hp'' [Hg'' [HA''ok [HG''ok [HA'' HG'']]]]]].
  exists <* rho_ccap, mk_Lccap p'' g'' A'' G'' *> .
  split.
    eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
    apply lift_lookup with psi; trivial.
  unfold nextS.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_weaken_A_G with A G; eauto.
  eapply ccapWFTQ_strengthen_q; eauto.
Qed.

Lemma ccap_addu : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd rs rt is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (addu rd rs rt) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (addu rd rs rt) is).
Proof.
  intros D OPsi psi p g A G rd rs rt is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_addiu : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd rs w is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} 
    / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (addiu rd rs w) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (addiu rd rs w) is).
Proof.
  intros D OPsi psi p g A G rd rs rt is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_subu : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd rs rt is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (subu rd rs rt) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (subu rd rs rt) is).
Proof.
  intros D OPsi psi p g A G rd rs rt is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_subiu : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd rs w is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (subiu rd rs w) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (subiu rd rs w) is).
Proof.
  intros D OPsi psi p g A G rd rs rt is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_move : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd rs is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (move rd rs) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (move rd rs) is).
Proof.
  intros D OPsi psi p g A G rd rs is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_li : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd w is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (li rd w) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (li rd w) is).
Proof.
  intros D OPsi psi p g A G rd w is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_lw : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd w rs is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (lw rd w rs) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (lw rd w rs) is).
Proof.
  intros D OPsi psi p g A G rd w rs is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  assert (Hx : exists v, Heap.get H1 (Regfile.get R rs + w) = Some v).
    destruct (Heap.get H1 (Regfile.get R rs + w)) as [ v | ];
      try contradiction.
    exists v; trivial.
  destruct Hx as [v Hv].
  rewrite Hv in Hp'.
  rewrite Hv in Hg'.
  rewrite (Heap.join_get_get_l HJ Hv).
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_sw : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rd w rs is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (sw rd w rs) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (sw rd w rs) is).
Proof.
  intros D OPsi psi p g A G rd w rs is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis.
  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  assert (Hx : exists v, Heap.get H1 (Regfile.get R rs + w) = Some v).
    destruct (Heap.get H1 (Regfile.get R rs + w)) as [ v | ];
      try contradiction.
    exists v; trivial.
  destruct Hx as [v Hv].
  rewrite Hv in Hp'.
  rewrite Hv in Hg'.
  rewrite (Heap.join_get_get_l HJ Hv).
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q.
  exists (Heap.set H1 (Regfile.get R rs + w) (Regfile.get R rd)). 
  exists H2.
  split.
    apply Heap.join_set_l; trivial.
    exists v; trivial.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_blt : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rs rt f is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (blt rs rt f) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (blt rs rt f) is).
Proof.
  intros D OPsi psi p g A G rs rt f is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis |- * .
  
  decbltnat (gR R rs) (gR R rt).

  destruct Hwfis as [p'' [g'' [A'' [G'' Hwfis]]]].
  destruct Hwfis as [Hfin [Hp'' [Hg'' [HA''ok [HG''ok [HA'' HG'']]]]]].
  exists <* rho_ccap, mk_Lccap p'' g'' A'' G'' *> .
  split.
    eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
    apply lift_lookup with psi; trivial.
  unfold nextS.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_weaken_A_G with A G; eauto.
  eapply ccapWFTQ_strengthen_q; eauto.

  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_beq : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rs rt f is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (beq rs rt f) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (beq rs rt f) is).
Proof.
  intros D OPsi psi p g A G rs rt f is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis |- * .
  
  decbeqnat (gR R rs) (gR R rt).

  destruct Hwfis as [p'' [g'' [A'' [G'' Hwfis]]]].
  destruct Hwfis as [Hfin [Hp'' [Hg'' [HA''ok [HG''ok [HA'' HG'']]]]]].
  exists <* rho_ccap, mk_Lccap p'' g'' A'' G'' *> .
  split.
    eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
    apply lift_lookup with psi; trivial.
  unfold nextS.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_weaken_A_G with A G; eauto.
  eapply ccapWFTQ_strengthen_q; eauto.

  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_bgtz : forall (D : idict) (OPsi : chspec) (psi : ccap_chspec)
  (p : stpred) (g : guarantee) (A : assume)
  (G : guarantee) rs f is,
  DSpec.lookup D rho_ccap <@itpccap@>
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> ccap_chspec_lift psi OPsi
  -> (forall p g A G, 
    ccap_wfis psi p g A G is
    -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]} / (OCHSpec.merge OPsi OPsis) >> is)
  -> ccap_wfis psi p g A G (iseq (bgtz rs f) is)
  -> wfis D << {[ <* rho_ccap, mk_cdspec p g A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >>
    (iseq (bgtz rs f) is).
Proof.
  intros D OPsi psi p g A G rs f is.
  intros HD HDscap Hpsi IHwfis.
  unfold ccap_wfis; fold ccap_wfis.
  unfold wfis; fold wfis.
  intros Hwfis.
  unfold lift, BigInterp.
  intros Psi s [Hsub [itp [HD' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear HD' Hitpeq.
  unfold itpccap, proj2_xcdspec in Hitp.
  destruct s as [H R pc].
  destruct Hitp as [Q [H1 [H2 [HJ [Hp [Hq HWFTQ]]]]]].
  destruct HWFTQ as [psi' [Hsink HWFTQ]].
  substH Hwfis with (Hwfis (mkst H1 R pc) Hp).
  unfold nextS in Hwfis |- * .
  
  decbltnat O (gR R rs).

  destruct Hwfis as [p'' [g'' [A'' [G'' Hwfis]]]].
  destruct Hwfis as [Hfin [Hp'' [Hg'' [HA''ok [HG''ok [HA'' HG'']]]]]].
  exists <* rho_ccap, mk_Lccap p'' g'' A'' G'' *> .
  split.
    eapply OCHSpec.lookup_sub_merge_l_lookup; eauto.
    apply lift_lookup with psi; trivial.
  unfold nextS.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_weaken_A_G with A G; eauto.
  eapply ccapWFTQ_strengthen_q; eauto.

  destruct Hwfis as [p' [g' [Hp' [Hg' Hwfis]]]].
  substH IHwfis with (IHwfis p' g' A G Hwfis).
  exists << {[ <* rho_ccap, mk_cdspec p' g' A G *> | D ]}
    / (OCHSpec.merge OPsi OPsis) >> .
  unfold nextS in * .
  split_r.
  unfold lift, BigInterp.
  split_l.
  exists itpccap.
  split_l.
  unfold itpccap, proj2_xcdspec.
  exists Q; exists H1; exists H2.
  split_l.
  split_l.
  split_l.
  exists psi'; split_l.
  apply ccapWFTQ_strengthen_q with (g <. H1, R, pc .>);
    trivial.
Qed.

Lemma ccap_wfis_ocap_wfis :
  forall D OPsi psi is p g A G,
    DSpec.lookup D rho_ccap <@itpccap@>
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> ccap_chspec_lift psi OPsi
    -> ccap_wfis psi p g A G is
    -> ocap_wfis D 
    << {[ <* rho_ccap, (mk_Lccap p g A G) *> | D]} 
    / (OCHSpec.merge OPsi OPsis)>> is.
Proof.
  intros D OPsi psi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f | r ];
    unfold ccap_wfis; fold ccap_wfis;
(*     unfold ocap_wfis; fold ocap_wfis; *)
    intros p g A G Hd Hd' Hlift Hccap;
      try contradiction.

  eapply ccap_j; eauto.
    destruct (beq_nat_dec f YIELD).
    substbnat f.
    eapply ccap_yield; eauto.
  eapply ccap_jal; eauto.
  eapply ccap_jr; eauto.

  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f | r ];
    unfold ccap_wfis; fold ccap_wfis;
    intros p g A G Hd Hd' Hlift Hscap;
      try contradiction.
  
  apply ccap_addu with psi; auto.
  apply ccap_addiu with psi; auto.
  apply ccap_subu with psi; auto.
  apply ccap_subiu with psi; auto.
  apply ccap_move with psi; auto.
  apply ccap_li with psi; auto.
  apply ccap_lw with psi; auto.
  apply ccap_sw with psi; auto.
  apply ccap_blt with psi; auto.
  apply ccap_beq with psi; auto.
  apply ccap_bgtz with psi; auto.
Qed.

Lemma D_ccap_describe_lift_psi :
  forall D OPsi psi,
    DSpec.lookup D rho_ccap <@itpccap@>
    -> ccap_chspec_lift psi OPsi
    -> describe D OPsi.
Proof.
  intros D OPsi psi Hd.
  unfold ccap_chspec_lift.
  intros H f rho lty theta Hin.
  destruct H as [H1 H2].
  destruct (H2 f <*rho,theta*> Hin) as [p [g [A [G [Hpi Hin']]]]].
  injection Hpi.
  intros Heq1 Heq2 Heq3.
  subst.
  exists itpccap; trivial.
Qed.

Lemma D_ccap_describe_lift_psi_with_s :
  forall D OPsi psi,
    DSpec.lookup D rho_ccap <@itpccap@>
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> ccap_chspec_lift psi OPsi
    -> describe D (OCHSpec.merge OPsi OPsis).
Proof.
  intros D OPsi psi Hd Hd'.
  unfold ccap_chspec_lift.
  intros H f rho lty theta Hin.
  destruct H as [H1 H2].
  destruct (OCHSpec.lookup_merge_elim Hin) as [Hf | Hf].
    destruct (H2 f <*rho,theta*> Hf) as [p [g [A [G [Hpi Hin']]]]].
    injection Hpi.
    intros Heq1 Heq2 Heq3.
    subst.
    exists itpccap; trivial.
  unfold OPsis in Hf.
  decbeqnat f (YIELD).
    substbnat f.
    unfold OCHSpec.lookup, In in Hf.
    destruct Hf; try contradiction.
    injection H; intros; subst.
  exists itpscap; trivial.
  unfold OCHSpec.lookup, In in Hf.
  destruct Hf; try contradiction.
  injection H; intros; subst.
  simplbnat.
Qed.

Lemma ccap_wfch_ocap_wfch :
  forall D ch psi psi' OPsi OPsi',
    DSpec.lookup D rho_ccap <@itpccap@>
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> ccap_chspec_lift psi OPsi
    -> ccap_chspec_lift psi' OPsi'
    -> ccap_wfch psi psi' ch
    -> ocap_wfch D (OCHSpec.merge OPsi OPsis) OPsi' ch.
Proof.
  intros D ch psi psi' OPsi OPsi'.
  intros Hd Hd' Hpsi Hpsi' Hch.
  unfold wfch.
  split.
    split.
      eapply D_ccap_describe_lift_psi_with_s; eauto.
      eapply D_ccap_describe_lift_psi; eauto.
  intros f pi Hin.
  unfold ccap_chspec_lift in Hpsi'.
  unfold wfch in Hch.
  destruct Hpsi' as [_ Hpsi'].
  destruct (Hpsi' f pi Hin) as [p [g [A [G [Hpi Hin']]]]].
  subst pi.
  
  destruct (Hch f p g A G (CHSpec.lookup_get Hin')) as [is [Hgis Hwfis]].
  exists is; split_l.
  apply ccap_wfis_ocap_wfis with psi; trivial.
Qed.

Theorem ccap_sound :
  forall D ch psi psi' OPsi OPsi',
    DSpec.lookup D rho_ccap <@itpccap@>
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> ccap_chspec_lift psi OPsi
    -> ccap_chspec_lift psi' OPsi'
    -> ccap_wfch psi psi' ch
    -> ocap_wfch D (OCHSpec.merge OPsi OPsis) OPsi' ch.
Proof ccap_wfch_ocap_wfch.

End CcapOcap.
