(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Link CCAP with SCAP                                         *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: link.v,v 1.1.1.1 2006/12/07 16:43:25 ssg05 Exp $ *)

Require Import machine.
Require Import schedule.
Require Import ccap_ocap.
Require Import scap.
Require Import scap_ocap.
Require Import scap_ck.
Require Import ocap.

Module CcapParam <: CCAPPARAM.
  
  Definition rho_ccap := 1.

  Definition rho_scap := 0.

  Lemma rho_ccap_neq_rho_scap : 
    beq_nat rho_scap rho_ccap = false.
  Proof.
    simpl; trivial.
  Qed.

  Definition p_ctxqueue := schedule.p_ctxqueue.

End CcapParam.

Module CcapOcap := CcapOcap CcapParam.

Import CcapOcap.

Lemma D_ccap0_disj_D_scap : DSpec.disj D_ccap0 D_scap.
Proof.
  unfold DSpec.disj, DSpec.get, DSpecDef.get.
  unfold D_ccap0, D_scap.
  intros id.
  destruct id.
  simpl in * .
  trivial.
  simpl in * ; trivial.
  destruct id.
  simpl in * ; trivial.
  rewbnat (beq_nat (S id) O = false).
  trivial.
Qed.

Lemma rho_scap_in_D_ccap : 
  DSpec.lookup D_ccap rho_scap (xinterp_pack itpscap).
Proof.
  unfold D_ccap.
  apply DSpec.lookup_merge_intro.
  right.
  unfold D_scap, DSpec.lookup, DSpec.get, DSpecDef.get.
  simplbnat.
  trivial.
  apply DSpec.disj_compat.
  apply D_ccap0_disj_D_scap.
Qed.

Definition OPsis : chspec :=
  (pairT SCHEDULE <* rho_scap, mk_Lscap schedule_pre schedule_guar *>) ::
  (pairT READPTR <* rho_scap, mk_Lscap readptr_pre readptr_guar *>) ::
  (pairT ENQUEUE <* rho_scap, mk_Lscap enqueue_pre enqueue_guar *>) ::
  (pairT DEQUEUE <* rho_scap, mk_Lscap dequeue_pre dequeue_guar *>) ::
  (pairT SWITCH<* rho_scap, mk_Lscap switch_pre switch_guar *>) ::
  (pairT WRITEPTR <* rho_scap, mk_Lscap writeptr_pre writeptr_guar *>) ::
  (pairT SRET <* rho_scap, mk_Lscap sret_pre sret_guar *>) ::
    nilT.

Ltac opsi_lookup_solve Psi :=
  unfold Psi, OCHSpec.lookup; simpl; tauto.

Lemma OPsis_lookup_SCHEDULE: OCHSpec.lookup OPsis SCHEDULE  <* rho_scap,  mk_Lscap schedule_pre schedule_guar*>.
Proof. opsi_lookup_solve OPsis. Qed.

Lemma OPsis_lookup_READPTR : OCHSpec.lookup OPsis READPTR  <* rho_scap, mk_Lscap readptr_pre readptr_guar*>.
Proof. opsi_lookup_solve OPsis. Qed.

Lemma OPsis_lookup_ENQUEUE : OCHSpec.lookup OPsis ENQUEUE <* rho_scap, mk_Lscap enqueue_pre enqueue_guar *>.
Proof. opsi_lookup_solve OPsis. Qed.

Lemma OPsis_lookup_DEQUEUE : OCHSpec.lookup OPsis DEQUEUE <* rho_scap, mk_Lscap dequeue_pre dequeue_guar *>.
Proof. opsi_lookup_solve OPsis. Qed.

Lemma OPsis_lookup_SWITCH : OCHSpec.lookup OPsis SWITCH <* rho_scap, mk_Lscap switch_pre switch_guar *>.
Proof. opsi_lookup_solve OPsis. Qed.

Lemma OPsis_lookup_WRITEPTR : OCHSpec.lookup OPsis WRITEPTR <* rho_scap, mk_Lscap writeptr_pre writeptr_guar *>.
Proof. opsi_lookup_solve OPsis. Qed.

Lemma OPsis_lookup_SRET : OCHSpec.lookup OPsis SRET <* rho_scap, mk_Lscap sret_pre sret_guar *>.
Proof. opsi_lookup_solve OPsis. Qed.

Require Import scap_2.

Lemma schedule_psi_lift_OPsis : scap_chspec_lift schedule.psi OPsis.
Proof.
  unfold scap_chspec_lift.
  change ScapCHSpec.lookup with CHSpec.lookup.
  split.

  intros.
  psi_lookup (SCHEDULE).
  apply OPsis_lookup_SCHEDULE.
  psi_lookup (READPTR).
  apply OPsis_lookup_READPTR.
  psi_lookup (ENQUEUE).
  apply OPsis_lookup_ENQUEUE.
  psi_lookup (DEQUEUE).
  apply OPsis_lookup_DEQUEUE.
  psi_lookup (SWITCH).
  apply OPsis_lookup_SWITCH.
  psi_lookup (WRITEPTR).
  apply OPsis_lookup_WRITEPTR.
  psi_lookup (SRET).
  apply OPsis_lookup_SRET.
  psi_lookup_not.
Require Import ocap_2.
Require Import extac.
  intros f pi Hpi.
  opsi_lookup (SCHEDULE).
  exists schedule.schedule_pre; exists schedule.schedule_guar.
  split_l; compute; trivial.
  opsi_lookup (READPTR).
  exists schedule.readptr_pre; exists schedule.readptr_guar.
  split_l; compute; trivial.
  opsi_lookup (ENQUEUE).
  exists schedule.enqueue_pre; exists schedule.enqueue_guar.
  split_l; compute; trivial.
  opsi_lookup (DEQUEUE).
  exists schedule.dequeue_pre; exists schedule.dequeue_guar.
  split_l; compute; trivial.
  opsi_lookup (SWITCH).
  exists schedule.switch_pre; exists schedule.switch_guar.
  split_l; compute; trivial.
  opsi_lookup (WRITEPTR).
  exists schedule.writeptr_pre; exists schedule.writeptr_guar.
  split_l; compute; trivial.
  opsi_lookup (SRET).
  exists schedule.sret_pre; exists schedule.sret_guar.
  split_l; compute; trivial.
  opsi_lookup_not.
Qed.

Definition chs := schedule.C.

Lemma schedule_wfch : wfch D_ccap OPsis OPsis chs.
Proof.
  assert (wfch (DSpec.merge D_ccap0 D_scap) OPsis OPsis chs).
  apply scap_wfch_ocap_wfch with schedule.psi schedule.psi.
  apply schedule_psi_lift_OPsis.
  apply schedule_psi_lift_OPsis.
  apply schedule.C_ok.
  trivial.
Qed.

Lemma ocap_wfis_schedule : wfis D_ccap (lift OPsis
  (BigInterp D_ccap <* rho_scap, mk_Lscap schedule_pre schedule_guar *>))
  schedule_code.
Proof.
  apply wfis_weaken_lift_Psi with OPsis.
  apply scap_wfis_ocap_wfis with schedule.psi.
  unfold ScapParam.D0, D_ccap.
  apply DSpec.sub_merge_l.
  apply DSpec.disj_compat.
  apply D_ccap0_disj_D_scap.
  apply rho_scap_in_D_ccap.
  apply schedule_psi_lift_OPsis.
  apply schedule_ok.
  unfold OCHSpec.sub. trivial.
Qed.

Require Import deptac.

Lemma wfch_chs : 
  wfch D_ccap OPsis (OCHSpec.merge CcapOcap.OPsis OPsis) chs.
Proof.
  unfold CcapOcap.OPsis.
  apply link_2 with schedule_code.
  apply schedule_wfch.
  unfold describe_pi.
  exists itpscap.
  unfold D_ccap, DSpec.lookup, DSpec.merge, DSpecDef.merge.
  unfold DSpec.get, DSpecDef.get, D_scap.
  simpl; trivial.
  unfold chs.
  unfold C.
  simpl.
  compute.
  trivial.
  apply wfis_weak_a_lift with 
    (BigInterp D_ccap <* rho_scap, mk_Lscap schedule_pre schedule_guar *>).
  apply ocap_wfis_schedule.
  unfold aimp.
  intros.
  destruct H as [itp [Hitp H]].
  unfold D_ccap, DSpec.lookup, DSpec.get, DSpecDef.get in Hitp.
  unfold DSpec.merge, DSpecDef.merge, D_ccap0, D_scap in Hitp.
  unfold rho_scap, rho_ccap, beq_nat in Hitp.
  unfold ScapParam.rho_scap, rho_scap in Hitp.
  inj_hyp Hitp.
  deprewrite <- Hitp in H.
  exists itpscap.
  split.
    apply rho_scap_in_D_ccap.
  simpl in  H |- * . 
  destruct H. 
  split. 
    apply yield_schedule_p.
    apply H.
  destruct H0 as [n H0].
  induction n; simpl in H0.
  exists O.
  simpl.
  intros s' Hg.
  apply H0.
  apply schedule_yield_g; trivial.
  exists (S n); simpl.
  intros.
  apply H0.
  apply schedule_yield_g; trivial.
Qed.

Lemma ccap_sound : forall OPsi OPsi' ch,
  CHeap.compat ch chs
  -> wfch D_ccap (OCHSpec.merge OPsi CcapOcap.OPsis) OPsi' ch
  -> wfch D_ccap 
  (OCHSpec.merge (OCHSpec.merge OPsi CcapOcap.OPsis) OPsis) 
  (OCHSpec.merge OPsi' (OCHSpec.merge CcapOcap.OPsis OPsis))
  (CHeap.merge ch chs).
Proof.
  intros OPsi OPsi' ch Hch H.
  apply link_D; trivial.
  apply wfch_chs.
Qed.

(* end *)