Require Import machine.
Require Import optionT.
Require Export Coq.Lists.List.
Require Export natpm.
Require Export ccap.
Require Export seplogic.
Set Implicit Arguments.
Unset Strict Implicit.

(* glob consts *)
Notation NULL := O.

Open Local Scope seplogic_scope.

Definition QNode (p x q : nat) :=
  !(p > 0) ** (p |-> x) ** (p + 1 |-> q).

Fixpoint TQ (p : nat) (Q: queue) {struct Q} : hpred :=
  match Q with
    | Queue.nil => !(p = NULL)
    | Queue.cons x Q' => lexists q,
      (QNode p x q) ** (TQ q Q')
  end.

Fixpoint TQEnd (p r : nat) (Q : queue) {struct Q} : hpred :=
  match Q with
    | Queue.nil => !(p = NULL /\ p = r)
    | Queue.cons x Queue.nil => !(p = r) ** (QNode p x NULL)
    | Queue.cons x Q' => lexists q,
      (QNode p x q) ** (TQEnd q r Q')
  end.

Lemma sep_intro: forall (A: hpred)  (h : heap),
  A h -> h ||- A.
Proof.
  intros.
  unfold hpred_true.
  trivial.
Qed.

Lemma sep_elim : forall (A: hpred) (h : heap),
  h ||- A -> A h.
Proof.
  intros.
  unfold hpred_true in H.
  trivial.
Qed.

Lemma extl : forall Q tlv h hdv A,
  h ||- TQEnd hdv tlv Q ** A -> Q <> Queue.nil ->
    exists v, Heap.get h (tlv + 1) = Some v.
Proof.
  induction Q.
  intros.
  tauto.
  intros.
  destruct Q.
  simpl in H.
  unfold QNode in H.
  sepelim H.
  rewrite A0 in H.
  sconjlifth 1 H.
  generalize (cont_get_eq1 H); intros.
  exists 0.
  trivial.
  simpl in H.
  sepelim H.
  sconjlifth 3 H.
  apply  (IHQ _ _ _ _ H).
  discriminate.
Qed.


Lemma exhd : forall (A : hpred) (h : heap) hdv tlv Q,
  h ||- TQEnd hdv tlv Q ** A -> Q <> Queue.nil ->
    exists v, Heap.get h (hdv + 1) = Some v.
  intros.
  unfold TQEnd in H.
  destruct Q.
  tauto.
  destruct Q.
  unfold QNode in H.
  sepelim H.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H); intros.
  exists 0; trivial.
  unfold QNode in H.
  sepelim H.
  assocl H.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H); intros.
  exists x; trivial.
Qed.


Lemma tl_neq_other : forall Q (A : hpred) (h : heap) hdv tlv c x,
  h ||- TQEnd hdv tlv Q ** c |-> x ** A -> Q <> Queue.nil
    -> tlv + 1 <> c.
Proof.
  induction Q.
  intros.
  tauto.
  intros.
  destruct Q.
  simpl in H.
  unfold QNode in H.
  sepelim H.
  rewrite A0 in H.
  intro.
  subst.
  apply cont_cont_false with h (tlv+1) 0 x top.
  simplsep 3 3.
  simpl in H.
  sepelim H.
  sconjlifth 3 H.
  apply IHQ with top h x0 x.
  simplsep 5 5.
  discriminate.
Qed.

Lemma hd_neq_other : forall (A : hpred) (h : heap) hdv tlv Q c x,
  h ||- TQEnd hdv tlv Q ** c |-> x ** A -> Q <> Queue.nil
    -> hdv + 1 <> c.
Proof.
  intros.
  unfold TQEnd in H.
  destruct Q.
  tauto.
  destruct Q.
  unfold QNode in H.
  sepelim H.
  assert (h ||- hdv + 1 |-> 0 ** c |-> x ** hdv |-> a ** A).
  simplsep 4 4.
  intro.
  subst c.
  hypreplace H1 (cont_cont_false H1).
  trivial.
  unfold QNode in H.
  sepelim H.
  assert (h ||- hdv + 1 |-> x0 ** c |-> x
    ** hdv |-> a **
      match Q with
        | Queue.nil =>
          !(x0 = tlv) ** !(x0 > 0) ** x0 |-> a0 ** x0 + 1 |-> 0
        | Queue.cons _ _ =>
          lexists q,
          (!(x0 > 0) ** x0 |-> a0 ** x0 + 1 |-> q) **
          (fix TQEnd (p r : label) (Q : queue) {struct Q} : hpred :=
            match Q with
              | Queue.nil => !(p = 0 /\ p = r)
              | Queue.cons x Q' =>
                match Q' with
                  | Queue.nil => !(p = r) ** QNode p x 0
                  | Queue.cons _ _ =>
                    lexists q0, QNode p x q0 ** TQEnd q0 r Q'
                end
            end) q tlv Q
      end ** A).
  simplsep 5 5.
  intro.
  subst c.
  assocl H1.
  hypreplace H1 (cont_cont_false H1).
  trivial.
Qed.
  

Lemma hdt : forall (A : hpred) (h : heap) p r Q,
  h ||- TQEnd p r Q ** A -> Q <> Queue.nil ->
    p > 0.
  intros.
  unfold TQEnd in H.
  destruct Q.
  tauto.
  destruct Q.
  unfold QNode in H.
  sepelim H.
  trivial.
  sepelim H.
  assocl H.
  sepelim H.
  trivial.
Qed.
(*
Lemma emp_cont_false : forall x v B A h,
  h ||- x |-> v ** B ** A -> h ||- A -> False.
Proof.
  intros.
  destruct H as [ha [hb H]].
  destruct H as [H1 [H2 H3]].
  destruct H3 as [hc [hd H3]].
  destruct H3 as [H3 [H4 H5]].
*)
Lemma deQ_enQ_t : forall Q Q' x,
  Q <> Queue.nil -> Queue.del (Queue.add x Q) 0 = Some Q' -> Q' <> Queue.nil.
Proof.
  intros.
  destruct Q.
  tauto.
  unfold Queue.add in H0.
  unfold Queue.del in H0.
  injection H0. intros.
  destruct Q.
  intro.
  rewrite H2 in H1.
  discriminate.
  intro.
  rewrite H2 in H1.
  discriminate.
Qed.

Lemma eq_Q: forall Q p r Q' A B h,
  h ||- TQEnd p r Q ** A -> h ||- TQEnd p r Q' ** B -> Q = Q'.
Proof.
  induction Q.
  intros.
  destruct Q'.
  trivial.
  simpl in * .
  destruct Q'.
  unfold QNode in H0.
  sepelim H; sepelim H0.
  elimtype False; omega.
  unfold QNode in H0; sepelim H0.
  sepelim H.
  elimtype False; omega.
  intros.
  destruct Q'.
  simpl in * .
  unfold QNode in H.
  destruct Q.
  sepelim H; sepelim H0.
  elimtype False; omega.
  unfold QNode in H; sepelim H0.
  sepelim H.
  elimtype False; omega.
  simpl in * .
  destruct Q; destruct Q'.
  unfold QNode in * .
  hvinject 4 4.
  trivial.
  rewrite Eq; trivial.
  sepelim H0.
  unfold QNode in * .
  simpl in H0.
  destruct Q'.
  unfold QNode in H0.
  sepelim H; sepelim H0.
  hvinject 4 4.
  subst; elimtype False; omega.
  unfold QNode in H0.
  sepelim H; sepelim H0.
  hvinject 4 4.
  subst; elimtype False; omega.
  sepelim H.
  unfold QNode in * .
  simpl in H.
  destruct Q.
  unfold QNode in H.
  sepelim H; sepelim H0.
  hvinject 4 4.
  subst; elimtype False; omega.
  unfold QNode in H.
  sepelim H; sepelim H0.
  hvinject 4 4.
  subst; elimtype False; omega.
  sepelim H; sepelim H0.
  sconjlifth 3 H. sconjlifth 3 H0.
  hvinject 4 4.
  subst; trivial.

  hypreplace IHQ (IHQ _ _ _ _ _ _ H H0).
  rewrite IHQ.
  trivial.
Qed.

Lemma add_Q: forall Q p r q x A h,
  h ||- TQEnd p r Q ** QNode q x 0 ** A ->
    Q <> Queue.nil ->
    Heap.set h (r + 1) q ||- TQEnd p q (Queue.add x Q) ** A.
Proof.
  induction Q.
  intros.
  destruct H0; trivial.
  intros.
  destruct Q.
  simpl in * .
  sepelim H.
  subst p.
  lexintro 1.
  exists q.
  unfold QNode in * .
  sepelim H.
  sconjlift 2.
  apply cont_update1 with 0.
  simplsep 5 5.
  simpl in * .
  sepelim H.
  lexintro 1.
  exists x0.
  assocr.
  sconjlift 1.
  apply IHQ.
  simplsep 4 4.
  discriminate.
Qed.

Lemma get_de_cons : forall v Q Q',
  Queue.del Q 0 = Some Q' -> Queue.get Q 0 = Some v
  -> Q = Queue.cons v Q'.
  intros.
  unfold Queue.del in H; unfold Queue.get in H0.
  destruct Q.
  discriminate.
  injection H; injection H0; intros.
  subst.
  trivial.
Qed.
(*
  intros.
  generalize Q p h H H0; clear Q p h H H0.
  induction Q .
  destruct Q' .
  intros.
  trivial.
  intros.
  unfold TQEnd in H, H0.
  destruct Q' .
  sepelim H. 
  unfold QNode in H0.
  sepelim H0.
  destruct A0.
  rewrite H1 in A2.
  apply False_ind.
  omega.

  sepelim H0.
  fold TQEnd in H0.
  assocl H0.
  sepelim H0; sepelim H.
  destruct A1.
  rewrite H1 in A0.
  apply False_ind.
  omega.
  intros.
  unfold TQEnd in H.
  fold TQEnd in H.
  destruct Q.
  unfold TQEnd in H0.
  fold TQEnd in H0.
Admitted.
*)  




Lemma TQEnd_TQ: forall p r Q h,
  h ||- TQEnd p r Q -> h ||- TQ p Q.
Proof.
  intros.
  generalize Q p h H; clear Q p h H.
  induction Q.
  intros.
  unfold TQEnd in H.
  unfold TQ.
  generalize (bang_elim1 H).
  generalize (bang_elim2 H).
  intros.
  apply bang_intro.
  trivial.
  tauto.
  intros.
  induction Q.
  unfold TQEnd in H.
  unfold TQ.
  exists 0.
  apply bang_intro_right.
  omega.
  sepelim H.
  trivial.
  unfold TQEnd in H, IHQ.
  fold TQEnd in H, IHQ.
  unfold TQ. unfold TQ in IHQ.
  fold TQ. fold TQ in IHQ.
  destruct Q.
  unfold TQ.
  destruct H as [q H].
  exists q.
  apply sep_elim.
  generalize (sep_intro H).
(*  generalize (sep_intro 
    (QNode p a q ** !(q = r) ** QNode q a0 0) h). *)
  intros.
(*  hypreplace H (H0 H). *)
(*  clear H0. *)
  sepelim H0.
  eapply sconj_mono_right.
  apply H0.
  intros.
  exists 0.
  apply bang_intro_right.
  trivial.
  trivial.
  destruct H as [q H].
  exists q.
  apply sep_elim.
  generalize (sep_intro H).
(*  generalize (sep_intro 
    (QNode p a q **
      (lexists q0, QNode q a0 q0 ** TQEnd q0 r (Queue.cons a1 Q))) h). *)
  intros.
(*  hypreplace H (H0 H). *)
(*  clear H0. *)
  simplsep 3 3.
  hypreplace IHQ (IHQ q h0).
  apply IHQ.
  trivial.
Qed.
