Require Import optionT.
Require Import prodT.
Require Import machine.
Require Export ccap.
Require Import scap.
Require Import scap_ck.
Require Export queue.
Require Import extac.
Require Import natpm2.

Open Local Scope tm_scope.
Open Local Scope scap_scope.
Open Local Scope seplogic_scope.
Open Local Scope listt_scope.

Notation tmp := a0.
Notation hd  := a1.
Notation tl := a2.

Notation SCHEDULE := 10.
Notation READPTR :=  20.
Notation ENQUEUE :=  30.
Notation DEQUEUE :=  40.
Notation SWITCH :=  50.
Notation WRITEPTR := 60.
Notation SRET := 70.

Parameter Stmp : nat.
Parameter Shd : nat.
Parameter Stl : nat.
Parameter cth   : nat.
Parameter tq    : nat.

Definition schedule_code :=
  sw tmp Stmp zero;
  sw hd Shd zero;
  sw tl Stl zero;
  j READPTR;; .

Definition readptr_code :=
  lw hd tq zero;
  addiu tmp zero tq;
  lw tl 1 tmp;
  lw tmp cth zero;
  sw ra 0 tmp;
  beq hd zero WRITEPTR;
  j ENQUEUE;; .

Definition enqueue_code :=
  sw ra 0 tmp;
  sw tmp 1 tl;
  move tl tmp;
  j DEQUEUE;; .

Definition dequeue_code :=
  move tmp hd;
  lw hd 1 hd;
  sw zero 1 tmp;
  j SWITCH;; .

Definition switch_code :=
  lw ra 0 tmp;
  j WRITEPTR;; .

Definition writeptr_code :=
  sw tmp cth zero;
  sw hd tq zero;
  addiu tmp zero tq;
  sw tl 1 tmp;
  j SRET;; .

Definition sret_code :=
  lw tmp Stmp zero;
  lw hd Shd zero;
  lw tl Stl zero;
  jr ra;; .

Import List.

Definition C : codeheap := 
  (* SCHEDULE *)
  (0010,   sw tmp Stmp zero             ) ::
  (0011,   sw hd Shd zero               ) ::
  (0012,   sw tl Stl zero               ) ::
  (0013,   j READPTR                    ) ::
  (* READPTR *)
  (0020,   lw hd tq zero                ) ::
  (0021,   addiu tmp zero tq            ) ::
  (0022,   lw tl 1 tmp                  ) ::
  (0023,   lw tmp cth zero              ) ::
  (0024,   sw ra 0 tmp                  ) ::
  (0025,   beq hd zero WRITEPTR         ) ::
  (0026,   j ENQUEUE                    ) ::
  (* ENQUEUE *)
  (0030,   sw ra 0 tmp                  ) ::
  (0031,   sw tmp 1 tl                  ) ::
  (0032,   move tl tmp                  ) ::
  (0033,   j DEQUEUE                    ) ::
  (* DEQUEUE *)
  (0040,   move tmp hd                  ) ::
  (0041,   lw hd 1 hd                   ) ::
  (0042,   sw zero 1 tmp                ) ::
  (0043,   j SWITCH                     ) ::
  (* SWITCH *)
  (0050,   lw ra 0 tmp                  ) ::
  (0051,   j WRITEPTR                   ) ::
  (* WRITEPTR *)
  (0060,   sw tmp cth zero              ) ::
  (0061,   sw hd tq zero                ) ::
  (0062,   addiu tmp zero tq            ) ::
  (0063,   sw tl 1 tmp                  ) ::
  (0064,   j SRET                       ) ::
  (* SRET *)
  (0070,   lw tmp Stmp zero             ) ::
  (0071,   lw hd Shd zero               ) ::
  (0072,   lw tl Stl zero               ) ::
  (0073,   jr ra                        ) ::
  nil.

Definition enQ (Q : queue) x :=
  Queue.add x Q.

Definition deQ (Q : queue) :=
  Queue.del Q 0.

Definition getQ (Q : queue) :=
  Queue.get Q 0.

Definition p_ctxq Q c1 c2 c3 hdv tlv:=
  TQEnd hdv tlv Q ** (Stmp |-> c1)
  ** (Shd |-> c2) ** (Stl |-> c3).

Definition p_ctxqueue Q :=
  lexists c1, lexists c2, lexists c3, lexists pcx,
    lexists cthv, lexists hdv, lexists tlv,
    (cth |-> cthv) ** QNode cthv pcx NULL ** QNode tq hdv tlv
    ** p_ctxq Q c1 c2 c3 hdv tlv.

Definition use_regs s s':=
  S_R (rset (rset (rset (rset s ra 0) tmp 0) hd 0) tl 0)
  = S_R (rset (rset (rset (rset s' ra 0) tmp 0) hd 0) tl 0).

Definition save_regs s s' :=
  hget s Stmp = Some (rget s' tmp)
  /\ hget s Shd = Some (rget s' hd)
  /\ hget s Stl = Some (rget s' tl).

Definition schedule_post s A Q pcx :=
  s |||- A ** (lexists c1, lexists c2, lexists c3, 
    lexists cthv, lexists hdv, lexists tlv,
    !(pcx = rget s ra) ** (cth |-> cthv) ** QNode cthv pcx NULL ** QNode tq hdv tlv
    ** p_ctxq Q c1 c2 c3 hdv tlv).

Definition schedule_pre s :=
  exists Q, s |||- top ** p_ctxqueue Q.

Definition schedule_guar s s' :=
  forall Q : queue, forall A,
    s |||- A ** p_ctxqueue Q ->
    exists pcx,
      getQ (enQ Q (rget s ra)) = Some pcx
        /\ exists Q', deQ (enQ Q (rget s ra)) = Some Q'
          /\ pcget s' = pcx
          /\ S_R (rset s ra pcx) = S_R s'
          /\ schedule_post s' A Q' pcx.

Definition readptr_pre s :=
  exists Q, exists pcx, exists cthv, exists hdv, exists tlv,
      s |||- top ** (cth |-> cthv) ** QNode cthv pcx NULL
      ** QNode tq hdv tlv
      ** p_ctxq Q (rget s tmp) (rget s hd) (rget s tl) hdv tlv.

Definition readptr_guar s s' :=
  forall Q : queue, forall A,
    (exists pcx, exists cthv, exists hdv, exists tlv,
        s |||- A ** (cth |-> cthv) ** QNode cthv pcx NULL
          ** QNode tq hdv tlv 
          ** p_ctxq Q (rget s tmp) (rget s hd) (rget s tl) hdv tlv) ->
    exists pcx,
      getQ (enQ Q (rget s ra)) = Some pcx
      /\ exists Q', deQ (enQ Q (rget s ra)) = Some Q'
        /\ pcget s' = pcx
        /\ save_regs s s'
        /\ use_regs s s'
        /\ schedule_post s' A Q' pcx.

Definition enqueue_pre s :=
  exists Q, exists c1, exists c2, exists c3, exists pcx,
    exists cthv, exists hdv, exists tlv,
      s |||- top ** !(cthv = rget s tmp
        /\ hdv = rget s hd /\ tlv = rget s tl
        /\ Q <> Queue.nil)
      ** (cth |-> cthv) ** QNode cthv pcx NULL
      ** QNode tq hdv tlv ** p_ctxq Q c1 c2 c3 hdv tlv.

Definition enqueue_guar s s' :=
  forall Q : queue, forall A,
    (exists c1, exists c2, exists c3, exists pcx,
      exists cthv, exists hdv, exists tlv,
          s |||- A ** !(cthv = rget s tmp
            /\ hdv = rget s hd /\ tlv = rget s tl
            /\ Q <> Queue.nil)
          ** (cth |-> cthv) ** QNode cthv pcx NULL
          ** QNode tq hdv tlv ** p_ctxq Q c1 c2 c3 hdv tlv) ->
    exists pcx,
      getQ (enQ Q (rget s ra)) = Some pcx
      /\ exists Q', deQ (enQ Q (rget s ra)) = Some Q'
        /\ pcget s' = pcx
        /\ save_regs s s'
        /\ use_regs s s'
        /\ schedule_post s' A Q' pcx.

Definition dequeue_pre s :=
  exists Q, exists Q1, exists c1, exists c2, exists c3, exists pcx,
    exists cthv, exists hdv, exists tlv,
      exists c, exists hv, exists tv,
      s |||- top ** !(cthv = rget s tmp
        /\ hdv = rget s hd /\ tlv = rget s tl
        /\ Q <> Queue.nil /\ deQ Q = Some Q1
        /\ Q1 <> Queue.nil /\ cthv = tlv
        /\ hget s cthv = Some pcx /\ hget s (cthv + 1) = Some NULL)
      ** (cth |-> c)
      ** QNode tq hv tv ** p_ctxq Q c1 c2 c3 hdv tlv.

Definition dequeue_guar (s s' : state) := 
  forall Q : queue, forall A,
    (exists Q1, exists c1, exists c2, exists c3, exists pcx,
      exists cthv, exists hdv, exists tlv,
        exists c, exists hv, exists tv,
          s |||- A ** !(cthv = rget s tmp
            /\ hdv = rget s hd /\ tlv = rget s tl
            /\ Q <> Queue.nil /\ deQ Q = Some Q1
            /\ Q1 <> Queue.nil /\ cthv = tlv
            /\ hget s cthv = Some pcx /\ hget s (cthv + 1) = Some NULL)
          ** (cth |-> c)
          ** QNode tq hv tv ** p_ctxq Q c1 c2 c3 hdv tlv) ->
    exists pcx,
      getQ Q = Some pcx
      /\ exists Q', deQ Q = Some Q'
        /\ pcget s' = pcx
        /\ save_regs s s'
        /\ use_regs s s'
        /\ schedule_post s' A Q' pcx.

Definition switch_pre s := 
  exists Q, exists c1, exists c2, exists c3, exists pcx,
    exists cthv, exists hdv, exists tlv,
      exists c, exists hv, exists tv,
        s |||- top ** !(cthv = rget s tmp
        /\ hdv = rget s hd /\ tlv = rget s tl)
      ** (cth |-> c) ** QNode cthv pcx NULL
      ** QNode tq hv tv ** p_ctxq Q c1 c2 c3 hdv tlv.

Definition switch_guar s s' :=
  forall Q : queue, forall A,
    (exists c1, exists c2, exists c3, exists pcx,
      exists cthv, exists hdv, exists tlv, 
        exists c, exists hv, exists tv,
          s |||- A ** !(cthv = rget s tmp
          /\ hdv = rget s hd /\ tlv = rget s tl)
            ** (cth |-> c) ** QNode cthv pcx NULL
            ** QNode tq hv tv ** p_ctxq Q c1 c2 c3 hdv tlv) ->
    exists pcx,
      pcget s' = pcx
      /\ hget s (rget s tmp) = Some pcx
      /\ save_regs s s'
      /\ use_regs s s'
      /\ schedule_post s' A Q pcx.

Definition writeptr_pre s := 
  exists Q, exists c1, exists c2, exists c3, exists pcx,
    exists cthv, exists hdv, exists tlv,
      exists c, exists hv, exists tv,
        s |||- top ** !(pcx = rget s ra /\ cthv = rget s tmp
        /\ hdv = rget s hd /\ tlv = rget s tl)
      ** (cth |-> c) ** QNode cthv pcx NULL
      ** QNode tq hv tv ** p_ctxq Q c1 c2 c3 hdv tlv.

Definition writeptr_guar s s' :=
  forall Q : queue, forall A,
    (exists c1, exists c2, exists c3, exists pcx,
      exists cthv, exists hdv, exists tlv, 
        exists c, exists hv, exists tv,
          s |||- A ** !(pcx = rget s ra /\ cthv = rget s tmp
          /\ hdv = rget s hd /\ tlv = rget s tl)
            ** (cth |-> c) ** QNode cthv pcx NULL
            ** QNode tq hv tv ** p_ctxq Q c1 c2 c3 hdv tlv) ->
    exists pcx,
      pcget s' = pcx
      /\ pcx = rget s ra
      /\ save_regs s s'
      /\ use_regs s s'
      /\ schedule_post s' A Q pcx.

Definition sret_pre s :=
  exists Q, exists c1, exists c2, exists c3, exists pcx,
    exists cthv, exists hdv, exists tlv,
      s |||- top ** !(pcx = rget s ra)
        ** (cth |-> cthv) ** QNode cthv pcx NULL
        ** QNode tq hdv tlv ** p_ctxq Q c1 c2 c3 hdv tlv.

Definition sret_guar s s' :=
  forall Q : queue, forall A,
    (exists c1, exists c2, exists c3, exists pcx,
      exists cthv, exists hdv, exists tlv, 
        s |||- A ** !(pcx = rget s ra)
          ** (cth |-> cthv) ** QNode cthv pcx NULL
          ** QNode tq hdv tlv ** p_ctxq Q c1 c2 c3 hdv tlv) ->
    exists pcx,
      pcget s' = pcx
      /\ pcx = rget s ra
      /\ save_regs s s'
      /\ use_regs s s'
      /\ schedule_post s' A Q pcx.

Import ListT.

Open Local Scope listt_scope.

Definition psi : chspec :=
  (pairT SCHEDULE (mk_cdspec schedule_pre schedule_guar)) ::
  (pairT READPTR (mk_cdspec readptr_pre readptr_guar)) ::
  (pairT ENQUEUE (mk_cdspec enqueue_pre enqueue_guar)) ::
  (pairT DEQUEUE (mk_cdspec dequeue_pre dequeue_guar)) ::
  (pairT SWITCH (mk_cdspec switch_pre switch_guar)) ::
  (pairT WRITEPTR (mk_cdspec writeptr_pre writeptr_guar)) ::
  (pairT SRET (mk_cdspec sret_pre sret_guar)) ::
   nilT.

Ltac unfoldpred :=
  unfold schedule_pre, schedule_guar;
  unfold readptr_pre, readptr_guar;
  unfold enqueue_pre, enqueue_guar;
  unfold dequeue_pre, dequeue_guar;
  unfold switch_pre, switch_guar;
  unfold writeptr_pre, writeptr_guar;
  unfold sret_pre, sret_guar.


Definition p_yield (s : state) := 
  exists Q, s |||- top ** p_ctxqueue Q.

Definition g_yield (s s' : state) := 
  forall Q : queue, forall A,
    (s |||- A ** p_ctxqueue Q) ->
    exists i, exists pcx,
      Queue.get (Queue.add (rget s ra) Q) i = Some pcx
      /\ exists Q', (Queue.del (Queue.add (rget s ra) Q) i) = Some Q'
        /\ pcget s' = pcx
        /\ S_R (rset s ra pcx) = S_R s'
        /\ (s' |||- A ** p_ctxqueue Q').

Lemma yield_schedule_p : forall s,
  p_yield s -> schedule_pre s.
Proof.
  unfold schedule_pre, p_yield.
  intros.
  trivial.
Qed.

Lemma schedule_yield_g : forall s s',
  schedule_guar s s' -> g_yield s s'.
Proof.
  unfold schedule_guar, g_yield.
  intros.
  exists 0.
  hypreplace H (H Q A H0).
  destruct H as [pcx [Ha [Q' [Hb [Hc [Hd He]]]]]].
  exists pcx.
  unfold getQ, enQ in Ha.
  unfold deQ, enQ in Hb.
  split; trivial.
  exists Q'.
  repeat (split; trivial).
  unfold schedule_post in He.
  unfold p_ctxqueue.
  destruct s' as [h' rf' pc'].
  simplsep 1 1.
  sepelim He.
  exists x; exists x0; exists x1; exists pcx.
    exists x2; exists x3; exists x4.
  trivial.
Qed.

Theorem sret_ok : wfis psi sret_pre sret_guar sret_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct H as [Q [c1 [c2 [c3 [pcx [cthv [hdv [tlv H]]]]]]]].
  destruct s as [h rf pc].
  sepelim H.
  simplst.
  unfold p_ctxq in H.
  do 9 (hypreplace H (sconj_sym H); assocl H).
  split.
  generalize (cont_get_eq1 H). intros. simpl. rewrite H0.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H). intros. simpl. rewrite H1.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H). intros. simpl. rewrite H2.
  trivial.
  intros.
  simpl in H0.
  generalize (cont_get_eq1 H); intros.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H); intros.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H); intros.
  rewrite H2 in H0.
  simpl in H0.
  rewrite H3 in H0.
  simpl in H0.
  rewrite H4 in H0.
  destruct s' as [h' rf' pc'].
  injection H0.
  intros.
  unfold Regfile.get in H5.
  compute in H5.
  exists pcx.
  subst.
  split_l.
  split; trivial.
  split_l; split_lr.
  split.
  simplst.
  do 11 (hypreplace H (sconj_sym H); assocl H).
  generalize (cont_get_eq1 H); intros.
  trivial.
  unfold save_regs.
  simplst.
  generalize (cont_get_eq1 H); intros.
  trivial.
  unfold use_regs.
  apply Regfile.extensionality.
  destruct r; simplst; trivial.
  unfold schedule_post.
  clear H2 H3 H4 H0 H.
  clear cthv hdv tlv.
  destruct H1 as [c4 [c5 [c6 [pcx [cthv [hdv [tlv H]]]]]]].
  simplsep 1 1.
  exists c4; exists c5; exists c6; exists cthv;
    exists hdv; exists tlv.
  sepelim H.
  apply sep_elim.
  subst.
  simplsep 5 5.
Qed.

Theorem writeptr_ok : wfis psi writeptr_pre writeptr_guar writeptr_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct H as [Q [c1 [c2 [c3 [pcx
    [cthv [hdv [tlv [c [hv [tv H]]]]]]]]]]].
  destruct s as [h rf pc].
  simplst.
  unfold p_ctxq in H.
  unfold QNode in H.
  sepelim H.
  hypreplace H (sconj_sym H); assocl H.
  split.
  generalize (cont_get_eq1 H). intros. simpl. rewrite H0.
  simplst.
  do 3 (hypreplace H (sconj_sym H); assocl H).
  generalize (cont_get_eq1 H). intros. simpl. rewrite H1.
  simplst.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H). intros. simpl. rewrite H2.
  exists Q; exists c1; exists c2; exists c3; exists pcx;
    exists cthv; exists hdv; exists tlv.
  unfold p_ctxq. unfold QNode.
  do 8 (apply sconj_sym; assocr).
  destruct A as [Ha [Hb [Hc Hd]]].
  rewrite Hd.
  apply cont_update1 with (v1 := tv).
  do 12 (apply sconj_sym; assocr).
  rewrite Hc.
  apply cont_update1 with (v1 := hv).
  do 8 (apply sconj_sym; assocr).
  rewrite Hb.
  apply cont_update1 with (v1 := c).
  subst.
  simplsep 10 10.
  simpl.
  intro.
  assert (h ||- cth |-> c ** tq + 1 |-> tv ** tq |-> hv
    ** TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 **
          top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  rewrite H2 in H3.
  hypreplace H3 (cont_cont_false H3).
  trivial.
  omega.
  simpl.
  intro.
  assert (h ||- cth |-> c ** tq |-> hv ** tq + 1 |-> tv
    ** TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 **
      top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  rewrite H1 in H2.
  hypreplace H2 (cont_cont_false H2).
  trivial.
  intros.
  simplst.
  simpl in H0.
  generalize (cont_get_eq1 H). intros. rewrite H2 in H0.
  destruct A as [Ha [Hb [Hc Hd]]].
  rewrite <- Hb in H0.
  rewrite <- Hc in H0.
  simpl in H0.
  do 3 (hypreplace H (sconj_sym H); assocl H).
  generalize (cont_get_eq1 H). intros.
  assert (Heap.get (Heap.set h cth cthv) tq = Some hv).
  simplst.
  trivial.
  intro.
  assert (h ||- cth |-> c ** tq |-> hv ** tq + 1 |-> tv
    ** TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 **
      top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  rewrite H4 in H5.
  hypreplace H5 (cont_cont_false H5).
  trivial.
  rewrite H4 in H0.
  simplst.
  simpl in H0.
  assert (Heap.get (Heap.set (Heap.set h cth cthv) tq hdv)
    (tq + 1) = Some tv).
  simplst.
  hypreplace H (sconj_sym H). assocl H.
  generalize (cont_get_eq1 H). intros.
  trivial.
  assert (h ||- cth |-> c ** tq + 1 |-> tv ** tq |-> hv
    ** TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 **
      top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H6 in H5.
  hypreplace H5 (cont_cont_false H5).
  trivial.
  omega.
  rewrite H5 in H0.
  hypreplace H0 (H0 Q0 A2).
  elim H0.
  intros pcx0 Heq.
  exists pcx0.
  destruct Heq as [He1 [Hg [He2 [He3 He4]]]].
  split; trivial.
  split.
  simplst.
  trivial.
  unfold save_regs. unfold save_regs in He2.
  unfold hget. unfold hget in He2.
  destruct He2 as [Hf1 [Hf2 Hf3]].
  destruct s' as [h' rf' pc'].
  split.
  rewrite <- Hf1.
  simplst.
  trivial.
  split; trivial.
  split.
  rewrite <- Hf2.
  simplst.
  trivial.
  assert (h ||- cth |-> c ** Shd |-> c2 ** tq + 1 |-> tv **
    TQEnd hdv tlv Q ** Stmp |-> c1 ** Stl |-> c3 ** tq |-> hv
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  assert (h ||- tq |-> hv ** Shd |-> c2 ** tq + 1 |-> tv **
    TQEnd hdv tlv Q ** Stmp |-> c1 ** Stl |-> c3 ** cth |-> c
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  assert (h ||- tq + 1 |-> tv ** Shd |-> c2 ** tq |-> hv **
    TQEnd hdv tlv Q ** Stmp |-> c1 ** Stl |-> c3 ** cth |-> c
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  rewrite <- Hf3.
  simplst. trivial.
  assert (h ||- cth |-> c ** Stl |-> c3 ** tq + 1 |-> tv **
    TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** tq |-> hv
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  assert (h ||- tq |-> hv ** Stl |-> c3 ** tq + 1 |-> tv **
    TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** cth |-> c
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  assert (h ||- tq + 1 |-> tv ** Stl |-> c3 ** tq |-> hv **
    TQEnd hdv tlv Q ** Stmp |-> c1 ** Shd |-> c2 ** cth |-> c
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.


  assert (h ||- cth |-> c ** Stmp |-> c1 ** tq + 1 |-> tv **
    TQEnd hdv tlv Q ** Shd |-> c2 ** Stl |-> c3 ** tq |-> hv
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  assert (h ||- tq |-> hv ** Stmp |-> c1 ** tq + 1 |-> tv **
    TQEnd hdv tlv Q ** Shd |-> c2 ** Stl |-> c3 ** cth |-> c
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  assert (h ||- tq + 1 |-> tv ** Stmp |-> c1 ** tq |-> hv **
    TQEnd hdv tlv Q ** Shd |-> c2 ** Stl |-> c3 ** cth |-> c
  ** top ** cthv |-> pcx ** cthv + 1 |-> 0).
  simplsep 10 10.
  intro.
  rewrite H7 in H6.
  hypreplace H6 (cont_cont_false H6).
  trivial.
  split.
  unfold use_regs. unfold use_regs in He3.
  simplst.
  rewrite <- He3.
  apply extensionality.
  repeat (destruct r; simplst; trivial).
  trivial.
  destruct H1 as [c4 [c5 [c6 [pcx0 [cthv0
    [hdv0 [tlv0 [c0 [hv0 [tv0 H1]]]]]]]]]].
  exists c4; exists c5; exists c6; exists pcx0;
    exists cthv0; exists hdv0; exists tlv0.
  sepelim H1.
  destruct A as [Ha1 [Hb1 [Hc1 Hd1]]].
  subst.
  unfold QNode.
  do 8 (apply sconj_sym; assocr).
  apply cont_update1 with (v1 := tv0).
  do 9 (apply sconj_sym; assocr).
  apply cont_update1 with (v1 := hv0).
  do 5 (apply sconj_sym; assocr).
  apply cont_update1 with (v1 := c0).
  unfold QNode in H1.
  sepelim H1.
  clear H.
  unfold p_ctxq.
  simplsep 10 10.
Qed.

Theorem switch_ok : wfis psi switch_pre switch_guar switch_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct H as [Q [c1 [c2 [c3 [pcx
    [cthv [hdv [tlv [c [hv [tv H]]]]]]]]]]].
  destruct s as [h rf pc].
  simplst.
  sepelim H.
  destruct A as [Ha [Hb Hc]].
  rewrite <- Ha.
  simpl.
  unfold QNode in H.
  sepelim H.
  do 2 (hypreplace H (sconj_sym H); assocl H).
  generalize (cont_get_eq1 H). intros.
  assert (cthv + 0 = cthv).
  omega.
  rewrite H1.
  rewrite H0.
  intros.
  split.
  exists Q; exists c1; exists c2; exists c3; exists pcx;
    exists cthv; exists hdv; exists tlv; exists c;
      exists hv; exists tv.
  unfold p_ctxq; unfold QNode.
  simplsep 10 10.
  intros.
  hypreplace H2 (H2 Q0 A1).
  elim H2.
  intros pcx0 Heq.
  exists pcx0.
  destruct s' as [h' rf' pc'].
  destruct Heq as [He1 [Hg [He2 [He3 He4]]]].
  split; trivial.
  split.
  simplst. auto.
  unfold save_regs. unfold save_regs in He2.
  destruct He2 as [Ha1 [Ha2 Ha3]].
  split.
  split.
  rewrite <- Ha1.
  simplst. trivial.
  split.
  rewrite <- Ha2.
  simplst. trivial.
  rewrite <- Ha3.
  simplst. trivial.
  split.
  unfold use_regs. unfold use_regs in He3.
  simplst.
  rewrite <- He3.
  apply extensionality.
  repeat (destruct r; simplst; trivial).
  trivial.
  destruct H3 as [c4 [c5 [c6 [pcx0
    [cthv0 [hdv0 [tlv0 [c0 [hv0 [tv0 H3]]]]]]]]]].
  exists c4; exists c5; exists c6; exists pcx0;
    exists cthv0; exists hdv0; exists tlv0; exists c0;
      exists hv0; exists tv0.
  unfold QNode in H3.
  sepelim H3.
  destruct A2 as [Ha1 [Ha2 Ha3]].
  subst.
  hvinject 7 7.
  subst.
  unfold QNode.
  simplsep 10 10.
Qed.

Theorem dequeue_ok : wfis psi dequeue_pre dequeue_guar dequeue_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct H as [Q [Q1 [c1 [c2 [c3 [pcx
    [cthv [hdv [tlv [c [hv [tv H]]]]]]]]]]]].
  destruct s as [h rf pc].
  simplst.
  unfold QNode, p_ctxq in H.
  sepelim H.
  destruct A as [Ha [Hb [Hc [Hd [Hh1 [Hh2 [He [Hf Hg]]]]]]]].
  rewrite <- Hb.
  do 4 (hypreplace H (sconj_sym H); assocl H).
  generalize (exhd H Hd).
  intros.
  destruct H0 as [hdv' H0].
  rewrite H0.
  simplst.
  rewrite H0.
  split.
  intros.
  assert (exists pcx', getQ Q = Some pcx').
  unfold getQ.
  unfold Queue.get.
  destruct Q.
  tauto.
  exists a; trivial.
  destruct H1 as [pcx' H1].
  exists Q1; exists c1; exists c2; exists c3;
    exists pcx'; exists hdv; exists hdv'; exists cthv; exists c;
      exists hv; exists tv.
  unfold QNode, p_ctxq.
  do 5 (apply sconj_sym; assocr).
  apply cont_update1 with (v1 := hdv').
  subst.
  apply sconj_sym; assocr.
  apply bang_intro_left.
  trivial.
  do 6 (apply sconj_sym; assocr).
  apply bang_intro_left.
  split. simplst. trivial.
  split. simplst. trivial.
  simplst. trivial.
  assocr.
  apply sconj_sym; assocr.
  apply bang_intro_left.
  generalize (hdt H Hd).
  intros; trivial.
  rewrite He.
  destruct Q.
  tauto.
  unfold deQ, Queue.del in Hh1.
  injection Hh1; clear Hh1; intro Hh1.
  unfold getQ, Queue.get in H1.
  injection H1; clear H1; intro H1.
  subst Q1; subst a.
  unfold TQEnd, QNode in H.
  destruct Q.
  tauto.
  fold TQEnd in H.
  unfold TQEnd.
  fold TQEnd.
  sepelim H.
  assert (x = hdv').
  assocl H.
  hypreplace H (sconj_sym H); assocl H.
  generalize (cont_get_eq1 H); intros.
  rewrite H0 in H1.
  injection H1; auto.
  subst x.
  simplsep 10 10.
  intros.
  hypreplace H1 (H1 Q1 A).
  elim H1.
  clear H1.
  intros pcx' H3.
  destruct s' as [h' rf' pc'].
  exists pcx'.
  split.
  destruct H2 as [Q2 [c4 [c5 [c6 [pcx0
    [cthv0 [hdv0 [tlv0 [c0 [hv0 [tv0 H2]]]]]]]]]]].
  sepelim H2.
  destruct H3 as [Ha1 [Ha2 [Ha3 [Ha4 Ha5]]]].
  simplst.
  unfold QNode, p_ctxq in H2.
  destruct A1 as [He1 [He2 [He3 [He4 [He5 [He6 [He7 He8]]]]]]].
  subst hdv0.
  assert (Heap.get h hdv = Some pcx').
  simplst.
  rewrite Heap.set_sem in Ha2.
  unfold HeapDef.beq in Ha2.
  assert (beq_nat (hdv + 1) hdv = false).
  apply neq_beq_false.
  omega.
  rewrite H1 in Ha2. trivial.
  destruct Q0. tauto.
  unfold TQEnd, QNode in H2.
  destruct Q0.
  unfold getQ, Queue.get.
  sepelim H2.
  do 4 (hypreplace H2 (sconj_sym H2); assocl H2).
  generalize (cont_get_eq1 H2); intros.
  rewrite H1 in H3; injection H3; auto.
  unfold getQ, Queue.get.
  sepelim H2.
  do 4 (hypreplace H2 (sconj_sym H2); assocl H2).
  generalize (cont_get_eq1 H2); intros.
  rewrite H1 in H3; injection H3; auto.
  destruct H2 as [Q2 [c4 [c5 [c6 [pcx0 [cthv0 [hdv0
    [tlv0 [c0 [hv0 [tv0 H2]]]]]]]]]]].
  exists Q2.
  sepelim H2.
  split.
  destruct A1 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
  trivial.
  split.
  destruct H3 as [H3 _].
  trivial.
  split.
  unfold save_regs.
  destruct H3 as [_ [_ [H3 _]]].
  unfold save_regs in H3.
  destruct H3 as [He1 [He2 He3]].
  split. rewrite <- He1. simplst. trivial.
  assert  (h ||- TQEnd hdv tlv Q **
    Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 **
    ((top ** cth |-> c) ** tq |-> hv) ** tq + 1 |-> tv).
  simplsep 7 7.
  apply (hd_neq_other H1 Hd).
  split. rewrite <- He2. simplst. trivial.
  assert  (h ||- TQEnd hdv tlv Q **
    Shd |-> c2 ** Stmp |-> c1 ** Stl |-> c3 **
    ((top ** cth |-> c) ** tq |-> hv) ** tq + 1 |-> tv).
  simplsep 7 7.
  apply (hd_neq_other H1 Hd).
  rewrite <- He3. simplst. trivial.
  assert  (h ||- TQEnd hdv tlv Q **
     Stl |-> c3 ** Stmp |-> c1 ** Shd |-> c2 **
    ((top ** cth |-> c) ** tq |-> hv) ** tq + 1 |-> tv).
  simplsep 7 7.
  apply (hd_neq_other H1 Hd).
  split.
  destruct H3 as [_ [_ [_ [H3 _]]]].
  unfold use_regs. unfold use_regs in H3.
  rewrite <- H3.
  simplst.
  apply extensionality.
  repeat (destruct r; simplst; trivial).
  unfold p_ctxq, QNode in H2.
  sepelim H2.
  destruct A1 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5
    [Ha6 [Ha7 [Ha8 Ha9]]]]]]]].
  rewrite Ha3 in H2.
  rewrite <- Hc in H2.
  generalize H; clear H; intro H.
  rewrite Ha2 in H2.
  do 4 (hypreplace H (sconj_sym H); assocl H).
  hvinject 12 12.
  subst c0; subst c4; subst c5; subst c6; subst tv0; subst hv0.
  do 4 (hypreplace H (sconj_sym H); assocl H).
  do 4 (hypreplace H2 (sconj_sym H2); assocl H2).
  generalize (eq_Q H H2).
  intros. subst Q0.
  rewrite Ha5 in Hh1.
  injection Hh1; intros.
  subst Q2.
  destruct H3 as [_ [_ [_ [_ H3]]]].
  trivial.
  clear H1.
  destruct H2 as [Q2 [c4 [c5 [c6 [pcx0 [cthv0 [hdv0
    [tlv0 [c0 [hv0 [tv0 H2]]]]]]]]]]].
  assert (exists pcx', (getQ Q) = Some pcx').
  unfold getQ, Queue.get.
  destruct Q. tauto.
  exists a. trivial.
  destruct H1 as [pcx' H1].
  exists c4; exists c5; exists c6; exists pcx'; exists hdv0;
    exists hdv'; exists tlv0; exists c0; exists hv0; exists tv0.
  sepelim H2.
  unfold p_ctxq, QNode in H2.
  sepelim H2.
  destruct A1 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5
    [Ha6 [Ha7 [Ha8 Ha9]]]]]]]].
  subst hdv0.
  rewrite <- Ha3 in Hc.
  subst tlv0.
  rewrite <- Hc in H2.
  do 4 (hypreplace H2 (sconj_sym H2); assocl H2).
  generalize (eq_Q H H2).
  intros. subst Q0.
  rewrite Ha5 in Hh1. 
  injection Hh1; intros.
  subst Q2.
  unfold p_ctxq, QNode.
  do 5 (apply sconj_sym; assocr).
  apply cont_update1 with (v1 := hdv').
  rewrite <- Hc.
  hvinject 3 3.
  subst c4; subst c0.
  hvinject 7 7.
  subst c5; subst c6; subst hv0; subst tv0.
  generalize (get_de_cons Ha5 H1).
  intros.
  rewrite H3 in H2.
  unfold TQEnd in H2.
  fold TQEnd in H2.
  destruct Q1.
  tauto.
  unfold QNode in H2; sepelim H2.
  assocl H2.
  hypreplace H2 (sconj_sym H2); assocl H2.
  generalize (cont_get_eq1 H2). intros.
  rewrite H0 in H4.
  injection H4; intros.
  subst x.
  simplsep 10 10.
Qed.

Theorem enqueue_ok : wfis psi enqueue_pre enqueue_guar enqueue_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct H as [Q [c1 [c2 [c3 [pcx
    [cthv [hdv [tlv H]]]]]]]].
  destruct s as [h rf pc].
  sepelim H.
  destruct A as [Ha [Hb [Hc Hd]]].
  simplst.
  rewrite <- Ha.
  assert (cthv + 0 = cthv).
  omega.
  rewrite H0.
  unfold QNode in H.
  sepelim H.
  sconjlifth 2 H.
  generalize (cont_get_eq1 H); intros.
  rewrite H1.
  simplst.
  unfold p_ctxq, QNode in H.
  sconjlifth 4 H.
  generalize (extl H Hd). intros.
  destruct H2 as [v H2].
  rewrite <- Hc.
  rewrite H2.
  simplst.
  split.
  exists (enQ Q (Regfile.get rf ra)).
  assert (exists Q1, deQ (enQ Q (Regfile.get rf ra)) = Some Q1).
  unfold deQ, enQ.
  unfold Queue.del, Queue.add.
  destruct Q.
  tauto.
  exists  ((fix add (a0 : QueueElem.A) (l : Queue.list) {struct l} :
           Queue.list :=
           match l with
           | Queue.nil => Queue.cons a0 Queue.nil
           | Queue.cons a' l' => Queue.cons a' (add a0 l')
           end) (Regfile.get rf ra) Q).
  trivial.
  destruct H3 as [Q1 H3].
  exists Q1.
  exists c1; exists c2; exists c3.
  exists (Regfile.get rf ra).
  exists cthv; exists hdv; exists cthv.
  exists cthv; exists hdv; exists tlv.
  simplst.
  unfold p_ctxq, QNode.
  sconjlifth 7 H.
  generalize (cont_get_eq1 H).
  intros.
  sconjlifth 9 H.
  hypreplace H (cont_update1 H (Regfile.get rf ra)).
  sconjlift 6.
  unfold enQ.
  rewrite <- Ha.
  apply add_Q.
  unfold QNode.
  simplsep 9 9.
  repeat (split; trivial).
  unfold Queue.add.
  destruct Q.
  tauto.
  discriminate.
  unfold deQ, enQ, Queue.del, Queue.add in H3.
  destruct Q.
  tauto.
  destruct Q.
  injection H3; intros.
  rewrite <- H5.
  discriminate.
  injection H3; intros.
  rewrite <- H5.
  discriminate.
  trivial.
  omega.
  assert (h ||- TQEnd hdv tlv Q ** cthv + 1 |-> 0 **
          (Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3) **
          ((((top ** cth |-> cthv) ** cthv |-> pcx) **
           tq |-> hdv ** tq + 1 |-> tlv))).
  simplsep 8 8.
  exact (tl_neq_other H4 Hd).
  assert (h ||- TQEnd hdv tlv Q ** cthv |-> pcx **
    cthv + 1 |-> 0 **
      (Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3) **
      ((((top ** cth |-> cthv) ** 
        tq |-> hdv ** tq + 1 |-> tlv)))).
  simplsep 8 8.
  exact (tl_neq_other H4 Hd).
  intros.
  hypreplace H3 (H3 (enQ Q (Regfile.get rf ra)) A1).
  elim H3.
  intros.
  exists x.
  destruct H5 as [He1 [Q' [He2 [He3 [He4 [He5 He6]]]]]].
  destruct s' as [h' rf' pc'].
  destruct H4 as [c4 [c5 [c6 [pcx0 [cthv0 [hdv0 [tlv0 H5]]]]]]].
  unfold p_ctxq, QNode in H5.
  sepelim H5.
  destruct A2 as [Ha1 [Ha2 [Ha3 Ha4]]].
  subst tlv0 cthv0.
  rewrite Ha2 in H5; rewrite <- Hb in H5.
  sconjlifth 6 H5.
  hvinject 2 2.
  subst c4.
  hvinject 6 6.
  subst c5; subst c6.
  sconjlifth 6 H; sconjlifth 6 H5.
  hvinject 1 1.
  subst pcx0.
  sconjlifth 4 H; sconjlifth 4 H5.
  generalize (eq_Q H H5).
  intros.
  subst Q0.
  split; trivial.
  exists Q'.
  unfold save_regs in He4.
  destruct He4 as [Hs1 [Hs2 Hs3]].
  repeat (split; trivial).
  rewrite <- Hs1. simplst. trivial.
  intro.
  rewrite H4 in H5.
  assert (h
       ||- Stmp |-> c1 ** Stmp |-> pcx ** TQEnd hdv tlv Q **
           ( Shd |-> c2 ** Stl |-> c3) **
           ((((A1 ** cth |-> Stmp) ** Stmp + 1 |-> 0) **
            tq |-> hdv) ** tq + 1 |-> tlv)).
  simplsep 9 9.
  exact (cont_cont_false H6).
  assert (h
       ||- TQEnd hdv tlv Q **
           Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 **
           ((((A1 ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
            tq |-> hdv) ** tq + 1 |-> tlv).
  simplsep 9 9.
  exact (tl_neq_other H4 Hd).
  rewrite <- Hs2. simplst. trivial.
  intro.
  rewrite H4 in H5.
  simplst.
  assert (h
       ||- Shd |-> c2 ** Shd |-> pcx ** TQEnd hdv tlv Q **
           (Stmp |-> c1 ** Stl |-> c3) **
           ((((A1 ** cth |-> Shd) ** Shd + 1 |-> 0) **
            tq |-> hdv) ** tq + 1 |-> tlv)).
  simplsep 9 9.
  exact (cont_cont_false H6).
  assert (h
       ||- TQEnd hdv tlv Q **
         Shd |-> c2 ** Stmp |-> c1 ** Stl |-> c3 **
           ((((A1 ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
             tq |-> hdv) ** tq + 1 |-> tlv).
  simplsep 9 9.
  exact (tl_neq_other H4 Hd).
  rewrite <- Hs3. simplst. trivial.
  intro.
  rewrite H4 in H5.
  simplst.
  assert (h
       ||- Stl |-> c3 ** Stl |-> pcx ** TQEnd hdv tlv Q **
           (Stmp |-> c1 ** Shd |-> c2) **
           ((((A1 ** cth |-> Stl) ** Stl + 1 |-> 0) **
            tq |-> hdv) ** tq + 1 |-> tlv)).
  simplsep 9 9.
  exact (cont_cont_false H6).
  assert (h
       ||- TQEnd hdv tlv Q **
         Stl |-> c3 ** Stmp |-> c1 ** Shd |-> c2 **
           ((((A1 ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
             tq |-> hdv) ** tq + 1 |-> tlv).
  simplsep 9 9.
  exact (tl_neq_other H4 Hd).
  unfold use_regs. unfold use_regs in He5.
  simplst.
  rewrite <- He5.
  apply extensionality.
  repeat (destruct r; simplst; trivial).
  destruct H4 as [c4 [c5 [c6 [pcx0 [cthv0 [hdv0 [tlv0 H5]]]]]]].
  unfold p_ctxq, QNode in H5.
  sepelim H5.
  destruct A2 as [He1 [He2 [He3 He4]]].
  subst cthv0 tlv0.
  sconjlifth 6 H5.
  hvinject 2 2.
  subst c4.
  hvinject 7 7.
  subst c5 c6.
  sconjlifth 6 H; sconjlifth 6 H5.
  hvinject 1 1.
  subst pcx0.
  sconjlifth 4 H; sconjlifth 4 H5.
  rewrite He2 in H5; rewrite <- Hb in H5.
  generalize (eq_Q H H5).
  intros.
  subst Q0.
  assert (exists Q', deQ (enQ Q (Regfile.get rf ra)) = Some Q').
  destruct Q.
  tauto.
  unfold deQ, enQ, Queue.del, Queue.add.
  exists  ((fix add (a0 : QueueElem.A) (l : Queue.list) {struct l} :
           Queue.list :=
           match l with
           | Queue.nil => Queue.cons a0 Queue.nil
           | Queue.cons a' l' => Queue.cons a' (add a0 l')
           end) (Regfile.get rf ra) Q).
  auto.
  destruct H4 as [Q' H4].
  unfold deQ, enQ in H4.
  generalize (deQ_enQ_t Hd H4).
  intros.
  exists Q'; exists c1; exists c2; exists c3; 
    exists (Regfile.get rf ra); exists cthv; exists hdv; exists cthv;
      exists cthv; exists hdv; exists tlv.
  unfold p_ctxq, QNode.
  simplst.
  rewrite <- Ha; rewrite <- Hb.
  sconjlifth 6 H5.
  generalize (cont_update1 H5 (Regfile.get rf ra)).
  intros.
  assert (Heap.set (Heap.set h cthv (Regfile.get rf ra)) (tlv + 1) cthv
   ||- TQEnd hdv cthv (enQ Q (Regfile.get rf ra)) **
      tq |-> hdv ** tq + 1 |-> tlv ** A1 **
       !(cthv = cthv /\
         hdv = hdv /\
         cthv = cthv /\
         enQ Q (Regfile.get rf ra) <> Queue.nil /\
         deQ (enQ Q (Regfile.get rf ra)) = Some Q' /\
         Q' <> Queue.nil /\
         cthv = cthv /\
         Some (Regfile.get rf ra) = Some (Regfile.get rf ra) /\
         Heap.get h (cthv + 1) = Some 0) **
       cth |-> cthv **
       !(tq > 0) **
       Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3).
  unfold enQ.
  apply add_Q.
  unfold QNode.
  simplsep 10 10.
  repeat (split; trivial).
  destruct Q.
  tauto.
  unfold Queue.add.
  discriminate.
  sconjlifth 1 H5.
  exact (cont_get_eq1 H5).
  trivial.
  simplsep 10 10.
  repeat (split; trivial).
  destruct Q.
  tauto.
  unfold Queue.add.
  discriminate.
  sconjlifth 1 H5.
  exact (cont_get_eq1 H5).
  sepelim H8.
  trivial.
  omega.
  assert (h ||- TQEnd hdv tlv Q ** cthv + 1 |-> 0 **
           (Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3) **
           ((((A1 ** cth |-> cthv) ** cthv |-> pcx ** 
            tq |-> hdv) ** tq + 1 |-> tlv))).
  simplsep 8 8.
  exact (tl_neq_other H7 Hd).
  assert (h ||- TQEnd hdv tlv Q ** cthv |-> pcx ** 
           (Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3) **
           ((((A1 ** cth |-> cthv) ** cthv + 1 |-> 0 ** 
            tq |-> hdv) ** tq + 1 |-> tlv))).
  simplsep 8 8.
  exact (tl_neq_other H7 Hd).
  rewrite <- Hc.
  unfold p_ctxq in H.
  assert (h ||- TQEnd hdv tlv Q ** 
    cthv |-> pcx **
      (cthv + 1 |-> 0 **
        tq |-> hdv **
          tq + 1 |-> tlv **
            Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3) **
      top ** cth |-> cthv).
  simplsep 8 8.
  generalize (tl_neq_other H2 Hd).
  intros.
  auto.
Qed.

Theorem readptr_ok : wfis psi readptr_pre readptr_guar readptr_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct s as [h rf pc].
  simplst.
  simpl.
  destruct H as [Q [pcx [cthv [hdv [tlv H]]]]].
  unfold p_ctxq, QNode in H.
  sepelim H.
  sconjlifth 4 H.
  generalize (cont_get_eq1 H).
  intros Heq1.
  rewrite Heq1.
  simplst.
  simpl.
  sconjlifth 1 H.
  generalize (cont_get_eq1 H).
  intros Heq2.
  rewrite Heq2.
  simplst.
  simpl.
  sconjlifth 6 H.
  generalize (cont_get_eq1 H).
  intros Heq3.
  rewrite Heq3.
  simplst.
  assert (cthv + 0 = cthv).
  omega.
  rewrite H0.
  clear H0.
  sconjlifth 1 H.
  generalize (cont_get_eq1 H).
  intros Heq4.
  rewrite Heq4.
  simplst.
  compare hdv 0.
  intros.
  rewrite e.
  simpl.
  rewrite e in H.
  unfold TQEnd in H.
  assert (Q = Queue.nil).
  destruct Q.
  trivial.
  unfold QNode in H.
  fold TQEnd in H.
  destruct Q.
  sepelim H.
  apply False_ind.
  omega.
  sepelim H.
  apply False_ind.
  omega.
  rewrite H0 in H.
  split.
  exists Q; exists (Regfile.get rf tmp);
    exists (Regfile.get rf hd); exists (Regfile.get rf tl);
      exists (Regfile.get rf ra); exists cthv; exists hdv; exists tlv;
        exists cthv; exists hdv; exists tlv.
  rewrite e; rewrite H0.
  unfold p_ctxq, QNode, TQEnd.
  sconjlift 4.
  apply cont_update1 with (v1 := pcx).
  simplsep 10 10.
  sepelim H.
  trivial.
  sepelim H.
  trivial.
  intros.
  hypreplace H1 (H1 Q0 A1).
  elim H1.
  intros.
  exists (Regfile.get rf ra).
  destruct H2 as [pcx0 [cthv0 [hdv0 [tlv0 H2]]]].
  unfold p_ctxq, QNode in H2.
  sepelim H2.
  sconjlifth 1 H2.
  sconjlifth 9 H.
  hvinject 1 1.
  subst cthv0.
  sepelim H.
  hvinject 7 7.
  subst pcx0 hdv0.
  assert (Q0 = Queue.nil).
  destruct Q0.
  trivial.
  unfold TQEnd in H2.
  destruct Q0.
  unfold QNode in H2.
  sepelim H2.
  apply False_ind.
  omega.
  unfold QNode in H2.
  sepelim H2.
  apply False_ind.
  omega.
  subst Q0.
  unfold getQ, enQ, deQ, Queue.get, Queue.del, Queue.add.
  split; trivial.
  exists Queue.nil.
  split; trivial.
  destruct s' as [h' rf' pc'].
  sconjlifth 1 H.
  generalize (cont_get_eq1 H).
  intros.
  destruct H3 as [Ha1 [Ha2 [Ha3 [Ha4 Ha5]]]].
  rewrite Ha2 in Ha1.
  split; trivial.
  split.
  unfold save_regs.
  unfold save_regs in Ha3.
  destruct Ha3 as [He1 [He2 He3]].
  split. rewrite <- He1; simplst; trivial.
  intro.
  assert (h ||- cthv |-> pcx **
    Stmp |-> Regfile.get rf tmp **      
    (cthv + 1 |-> 0 **
           tq |-> 0 **
           tq + 1 |-> tlv **
           
           Shd |-> Regfile.get rf hd ** Stl |-> Regfile.get rf tl ** top) **
          cth |-> cthv).
  simplsep 7 7.
  rewrite H3 in H5.
  exact (cont_cont_false H5).
  split. rewrite <- He2; simplst; trivial.
  intro.
  assert (h ||- cthv |-> pcx **
    Shd |-> Regfile.get rf hd **
      Stmp |-> Regfile.get rf tmp **      
        (cthv + 1 |-> 0 **
          tq |-> 0 **
            tq + 1 |-> tlv **
            Stl |-> Regfile.get rf tl ** top) **
          cth |-> cthv).
  simplsep 7 7.
  rewrite H3 in H5.
  exact (cont_cont_false H5).
  rewrite <- He3; simplst; trivial.
  intro.
  assert (h ||- cthv |-> pcx **
    Stl |-> Regfile.get rf tl ** top **
    Shd |-> Regfile.get rf hd **
      Stmp |-> Regfile.get rf tmp **      
        (cthv + 1 |-> 0 **
          tq |-> 0 **
            tq + 1 |-> tlv **
          cth |-> cthv)).
  simplsep 8 8.
  rewrite H3 in H5.
  exact (cont_cont_false H5).
  split.
  unfold use_regs; unfold use_regs in Ha4.
  simplst.
  rewrite <- Ha4.
  apply extensionality.
  repeat (destruct r; simplst; trivial).
  rewrite <- Ha2.
  trivial.
  destruct H2 as [pcx0 [cthv0 [hdv0 [tlv0 H2]]]].
  unfold p_ctxq, QNode in H2.
  sepelim H2.
  sconjlifth 1 H2.
  sconjlifth 9 H.
  hvinject 1 1.
  subst cthv0.
  hvinject 7 7.
  subst pcx0.
  hvinject 7 7.
  subst hdv0 tlv0.
  assert (Q0 = Queue.nil).
  unfold TQEnd in H2.
  destruct Q0.
  trivial.
  unfold QNode in H2.
  destruct Q0.
  sepelim H2.
  apply False_ind.
  omega.
  sepelim H2.
  apply False_ind.
  omega.
  exists (Regfile.get rf tmp); exists (Regfile.get rf hd);
    exists (Regfile.get rf tl); exists (Regfile.get rf ra);
      exists cthv; exists hdv; exists tlv;
        exists cthv; exists hdv; exists tlv.
  unfold p_ctxq, QNode.
  subst hdv.
  sconjlift 4.
  apply cont_update1 with (v1 := pcx).
  generalize H; clear H; intro H.
  simplsep 7 7.
  intros.
  generalize (neq_beq_false hdv 0 n).
  intros.
  rewrite H0.
  assert (Q <> Queue.nil).
  destruct Q.
  unfold TQEnd in H.
  sepelim H.
  destruct A1 as [Ha Hb].
  apply False_ind.
  omega.
  discriminate.
  split.
  exists Q; exists (Regfile.get rf tmp); exists (Regfile.get rf hd);
    exists (Regfile.get rf tl); exists (Regfile.get rf ra).
  exists cthv; exists hdv; exists tlv.
  unfold p_ctxq, QNode.
  sconjlift 4.
  apply cont_update1 with (v1 := pcx).
  simplsep 7 7.
  intros.
  hypreplace H2 (H2 Q0 A1).
  elim H2.
  intros.
  exists x.
  destruct H4 as [He1 [Q' [He2 [He3 [He4 [He5 He6]]]]]].
  repeat (split; trivial).
  exists Q'.
  destruct H3 as [pcx0 [cthv0 [hdv0 [tlv0 H3]]]].
  unfold p_ctxq, QNode in H3.
  sconjlifth 1 H3.
  sconjlifth 9 H.
  hvinject 1 1.
  subst cthv0.
  sepelim H3.
  hvinject 3 3.
  subst pcx0.
  hvinject 7 7.
  subst hdv0 tlv0.
  sconjlifth 5 H; sconjlifth 5 H3.
  generalize (eq_Q H H3).
  intros.
  subst Q0.
  destruct s' as [h' rf' pc'].
  unfold save_regs in He4.
  destruct He4 as [Hs1 [Hs2 Hs3]].
  repeat (split; trivial).
  rewrite <- Hs1.
  simplst.
  trivial.
  assert (h ||- cthv |-> pcx ** Stmp |-> Regfile.get rf tmp **
    TQEnd hdv tlv Q **
    Shd |-> Regfile.get rf hd ** Stl |-> Regfile.get rf tl ** A1 **
  (((cth |-> cthv ** cthv + 1 |-> 0) ** tq |-> hdv) **
    tq + 1 |-> tlv)).
  simplsep 9 9.
  intro.
  rewrite H5 in H4.
  exact (cont_cont_false H4).
  rewrite <- Hs2.
  simplst.
  trivial.
  assert (h ||- cthv |-> pcx ** Shd |-> Regfile.get rf hd **
    Stmp |-> Regfile.get rf tmp **
    TQEnd hdv tlv Q **
    Stl |-> Regfile.get rf tl ** A1 **
  (((cth |-> cthv ** cthv + 1 |-> 0) ** tq |-> hdv) **
    tq + 1 |-> tlv)).
  simplsep 9 9.
  intro.
  rewrite H5 in H4.
  exact (cont_cont_false H4).
  rewrite <- Hs3.
  simplst.
  trivial.
  assert (h ||- cthv |-> pcx ** Stl |-> Regfile.get rf tl **
    Shd |-> Regfile.get rf hd **
    Stmp |-> Regfile.get rf tmp **
    TQEnd hdv tlv Q **
     A1 **
  (((cth |-> cthv ** cthv + 1 |-> 0) ** tq |-> hdv) **
    tq + 1 |-> tlv)).
  simplsep 9 9.
  intro.
  rewrite H5 in H4.
  exact (cont_cont_false H4).
  unfold use_regs; unfold use_regs in He5.
  rewrite <- He5.
  simplst.
  apply extensionality.
  repeat (destruct r; simplst; trivial).
  destruct H3 as [pcx0 [cthv0 [hdv0 [tlv0 H3]]]].
  unfold p_ctxq, QNode in H3.
  sepelim H3.
  sconjlifth 1 H3.
  sconjlifth 9 H.
  hvinject 1 1.
  subst cthv0.
  hvinject 7 7.
  subst pcx0.
  hvinject 7 7.
  subst hdv0 tlv0.
  exists (Regfile.get rf tmp); exists (Regfile.get rf hd);
    exists (Regfile.get rf tl); exists (Regfile.get rf ra);
      exists cthv; exists hdv; exists tlv.
  unfold p_ctxq, QNode.
  sconjlift 4.
  apply cont_update1 with (v1 := pcx).
  generalize H; clear H; intro H.
  simplsep 7 7.
  repeat (split; trivial).
  sconjlifth 4 H; sconjlifth 5 H3.
  generalize (eq_Q H H3).
  intros.
  subst Q0; trivial.
Qed.

Theorem schedule_ok: wfis psi schedule_pre schedule_guar schedule_code.
Proof.
  simplvc.
  unfoldpred.
  intros.
  destruct s as [h rf pc].
  simplst.
  simpl.
  unfold p_ctxqueue in H.
  destruct H as [Q H].
  sepelim H.
  generalize H; clear H.
  generalize x; clear x; intro c1.
  generalize x0; clear x0; intro c2.
  generalize x1; clear x1; intro c3.
  generalize x2; clear x2; intro pcx.
  generalize x3; clear x3; intro cthv.
  generalize x4; clear x4; intro hdv.
  generalize x5; clear x5; intro tlv.
  intros.
  unfold p_ctxq, QNode in H.
  sepelim H.
  sconjlifth 7 H.
  generalize (cont_get_eq1 H).
  intro Heq1.
  rewrite Heq1.
  simplst.
  simpl.
  sconjlifth 1 H.
  generalize (cont_get_eq1 H).
  intro Heq2.
  rewrite Heq2.
  simplst.
  simpl.
  sconjlifth 1 H.
  generalize (cont_get_eq1 H).
  intro Heq3.
  rewrite Heq3.
  split.
  exists Q; exists pcx; exists cthv; exists hdv; exists tlv.
  unfold p_ctxq, QNode.
  sconjlift 11.
  apply cont_update1 with (v1 := c3).
  sconjlift 11.
  apply cont_update1 with (v1 := c2).
  sconjlift 11.
  apply cont_update1 with (v1 := c1).
  simplsep 9 9.
  intros.
  hypreplace H0 (H0 Q0 A1).
  elim H0.
  intros.
  exists x.
  destruct H2 as [Ha [Q' [Hb [Hc [Hd [He Hf]]]]]].
  split; trivial.
  exists Q'.
  repeat (split; trivial).
  destruct s' as [h' rf' pc'].
  unfold save_regs in Hd.
  unfold use_regs in He.
  apply extensionality.
  simplst.
  intros.
  assert 
(Regfile.get (set (set (set (set rf ra 0) tmp 0) hd 0) tl 0) r
    = Regfile.get (set (set (set (set rf' ra 0) tmp 0) hd 0) tl 0) r).
  rewrite He. trivial.
  repeat (destruct r; simplst; trivial).
  destruct Hd as [He1 [He2 He3]].
  assert (Some (Regfile.get rf tmp) = Some (Regfile.get rf' tmp)).
  rewrite <- He1.
  simplst.
  trivial.
  assert (h ||- Shd |-> c2 ** Stmp |-> c1 ** Stl |-> c3 **
    (((((((top ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
      tq |-> hdv) ** tq + 1 |-> tlv) ** TQEnd hdv tlv Q))).
  simplsep 9 9.
  intro.
  rewrite H4 in H3. exact (cont_cont_false H3).
  assert (h ||- Stl |-> c3 ** Stmp |-> c1 ** Shd |-> c2 ** 
    (((((((top ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
      tq |-> hdv) ** tq + 1 |-> tlv) ** TQEnd hdv tlv Q))).
  simplsep 9 9.
  intro.
  rewrite H4 in H3. exact (cont_cont_false H3).
  injection H3; intros; trivial.
  destruct Hd as [He1 [He2 He3]].
  assert (Some (Regfile.get rf hd) = Some (Regfile.get rf' hd)).
  rewrite <- He2.
  simplst.
  trivial.
  assert (h ||- Stl |-> c3 ** Shd |-> c2 ** Stmp |-> c1 ** 
    (((((((top ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
      tq |-> hdv) ** tq + 1 |-> tlv) ** TQEnd hdv tlv Q))).
  simplsep 9 9.
  intro.
  rewrite H4 in H3. exact (cont_cont_false H3).
  injection H3; intros; trivial.
  destruct Hd as [He1 [He2 He3]].
  assert (Some (Regfile.get rf tl) = Some (Regfile.get rf' tl)).
  rewrite <- He3.
  simplst.
  trivial.
  injection H3; intros; trivial.
  unfold schedule_post in Hf.
  sepelim Hf.
  unfold rget in A2.
  trivial.
  exists pcx; exists cthv; exists hdv; exists tlv.
  unfold p_ctxq, QNode.
  sconjlift 11.
  apply cont_update1 with (v1 := c3).
  sconjlift 11.
  apply cont_update1 with (v1 := c2).
  sconjlift 11.
  apply cont_update1 with (v1 := c1).
  unfold p_ctxqueue in H1.
  unfold p_ctxq, QNode in H1.
  sepelim H1.
  sconjlifth 1 H1.
  sconjlifth 2 H.
  hvinject 1 1.
  subst x3.
  hvinject 7 7.
  subst x2 x0 x1.
  hvinject 7 7.
  subst x4 x5.
  sconjlifth 5 H; sconjlifth 5 H1.
  generalize (eq_Q H H1).
  intros; subst Q0.
  hvinject 7 7.
  subst x.
  clear H.
  simplsep 7 7.
  simpl.
  assert (h ||- Stmp |-> c1 ** Stl |-> c3 ** Shd |-> c2 **
          ((((((top ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
             tq |-> hdv) ** tq + 1 |-> tlv) ** TQEnd hdv tlv Q)).
  simplsep 7 7.
  intro.
  rewrite H1 in H0.
  exact (cont_cont_false H0).
  simpl.
  assert (h ||- Shd |-> c2 ** Stl |-> c3 ** Stmp |-> c1 **
          ((((((top ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
             tq |-> hdv) ** tq + 1 |-> tlv) ** TQEnd hdv tlv Q)).
  simplsep 7 7.
  intro.
  rewrite H1 in H0.
  exact (cont_cont_false H0).
  simpl.
  assert (h ||- Stmp |-> c1 ** Shd |-> c2 ** Stl |-> c3 ** 
          ((((((top ** cth |-> cthv) ** cthv |-> pcx) ** cthv + 1 |-> 0) **
             tq |-> hdv) ** tq + 1 |-> tlv) ** TQEnd hdv tlv Q)).
  simplsep 7 7.
  intro.
  rewrite H1 in H0.
  exact (cont_cont_false H0).
Qed.

Theorem C_ok : wfch psi psi C.
Proof.
  prove_wfch.
  apply schedule_ok.
  apply readptr_ok.
  apply enqueue_ok.
  apply dequeue_ok.
  apply switch_ok.
  apply writeptr_ok.
  apply sret_ok.
Qed.
