(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Map                                                         *)
(*    using listT                                              *)
(*                                                             *)
(***************************************************************)

(* $Id: ListB_MULMAP.v,v 1.1.1.1 2006/12/07 16:43:24 ssg05 Exp $ *)

Require Import Coq.Bool.Bool.
Require Import ListT.
Require Import prodT.
Require Import bnat.

Require Export MULMAP.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Open Local Scope listt_scope.
Open Local Scope t_type_scope.

Module MultiMap (BDef : ONE_TYPE) <: MAPDEF.

Definition A := nat.

Definition A_eq_dec := nat_eq_dec.

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := listT (prodT A B).

Definition emp_map : map := nil.

Lemma beq_B (b b' : B) :=
  match B_eq_dec b b' with
    | left _ => 
    | 
Fixpoint min (m : map) (a : A) (b : B) : bool := 
  match m with 
    | nil => false
    | (a', b') :: m' => orb (beq_nat a a') (beq_B b b') 

Fixpoint In (a:A) (l:listT) {struct l} : Prop :=
  match l with
  | nil => False
  | b :: m => b = a \/ In a m
  end.

Definition sig_map (a : A) (b : B) := (a, b) :: nil.

Definition merge (m1 m2 : map) : map := 
  app m1 m2.

Definition lookup (m1 : map) (a : A) (b : B) :=
  In (a, b) m1.

Definition indom (m : map) (a : A) :=
  exists b, lookup m a b.

Definition sub (m1 m2 : map) :=
  forall (a : A) (b : B), 
    lookup m1 a b -> lookup m2 a b.

Lemma sub_refl : forall m, sub m m.
Proof.
  unfold sub; auto.
Qed.

Lemma sub_trans : forall m1 m2 m3,
  sub m1 m2 -> sub m2 m3 -> sub m1 m3.
Proof.
  unfold sub, lookup; intros m1 m2 m3 H12 H23.
  intros a b.
  assert (H := (@incl_tran _ m1 m2 m3)).
  unfold incl in H.
  generalize (a, b).
  apply H.
  intros a0.
  destruct a0.
  apply H12; trivial.
  intros a0.
  destruct a0.
  apply H23; trivial.
Qed.

Lemma lookup_merge_elim : 
  forall (m1 m2 : map) (a : A) (b : B),
    lookup (merge m1 m2) a b 
    -> lookup m1 a b \/ lookup m2 a b.
Proof.
  unfold lookup, merge.
  intros m1 m2 a b H.
  apply in_app_or; trivial.
Qed.
  
Lemma lookup_merge_intro :
  forall (m1 m2 : map) (a : A) (b : B), 
    lookup m1 a b \/ lookup m2 a b
    -> lookup (merge m1 m2) a b.
Proof.
  unfold lookup, merge; intros.
  apply in_or_app; trivial.
Qed.

Lemma sub_merge_elim_l :
  forall (m1 m2 m3: map),
    sub (merge m1 m2) m3
    -> sub m1 m3.
Proof.
  unfold sub.
  intros m1 m2 m3 Hsub a b Hin1.
  apply Hsub.
  apply lookup_merge_intro.
  left; trivial.
Qed.
Implicit Arguments sub_merge_elim_l [m1 m2 m3].

Lemma sub_merge_elim_r :
  forall (m1 m2 m3: map),
    sub (merge m1 m2) m3
    -> sub m2 m3.
Proof.
  unfold sub.
  intros m1 m2 m3 Hsub a b Hin2.
  apply Hsub.
  apply lookup_merge_intro.
  right; trivial.
Qed.
Implicit Arguments sub_merge_elim_r [m1 m2 m3].

Lemma sub_merge_l :
  forall (m1 m2 : map),
    sub m1 (merge m1 m2).
Proof.
  intros m1 m2.
  unfold sub, merge, lookup.
  intros a b H.
  apply in_or_app.
  left; trivial.
Qed.

Lemma sub_merge_r :
  forall (m1 m2 : map),
    sub m2 (merge m1 m2).
Proof.
  intros m1 m2.
  unfold sub, merge, lookup.
  intros a b H.
  apply in_or_app.
  right; trivial.
Qed.

Lemma lookup_sub_lookup :
  forall (m1 m2 : map) (a : A) (b : B),
    lookup m1 a b 
    -> sub m1 m2 
    -> lookup m2 a b.
Proof.
  unfold sub; intros; auto.
Qed.

End MultiMap.

Unset Implicit Arguments.
