(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* MAP Interface                                               *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: MULMAP.v,v 1.1.1.1 2006/12/07 16:43:24 ssg05 Exp $ *)

Require Import Coq.Bool.Bool.
Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Module Type ONE_TYPE.

  Parameter B : Type.

End ONE_TYPE.

Module Type MAPDEF.

  Parameter A : Set.

(*   Parameter beq_A : A -> A -> bool. *)

(*   Axiom beq_true_eq : forall a a' : A,  *)
(*     beq_A a a' = true -> a = a'. *)

(*   Axiom beq_false_neq : forall a a' : A, *)
(*     beq_A a a' = false -> a <> a'. *)

(*   Axiom eq_beq_true : forall a a' : A,  *)
(*     a = a' -> beq_A a a' = true. *)

(*   Axiom neq_beq_false : forall a a' : A,  *)
(*     a <> a' -> beq_A a a' = false. *)
  
  Axiom A_eq_dec : forall a a' : A, {a=a'}+{a<>a'}.

  Parameter B : Type.

(***********************)
(** map                *)
(***********************)

  Parameter map : Type.
  
  Parameter emp_map : map.

(***********************)
(** lookup             *)
(***********************)

  Parameter min : map -> A -> B -> bool.

  Axiom emp_sem : forall (a : A) (b : B), 
    min emp_map a b = false.
  
(************************)
(** merge               *)
(************************)

  Parameter merge : map -> map -> map.

  Axiom merge_sem : forall (m1 m2 : map) (a : A) (b : B),
    min (merge m1 m2) a b
    = match min m1 a b, min m2 a b with
        | true, true => true
        | true, false => true
        | false, true => true
        | false, false => false
      end.

End MAPDEF.

Module MapEx (MapDef : MAPDEF).

Import MapDef.

Definition lookup (m : map) (a : A) (b : B) : Prop :=
  min m a b = true.

Definition indom (m : map) (a : A) :=
  exists b, min m a b = true.

Definition sub (m1 m2 : map) :=
  forall (a : A) (b : B), 
    min m1 a b = true 
    -> min m2 a b = true.

Definition meq (m1 m2 : map) :=
  sub m1 m2 /\ sub m2 m1.

Definition disj (m1 m2 : map) : Prop :=
  forall (a : A) (b : B), 
    match min m1 a b, min m2 a b with
      | true, false => True
      | false, true => True
      | false, false => True
      | _ , _ => False
    end.

Definition join (m1 m2 m: map) :=
  forall (a : A) (b : B),
    match min m1 a b, min m2 a b, min m a b with
      | false, true, true => True
      | true, false, true => True
      | false, false, false => True
      | _, _, _ => False
    end.

(***********************************************************)
(**                                                        *)
(**    Lemmas                                              *)
(**                                                        *)
(**    Tactics                                             *)
(**                                                        *)
(***********************************************************)

Lemma sub_refl : forall m, sub m m.
Proof.
  unfold sub; auto.
Qed.

Lemma sub_trans : forall m1 m2 m3,
  sub m1 m2 -> sub m2 m3 -> sub m1 m3.
Proof.
  unfold sub, lookup; intros m1 m2 m3 H12 H23.
  intros a b H1.
  apply H23; apply H12; trivial.
Qed.

Lemma lookup_2_min : forall m a b,
  lookup m a b = (min m a b = true).
Proof.
  intros; auto.
Qed.

Lemma lookup_min : forall m a b,
  min m a b = true -> lookup m a b.
Proof.
  intros; auto.
Qed.

Lemma min_lookup : forall m a b,
  lookup m a b -> min m a b = true.
Proof.
  intros; auto.
Qed.

Ltac minlize :=
  match goal with
    | |- context[lookup ?m ?a ?b] =>
      unfold lookup; minlize
    | H : context[lookup ?m ?a ?b] |- _ =>
      unfold lookup in H; minlize

    | |- context[sub ?m1 ?m2] =>
      unfold sub; minlize
    | H : context[sub ?m1 ?m2] |- _ =>
      unfold sub in H; minlize

    | |- context[min (merge ?m1 ?m2) ?a ?b] =>
      rewrite (merge_sem m1 m2 a b); minlize
    | H : context[min (merge ?m1 ?m2) ?a ?b] |- _ =>
      rewrite (merge_sem m1 m2 a b) in H; minlize

    | |- context[min emp_map ?a ?b] =>
      rewrite (emp_sem a b); minlize
    | H : context[min emp_map ?a ?b] |- _ =>
      rewrite (emp_sem a b) in H; minlize

    | _ => idtac
  end.

Lemma lookup_merge_elim : 
  forall (m1 m2 : map) (a : A) (b : B),
    lookup (merge m1 m2) a b 
    -> lookup m1 a b \/ lookup m2 a b.
Proof.
  intros; minlize.
  destruct (min m1 a b); destruct (min m2 a b); tauto.
Qed.
  
Lemma lookup_merge_intro :
  forall (m1 m2 : map) (a : A) (b : B), 
    lookup m1 a b \/ lookup m2 a b
    -> lookup (merge m1 m2) a b.
Proof.
  intros; minlize.
  destruct (min m1 a b); destruct (min m2 a b); tauto.
Qed.

Lemma sub_merge_elim_l :
  forall (m1 m2 m3: map),
    sub (merge m1 m2) m3
    -> sub m1 m3.
Proof.
  intros; minlize; intros.
  substH H with (H a b).
  minlize.
  destruct (min m1 a b); 
    destruct (min m2 a b); 
      destruct (min m3 a b); try tauto.
Qed.
Implicit Arguments sub_merge_elim_l [m1 m2 m3].

Lemma sub_merge_elim_r :
  forall (m1 m2 m3: map),
    sub (merge m1 m2) m3
    -> sub m2 m3.
Proof.
  intros; minlize; intros.
  substH H with (H a b).
  minlize.
  destruct (min m1 a b); 
    destruct (min m2 a b); 
      destruct (min m3 a b); try tauto.
Qed.
Implicit Arguments sub_merge_elim_r [m1 m2 m3].

Lemma sub_merge_l :
  forall (m1 m2 : map),
    sub m1 (merge m1 m2).
Proof.
  intros; minlize. intros; minlize.
  destruct (min m1 a b); 
      destruct (min m2 a b); try tauto.
Qed.

Lemma sub_merge_r :
  forall (m1 m2 : map),
    sub m2 (merge m1 m2).
Proof.
  intros; minlize. intros; minlize.
  destruct (min m1 a b); 
      destruct (min m2 a b); try tauto.
Qed.

Lemma lookup_sub_lookup :
  forall (m1 m2 : map) (a : A) (b : B),
    lookup m1 a b 
    -> sub m1 m2 
    -> lookup m2 a b.
Proof.
  intros; minlize. 
  substH H0 with (H0 a b). 
  destruct (min m1 a b); 
      destruct (min m2 a b); try tauto.
Qed.

End MapEx.
