(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Tiny Machine                                                *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: Regfile.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Word.
Require Import Reg.

Definition regfile := reg -> word.

Definition get (rf : regfile) r := 
  if beq_reg r0 r then 0 else (rf r).

Definition set (rf : regfile) r w := 
  fun (r' : reg) => 
    if beq_reg r' r then w else get rf r'.

Implicit Types rf : regfile.

Lemma set_get_eq : 
  forall rf r w, beq_reg r zero = false
  -> get (set rf r w) r = w.
Proof.
  intros.
  unfold get, set.
  rewrite (beq_req_true r r (refl_equal _)).
  rewrite (beq_sym r zero) in H.
  rewrite H.
  trivial.
Qed.

Lemma get_zero : forall rf, get rf r0 = 0.
Proof.
  simpl.
  unfold get.
  rewrite (beq_req_true r0 r0 (refl_equal _)).
  trivial.
Qed.

Lemma get_zero_2 : forall r rf, 
  beq_reg r r0 = true -> get rf r = 0.
Proof.
  unfold get.
  intros.
  rewrite (beq_sym zero r).
  rewrite H.
  trivial.
Qed.

Lemma set_get_neq : forall rf r r' w,
  beq_reg r r' = false 
  -> get (set rf r' w) r 
  = get rf r.
Proof.
  intros.
  unfold set, get.
  decbeqreg r0 r; trivial.
Qed.

Axiom extensionality : 
  forall rf rf',
    (forall r, get rf r = get rf' r)
    -> rf = rf'.

Lemma set_set_eq : forall rf r r' w w',
  beq_reg r r' = true 
  -> set (set rf r' w') r w
  = set rf r w.
Proof.
  intros.
  simplbreg.
  apply extensionality.
  intro r.
  decbeqreg r r'; trivial.
  decbeqreg r' r0; trivial.
  rewrite set_get_eq; trivial.
  rewrite set_get_eq; trivial.
  rewrite set_get_neq; trivial.
  rewrite set_get_neq; trivial. 
  rewrite set_get_neq; trivial.
Qed.

Ltac simplgs :=
  match goal with
    | [ |- context [(get ?rf r0)] ] =>
      rewrite get_zero; simplgs
    | [ H : context [(get ?rf r0)] |- _ ] =>
      rewrite get_zero in H; simplgs

    | [ H : beq_reg ?r r0 = true
      |- context [(get ?rf ?r)] ] =>
      rewrite (get_zero_2 r rf H); simplgs

    | [ H : beq_reg ?r r0 = false
      |- context [(get (set ?rf ?r ?v) ?r)] ] =>
      rewrite (set_get_eq rf r v H); simplgs

    | [ |- context [(get (set ?s ?r ?v) ?r)] ] =>
      rewrite set_get_eq; [idtac | solvebreg]; simplgs
    | [ H : context [(get (set ?s ?r ?v) ?r)] |- _ ] =>
      rewrite set_get_eq in H; [idtac | solvebreg]; simplgs

    | [ |- context [(get (set ?s ?r2 ?v) ?r1)] ] =>
      rewrite set_get_neq; [idtac | solvebreg]; simplgs
    | [ H: context [(get (set ?s ?r2 ?v) ?r1)] |- _ ] =>
      rewrite set_get_neq in H; [idtac | solvebreg]; simplgs

    | [ |- context [(set (set ?s ?r2 ?v2) ?r1 ?v1)] ] =>
      rewrite set_set_eq; [idtac | solvebreg]; simplgs
    | [ H: context [(set (set ?s ?r2 ?v2) ?r1 ?v1)] |- _ ] =>
      rewrite set_set_eq in H; [idtac | solvebreg]; simplgs

    | _ => idtac
  end.
