(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Inter Operation                                             *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: link.v,v 1.1.1.1 2006/12/07 16:43:18 ssg05 Exp $ *)

Require Import optionT.
Require Import prodT.

Require Import bnat.
Require Import machine.
Require Import Talp.
Require Import scap.
Require Import ocap.
Require Import scap_ocap.
Require Import talp_ocap.

Require Import scap_newpair.
Require Import talp_main.

Require Import cdsa.
Require Import seplogic.

Require Import ocap_2.
Require Import sep_talp.

Require Import extac.
Require Import deptac.

Open Local Scope ocap_scope.
Open Local Scope tm_scope.
Open Local Scope tal_scope.
Open Local Scope seplogic_scope.

(***************************************************************)

Definition C : codeheap := 
  CHeap.merge talp_main.C scap_newpair.C.

Definition rho_tal := 1.
Definition rho_scap := 2.

Definition open_r (psi : ocap.chspec) (s : state) := 
  exists fv, 
    s |||- FLIST |-> fv ** Good fv.

Lemma open_r_only_assert_heap :
    forall Psi H R R' pc pc',
      open_r Psi <.H, R, pc.>
      -> open_r Psi <.H, R', pc'.>.
Proof.
  unfold open_r.
  intros Psi H R R' pc pc' Hr.
  trivial.
Qed.

Module Tp <: TALPARAM.
  Definition rho_tal := rho_tal.
  Definition open_r := open_r.
  Definition open_r_only_assert_heap := 
    open_r_only_assert_heap.
End Tp.

Module TalM := TalOcap Tp.
Import TalM.

Definition D_tal : idict := fun id =>
  if (beq_nat id rho_tal) then SomeT <@ itptal @> else
    NoneT.

Module Sp <: SCAPPARAM.
  Definition rho_scap := rho_scap.
  Definition D0 := D_tal.
  Lemma rho_scap_not_indom_D0 :
    ~ DSpec.indom D0 rho_scap.
  Proof.
    intros [b H].
    unfold DSpec.get, DSpecDef.get, D0, D_tal in H.
    unfold rho_scap, rho_tal in H.
    simpl in H.
    discriminate H.
  Qed.
  
End Sp.

Module ScapM := ScapOcap Sp.
Import ScapM.

(**********************************************************************)

Definition D : idict := fun id =>
  if (beq_nat id rho_tal) then SomeT <@ itptal @> else
  if (beq_nat id rho_scap) then SomeT <@ itpscap @> else
    NoneT.

Lemma D_tal_sub_D : DSpec.sub D_tal D.
Proof.
  unfold DSpec.sub, DSpec.lookup, DSpec.get, DSpecDef.get.
  unfold D, D_tal.
  intros id rho H.
  destruct id.
  simpl in * .
  trivial.
  destruct id.
  simpl in * ; trivial.
  destruct id.
  simpl in * ; trivial.
  discriminate.
  simpl in * ; trivial.
Qed.

Lemma D_scap_sub_D : DSpec.sub D_scap D.
Proof.
  unfold DSpec.sub, DSpec.lookup, DSpec.get, DSpecDef.get.
  unfold D, D_scap.
  intros id rho H.
  destruct id.
  simpl in * .
  trivial.
  destruct id.
  simpl in * ; trivial.
  discriminate.
  destruct id.
  simpl in * ; trivial.
  discriminate.
Qed.

Lemma D0_plus_D_tal_sub_D : DSpec.sub (DSpec.merge D_tal D_scap) D.
Proof.
  apply DSpec.merge_sub_intro.
  apply D_tal_sub_D.
  apply D_scap_sub_D.
Qed.

(***************************************************************)

Theorem newpair_weakening :
  (itptal (mk_Ltal talp_main.newpair_spec))
  ===>
  (itpscap (mk_Lscap scap_newpair.newpair_pre scap_newpair.newpair_guar)).
Proof.
  unfold aimp.
  intros OPsi [h rf pc] Htal.
  unfold newpair_pre.
  destruct Htal as [h1 H1].
  destruct H1 as [h2 H1].
  destruct H1 as [Hh H1].
  destruct H1 as [H1 Hh2].
  simpl.
  destruct H1 as [talpsi [Hsink Hs]].

  assert (Hh' : exists fv : Heap.B, 
    h ||- FLIST |-> fv ** Good fv ** top).
    destruct Hh2 as [fv Hh2].
    exists fv.
    sconjlift 2.
    eapply sconj_intro.
    2: apply Hh2.
    2: apply Hh.
    apply top_intro.

(* tal_spec -> scap_spec_pre *)
  split_l.
(* scap_spec_guar -> tal_spec *)
  exists 0.
  simpl.
  unfold newpair_guar.
  intros [h' rf' pc'] Hg.
  destruct Hg as [Hra [Hsx Hg]].
  destruct Hsx as [Hs0 [Hs1 [Hs2 [Hs3 [Hs4 [Hs5 [Hs6 Hs7]]]]]]].
  destruct (Hg top Hh') as [fv Hpair].
  simpl in Hpair.
  simpl in Hra.
  simpl in Hs0.
  simpl in Hs1.
  simpl in Hs2.
  simpl in Hs3.
  simpl in Hs4.
  simpl in Hs5.
  simpl in Hs6.
  simpl in Hs7.
  assert (Hraok : regok 9
    [{(ast, int), 
      (s0, ^0),
      (s1, ^1),
      (s2, ^2),
      (s3, ^3),
      (s4, ^4),
      (s5, ^5),
      (s6, ^6),
      (s7, ^7),
      (ra,
        [0]- [{
          (ast, int), 
          (s0, ^0),
          (s1, ^1),
          (s2, ^2),
          (s3, ^3),
          (s4, ^4),
          (s5, ^5),
          (s6, ^6),
          (s7, ^7),
          (v0, [<(junk, ^8), (junk, ^8)>])}])}]
    ra (valTy_code 0 
      [{
        (ast, int), 
        (s0, ^0),
        (s1, ^1),
        (s2, ^2),
        (s3, ^3),
        (s4, ^4),
        (s5, ^5),
        (s6, ^6),
        (s7, ^7),
        (v0, [<(junk, ^8), (junk, ^8)>])}])).
    constructor.
    simpl; trivial.
    apply valTy_bine_valTyok; simpl; trivial.
(**********************)
  generalize Hs; intro Hs'.
  destruct Hs' as [Phi [HH [Y' [Hsub' HR]]]].
  destruct Hsub' as [tl [Ha0 [Ha'' [Htlok Hsub'']]]]; simplenv.
  rewrite shiftY_emp in Hsub''.
  
  inversion_clear Hraok as [ex Yx rx tx Hsel Htok].
  simpl in Hsub''.
  rewrite shiftL_emp in Hsub''.
(*   assert (Hsel' := RfTy_sel_some_instYL tl 1 0 Hsel). *)
  set (e'' := emp).
  set (Y'' := 
    [{(ast, int), 
      (s0, ^0),
      (s1, ^1),
      (s2, ^2),
      (s3, ^3),
      (s4, ^4),
      (s5, ^5),
      (s6, ^6),
      (s7, ^7),
      (v0, [<(junk, ^8), (junk, ^8)>])}]) in * .
  assert (Hsel' : RfTy.sel Y' ra = Some 
    ([0]- [{(ast, int), 
      (s0, get2 tl int 0),
      (s1, get2 tl int 1),
      (s2, get2 tl int 2),
      (s3, get2 tl int 3),
      (s4, get2 tl int 4),
      (s5, get2 tl int 5),
      (s6, get2 tl int 6),
      (s7, get2 tl int 7),
      (v0, [<(junk, get2 tl int 8), (junk, get2 tl int 8)>])}])).
    assert (Hsel'' : RfTy.sel 
      [{(ast, int), 
        (s0, get2 tl int 0),
        (s1, get2 tl int 1),
        (s2, get2 tl int 2),
        (s3, get2 tl int 3),
        (s4, get2 tl int 4),
        (s5, get2 tl int 5),
        (s6, get2 tl int 6),
        (s7, get2 tl int 7),
        (ra,
          [0]- [{(ast, int), 
            (s0, get2 tl int 0),
            (s1, get2 tl int 1),
            (s2, get2 tl int 2),
            (s3, get2 tl int 3),
            (s4, get2 tl int 4),
            (s5, get2 tl int 5),
            (s6, get2 tl int 6),
            (s7, get2 tl int 7),
            (v0, [<(junk, get2 tl int 8), (junk, get2 tl int 8)>])}])}]
      ra = Some 
      ([0]- [{(ast, int), 
        (s0, get2 tl int 0),
        (s1, get2 tl int 1),
        (s2, get2 tl int 2),
        (s3, get2 tl int 3),
        (s4, get2 tl int 4),
        (s5, get2 tl int 5),
        (s6, get2 tl int 6),
        (s7, get2 tl int 7),
        (v0, [<(junk, get2 tl int 8), (junk, get2 tl int 8)>])}])).
      simpl; trivial.
    apply (rfTy_sub_sel Hsub'' Hsel'').
  assert (Hraok := HR ra 
    (valTy_code TyEnv.emp
      [{(ast, int), 
        (s0, get2 tl int 0),
        (s1, get2 tl int 1),
        (s2, get2 tl int 2),
        (s3, get2 tl int 3),
        (s4, get2 tl int 4),
        (s5, get2 tl int 5),
        (s6, get2 tl int 6),
        (s7, get2 tl int 7),
        (v0, [<(junk, get2 tl int 8), (junk, get2 tl int 8)>])}]) 
    Hsel').
  inversion_clear Hraok as 
    [ | px hx fx ex Yx a Hrain Hsub | | ].
  clear Hsel' Hsel Htok Htlok Ha'' Ha0.
(**********************)
  exists <* rho_tal, mk_Ltal a *> .
  change CHSpec.lookup with TalCHSpec.lookup in Hrain.
  split.
    subst pc'.
    eapply sink_lookup; eauto.
  (* prove return point in the talp world *)
  simpl.
  exists itptal.
  split.
    unfold Sp.D0.
    apply DSpec.get_lookup.
    unfold D_tal.
    simpl.
    unfold DSpec.get, DSpecDef.get.
    simpl; trivial.
  unfold itptal.
  unfold Tp.open_r, open_r.
  substH Hg with (Hg (fun H => H = h1)).
  simpl in Hg.
  assert (Hh'' : exists fv : Heap.B,
    h ||- FLIST |-> fv ** Good fv 
      ** (fun H : heap => H = h1)).
    destruct Hh2 as [fv' Hh2].
    exists fv'.
    sconjlift 2.
    eapply sconj_intro; eauto.
    unfold hpred_true.
    reflexivity.
  destruct (Hg Hh'') as [fv' Hg'].
  clear Hg.
  destruct Hg' as [lnk [siz [Hsiz Hg]]].
  
  assert (Hhx : h' ||- ((fun H : heap => H = h1) **
    MBlk (Regfile.get rf' v0) lnk siz) ** 
      FLIST |-> fv' ** Good fv').
    simplsep 4 4.
  clear Hg.
  destruct Hhx as [h1' [h2' [J [Hh1' Hh2']]]].
  exists h1'; exists h2'.
  split_l.
  split_r by (exists fv'; trivial).
  exists talpsi.
  split_l.
  exists (DHSpec.set
    (DHSpec.set Phi 
      (Regfile.get rf' v0) (junk, get2 tl int 8))
    (S (Regfile.get rf' v0)) (junk, get2 tl int 8)).
  unfold MBlk in Hh1'.
  unfold buffer in Hh1'.
  sepelim Hh1'.
  destruct x.
    simpl in A0.
    destruct A.
    elimtype False.
    omega.
  destruct x.
    simpl in A0.
    elimtype False.
    omega.
  simpl in Hh1'.
  set (addr := Regfile.get rf' v0)  in * .
  destruct x.
  (*>>>>>>>>>>>>>>>>>>>>>>>>>>*)
    assert (Hx : h1' ||- (fun h => 
      h = h1) ** addr |-> n ** S addr |-> n0 ** top).
      simplsep 4 6.
    split.
      eapply heap_extend_pair_wfdh; eauto.

    exists [{(ast, int), 
      (s0, get2 tl int 0),
      (s1, get2 tl int 1),
      (s2, get2 tl int 2),
      (s3, get2 tl int 3),
      (s4, get2 tl int 4),
      (s5, get2 tl int 5),
      (s6, get2 tl int 6),
      (s7, get2 tl int 7),
      (v0, [<(junk, get2 tl int 8), (junk, get2 tl int 8)>])}].
    split_l.
    intros r t.
    assert (H1 : h1' ||- addr |-> n 
      ** S addr |-> n0 ** top).
      simplsep 3 4.
    destruct r; intro Hsel; 
      try discriminate Hsel;
        inj_hyp Hsel; subst t.
      constructor.
      subst addr;
        eapply heap_extend_pair_regtype; eauto.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs0.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs1.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs2.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs3.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs4.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs5.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs6.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.
    simpl in Hs.
    substH HR with (regfile_weakening HR Hsub'').
    rewrite <- Hs7.
    apply heap_extend_pair_valok with h1' h1 n n0;
      trivial.
    apply HR; simpl; trivial.

  (*>>>>>>>>>>>>>>>>>>>>>>>>>>*)
  assert (Hx : h1' ||- (fun h => 
    h = h1) ** addr |-> n ** S addr |-> n0 ** top).
    simplsep 4 6.
  split.
    eapply heap_extend_pair_wfdh; eauto.

  exists [{(ast, int), 
    (s0, get2 tl int 0),
    (s1, get2 tl int 1),
    (s2, get2 tl int 2),
    (s3, get2 tl int 3),
    (s4, get2 tl int 4),
    (s5, get2 tl int 5),
    (s6, get2 tl int 6),
    (s7, get2 tl int 7),
    (v0, [<(junk, get2 tl int 8), (junk, get2 tl int 8)>])}].
  split_l.
  intros r t.
  assert (H1 : h1' ||- addr |-> n 
    ** S addr |-> n0 ** top).
    simplsep 3 4.
  destruct r; intro Hsel; 
    try discriminate Hsel; inj_hyp Hsel; subst t.
  apply valok_int.
  subst addr.
  eapply heap_extend_pair_regtype; eauto.
  rewrite <- Hs0.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs1.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs2.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs3.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs4.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs5.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs6.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
  rewrite <- Hs7.
  simpl in Hs.
  substH HR with (regfile_weakening HR Hsub'').
  apply heap_extend_pair_valok with h1' h1 n n0;
    trivial.
  apply HR; simpl; trivial.
Qed.

(**********************************************************************)

Require Import ListT.
Open Local Scope listt_scope.

Definition FPsi_main : chspec := 
  (pairT MAIN <* rho_tal, mk_Ltal main_spec*>) ::
  (pairT CONT <* rho_tal, mk_Ltal cont_spec*>) ::
  (pairT NEWPAIR <* rho_tal, mk_Ltal talp_main.newpair_spec *>) ::
    nilT.

Definition FPsi'_main : chspec := 
  (pairT MAIN <* rho_tal, mk_Ltal main_spec*>) ::
  (pairT CONT <* rho_tal, mk_Ltal cont_spec*>) ::
    nilT.

Definition FPsi_newpair : chspec := 
  (pairT NEWPAIR <* rho_scap, mk_Lscap newpair_pre newpair_guar *>) ::
  (pairT MALLOC <* rho_scap, mk_Lscap malloc_pre malloc_guar *>) ::
  (pairT MITER <* rho_scap, mk_Lscap miter_pre miter_guar *>) ::
  (pairT MNEXT <* rho_scap, mk_Lscap mnext_pre mnext_guar *>) ::
  (pairT MRETPTR <* rho_scap, mk_Lscap mretptr_pre mretptr_guar *>) ::
  (pairT MLPRV <* rho_scap, mk_Lscap mlprv_pre mlprv_guar *>) ::
  (pairT MSPLIT <* rho_scap, mk_Lscap msplit_pre msplit_guar *>) ::
  (pairT MCOMP <* rho_scap, mk_Lscap mcomp_pre mcomp_guar *>) ::
  (pairT MRET <* rho_scap, mk_Lscap mret_pre mret_guar *>) ::
    nilT.

Ltac opsi_lookup_solve Psi :=
  unfold Psi, OCHSpec.lookup; simpl; tauto.

Lemma FPsi_main_lookup_MAIN : OCHSpec.lookup FPsi_main MAIN  <* rho_tal, mk_Ltal main_spec*>.
Proof. opsi_lookup_solve FPsi_main. Qed.

Lemma FPsi_main_lookup_CONT : OCHSpec.lookup FPsi_main CONT  <* rho_tal, mk_Ltal cont_spec*>.
Proof. opsi_lookup_solve FPsi_main. Qed.

Lemma FPsi_main_lookup_NEWPAIR : OCHSpec.lookup FPsi_main NEWPAIR  <* rho_tal, mk_Ltal talp_main.newpair_spec*>.
Proof. opsi_lookup_solve FPsi_main. Qed.

Lemma FPsi'_main_lookup_MAIN : OCHSpec.lookup FPsi'_main MAIN  <* rho_tal, mk_Ltal main_spec*>.
Proof. opsi_lookup_solve FPsi_main. Qed.

Lemma FPsi'_main_lookup_CONT : OCHSpec.lookup FPsi'_main CONT  <* rho_tal, mk_Ltal cont_spec*>.
Proof. opsi_lookup_solve FPsi_main. Qed.


Lemma FPsi_newpair_lookup_NEWPAIR : OCHSpec.lookup FPsi_newpair NEWPAIR  <* rho_scap,  mk_Lscap newpair_pre newpair_guar*>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MALLOC : OCHSpec.lookup FPsi_newpair MALLOC  <* rho_scap, mk_Lscap malloc_pre malloc_guar*>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MITER : OCHSpec.lookup FPsi_newpair MITER <* rho_scap, mk_Lscap miter_pre miter_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MNEXT : OCHSpec.lookup FPsi_newpair MNEXT <* rho_scap, mk_Lscap mnext_pre mnext_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MRETPTR : OCHSpec.lookup FPsi_newpair MRETPTR <* rho_scap, mk_Lscap mretptr_pre mretptr_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MLPRV : OCHSpec.lookup FPsi_newpair MLPRV <* rho_scap, mk_Lscap mlprv_pre mlprv_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MSPLIT : OCHSpec.lookup FPsi_newpair MSPLIT <* rho_scap, mk_Lscap msplit_pre msplit_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MCOMP : OCHSpec.lookup FPsi_newpair MCOMP <* rho_scap, mk_Lscap mcomp_pre mcomp_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

Lemma FPsi_newpair_lookup_MRET : OCHSpec.lookup FPsi_newpair MRET <* rho_scap, mk_Lscap mret_pre mret_guar *>.
Proof. opsi_lookup_solve FPsi_newpair. Qed.

(**********************************************************************)

Import Talp.
Require Import talp_2.

Lemma talpsi_lift_FPsi_main : tal_chspec_lift talp_main.psi FPsi_main.
Proof.
  unfold tal_chspec_lift, Tp.rho_tal.
  change TalCHSpec.lookup with Ty.CHSpec.lookup.
  split.
  
  intros.
  psi_lookup (MAIN).
  apply FPsi_main_lookup_MAIN.
  psi_lookup (CONT).
  apply FPsi_main_lookup_CONT.
  psi_lookup (NEWPAIR).
  apply FPsi_main_lookup_NEWPAIR.
  psi_lookup_not.

  intros f pi Hpi.
  opsi_lookup (NEWPAIR).
  exists talp_main.newpair_spec.
  split_l; compute; trivial.
  opsi_lookup (MAIN).
  exists talp_main.main_spec.
  split_l; compute; trivial.
  opsi_lookup (CONT).
  exists talp_main.cont_spec.
  split_l; compute; trivial.
  opsi_lookup_not.
Qed.

Lemma talpsi'_lift_FPsi'_main : tal_chspec_lift talp_main.psi' FPsi'_main.
Proof.
  unfold tal_chspec_lift, Tp.rho_tal.
  change TalCHSpec.lookup with Ty.CHSpec.lookup.
  split.

  intros.
  psi_lookup (MAIN).
  apply FPsi'_main_lookup_MAIN.
  psi_lookup (CONT).
  apply FPsi'_main_lookup_CONT.
  psi_lookup_not.

  intros f pi Hpi.
  opsi_lookup (MAIN).
  exists talp_main.main_spec.
  split_l; compute; trivial.
  opsi_lookup (CONT).
  exists talp_main.cont_spec.
  split_l; compute; trivial.
  opsi_lookup_not.
Qed.

Lemma talp_main_wfch : ocap_wfch D_tal FPsi_main FPsi'_main talp_main.C.
Proof.
  apply tal_wfch_ocap_wfch with talp_main.psi talp_main.psi'.
  apply talpsi_lift_FPsi_main.
  apply talpsi'_lift_FPsi'_main.
  apply talp_main.C_wf.
Qed.

(**********************************************************************)

Import scap.
Require Import scap_2.

Lemma scappsi_lift_FPsi_newpair : scap_chspec_lift scap_newpair.psi FPsi_newpair.
Proof.
  unfold scap_chspec_lift, Sp.rho_scap.
  change ScapCHSpec.lookup with CHSpec.lookup.
  split.

  intros.
  psi_lookup (NEWPAIR).
  apply FPsi_newpair_lookup_NEWPAIR.
  psi_lookup (MALLOC).
  apply FPsi_newpair_lookup_MALLOC.
  psi_lookup (MITER).
  apply FPsi_newpair_lookup_MITER.
  psi_lookup (MNEXT).
  apply FPsi_newpair_lookup_MNEXT.
  psi_lookup (MRETPTR).
  apply FPsi_newpair_lookup_MRETPTR.
  psi_lookup (MLPRV).
  apply FPsi_newpair_lookup_MLPRV.
  psi_lookup (MSPLIT).
  apply FPsi_newpair_lookup_MSPLIT.
  psi_lookup (MCOMP).
  apply FPsi_newpair_lookup_MCOMP.
  psi_lookup (MRET).
  apply FPsi_newpair_lookup_MRET.
  psi_lookup_not.

  intros f pi Hpi.
  opsi_lookup (NEWPAIR).
  exists scap_newpair.newpair_pre; exists scap_newpair.newpair_guar.
  split_l; compute; trivial.
  opsi_lookup (MALLOC).
  exists scap_newpair.malloc_pre; exists scap_newpair.malloc_guar.
  split_l; compute; trivial.
  opsi_lookup (MITER).
  exists scap_newpair.miter_pre; exists scap_newpair.miter_guar.
  split_l; compute; trivial.
  opsi_lookup (MNEXT).
  exists scap_newpair.mnext_pre; exists scap_newpair.mnext_guar.
  split_l; compute; trivial.

  opsi_lookup (MRETPTR).
  exists scap_newpair.mretptr_pre; exists scap_newpair.mretptr_guar.
  split_l; compute; trivial.
  opsi_lookup (MLPRV).
  exists scap_newpair.mlprv_pre; exists scap_newpair.mlprv_guar.
  split_l; compute; trivial.
  opsi_lookup (MSPLIT).
  exists scap_newpair.msplit_pre; exists scap_newpair.msplit_guar.
  split_l; compute; trivial.
  opsi_lookup (MCOMP).
  exists scap_newpair.mcomp_pre; exists scap_newpair.mcomp_guar.
  split_l; compute; trivial.
  opsi_lookup (MRET).
  exists scap_newpair.mret_pre; exists scap_newpair.mret_guar.
  split_l; compute; trivial.
  opsi_lookup_not.
Qed.

Lemma scap_newpair_wfch : ocap_wfch D FPsi_newpair FPsi_newpair scap_newpair.C.
Proof.
  assert (ocap.wfch (DSpec.merge D_tal D_scap) FPsi_newpair FPsi_newpair scap_newpair.C).
  apply scap_wfch_ocap_wfch with scap_newpair.psi scap_newpair.psi.
  apply scappsi_lift_FPsi_newpair.
  apply scappsi_lift_FPsi_newpair.
  apply scap_newpair.C_ok.
  apply ocap.wfch_weaken_D with (DSpec.merge D_tal D_scap); trivial.
  apply D0_plus_D_tal_sub_D.
Qed.

(**********************************************************************)

Import ocap.

Lemma C_main_compat_C_newpair :
  CHeap.compat talp_main.C scap_newpair.C.
Proof.
  unfold CHeap.compat.
  intro f.
  unfold talp_main.C.
  unfold CHeap.get at 1, CHeapDef.get, CHeapDef.beq.
  decbeqnat f (NEWPAIR).
  substbnat f.
  simpl; trivial.
  
  decbeqnat f (6).
  substbnat f.
  simpl; trivial.

  decbeqnat f 100.
  substbnat f.
  simpl; trivial.

  decbeqnat f 101.
  substbnat f.
  simpl; trivial.

  decbeqnat f 102.
  substbnat f.
  simpl; trivial.

  trivial.
Qed.

Lemma ocap_wfch_0 : wfch (DSpec.merge D_tal D)
  (OCHSpec.merge FPsi_main FPsi_newpair)
  (OCHSpec.merge FPsi'_main FPsi_newpair) 
  (CHeap.merge talp_main.C scap_newpair.C).
Proof.
  apply link.
  apply talp_main_wfch.
  apply scap_newpair_wfch.
  apply DSpec.sub_compat.  
  apply D_tal_sub_D.
  apply C_main_compat_C_newpair.
Qed.

Definition FPsi := (OCHSpec.merge FPsi_main FPsi_newpair).

Definition FPsi' := (OCHSpec.merge FPsi'_main FPsi_newpair).

Lemma ocap_wfis_newpair : wfis D << 
  {[ <* rho_scap, mk_Lscap newpair_pre newpair_guar *> | D ]}
  / FPsi >> newpair_code.
Proof.
  apply wfis_weaken_lift_Psi with FPsi_newpair.
  apply scap_wfis_ocap_wfis with scap_newpair.psi.
  unfold Sp.D0.
  apply D_tal_sub_D.
  unfold DSpec.lookup.
  intuition.
  apply scappsi_lift_FPsi_newpair.
  apply scap_newpair.newpair_ok.
  unfold FPsi.
  apply OCHSpec.sub_merge_r.
Qed.

Lemma ocap_wfis_newpair_2 : wfis D << 
  {[ <* rho_tal, mk_Ltal talp_main.newpair_spec *> | D ]}
  / FPsi >> newpair_code.
Proof.
  apply wfis_weak_a_lift with 
    {[ <* rho_scap, mk_Lscap newpair_pre newpair_guar *> | D ]}.
  apply ocap_wfis_newpair.
  simpl.
  unfold aimp.
  intros.
  destruct H as [itp [Hitp H]].
  unfold D, DSpec.lookup, DSpec.get, DSpecDef.get in Hitp.
  simpl in Hitp.
  inj_hyp Hitp.
  deprewrite <- Hitp in H.
  exists itpscap.
  unfold D, DSpec.lookup, DSpec.get, DSpecDef.get.
  split_l.
  apply newpair_weakening; trivial.
Qed.

Theorem ocap_wfch : wfch D FPsi 
  (OCHSpec.merge 
    (OCHSpec.sig_map NEWPAIR 
      <* rho_tal, mk_Ltal talp_main.newpair_spec *>) FPsi')
  (CHeap.merge talp_main.C scap_newpair.C).
Proof.
  apply link_2 with newpair_code.
  apply ocap.wfch_weaken_D with (DSpec.merge D_tal D).
  apply DSpec.merge_sub_intro.
  apply D_tal_sub_D.
  apply DSpec.sub_refl.
  apply ocap_wfch_0.
  unfold describe_pi.
  exists itptal.
  unfold DSpec.lookup, DSpec.get, DSpecDef.get, D.
  simpl.
  trivial.
  apply getis_compat_merge_intro.
  right.
  compute; trivial.
  apply C_main_compat_C_newpair.
  apply ocap_wfis_newpair_2.
Qed.
