(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* newpair                                                     *)
(*     in SCAP world                                           *)
(*                                                             *)
(***************************************************************)

(* $Id: scap_newpair.v,v 1.1.1.1 2006/12/07 16:43:18 ssg05 Exp $ *)

Require Import optionT.
Require Import prodT.
Require Import machine.
Require Import scap.
Require Import scap_ck.
Require Import cdsa.
Require Import extac.
Require Import natpm2.

Open Local Scope tm_scope.
Open Local Scope scap_scope.
Open Local Scope seplogic_scope.
Open Local Scope listt_scope.
(* lables *)

Notation NEWPAIR := 0005.

Notation MALLOC := 0009.
Notation MITER := 0015.
Notation MCOMP := 0017.
Notation MRETPTR := 0025.
Notation MSPLIT := 0027.
Notation MNEXT := 0032.
Notation MLPRV := 0035.
Notation MRET := 0037.

(* code *)

(* Notation FLIST := 100. *)
Parameter FLIST : nat.
(* reg alias *)
Notation Local reqsize := a0.
Notation Local actsize := t4.
Notation Local retptr := v0.
Notation Local tmp_actsize := t5.

Definition newpair_code :=
(* 0005 *)  addiu a0 zero 2; 
(* 0006 *)  j MALLOC;;.

Definition malloc_code :=
(* 0009 *)  move tmp_actsize actsize;
(* 0010 *)  addiu actsize reqsize HDSIZE;
(* 0011 *)  addiu prev zero NULL;
(* 0012 *)  lw ptrp FLIST zero;
(* 0013 *)  addiu retptr zero NULL;
(* 0014 *)  j MITER;;.

Definition miter_code :=
(* 0015 *)  blt zero ptrp MCOMP;
(* 0016 *)  j MITER;;.

Definition mcomp_code :=
(* 0017 *)  lw t0 SZ ptrp;
(* 0018 *)  addiu t1 actsize HDSIZE;    
(* 0019 *)  blt t1 t0 MSPLIT;
(* 0020 *)  blt t0 actsize MNEXT;
(* 0021 *)  lw t0 NXT ptrp;
(* 0022 *)  blt zero prev MLPRV;
(* 0023 *)  sw t0 FLIST zero;
(* 0024 *)  j MRETPTR;;.

Definition mretptr_code :=
(* 0025 *)  addiu retptr ptrp HDSIZE;
(* 0026 *)  j MRET;;.

Definition msplit_code :=
(* 0027 *)  subu t0 t0 actsize;
(* 0028 *)  sw t0 SZ ptrp;
(* 0029 *)  addu ptrp ptrp t0;
(* 0030 *)  sw actsize SZ ptrp;
(* 0031 *)  j MRETPTR;;.

Definition mnext_code :=
(* 0032 *)  addu prev ptrp zero;
(* 0033 *)  lw ptrp NXT ptrp;
(* 0034 *)  j MITER;;.

Definition mlprv_code :=
(* 0035 *)  sw t0 NXT prev;
(* 0036 *)  j MRETPTR;;.

Definition mret_code :=
(* 0037 *)  move actsize tmp_actsize;
(* 0038 *)  jr ra;;.


(* code heap *)
Import List.

Definition C : codeheap := 
(* NEWPAIR *)
  (0005,   addiu a0 zero 2              ) ::
  (0006,   j MALLOC                     ) ::

(* MALLOC *)
  (0009,   move tmp_actsize actsize     ) ::
  (0010,   addiu actsize reqsize HDSIZE ) ::
  (0011,   addiu prev zero NULL         ) ::              
  (0012,   lw ptrp FLIST zero           ) ::
  (0013,   addiu retptr zero NULL       ) ::
  (0014,   j MITER                      ) ::

(* MITER *)
  (0015,   blt zero ptrp MCOMP          ) ::
  (0016,   j MITER                      ) ::

(* MCOMP *) 
  (0017,   lw t0 SZ ptrp                ) ::                      
  (0018,   addiu t1 actsize HDSIZE      ) ::
  (0019,   blt t1 t0 MSPLIT             ) ::
  (0020,   blt t0 actsize MNEXT         ) ::
  (0021,   lw t0 NXT ptrp               ) ::                     
  (0022,   blt zero prev MLPRV          ) ::
  (0023,   sw t0 FLIST zero             ) ::
  (0024,   j MRETPTR                    ) ::

(* MRETPTR *)
  (0025,   addiu retptr ptrp HDSIZE     ) ::
  (0026,   j MRET                       ) ::

(* MSPLIT *)
  (0027,   subu t0 t0 actsize           ) ::
  (0028,   sw t0 SZ ptrp                ) ::
  (0029,   addu ptrp ptrp t0            ) ::
  (0030,   sw actsize SZ ptrp           ) ::
  (0031,   j MRETPTR                    ) ::

(* MNEXT *)
  (0032,   addu prev ptrp zero          ) ::
  (0033,   lw ptrp NXT ptrp             ) ::
  (0034,   j MITER                      ) ::

(* MLPRV *)
  (0035,   sw t0 NXT prev               ) ::
  (0036,   j MRETPTR                    ) ::

(* MRET *)
  (0037,   move actsize tmp_actsize     ) ::
  (0038,   jr ra                        ) ::
  nil.

(* specs *)

(* Definition FList := *)
(*   lexists n, flist fl n. *)

Definition eq_sn s s' :=
  rget s s0 = rget s' s0
  /\ rget s s1 = rget s' s1
  /\ rget s s2 = rget s' s2
  /\ rget s s3 = rget s' s3
  /\ rget s s4 = rget s' s4
  /\ rget s s5 = rget s' s5
  /\ rget s s6 = rget s' s6
  /\ rget s s7 = rget s' s7.

Definition newpair_pre s :=
  exists fv, 
    s |||- FLIST |-> fv ** Good fv ** top.

Definition newpair_guar s s' :=
  rget s ra = S_pc s' /\
  eq_sn s s' /\
  forall A,
    (exists fv,
      s |||- FLIST |-> fv ** Good fv ** A)
    -> (exists fv, exists lnk, exists siz,
      siz >= (2 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** FLIST |-> fv ** 
        Good fv ** A).
  
Definition newpair_spec := [newpair_pre & newpair_guar].

Definition malloc_pre s :=
  rget s a0 > 0 
  /\ exists fv, 
    s |||- FLIST |-> fv ** Good fv ** top.

Definition malloc_guar s s' :=
  rget s ra = S_pc s' /\
  eq_sn s s' /\
  forall A,
    (exists fv,
      s |||- FLIST |-> fv ** Good fv ** A)
    -> (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition miter_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s retptr = NULL 
  /\ exists fv, 
    s |||- FLIST |-> fv ** MidL fv (rget s prev) (rget s ptrp) ** top.

Definition miter_guar s s' :=
  rget s ra = S_pc s' /\
  eq_sn s s' /\
  rget s tmp_actsize = rget s' actsize 
  /\ forall A,
    (exists fv,  
      s |||- FLIST |-> fv ** MidL fv (rget s prev) (rget s ptrp) ** A) 
    -> 
    (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition mcomp_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s retptr = NULL 
  /\ rget s ptrp > NULL 
  /\ exists fv,  
    s |||- FLIST |-> fv ** MidL fv (rget s prev) (rget s ptrp) ** top.

Notation mcomp_guar := miter_guar.

Definition msplit_pre s :=
  rget s a0 > 0
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s ptrp > NULL 
  /\ rget s t0 > rget s actsize + 2 
  /\ exists fv,  
    exists q, exists n,
      (rget s ptrp < q \/ q = NULL) /\
      s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s ptrp) ** 
        MBlk (rget s ptrp + 2) q (rget s t0) ** MBlkLst n q NULL ** top.

Definition msplit_guar s s' :=
  eq_sn s s' /\
  rget s tmp_actsize = rget s' actsize /\
  rget s ra = S_pc s'/\ 
  forall A,
    (exists fv,  
      exists q, exists n,
        (rget s ptrp < q \/ q = NULL) /\
        s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s ptrp) ** 
	  MBlk (rget s ptrp + 2) q (rget s t0) ** MBlkLst n q NULL ** A) 
    ->
    (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition mnext_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s ptrp > NULL 
  /\ rget s retptr = NULL 
  /\ exists fv,  
    exists q, exists n,
      (rget s ptrp < q \/ q = NULL) /\
      s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s ptrp) ** 
        MBlk (rget s ptrp + 2) q (rget s t0) ** MBlkLst n q NULL ** top.

Notation mnext_guar := msplit_guar.

Definition mlprv_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2 
  /\ rget s ptrp > NULL 
  /\ rget s prev > NULL 
  /\ exists fv,  
    exists n, exists siz,
      (rget s ptrp < rget s t0 \/ rget s t0 = NULL) /\ 
      (siz >= rget s actsize) /\
      s |||- FLIST |-> fv ** EndL fv (rget s prev) (rget s ptrp) ** 
        MBlk (rget s ptrp + 2) (rget s t0) siz ** 
        MBlkLst n (rget s t0) NULL ** top.

Definition mlprv_guar s s' :=
  eq_sn s s' /\
  rget s tmp_actsize = rget s' actsize /\
  rget s ra = S_pc s'/\ 
  forall A,
    (exists fv, 
      exists n, exists siz,
      (rget s ptrp < rget s t0 \/ rget s t0 = NULL) /\ 
      (siz >= rget s actsize) /\
      s |||- FLIST |-> fv ** 
        EndL fv (rget s prev) (rget s ptrp) ** 
	MBlk (rget s ptrp + 2) (rget s t0) siz ** 
        MBlkLst n (rget s t0) NULL ** A) 
    -> 
    (exists fv, exists lnk, exists siz,
    siz >= (rget s a0 + 2) /\
    s' |||- MBlk (rget s' retptr) lnk siz ** 
      FLIST |-> fv ** Good fv ** A).

Definition mretptr_pre s :=
  rget s a0 > 0 
  /\ rget s actsize = rget s a0 + 2
  /\ exists fv, exists siz, exists lnk,
    (siz >= rget s actsize) /\
    s |||- FLIST |-> fv ** 
      MBlk (rget s ptrp + 2) lnk siz ** Good fv ** top.

Definition mretptr_guar s s' :=
  eq_sn s s' /\
  rget s tmp_actsize = rget s' actsize /\
  rget s ra = S_pc s' /\ 
  forall A,
    (exists fv, exists siz, exists lnk,
      (siz >= rget s actsize) /\
      s |||- FLIST |-> fv ** 
        MBlk (rget s ptrp + 2) lnk siz ** 
        Good fv ** A) 
    ->
    (exists fv, exists lnk, exists siz,
      siz >= (rget s a0 + 2) /\
      s' |||- MBlk (rget s' retptr) lnk siz ** 
        FLIST |-> fv ** Good fv ** A).

Definition mret_pre (s : state) := True.

Definition mret_guar (s s' : state) :=
  eq_sn s s' /\
  rget s tmp_actsize = rget s' actsize /\
  let (h, rf, pc) := s in 
    let (h, rf, pc) := rset s actsize (rget s tmp_actsize) in
      <.h, rf, rget s ra.> = s'.

(* code heap spec *)
Import ListT.

Definition psi : chspec :=
  (pairT NEWPAIR (mk_cdspec newpair_pre newpair_guar)) ::
  (pairT MALLOC (mk_cdspec malloc_pre malloc_guar)) ::
  (pairT MITER (mk_cdspec miter_pre miter_guar)) :: 
  (pairT MNEXT (mk_cdspec mnext_pre mnext_guar)) ::
  (pairT MRETPTR (mk_cdspec mretptr_pre mretptr_guar)) ::
  (pairT MLPRV (mk_cdspec mlprv_pre mlprv_guar)) ::
  (pairT MSPLIT (mk_cdspec msplit_pre msplit_guar)) :: 
  (pairT MCOMP (mk_cdspec mcomp_pre mcomp_guar)) ::
  (pairT MRET (mk_cdspec mret_pre mret_guar)) ::
    nilT.

(* proof *)

Ltac unfoldpred :=
  unfold newpair_pre, newpair_guar;
  unfold malloc_pre, malloc_guar; 
  unfold miter_pre, miter_guar; 
  unfold mnext_pre, mretptr_pre; 
  unfold msplit_pre, msplit_guar; 
  unfold mlprv_pre, mlprv_guar; 
  unfold mcomp_pre, mretptr_guar; 
  unfold mret_pre, mret_guar.

(* Axiom newpair_ok : wfis psi newpair_pre newpair_guar newpair_code. *)
(* Axiom malloc_ok : wfis psi malloc_pre malloc_guar malloc_code. *)
(* Axiom miter_ok : wfis psi miter_pre miter_guar miter_code. *)
(* Axiom mcomp_ok : wfis psi mcomp_pre mcomp_guar mcomp_code. *)
(* Axiom mretptr_ok : wfis psi mretptr_pre mretptr_guar mretptr_code. *)
(* Axiom msplit_ok : wfis psi msplit_pre msplit_guar msplit_code. *)
(* Axiom mnext_ok : wfis psi mnext_pre mnext_guar mnext_code. *)
(* Axiom mlprv_ok : wfis psi mlprv_pre mlprv_guar mlprv_code. *)
(* Axiom mret_ok : wfis psi mret_pre mret_guar mret_code. *)

Theorem newpair_ok : wfis psi newpair_pre newpair_guar newpair_code.
Proof.
  simplvc.
  unfoldpred.
  intros s. 
  destruct s as [hx rfx pcx].
  intros.
  split.
    split_r.
    simplst.
    auto with arith.
  intros [hx' rfx' pcx'].
  unfold newpair_guar, rget.
  simplst.
  intros [Hra H1].
  split; tauto.
Qed.

Ltac substreg H :=
  destruct H as [Ha [Hb [Hc [Hd [He [Hf [Hg Hh]]]]]]];
    trivial.

Theorem malloc_ok : wfis psi malloc_pre malloc_guar malloc_code.
Proof.
  simplvc.

  unfoldpred.
  intros s. 
  destruct s as [hx rfx pcx].
  simplst.
  intros [R [fv Hp]].
  simpllw fv 1.
  simplst.
  split; auto.
  split; auto.
  split; auto.
  split; auto.
  exists fv.
  simplsep 3 3.
  apply (Good_MidL Hp).
  intros.
  elim H. intros.
  destruct H as [Ha [Hb [Hc Hd]]].
  split; auto.
  unfold eq_sn; unfold eq_sn in Hb.
  simplst.
  destruct Hb as [He1 [He2 [He3 [He4 [He5 [He6 [He7 He8]]]]]]].
  repeat split_l.
  trivial.
  clear He1 He2 He3 He4 He5 He6 He7 He8.
  intros.
  generalize (Hd A). intros. clear Hd.
  elim H2. intros.
  exists x; trivial.
  destruct H. exists fv.
  hvinject 1 1. subst.
  clear Hp.
  simplsep 3 3.
  simpl.
  apply (Good_MidL H).
Qed.

Lemma miter_ok : wfis psi miter_pre miter_guar miter_code.
Proof.
  simplvc.
  
  unfoldpred.
  intros [hx rfx pcx].
  unfold rget, rset.
  simplst.
  intros [R1 [R2 [R3 [fv Hp]]]].
  destruct (nat_eq_dec (Regfile.get rfx ptrp) O)
    as [Heq | Hneq].
    rewrite Heq.
    rewrite Heq in Hp.
    rewbnat (blt_nat O O = false).
    split.
      split_l; split_l; split_l; exists fv; trivial.
    intros [hx' rfx' pcx'].
    intros H.
    intuition.
(*   reflexivity. *)
(*   destruct H0. *)
(*   right; exists x; split; auto. *)
(*   simplsep 3 3. *)
(*   apply (MidL_Good H). *)
  rewbnat (blt_nat O (Regfile.get rfx ptrp) = true).
  assert (H' : exists n, Regfile.get rfx ptrp = S n).
    destruct (Regfile.get rfx ptrp).
    tauto.
    exists n; trivial.
  destruct H' as [n0 H'].
  rewrite H'.
  intuition.
  rewrite H' in Hp.
  exists fv; apply Hp.
Qed.

Lemma mretptr_ok : wfis psi mretptr_pre mretptr_guar mretptr_code.
Proof. 
  simplvc.
  intros [h rf pc].
  unfold mretptr_pre, mretptr_guar, mret_pre, mret_guar.
  simplst.
  intuition.
  destruct s' as [h' rf' pc'].
  simplst.
  injection H4.
  intros; trivial.
  injection H5; intros. trivial.
  destruct H4 as [fv [siz [lnk [R Hp]]]].
  exists fv; exists lnk; exists siz.
  split.
  omega.
  destruct s' as [h' rf' pc'].
  injection H5.
  intros.
  subst.
  simplst.
  simplsep 7 7.
Qed.

Lemma msplit_ok : wfis psi msplit_pre msplit_guar msplit_code.
Proof.
  simplvc.
  intros [h rf pc] [R1 [R2 [R3 [R4 [fv [q [n [R5 Hp]]]]]]]].
  simplst.
  unfold MBlk in Hp; sepelim Hp; simplpm2h Hp.
  simpllw (Regfile.get rf t0) 3.
    assert ((S (Regfile.get rf ptrp)) <> (S (Regfile.get rf ptrp
      + (Regfile.get rf t0 - Regfile.get rf actsize)))).
      omega.
    clear H.
    sconjlifth 4 Hp.
    assert (Regfile.get rf t0 - 2 > Regfile.get rf reqsize + 2).
      lwomega.
    hypreplace Hp (buffer_split Hp H).
    destruct Hp as [y [z Hp]].
    rewrite <- R2 in Hp.
    simplpm.
    replace (Regfile.get rf ptrp + 2 
      + (Regfile.get rf t0 - 2 - Regfile.get rf actsize)) 
    with
    (Regfile.get rf ptrp + (Regfile.get rf t0 - Regfile.get rf actsize)) in Hp;
      [ simpllw z 3| lwomega].
    repeat (split; trivial).
        exists fv; exists (Regfile.get rf actsize); exists y.
        split; auto.
        unfold MBlk; simplpm2.
        simplst.
        simplsw' 4 z.
        sconjlifth 9 Hp.
        hypreplace Hp (cont_update1 Hp (Regfile.get rf t0 
          - Regfile.get rf actsize)).
        replace (Regfile.get rf ptrp + 2 + (Regfile.get rf t0 - 2 
          - Regfile.get rf reqsize)) with
        (Regfile.get rf ptrp + (Regfile.get rf t0 - Regfile.get rf actsize) + 2) in Hp;
        [replace (Regfile.get rf reqsize) with (Regfile.get rf actsize - 2) in Hp;
          [simplsep 9 9 | lwomega] | lwomega].
          lwomega.
        apply MidL_Good with (Regfile.get rf prev) (Regfile.get rf ptrp).
        unfold MidL.
        lexintro 2; exists (S n).
        simpl.
        lexintro 2; exists q.
        lexintro 2; exists (Regfile.get rf t0 - Regfile.get rf actsize).
        unfold MBlk; simplpm2.
        replace (Regfile.get rf t0 - Regfile.get rf actsize - 2) with
          (Regfile.get rf t0 - 2 - Regfile.get rf actsize);
          [simplsep  5 5; omega| lwomega].
        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.

        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.


        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.

        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.

        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.

        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.


        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.

        unfold mretptr_guar in H0; destruct H0 as [H0 _].
        unfold eq_sn in H0.
        destruct H0 as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8
          ]]]]]]].
        simplst. trivial.
        unfold mretptr_guar in H0.
        destruct H0 as [_ H0].
        elim H0; intros.
        simplst. trivial.
        unfold mretptr_guar in H0.
        destruct H0 as [_ [_ [H0 _]]].
        simplst. trivial.

    unfold mretptr_guar in H0; simplst.
    intros P [fl' [q' [n' [R' Hp']]]].
    destruct H0 as [Ha [Hb [Hc H0]]].
    hypreplace H0 (H0 P).
    apply H0; clear H0.

    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    sconjlifth 4 Hp'.
    hypreplace Hp' (buffer_split Hp' H).
    destruct Hp' as [y' [z' Hp']].
    unfold rget, hget in Hp'.
    rewrite <- R2 in Hp'.
    replace (Regfile.get rf ptrp + 2 + (Regfile.get rf t0 - 2 
      - Regfile.get rf actsize)) with
      (Regfile.get rf ptrp + (Regfile.get rf t0 - Regfile.get rf actsize)) in Hp';
      [idtac | lwomega].
    hvinject 10 10.
    hypreplace Eq2 (sym_equal Eq2).
    subst; clear Hp.
    exists fv; exists (Regfile.get rf actsize); exists y'.
    split; auto.
    unfold MBlk; simplpm2.
    unfold hset; simpl.
(*   sconjlift 3. *)
(*   apply cont_update1 with (v1 := z). *)
    simplst.
    simplsw' 6 z.
    sconjlifth 9 Hp'.
    hypreplace Hp' (cont_update1 Hp' (Regfile.get rf t0 - Regfile.get rf actsize)).
    replace (Regfile.get rf ptrp + 2 + (Regfile.get rf t0 - 2 
      - Regfile.get rf reqsize)) with
      (Regfile.get rf ptrp + (Regfile.get rf t0 - Regfile.get rf actsize) + 2) in Hp'.
       replace (Regfile.get rf reqsize) with (Regfile.get rf actsize - 2) in Hp'.
         simplsep 9 9.
      lwomega.
    apply MidL_Good with (Regfile.get rf prev) (Regfile.get rf ptrp).
    unfold MidL.
    lexintro 2; exists (S n').
    simpl.
    lexintro 2; exists q.
    lexintro 2; exists (Regfile.get rf t0 - Regfile.get rf actsize).
    unfold MBlk; simplpm2.
    replace (Regfile.get rf t0 - Regfile.get rf actsize - 2) with
      (Regfile.get rf t0 - 2 - Regfile.get rf actsize);
      [simplsep  5 5; omega| lwomega].
  lwomega. lwomega. lwomega.
Qed.

Lemma mnext_ok : wfis psi mnext_pre mnext_guar mnext_code.
Proof.
  simplvc.
  intros [h rf pc] 
    [R1 [R2 [R3 [R4 [fv [q [n [R5 Hp]]]]]]]].
  simplst.
  assert (Hp' := Hp).
  unfold MBlk in Hp; simplpm2h Hp; sepelim Hp.
  simpllw q 4.
  repeat (split; trivial).
      exists fv.
      clear Hp; unfold MidL.
      lexintro 3; exists n.
      simplst.
      apply sconj_sym.
      assocr.
      apply sconj_sym.
      assocr.
      eapply sconj_mono_right.
      apply Hp'.
      clear Hp'; intros h0 Hp'.
      assocr.
      hypreplace Hp' (sconj_assoc_right Hp').
      eapply sconj_mono_left.
      apply Hp'.
      clear Hp' h0; intros h0 Hp'.
      eapply EndL_add; auto.
      apply Hp'.
    unfold miter_guar in H; simplst.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [_ [H _]].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
  destruct H as [Ha G].
  tauto.
  unfold miter_guar in H.
  destruct H as [H _].
  simplst. trivial.
  intros P [fv' [q' [n' [R' Hp1]]]].
  destruct H as [Ha [Hb [Hc G]]].
  hypreplace G (G P).
  apply G; clear G.
  generalize Hp1.
  unfold MBlk in Hp1; simplpm2h Hp1; sepelim Hp1.
  clear Hp'; hvinject 6 6; subst.
  clear Hp1 Hp.
  exists fv'; unfold MidL.
  lexintro 3; exists n'.
  simplst.
      apply sconj_sym.
      assocr.
      apply sconj_sym.
      assocr.
      eapply sconj_mono_right.
      apply Hp0.
      clear Hp0; intros h0 Hp0.
      assocr.
      hypreplace Hp0 (sconj_assoc_right Hp0).
      eapply sconj_mono_left.
      apply Hp0.
      clear Hp0 h0; intros h0 Hp0.
      eapply EndL_add; auto.
      apply Hp0.
Qed.

Lemma mlprv_ok : wfis psi mlprv_pre mlprv_guar mlprv_code.
Proof.
  simplvc.
  simplst.
  intros [h rf pc] 
    [R1 [R2 [R3 [R4 [fv [n [siz [R5 [R6 Hp]]]]]]]]].
  simplst.
  unfold EndL in Hp.
  rewrite neq_beq_false in Hp; [unfold MBlk in Hp; sepelim Hp | omega].
  simplpm2h Hp.
  simpllw (Regfile.get rf ptrp) 3.
  repeat (split; trivial).
      exists fv; exists siz; exists (Regfile.get rf t0).
      split; trivial.
      sconjlifth 2 Hp.
      hypreplace Hp (cont_update1 Hp (Regfile.get rf t0)).
      unfold hset; simpl.
      unfold MBlk; simplpm2; simplsep 6 7.
      assert (h0 ||- MBlkLst x fv (Regfile.get rf prev) ** 
        MBlk (Regfile.get rf prev + 2) (Regfile.get rf t0) x0 ** MBlkLst n (Regfile.get rf t0) 0).
        unfold MBlk; simplpm2; simplsep 5 5.
      clear Hp.
      hypreplace H (MBlkLst_add H).
      assert (Regfile.get rf prev < Regfile.get rf t0 \/ Regfile.get rf t0 = 0).
        lwomega.
      substH H with (unit_intro_right (H H0)).
      assocl H.
      substH H with (MBlkLst_append H).
      unfold Good.
      lexintro 1; exists (S x + n).
      apply (unit_elim_right H).
    unfold mretptr_guar in H; simplst.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

  destruct H as [Ha [Hb [Hc G]]].
  tauto.
  destruct H as [Ha [Hb [Hc G]]].
  tauto.
  intros P [fv' [n' [siz' [R1' [R2' Hp1]]]]].
  destruct H as [Ha [Hb [Hc G]]].
  apply G; clear G.
  unfold EndL in Hp1.
  simplst.
  rewrite neq_beq_false in Hp1; [unfold MBlk in Hp1; sepelim Hp1 | omega].
  simplpm2h Hp1.
  hvinject 8 8; subst; clear Hp.
  clear R1' R2' A2.
  exists fv'; exists siz; exists (Regfile.get rf t0).
  split; trivial.
  sconjlifth 2 Hp1.
  substH Hp1 with (cont_update1 Hp1 (Regfile.get rf t0)).
  unfold hset; simpl.
  unfold MBlk; simplpm2; simplsep 6 7.
  assert (h0 ||- MBlkLst x1 fv' (Regfile.get rf prev) ** 
    MBlk (Regfile.get rf prev + 2) (Regfile.get rf t0) x2 ** MBlkLst n' (Regfile.get rf t0) 0).
    unfold MBlk; simplpm2; simplsep 5 5.
  clear Hp1.
  substH H with (MBlkLst_add H).
  assert (Regfile.get rf prev < Regfile.get rf t0 \/ Regfile.get rf t0 = 0).
    lwomega.
  substH H with (unit_intro_right (H H0)).
  assocl H.
  substH H with (MBlkLst_append H).
  unfold Good.
  lexintro 1; exists (S x1 + n').
  apply (unit_elim_right H).
Qed.

Lemma mret_ok : wfis psi mret_pre mret_guar mret_code.
Proof.
  simplvc.
  intros [h rf pc].
  unfold mret_pre, mret_guar.
  simplst.
  intuition.
  unfold eq_sn.
  rewrite <- H0.
  repeat (split; trivial).
  destruct s'.
  injection H0.
  intros.
  rewrite <- H2.
  destruct (beq_reg_dec actsize tmp_actsize).

  apply sym_equal.
  apply set_get_eq.
  trivial.
  apply sym_equal.
  apply set_get_eq.
  trivial.
Qed.

Lemma mcomp_ok : wfis psi mcomp_pre mcomp_guar mcomp_code.
Proof.
  simplvc.
  unfoldpred.
  intros [h rf pc] [R1 [R2 [R3 [R4 [fv Hp]]]]].
  simplst.
  unfold MidL in Hp; sepelim Hp.
  destruct x as [ | n]; simpl in Hp; sepelim Hp.
    elimtype False; omega.
  unfold MBlk in Hp; sepelim Hp; simplpm2h Hp.
  simpllw x0 4.
  Regfile.simplgs.
  decbltnat  (Regfile.get rf actsize + 2) x0.
    repeat (split; trivial).
          desbnat.
          auto.
        exists fv; exists x; exists n.
        split; auto.
        unfold MBlk; simplpm2; simplsep 5 5.
      unfold msplit_guar in H.
      tauto.
    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.


    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.


    destruct H as [H _].
    unfold eq_sn in H.
    destruct H as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [Ha [Hc G]].
    trivial.
    intros P [fv' Hp'].
    destruct H as [Ha [Hc [Hd G]]].
    apply G; clear G.
    unfold MidL in Hp'; sepelim Hp'.
    destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
      elimtype False; omega.
    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    hvinject 7 7. subst; clear Hp.
    exists fv'; exists x1; exists n1.
    split; trivial.
    unfold MBlk; simplpm2. 
    abstract simplsep 5 5.

  simplst.
  decbltnat x0 (Regfile.get rf actsize).
    simplst.
    repeat (split; trivial).
        exists fv; exists x; exists n.
        split; trivial.
        unfold MBlk; simplpm2; simplsep 5 5.
      unfoldgs.
      destruct s' as [h' rf' pc'].
      unfold msplit_guar in H.
      tauto.
    destruct H as [Ha [Hd [Hc G]]].
    trivial.
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.

    destruct H as [Ha [Hd [Hc G]]].
    trivial.

    intros P [fv' Hp'].
    destruct H as [Ha [Hd [Hc G]]].
    apply G; clear G.
    unfold MidL in Hp'; sepelim Hp'.
    destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
      elimtype False; omega.
    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    hvinject 7 7; subst; clear Hp.
    exists fv'; exists x1; exists n1.
    split; trivial.
    unfold MBlk; simplpm2; abstract simplsep 5 5.
  
  simplpm.
  simpllw x 3.
  destruct (nat_eq_dec (Regfile.get rf prev) O) 
    as [Heq | Hneq].
    Regfile.simplgs.
    rewbnat (blt_nat 0 (Regfile.get rf prev) = false).
    simplpm.
    unfold EndL in Hp. 
    rewrite Heq in Hp.
    rewbnat (beq_nat O O = true) in Hp.
    simpl in Hp. 
    sepelim Hp.
    subst fv.
    simpllw (Regfile.get rf ptrp) 1.
    repeat (split; trivial).
        exists x; exists x0; exists x.
        split; trivial.
        unfold rget.
        Regfile.simplgs.
        desbnat.
        trivial.
        unfold Good; lexintro 4; exists n.
        unfold MBlk; simplpm2.
        simpl; abstract simplsw 6 6 (Regfile.get rf ptrp).
      unfoldgs.
      unfold mretptr_guar in H.
      tauto.

    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    trivial.
    intros P [fv' Hp'].
    destruct H as [Ha [Hd [Hc G]]].
    apply G; clear G.
    unfold MidL in Hp'; sepelim Hp'.
    unfold EndL in Hp'; simpl in Hp'; sepelim Hp'.
    rewrite Heq in Hp'.
    rewbnat (beq_nat O O = true) in Hp'.
    sepelim Hp'.
    subst fv'.
    destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
      elimtype False; omega.
    unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
    hvinject 7 7. subst x x0; clear Hp.
    exists x1; exists x2; exists x1.
    split; trivial.
      unfold rget.
      Regfile.simplgs.
      desbnat.
      trivial.
    unfold rget.
    Regfile.simplgs.
    unfold Good; lexintro 4; exists n1.
    unfold MBlk; simplpm2.

  simpl; abstract simplsw 6 6 (Regfile.get rf ptrp).
  Regfile.simplgs.
  rewbnat (blt_nat 0 (Regfile.get rf prev) = true).

  assert (H' : exists n, Regfile.get rf prev = S n).
    destruct (Regfile.get rf prev).
      tauto.
    exists n0; trivial.
  destruct H' as [n0 H'].
  unfold miter_guar.
  unfoldgs; Regfile.simplgs.
  rewrite H'.
  repeat (split; trivial).
         desbnat.
         desbnat.
         unfoldgs; Regfile.simplgs.
         lwomega.
      exists fv; exists n; exists x0.
      repeat (split; trivial).
        desbnat.
        desbnat.
        unfoldgs; Regfile.simplgs.
        omega.
      unfold MBlk; simplpm2; simplsep 5 5.
    rewrite <- H'; trivial.
    unfold mlprv_guar in H.
    tauto.

    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
    destruct H as [Ha [Hd [Hc G]]].
    unfold eq_sn in Ha.
    destruct Ha as [Ha1 [Ha2 [Ha3 [Ha4 [Ha5 [Ha6 [Ha7 Ha8]]]]]]].
    trivial.
  destruct H as [Ha [Hd [Hc G]]].
  trivial.
  intros P [fv' Hp'].
  destruct H as [Ha [Hd [Hc G]]].
  apply G; clear G.
  unfold MidL in Hp'; sepelim Hp'.
  destruct x1 as [ | n1]; simpl in Hp'; sepelim Hp'.
    elimtype False; omega.
  unfold MBlk in Hp'; sepelim Hp'; simplpm2h Hp'.
  hvinject 8 8. subst; clear Hp.
  exists fv'; exists n1; exists x0.
  repeat (split; trivial).
    desbnat.
    desbnat.
    unfoldgs; Regfile.simplgs.
    omega.
  unfoldgs.
  Regfile.simplgs.
  unfold MBlk; simplpm2.
  simplsep 5 5.
Qed.

Theorem C_ok : wfch psi psi C.
Proof.
  prove_wfch.
  apply newpair_ok.
  apply malloc_ok.
  apply miter_ok.
  apply mnext_ok.
  apply mretptr_ok.
  apply mlprv_ok.
  apply msplit_ok.
  apply mcomp_ok.
  apply mret_ok.
Qed.
