(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(*                                                             *)
(*  Additional TAL lemmas                                      *)
(*                                                             *)
(***************************************************************)

(* $Id: sep_talp.v,v 1.1.1.1 2006/12/07 16:43:18 ssg05 Exp $ *)

Require Import machine.
Require Import Talp.
Require Import seplogic.
Require Import natpm.
Require Import extac.

Open Local Scope seplogic_scope.
Open Local Scope tal_scope.

Lemma sep_heap_extend_heap_get_none :
  forall h h' l v,
    (h' ||- (fun H : heap => H = h) ** l |-> v ** top)
    -> Heap.get h l = None.
Proof.
  intros h h' l v H.
  unfold hpred_true, sconj in H.
  destruct H as [ha [hb [J [Hha Hhb]]]].
  unfold hpred_true in Hha.
  unfold hpred_true in Hhb.
  unfold cont in Hhb.
  subst ha.
  destruct Hhb as [ha [hc [J' [Hha Hhc]]]].
  substH J with (J l).
  substH J' with (J' l).
  rewrite Hha in J'.
  Heap.simplgs.
  destruct (Heap.get_dec h l) as [[b Hs] | Hs];
    rewrite Hs in J; trivial.
  clear Hs.
  destruct (Heap.get_dec hc l) as [[b' Hs] | Hs];
    rewrite Hs in J'; try contradiction.
  clear Hs.
  destruct (Heap.get_dec hb l) as [[b'' Hs] | Hs];
    rewrite Hs in J'; 
    rewrite Hs in J; 
    try contradiction.
Qed.

Lemma sep_heap_extend_heapty_get_none :
  forall psi phi h h' l v,
    (h' ||- (fun H : heap => H = h) ** l |-> v ** top)
    -> wfdh psi phi phi h
    -> DHSpec.get phi l = None.
Proof.
  intros psi phi h h' l v.
  intros Hh' Hh.
  substH Hh' with (sep_heap_extend_heap_get_none _ _ _ _ Hh').
  unfold wfdh in Hh.
  destruct (DHSpec.get_dec phi l) 
    as [[tx Hs] | Hs]; trivial.
  substH Hs with (DHSpec.get_lookup Hs).
  substH Hh with (Hh l tx Hs).
  destruct Hh as [v' [Hin Hce]].
  substH Hin with (Heap.lookup_get Hin).
  rewrite Hh' in Hin; discriminate.
Qed.

Lemma sep_eq : 
  forall h h' l v,
    (h' ||- (fun H : heap => H = h) ** l |-> v)
    -> h' = (Heap.set h l v).
Proof.
  intros h h' l v H.
  unfold hpred_true, sconj in H.
  destruct H as [ha [hb [J [Hha Hhb]]]].
  unfold hpred_true in Hha.
  unfold hpred_true in Hhb.
  unfold cont in Hhb.
  rewrite Hha in J.
  rewrite Hhb in J.
  rewrite (Heap.join_sig_r J).
  trivial.
Qed.

Lemma wfdh_weaken_h : forall psi phi h h1 h2,
  Heap.join h1 h2 h
  -> wfdh psi phi phi h1 
  -> wfdh psi phi phi h.
Proof.
  intros psi phi h h1 h2 Hj Hh1.
  unfold wfdh.
  intros l ct Hin.
  substH Hh1 with (Hh1 l ct Hin).
  destruct Hh1 as [v [Hvin Hctoke]].
  exists v.
  split_l.
  eapply Heap.join_lookup_lookup_l; eauto.
  trivial.
Qed.

Lemma heap_extend_sig_wfdh :
  forall h h' l v talpsi Phi t,
    (h' ||- (fun H : heap => H = h) ** l |-> v ** top)
    -> wfdh talpsi Phi Phi h
    -> wfdh talpsi
    (DHSpec.set Phi l (junk, t))
    (DHSpec.set Phi l (junk, t))
    h'.
Proof.
  intros h h' l v psi phi t Hsep Hh.
  assert (Hsep' : h' ||- ((fun H : heap => H = h) 
    ** l |-> v) ** top).
  simplsep 3 3. clear Hsep.
  destruct Hsep' as [h1 [h2 [J [Hh1 Hh2]]]].
  apply wfdh_weaken_h with h1 h2; trivial.
  rewrite (sep_eq h h1 l v Hh1).
  apply heap_extend_wfdh; trivial.
  destruct Hh1 as [h1a [h1b [h1J [Hh1a Hh1b]]]].
  unfold hpred_true in Hh1a.
  subst h1a.
  unfold wfdh in Hh.
  destruct (DHSpec.get_dec phi l) as [Hsome | Hnone].
  destruct Hsome as [b Hsome].
  substH Hsome with (DHSpec.get_lookup Hsome).
  destruct (Hh l b Hsome) as [vv [Hvvin Hct]].
  substH Hh1b with (cont_get_eq Hh1b).
  substH Hvvin with (Heap.lookup_get Hvvin).
  unfold Heap.join in h1J.
  substH h1J with (h1J l).
  rewrite Hh1b in h1J.
  rewrite Hvvin in h1J.
  destruct h1J.
  trivial.
  constructor.
Qed.

Lemma heap_extend_pair_wfdh :
  forall h h' l1 l2 v1 v2 talpsi Phi t,
    (h' ||- (fun H : heap => H = h) 
      ** l1 |-> v1 ** l2 |-> v2 ** top)
    -> wfdh talpsi Phi Phi h
    -> wfdh talpsi
    (DHSpec.set 
      (DHSpec.set Phi l1 (junk, t))
      l2 (junk, t)) 
    (DHSpec.set 
      (DHSpec.set Phi l1 (junk, t))
      l2 (junk, t))    h'.
Proof.
  intros h h' l1 l2 v1 v2 psi phi t.
  intros Hsep Hh.
  assert (H0 : h' ||- ((fun H : heap => H = h) 
    ** l1 |-> v1) ** l2 |-> v2 ** top).
  simplsep 4 4.
  destruct H0 as [h1' [h2' [Jh' [Hh1' Hh2']]]].
  assert (H : wfdh psi 
    (DHSpec.set phi l1 (junk,t)) 
    (DHSpec.set phi l1 (junk,t)) h1').
  assert (H1 : h1' ||- ((fun H : heap => H = h) 
    ** l1 |-> v1) ** top).
  apply top_intro_right; trivial.
  assert (H2 : h1' ||- (fun H : heap => H = h) 
    ** l1 |-> v1 ** top).
  simplsep 3 3.
  eapply heap_extend_sig_wfdh; eauto.
  assert (H3 : h' ||- (fun H : heap => H = h1') 
    ** l2 |-> v2 ** top).
  simplsep 3 4.
  destruct Hsep as [hb [hc [J1 [Hhb Hhc]]]].
  destruct Hh1' as [ha [hc' [J2 [Hha Hhc']]]].
  unfold hpred_true in * .
  subst ha hb.
  unfold cont in Hhc', Hhc.
  rewrite <- Hhc' in Hhc.
  clear Hhc'.
  subst hc'.
  eapply Heap.join_eq; eauto.
  eapply heap_extend_sig_wfdh; eauto.
Qed. 

Lemma heap_extend_pair_regtype :
  forall h addr v1 v2 talpsi talphi t,
    (h ||- addr |-> v1 ** S addr |-> v2 ** top)
    -> valok talpsi 
    (DHSpec.set 
      (DHSpec.set talphi
        addr (junk, t))
        (S addr) (junk, t)) addr
    [<(junk, t), (junk, t)>].
Proof.
  intros h addr v1 v2 psi phi t Hsep.
  apply valok_tup.
  intros i Hiin.
  exists (junk, t).
  exists (junk, t).
  destruct Hiin as [ct Hlook].
  destruct i.
  simpl in Hlook |- * .
  split_l.
  split_r by (constructor; trivial).
  simplpm.
  apply DHSpec.get_lookup.
  DHSpec.simplgs.
  trivial.
  omega.
  destruct i.
  simpl in Hlook |- * .
  split_l.
  split_r by (constructor; trivial).
  simplpm.
  apply DHSpec.get_lookup.
  DHSpec.simplgs.
  trivial.
  destruct i.
  simpl in Hlook |- * .
  destruct Hlook.
  simpl in Hlook |- * .
  destruct Hlook.
Qed.

Lemma heap_extend_pair_valok :
  forall psi phi h h1 addr n n0 w t t',
    (h ||- (fun h : heap => h = h1) 
      ** addr |-> n ** S addr |-> n0 ** top)
    -> wfdh psi phi phi h1
    -> valok psi phi w t
    -> valok psi 
    (DHSpec.set 
      (DHSpec.set phi addr (junk, t')) 
      (S addr) (junk, t'))
    w t.
Proof.
  intros psi phi h h1 addr n n0 w t t'.
  intros Hh Hh1 Hw.
  inversion_clear Hw.
  constructor.
  constructor 2 with a; trivial.
  constructor.
  constructor.

  assert (Hh' : h ||- (fun h : heap => h = h1) 
    ** addr |-> n  ** top).
    simplsep 3 4.
  assert (Haddr := sep_heap_extend_heapty_get_none
    psi phi h1 h addr n Hh' Hh1).
  assert (Hh'' : h ||- (fun h : heap => h = h1) 
    ** S addr |-> n0  ** top).
    clear Hh'.
    simplsep 3 4.
  assert (Hsaddr := sep_heap_extend_heapty_get_none
    psi phi h1 h (S addr) n0 Hh'' Hh1).
  clear Hh' Hh''.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  destruct (nat_eq_dec addr (w + i)) as [Heq | Hneq].
(* case addr = w + i *)
  subst addr.
  substH Hctx' with (DHSpec.lookup_get Hctx').
  rewrite Hctx' in Haddr.
  discriminate Haddr.
(* case addr <> v + i *)
  destruct (nat_eq_dec (S addr) (w + i)) as [Heq' | Hneq'].
  rewrite Heq' in Hsaddr.
  substH Hctx' with (DHSpec.lookup_get Hctx').
  rewrite Hctx' in Hsaddr.
  discriminate Hsaddr.
  exists ctx; exists ctx'.
  split_l.
  split_r.  
  apply DHSpec.get_lookup.
  DHSpec.simplgs.
  apply DHSpec.lookup_get; trivial.
Qed.
