(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* main                                                        *)
(*     in TAL world                                            *)
(*                                                             *)
(***************************************************************)

(* $Id: talp_main.v,v 1.1.1.1 2006/12/07 16:43:18 ssg05 Exp $ *)

Require Import machine.
Require Import Talp.
Require Import extac.

Import TyEnv.

Open Local Scope tm_scope.
Open Local Scope tal_scope.

(* labels *)
Notation MAIN := 0100.
Notation CONT := 0102.

Notation NEWPAIR := 0005.
Notation MALLOC := 0009.

Definition main_code :=
(* 0100 *)  li r31 CONT;
(* 0101 *)  j NEWPAIR;;.

Definition cont_code :=
(* 0102 *)  j CONT;;.

Definition newpair_code :=
(* 0005 *)  addiu a0 r0 2 ; 
(* 0006 *)  j MALLOC;;.

(* Definition C2 : codeheap2 := *)
(*   (0005, newpair_code) :: *)
(*   (0100, main_code) :: *)
(*   (0102, cont_code) :: *)
(*   nil. *)

Definition C : codeheap := 
(* NEWPAIR *)
  (0005,   addiu a0 r0 2                ) ::
  (0006,   j MALLOC                     ) ::

(* MAIN *)
  (0100,   li r31 CONT                  ) ::
  (0101,   j NEWPAIR                    ) ::

(* CONT *)
  (0102,   j CONT                       ) ::
  nil.

(* specs *)

Definition main_spec :=
  (emp, [{ 
    (r1, int), 
    (s0, int),
    (s1, int),
    (s2, int),
    (s3, int),
    (s4, int),
    (s5, int),
    (s6, int),
    (s7, int)
  }]).

Definition cont_spec :=
  (emp, [{ 
    (r1, int),
    (s0, int),
    (s1, int),
    (s2, int),
    (s3, int),
    (s4, int),
    (s5, int),
    (s6, int),
    (s7, int),
    (v0, [< (junk, int), (junk, int) >] )
  }]).

Definition newpair_spec :=
  (9, [{
    (r1, int),
    (s0, ^0),
    (s1, ^1),
    (s2, ^2),
    (s3, ^3),
    (s4, ^4),
    (s5, ^5),
    (s6, ^6),
    (s7, ^7),
    (r31, [emp]- 
      [{
        (r1, int),
        (s0, ^0),
        (s1, ^1),
        (s2, ^2),
        (s3, ^3),
        (s4, ^4),
        (s5, ^5),
        (s6, ^6),
        (s7, ^7),
        (v0, [< (junk, ^8), (junk, ^8) >])
      }])
  }]).

(* codeheap spec *)
Definition psi : chspec :=
  (NEWPAIR, newpair_spec) ::
  (MAIN, main_spec) ::
  (CONT, cont_spec) :: nil.

Definition psi' : chspec :=
  (MAIN, main_spec) ::
  (CONT, cont_spec) :: nil.

(* proof *)

Theorem C_wf : wfch psi psi' C.
Proof.
  tyck_ch C.
Qed.
