(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Map                                                         *)
(*    using coq list                                           *)
(*                                                             *)
(***************************************************************)

(* $Id: ListB_MAP.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Require Import Coq.Bool.Bool.
Require Import Coq.Lists.List.
Require Import bnat.

Require Export MAP.

Notation Local optionX := option.

Notation Local SomeX := Some.

Notation Local NoneX := None.
Require Import extac.

Set Implicit Arguments.

Module ListMap (BDef : ONE_SET) <: MAPDEF.

Definition A := nat.

Definition beq : A -> A -> bool := beq_nat.

Lemma beq_true_eq : forall a a' : A, 
  beq a a' = true -> a = a'.
Proof.
  apply beq_true_eq.
Qed.

Lemma beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.
Proof.
  apply beq_false_neq.
Qed.

Lemma eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.
Proof.
  apply eq_beq_true.
Qed.

Lemma neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
Proof.
  apply neq_beq_false.
Qed.

Lemma eq_dec : forall a a' : A, {a=a'}+{a<>a'}.
Proof.
  apply nat_eq_dec.
Qed.

Definition B := BDef.B.

(***********************)
(** map                *)
(***********************)

Definition map := list (prod A B).

Definition emp_map : map := nil.

Definition sig_map (a : A) (b : B) := (a, b) :: nil.

(***********************)
(** set                *)
(** get                *)
(***********************)

Fixpoint get (m : map) (a : A) {struct m} : option B :=
  match m with 
    | nil => None
    | (a', b') :: m' =>
      match beq a a' with
        | true => Some b'
        | false => get m' a
      end
  end.

Lemma get_dec :
  forall (m : map) (a : A),
    {exists b, get m a = Some b} + {get m a = None}.
Proof.
  intros m a.
  destruct (get m a).
  left.
  exists b.
  trivial.
  right; trivial.
Qed.

Fixpoint set (m : map) (a : A) (b : B) {struct m} : map :=
  match m return map with
    | nil => (a, b) :: nil
    | (a', b') :: m' =>
      match beq a a' with
        | true => (a, b) :: m'
        | false => (a', b') :: (set m' a b)
      end
  end.

Fixpoint upd (m : map) (a : A) (b : B) {struct m} : option map :=
  match m return option map with
      | nil => None
      | (a', b') :: m' =>
        match beq a a' with
          | true => Some ((a, b) :: m')
          | false => 
              let m'' := (upd m' a b) in
                match m'' with 
                  | None => None
                  | Some m' => Some ((a', b') :: m')
                end
        end
  end.

(*
Lemma eq_not_lt : forall a b, a = b -> lt a b -> False.
Proof.
  intros a b.
  intros H1 H2.
  subst a.
  apply (lt_irrefl b H2).
Qed.

Ltac eqdec_eq :=  
  match goal with
    | H: context [eq_dec ?a ?a] |- _ 
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a a) as [He | Hne ]; 
              [ clear He | discri ])
    | |- context [eq_dec ?a ?a]
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a a) as [He | Hne ]; 
              [ clear He | discri ])
    | _ => fail 1 "no eq_dec term found "
  end.

Ltac eqdec_neq :=  
  match goal with
    | H: context [eq_dec ?a ?b] |- _ 
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a b) as [He | Hne];
              [
                match goal with 
                  | Hlt : lt a b |- _ =>
                    destruct (eq_not_lt a b He Hlt)
                  | Hlt : lt b a |- _ => 
                    destruct (eq_not_lt b a He Hlt)
                  | Hneq : not (eq a b) |- _ =>
                    destruct (Hneq He)
                  | Hneq : not (eq b a) |- _ =>
                    destruct (Hneq (sym_eq He))
                  | _ => fail 1 "xx"
                end
              | idtac ])
    | |- context [eq_dec ?a ?b]
      =>  let He := fresh "He" in
          let Hne := fresh "Hne" in
            (destruct (eq_dec a b) as [He | Hne];
              [
                match goal with 
                  | Hlt : lt a b |- _ =>
                    destruct (eq_not_lt a b He Hlt)
                  | Hlt : lt b a |- _ => 
                    destruct (eq_not_lt b a He Hlt)
                  | Hneq : not (eq a b) |- _ =>
                    destruct (Hneq He)
                  | Hneq : not (eq b a) |- _ =>
                    destruct (Hneq (sym_eq He))
                  | _ => fail
                end
              | idtac])
    | _ => fail 1 "no eq_dec term found "
  end.
*)
(*
Lemma set_sem : forall m a a' b,
  get (set m a b) a' =
  match (beq a a') with
    | true => Some b
    | false => get m a'
  end.
Proof.
  intros.
  unfold get, set, beq.
*)
Lemma set_sem_old : forall m a a' b, 
  match (beq a a') with
    | true => get (set m a b) a = Some b
    | false => get (set m a b) a' = get m a'
  end.
Proof.
  intros m a a0 b.
  induction m as [|(a',b') m' IHm']; 
    unfold get, set, beq in * ; simplbnat.
  decbeqnat a0 a; trivial.
  fold get in * .
  fold set in * .
  decbeqnat a a0; 
    decbeqnat a a'; 
    decbeqnat a0 a'; 
    simpl; unfold beq; simplbnat; trivial.
  discribnat.
Qed.

Lemma set_sem : forall m a a' b,
  get (set m a b) a' =
  match (beq a a') with
  | true => Some b
  | false => get m a'
  end.
Proof.
  assert (H:=set_sem_old).
  intros.
  substH H with (H m a a' b).
  assert (H1:=beq_true_eq a a').
  assert (H2:=beq_false_neq). substH H2 with (H2 a a').
  destruct beq.
  assert (H3 := refl_equal).
  substH H3 with (H3 bool true).
  substH H1 with (H1 H3).
  rewrite H1; rewrite H1 in H; trivial.
  trivial.
Qed.

Lemma upd_sem_0 : forall m a b b0, 
  (get m a = Some b -> upd m a b0 = Some (set m a b0))
  /\ (get m a = None -> upd m a b0 = None).
Proof.
  intros.
  split.
  intros.
  induction m as [|(a',b') m' IHm'].
  discriminate.
  simpl in * ; unfold beq in * .
  decbeqnat a a'; trivial.
  rewrite (IHm' H).
  trivial.
  induction m as [|(a',b') m' IHm'].
  trivial.
  simpl in * ; unfold beq in * .
  decbeqnat a a'; trivial.
  intro Hf; discriminate Hf.
  intros H.
  rewrite (IHm' H).
  trivial.
Qed.

Lemma upd_sem_old : forall m a b,
    match get_dec m a with 
      | left _ => upd m a b = Some (set m a b)
      | right _ => upd m a b = None
    end.
Proof.
  intros.
  destruct (get_dec m a) as [[b' Hi] | Hni].
  assert (H := proj1 (upd_sem_0 m a b' b)).
  tauto.
  assert (H := proj2 (upd_sem_0 m a b b)).
  tauto.
Qed.

Lemma upd_sem : forall m a b,
  upd m a b =
  match get_dec m a with
    | left _ => Some (set m a b)
    | right _ => None
  end.
Proof.
  intros.
  assert (H := upd_sem_old).
  substH H with (H m a b).
  destruct get_dec; trivial.
Qed.

Lemma emp_sem : 
  forall a : A, get emp_map a = NoneX.
Proof.
  intros; unfold get, emp_map; trivial.
Qed.

Lemma sig_sem : forall (a a' : A) (b : B),
  get (sig_map a b) a' = 
  match (beq a a') with
    | true => Some b
    | false => None
  end.
Proof.
  unfold get, beq, sig_map, beq.
  intros a a' b.
  decbeqnat a' a; trivial.
Qed.

(************************)
(** merge               *)
(************************)

Fixpoint merge (m1 m2 : map) {struct m1} : map := 
  match m1 with 
    | nil => m2
    | (a, b) :: m1' => (a, b) :: (merge m1' m2)
  end.

Lemma merge_sem : forall (m1 m2 : map) (a : A),
  get (merge m1 m2) a 
  = match get m1 a, get m2 a with
      | Some b1, Some b2 => Some b1
      | Some b1, None => Some b1
      | None, Some b2 => Some b2
      | None, None => None
    end.
Proof.
  intros m1 m2 a. unfold merge.
  induction m1 as [| (a1, b1) m1' IHm1'].
  unfold get at 2.
  destruct (get m2 a); trivial.
  unfold merge in *; fold merge in *.
  unfold get; fold get.
  destruct (eq_dec a a1) as [Heq | Hneq].
  subst a1.
  unfold beq; simplbnat.
  destruct (get m2 a).
  trivial.
  trivial.
  unfold beq; simplbnat.
  assumption.
Qed.

End ListMap.

Unset Implicit Arguments.
