(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* MAP Interface                                               *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: SETMAP.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Notation Local optionX := option.

Notation Local SomeX := Some.

Notation Local NoneX := None.

Notation Local B_U := Set.

Require Import extac. 

Set Implicit Arguments.
Unset Strict Implicit.

Module Type ONE_SET.

  Parameter B : Set.

End ONE_SET.

Module Type MAPDEF.
  
  Parameter A : Set.
  
  Parameter beq : A -> A -> bool.

  Axiom beq_true_eq : forall a a' : A, 
    beq a a' = true -> a = a'.

  Axiom beq_false_neq : forall a a' : A,
    beq a a' = false -> a <> a'.

  Axiom eq_beq_true : forall a a' : A, 
    a = a' -> beq a a' = true.

  Axiom neq_beq_false : forall a a' : A, 
    a <> a' -> beq a a' = false.
  
  Axiom eq_dec : forall a a' : A, {a=a'}+{a<>a'}.

  Parameter B : B_U.

(***********************)
(** map                *)
(***********************)

  Parameter map : B_U.
  
  Parameter emp_map : map.

  Parameter sig_map : A -> B -> map.

(***********************)
(** get set upd        *)
(***********************)

  Parameter get : map -> A -> optionX B.

  Axiom get_dec : forall (m : map) (a : A), 
    {exists b, get m a = SomeX b} + {get m a = NoneX}.

  Parameter set : map -> A -> B -> map.

  Axiom set_sem : forall m a a' b,
    get (set m a b) a' = 
    match (beq a a') with
      | true => SomeX b
      | false => get m a'
    end.

(*   Parameter upd : map -> A -> B -> optionX map. *)

(*   Axiom upd_sem : forall m a b, *)
(*     match get_dec m a with  *)
(*       | left _ => upd m a b = SomeX (set m a b) *)
(*       | right _ => upd m a b = NoneX *)
(*     end. *)

  Axiom emp_sem : forall a : A, get emp_map a = NoneX.
  
  Axiom sig_sem : forall (a a' : A) (b : B), 
    get (sig_map a b) a' = 
    match (beq a a') with
      | true => SomeX b
      | false => NoneX
    end.

  Axiom extensionality: forall m1 m2, 
    (forall a, get m1 a = get m2 a) -> m1 = m2.

(************************)
(** merge               *)
(************************)

  Parameter merge : map -> map -> map.

  Axiom merge_sem : forall (m1 m2 : map) (a : A),
    get (merge m1 m2) a 
    = match (get m1 a, get m2 a) with
        | (SomeX b1, SomeX b2) => SomeX b1
        | (SomeX b1, NoneX) => SomeX b1
        | (NoneX, SomeX b2) => SomeX b2
        | (NoneX, NoneX) => NoneX
      end.

End MAPDEF.


Module MapEx (MapDef : MAPDEF).

Import MapDef.

Load "common_map.v".

(* Definition upd : map -> A -> B -> optionX map *)
(*   := MapDef.upd. *)

(* Lemma upd_sem : forall m a b, *)
(*   match get_dec m a with  *)
(*     | left _ => upd m a b = SomeX (set m a b) *)
(*     | right _ => upd m a b = NoneX *)
(*   end. *)
(* Proof MapDef.upd_sem. *)

(* Definition canupd (m : map) (a : A) (b : B) (m' : map) := *)
(*   upd m a b = SomeX m'. *)

Lemma extensionality: forall m1 m2, 
  (forall a, get m1 a = get m2 a) -> m1 = m2.
Proof MapDef.extensionality.

Lemma set_sig_a_sig : forall (a : A) (b b' : B),
  set (sig_map a b) a b' = sig_map a b'.
Proof.
  intros a b b'.
  apply extensionality.
  intros a0.
  assert (H := set_sem (sig_map a b) a a0 b').
  rewrite (sig_sem a a0 b').
  rewrite (sig_sem a a0 b) in H.
  destruct (eq_dec a a0); simplbeq;
    trivial.
Qed.

(* Lemma upd_a_get_a : forall (m m' : map) (a : A) (b : B), *)
(*   upd m a b = SomeX m' *)
(*   -> exists b', get m a = SomeX b'. *)
(* Proof. *)
(*   intros m m' a b Hu. *)
(*   assert (H := upd_sem m a b). *)
(*   destruct (get_dec m a). *)
(*   assumption. *)
(*   rewrite H in Hu; discriminate. *)
(* Qed. *)

(* Lemma get_a_upd_a : forall (m : map) (a : A) (b b' : B), *)
(*   get m a = SomeX b *)
(*   -> upd m a b' = SomeX (set m a b'). *)
(* Proof. *)
(*   intros. *)
(*   assert (H0 := upd_sem m a b'). *)
(*   destruct (get_dec m a). *)
(*   assumption. *)
(*   rewrite e in H. *)
(*   discriminate. *)
(* Qed. *)

(* Lemma upd_a_get_a' : forall (m m' : map) (a a' : A) (b : B), *)
(*   a <> a' *)
(*   -> upd m a b = SomeX m' *)
(*   -> get m' a' = get m a'. *)
(* Proof. *)
(*   intros m m' a a' b H H0. *)
(*   assert (H1 := upd_sem m a b). *)
(*   destruct (get_dec m a). *)
(*   rewrite H0 in H1. *)
(*   injection H1; intro He; subst m'. *)
(*   apply set_a_get_a'; trivial. *)
(*   rewrite H0 in H1. *)
(*   discriminate. *)
(* Qed. *)

(* Lemma canupd_get : forall (m m' : map) (a : A) (b : B), *)
(*   canupd m a b m' *)
(*   -> exists b', get m a = SomeX b'. *)
(* Proof. *)
(*   unfold canupd; apply upd_a_get_a. *)
(* Qed. *)

(* Lemma upd_lookup : forall (m m' : map) (a : A) (b : B), *)
(*   upd m a b = SomeX m' *)
(*   -> lookup m' a b. *)
(* Proof. *)
(*   intros m m' a b H. *)
(*   apply get_lookup. *)
(*   assert (H0 := upd_sem m a b). *)
(*   destruct (get_dec m a). *)
(*   rewrite H in H0.  *)
(*   injection H0; intro H1.  *)
(*   rewrite H1.  *)
(*   apply set_a_get_a. *)
(*   rewrite H in H0; discriminate. *)
(* Qed. *)

(* Lemma canupd_lookup : forall (m m' : map) (a : A) (b : B), *)
(*   canupd m a b m' -> lookup m' a b. *)
(* Proof. *)
(*   intros; unfold canupd in H. eapply upd_lookup; eauto. *)
(* Qed. *)

(* Lemma canupd_lookup_ex : forall (m m' : map) (a : A) (b : B), *)
(*   canupd m a b m' -> exists b', lookup m a b'. *)
(* Proof. *)
(*   intros. *)
(*   assert(H0 := canupd_get H). *)
(*   destruct H0 as [b' Hget]. *)
(*   exists b'; apply get_lookup; trivial. *)
(* Qed. *)

(* Lemma lookup_canupd : forall (m : map) (a : A) (b b': B), *)
(*   lookup m a b -> canupd m a b' (set m a b'). *)
(* Proof. *)
(*   intros. *)
(*   unfold canupd. *)
(*   eapply get_a_upd_a. *)
(*   apply lookup_get. *)
(*   eauto. *)
(* Qed. *)

(* Lemma canupd_lookup_neq :  *)
(*   forall (m m' : map) (a a' : A) (b b' : B), *)
(*     a <> a'  *)
(*     -> canupd m a' b' m'  *)
(*     -> lookup m a b  *)
(*     -> lookup m' a b. *)
(* Proof. *)
(*   unfold canupd, lookup; intros. *)
(*   assert (H2 : a' <> a). *)
(*   intro Hf; subst a; apply H; trivial. *)
(*   substH H0 with (upd_a_get_a' H2 H0). *)
(*   rewrite H0; trivial. *)
(* Qed. *)

Lemma meq_eq : forall (m1 m2 : map),
  meq m1 m2 -> m1 = m2.
Proof.
  intros m1 m2 Hmeq.
  apply extensionality.
  intros a.
  substH Hmeq with (meq_sem Hmeq a).
  destruct (get m1 a);
    destruct (get m2 a);
      trivial.
  subst b; trivial.
  discri.
  discri.
Qed.

Lemma sub_asym' : forall (m1 m2 : map),
  sub m1 m2 -> sub m2 m1 -> m1 = m2.
Proof.
  intros m1 m2 H1 H2.
  apply meq_eq.
  apply sub_asym; trivial.
Qed.

(***********************************************************)
(**                                                        *)
(**  eq merge                                              *)
(**                                                        *)
(***********************************************************)

Lemma eq_merge_sym : forall (m1 m2 : map),
  compat m1 m2
  -> (merge m1 m2) = (merge m2 m1).
Proof.
  intros m1 m2 Hcomp.
  apply meq_eq.
  apply meq_merge_sym.
  trivial.
Qed.

Lemma eq_merge_assoc_l : forall (m1 m2 m3 : map),
  (merge (merge m1 m2) m3) =
  (merge m1 (merge m2 m3)).
Proof.
  intros m1 m2 m3.
  apply meq_eq.
  apply meq_merge_assoc_l; trivial.
Qed.

Lemma eq_merge_assoc_r : forall (m1 m2 m3 : map),
  (merge m1 (merge m2 m3)) =
  (merge (merge m1 m2) m3).
Proof.
  intros m1 m2 m3.
  apply meq_eq.
  apply meq_merge_assoc_r; trivial.
Qed.

Lemma eq_merge_emp_l : forall (m : map),
  (merge emp_map m) = m.
Proof.
  intros m. apply meq_eq.
  apply meq_merge_emp_l; trivial.
Qed.

Lemma eq_merge_emp_r : forall (m : map),
  (merge m emp_map) = m.
Proof.
  intros m. apply meq_eq.
  apply meq_merge_emp_r; trivial.
Qed.

Lemma eq_merge_eq : 
  forall m1 m2 m3 m4,
    eq m1 m3 -> eq m2 m4
    -> eq (merge m1 m2) (merge m3 m4).
Proof.
  intros m1 m2 m3 m4 H1 H2.
  apply meq_eq.
  apply meq_merge_meq.
  rewrite H1; apply meq_refl.
  rewrite H2; apply meq_refl.
Qed.

Lemma join_disj_eq :
  forall (m1 m2 m : map),
    join m1 m2 m
    -> disj m1 m2 /\ (merge m1 m2) = m.
Proof.
  intros m1 m2 m H.
  split.
  apply (proj1 (join_disj_meq H)).
  apply meq_eq.
  apply (proj2 (join_disj_meq H)).
Qed.

Lemma disj_eq_join :
  forall (m1 m2 m: map),
    disj m1 m2 /\ (merge m1 m2) = m
    -> join m1 m2 m.
Proof.
  intros m1 m2 m [Hd Hm].
  apply disj_meq_join.
  split_l.
  rewrite Hm; apply meq_refl.
Qed.

Lemma join_id_eq_r :
  forall m1 m2 m3 m, 
  join m1 m2 m 
  -> join m1 m3 m
  -> m2 = m3.
Proof.
  intros m1 m2 m3 m.
  unfold join; intros H1 H2.
  apply extensionality.
  intros a0.
  substH H1 with (H1 a0).
  substH H2 with (H2 a0).
  destruct (get m1 a0);
    destruct (get m2 a0);
      destruct (get m3 a0); 
        destruct (get m a0); 
          tauto || auto.
  rewrite H1; rewrite H2.
  trivial.
Qed.

Lemma join_id_eq_l :
  forall m1 m2 m3 m, 
  join m1 m2 m 
  -> join m3 m2 m
  -> m1 = m3.
Proof.
  intros m1 m2 m3 m.
  unfold join; intros H1 H2.
  apply extensionality.
  intros a0.
  substH H1 with (H1 a0).
  substH H2 with (H2 a0).
  destruct (get m1 a0);
    destruct (get m2 a0);
      destruct (get m3 a0); 
        destruct (get m a0); 
          tauto || auto.
  rewrite H1; rewrite H2.
  trivial.
Qed.

Lemma join_emp_l : forall m1 m,
  join emp_map m1 m -> m1 = m.
Proof.
  unfold join.
  intros m1 m H.
  apply extensionality.
  intro a.
  substH H with (H a).
  rewrite (emp_sem a) in H.
  destruct (get m1 a); destruct (get m a); simpl;
    intros; try tauto ; try auto.
  subst; trivial.
Qed.

Lemma join_emp_r : forall m1 m,
  join m1 emp_map m -> m1 = m.
Proof.
  unfold join.
  intros m1 m H.
  apply extensionality.
  intro a.
  substH H with (H a).
  rewrite (emp_sem a) in H.
  destruct (get m1 a); destruct (get m a); simpl;
    intros; try tauto ; try auto.
  subst; trivial.
Qed.

Lemma join_sig_r :
  forall m m1 a b,
    join m1 (sig_map a b) m
    -> set m1 a b = m.
Proof.
  unfold join.
  intros m m1 a b Hj.
  apply extensionality.
  intro a0.
  substH Hj with (Hj a0).
  rewrite (sig_sem a a0 b) in Hj.
  destruct (eq_dec a a0); simplbeq; trivial.
  subst a0.
  destruct (get m1 a); destruct (get m a); 
    tauto|| auto.
  rewrite Hj. rewrite set_sem. simplbeq. trivial.
  rewrite (set_a_get_a' m1 b n).
  destruct (get m1 a0); destruct (get m a0);
    tauto || auto.
  rewrite Hj; auto.
Qed.

Lemma join_eq : forall m1 m2 m m',
  join m1 m2 m 
  -> join m1 m2 m'
  -> m = m'.
Proof.
  unfold join.
  intros m1 m2 m m' H1 H2.
  apply extensionality.
  intro a.
  substH H1 with (H1 a).
  substH H2 with (H2 a).
  destruct (get m1 a); destruct (get m2 a); 
    destruct (get m a); destruct (get m' a); 
      try subst; try tauto; auto.
  subst; trivial.
  subst; trivial.
Qed.

(***********************************************************)
(**                                                        *)
(**    Tactics                                             *)
(**                                                        *)
(***********************************************************)

End MapEx.

Unset Implicit Arguments.
