(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* OCAP                                                        *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: ocap.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Export Eqdep.
Require Export optionT.
Require Export prodT.

Require Import ListT.
Require Import machine.
Require Import FuncB_MAPT.
Require Import MultiMapT.

Require Import extac.
Require Import deptac.

Open Local Scope tm_scope.

(* rho : langid *)
Definition langid := nat.

(* (* L : langty *) *)
Definition langty := Type.

(* theta : xcdspec *)
Inductive xcdspec : Type :=
  | xcdspec_pack : forall (lty : Type) (theta : lty), xcdspec.
Implicit Arguments xcdspec_pack [lty].

Notation "<* theta *>" := (xcdspec_pack theta)
  (at level 0) : ocap_scope.

Definition proj1_xcdspec (xtheta : xcdspec) : Type := 
  match xtheta with
    | xcdspec_pack lty theta => lty
  end.

Definition proj2_xcdspec
  : forall xtheta : xcdspec, (proj1_xcdspec xtheta) :=
  fun xtheta : xcdspec => 
    match xtheta return (proj1_xcdspec xtheta) with
      | xcdspec_pack lty theta => theta
    end.

(* pi : ocdspec *)
Inductive ocdspec : Type :=
  | mk_pi : langid -> xcdspec -> ocdspec.
(* Definition ocdspec := (prodT langid xcdspec). *)

Notation "<* langid , theta *> " := 
  (mk_pi langid (xcdspec_pack theta)) (at level 0).

Module OCdSpec <: ONE_TYPE.
  Definition B := ocdspec.
End OCdSpec.

Module OCHSpec := MultiMap OCdSpec.
(* psi : chspec *)
Definition chspec := OCHSpec.map.

(* a : assert *)
Definition assert := chspec -> state -> Prop.

Definition aimp (a a' : assert) : Prop := 
  forall (psi : chspec) (s : state),
    a psi s -> a' psi s.

Notation " a ===> a' " := (aimp a a')
  (at level 75, right associativity) : ocap_scope.

(* Definition interp (lty : Type) : Type := lty -> assert. *)

(* id_interp *)
Definition Locap := assert.
Definition itpocap (a : Locap) : assert:= a.

Inductive xinterp : Type := 
  | xinterp_pack : forall (lty : Type) (itp : lty -> assert), xinterp.
Implicit Arguments xinterp_pack [lty].

Notation "<@ itp @>" := (xinterp_pack itp)
  (at level 0) : ocap_scope.

Definition proj1_xinterp (xitp : xinterp) : Type := 
  match xitp with
    | xinterp_pack lty itp => lty
  end.
   
Definition proj2_xinterp 
  : forall xitp : xinterp, (proj1_xinterp xitp) -> assert:=
  fun xitp : xinterp => 
    match xitp return (proj1_xinterp xitp) -> assert with
      | xinterp_pack lty itp => itp
    end.

Open Local Scope ocap_scope.

Module MInterp <: ONE_TYPE.
  Definition B := xinterp.
End MInterp.

Module DSpecDef := FuncMapT MInterp.
Module DSpec := MapEx DSpecDef.

Definition idict : Type := DSpec.map.

Definition BigInterp (D : idict) (pi : ocdspec) : assert :=
  let (id, xtheta) := pi in
    let lty := proj1_xcdspec xtheta in
      let theta := proj2_xcdspec xtheta in
        fun (psi : chspec) (s : state) =>
          exists itp : lty -> assert,
            DSpec.lookup D id <@itp@>
            /\ itp theta psi s.
  
Notation " {[ pi | d ]} " := (BigInterp d pi) 
  (at level 0) : ocap_scope.

Lemma biginterp_enlarge_D :
  forall (D1 D2 : idict) (pi : ocdspec),
    DSpec.sub D1 D2
    -> forall (psi : chspec) (s : state), 
      {[ pi | D1 ]} psi s -> {[ pi | D2 ]} psi s.
Proof.
  intros D1 D2 pi Hsub.
  unfold BigInterp.
  intros psi s.
  intros H.
  destruct pi as [id xtheta].
  destruct xtheta as [lty theta].
  destruct H as [itp [Hin Hitp]].
  exists itp.
  split.
  apply (DSpec.lookup_sub_lookup Hin Hsub).
  apply Hitp.
Qed.
Implicit Arguments biginterp_enlarge_D [D1 D2 pi].

Definition describe (D : idict) (psi : chspec) : Prop :=
  forall (f : label) (id : langid) (lty : Type) (theta : lty),
    OCHSpec.lookup psi f <* id , theta *>
    -> exists itp : lty -> assert,
      DSpec.lookup D id <@itp@> .

Notation " D ## psi " := (describe D psi) 
  (at level 40) : ocap_scope.

Definition lift (psi : chspec) (a : assert) : assert :=
  fun (psi' : chspec) (s : state) =>
    OCHSpec.sub psi psi' /\ a psi' s.

Notation " << a / psi >> " := (lift psi a)
  (at level 0, a at level 30, psi at level 30) : ocap_scope.

Lemma describe_sub : 
  forall (D D' : idict) (psi : chspec),
    DSpec.sub D D'
    -> describe D psi
    -> describe D' psi.
Proof.
  intros D D' psi Hsub HD.
  unfold describe in *.
  intros f id lty theta Hinpsi.
  substH HD with (HD f id lty theta Hinpsi).
  destruct HD as [itp Hin].
  exists itp.
  apply DSpec.lookup_sub_lookup with (m1:=D);
    assumption.
Qed.

Definition describe_pi (D : idict) (pi : ocdspec) : Prop :=
  let (id, xtheta) := pi in
    let lty := proj1_xcdspec xtheta in
      exists itp : lty -> assert,
        DSpec.lookup D id <@itp@> .

Lemma describe_psi_describe_pi :
  forall (D : idict) (psi : chspec) (f : label) (pi : ocdspec),
    describe D psi 
    -> OCHSpec.lookup psi f pi
    -> describe_pi D pi.
Proof.
  intros D psi f pi Hdes Hin.
  unfold describe, describe_pi in *.
  destruct pi as [id xtheta].
  destruct xtheta as [lty theta].
  substH Hdes with (Hdes f id lty theta Hin).
  destruct Hdes as [itp HinD].
  exists itp.
  simpl.
  trivial.
Qed.

Lemma describe_merge :
  forall (D1 D2 : idict) (Psi1 Psi2 : chspec),
    DSpec.compat D1 D2
    -> describe D1 Psi1 
    -> describe D2 Psi2
    -> describe (DSpec.merge D1 D2) (OCHSpec.merge Psi1 Psi2).
Proof. 
  intros D1 D2 Psi1 Psi2 HDcomp HD1 HD2.
  unfold describe in *.
  intros f id lty theta Hfinpsi.
  destruct (OCHSpec.lookup_merge_elim Hfinpsi) 
    as [Hfinpsi1 | Hfinpsi2].
  substH HD1 with (HD1 f id lty theta Hfinpsi1).
  destruct HD1 as [itp HidinD1].
  exists itp.
  refine (DSpec.lookup_merge_intro _ HDcomp).
  left; assumption.
  substH HD2 with (HD2 f id lty theta Hfinpsi2).
  destruct HD2 as [itp HidinD2].
  exists itp.
  refine (DSpec.lookup_merge_intro _ HDcomp).
  right; assumption.
Qed.

Lemma describe_merge_2 :
  forall D (Psi1 Psi2 : chspec),
    describe D Psi1 
    -> describe D Psi2
    -> describe D (OCHSpec.merge Psi1 Psi2).
Proof. 
  intros D Psi1 Psi2 HD1 HD2.
  unfold describe in *.
  intros f id lty theta Hfinpsi.
  destruct (OCHSpec.lookup_merge_elim Hfinpsi) 
    as [Hfinpsi1 | Hfinpsi2].
  substH HD1 with (HD1 f id lty theta Hfinpsi1).
  destruct HD1 as [itp HidinD1].
  exists itp. trivial.
  substH HD2 with (HD2 f id lty theta Hfinpsi2).
  destruct HD2 as [itp HidinD2].
  exists itp. trivial.
Qed.

Lemma biginterp_enlarge_2 :
  forall (D1 D2 : idict) (pi1 pi2: ocdspec),
    DSpec.sub D1 D2
    -> (forall (psi : chspec) (s : state),
      ({[ pi1 | D1 ]} psi s -> {[ pi2 | D1 ]} psi s)
      /\ describe_pi D1 pi1
      /\ describe_pi D1 pi2)
    -> (forall (psi : chspec) (s : state),
      {[ pi1 | D2 ]} psi s -> {[ pi2 | D2 ]} psi s).
Proof.
  intros D1 D2 pi1 pi2 Hsub HD1.
  intros psi s.
  destruct (HD1 psi s) as [Himp [Hdpi1 Hdpi2]].
  clear HD1.
  unfold BigInterp in *.
  destruct pi1 as [id1 xtheta1].
  destruct pi2 as [id2 xtheta2].
  intro H.
  destruct H as [itp1 [Hin1 Hpi1]].
  (* prove D2 -> D1 *)
  assert (H : exists itp : proj1_xcdspec xtheta1 -> assert,
    DSpec.lookup D1 id1 <@itp@> /\
    itp (proj2_xcdspec xtheta1) psi s).
  destruct Hdpi1 as [itp1' Hdpi1].
  assert (H1 : DSpec.lookup D2 id1 <@itp1'@>).
  apply (DSpec.lookup_sub_lookup Hdpi1 Hsub).
  assert (H2 : itp1'=itp1).
  assert (H3 := DSpec.lookup_unique Hin1 H1).
  injection H3.
  intros He.
  deprewrite He.
  trivial.
  subst itp1'.
  exists itp1.
  split; assumption.
  (* prove D1 -> D2 *)
  substH Himp with (Himp H).
  destruct Himp as [itp2 [Hin2 Hpi2]].
  exists itp2.
  assert (H1 : DSpec.lookup D2 id2 <@itp2@>).
  apply (DSpec.lookup_sub_lookup Hin2 Hsub).
  split; assumption.
Qed.

Lemma biginterp_contract_D :
  forall (D D' : idict) (pi : ocdspec),
    DSpec.sub D D'
    -> describe_pi D pi
    -> forall (psi : chspec) (s : state),
      {[ pi | D' ]} psi s -> {[ pi | D ]} psi s.
Proof.
  intros D D' pi Hsub Hdes.
  unfold BigInterp.
  intros psi s.
  destruct pi as [id xtheta].
  intros [itp [Hin Hs]].
  exists itp.
  destruct Hdes as [itp' Hdes].
  assert (H : DSpec.lookup D' id <@itp'@>).
  apply (DSpec.lookup_sub_lookup Hdes Hsub).
  assert (H1 : itp' = itp).
  assert (H2 := DSpec.lookup_unique Hin H).
  injection H2.
  intros He.
  deprewrite He.
  trivial.
  subst itp'.
  split; assumption.
Qed.

Fixpoint wfis (D : idict) (a : assert)
  (is : instrseq) {struct is} : Prop :=
  match is with
    | ione i =>
      match i with
        | j f =>
          (forall Psi s, a Psi s ->
            (exists pi'' : ocdspec, 
              OCHSpec.lookup Psi f pi''
              /\ optp ({[ pi'' | D ]} Psi) (nextS (j f) s)))
        | jal f fret =>
          (forall Psi s, a Psi s ->
            (exists pi'' : ocdspec, 
              OCHSpec.lookup Psi f pi''
            /\ optp ({[ pi'' | D ]} Psi) (nextS (jal f fret) s)))
        | jr r =>
          (forall psi s, a psi s ->
            (exists pi'' : ocdspec,
              OCHSpec.lookup psi (S_gR s r) pi''
              /\ optp ({[ pi'' | D ]} psi) (nextS (jr r) s)))
        | _ => False
      end
    | iseq i is' =>
      match i with
        | j f => False
        | jal f fret => False
        | jr r => False
        | blt rs rt f =>
          (forall psi s, a psi s ->
            if blt_nat (S_gR s rs) (S_gR s rt) 
              then 
                exists pi'' : ocdspec,
                  OCHSpec.lookup psi f pi''
                  /\ optp ({[ pi'' | D ]} psi) (nextS (blt rs rt f) s)
              else 
                exists a' : assert,
                  optp (a' psi) (nextS (blt rs rt f) s)
                  /\ wfis D a' is')
          
        | beq rs rt f =>
          (forall psi s, a psi s ->
            if beq_nat (S_gR s rs) (S_gR s rt) 
              then 
                exists pi'' : ocdspec,
                  OCHSpec.lookup psi f pi''
                  /\ optp ({[ pi'' | D ]} psi) (nextS (beq rs rt f) s)
              else 
                exists a' : assert, 
                  optp (a' psi) (nextS (beq rs rt f) s)
                  /\ wfis D a' is')
          
        | bgtz rs f =>
          (forall psi s, a psi s ->
            if blt_nat O (S_gR s rs) 
              then 
                exists pi'' : ocdspec,
                  OCHSpec.lookup psi f pi''
                  /\ optp ({[ pi'' | D ]} psi) (nextS (bgtz rs f) s)
              else 
                exists a' : assert, 
                  optp (a' psi) (nextS (bgtz rs f) s)
                  /\ wfis D a' is')
        | i =>
          (forall Psi s, a Psi s ->
            exists a',
              optp (a' Psi) (nextS i s)
              /\ wfis D a' is')
      end
  end.

(*
Inductive wfis : idict -> assert -> instrseq -> Prop :=
  | wfis_j :
      forall (D : idict) (a : assert) f,
        (forall Psi s, a Psi s ->
          (exists pi'' : ocdspec, 
            OCHSpec.lookup Psi f pi''
            /\ optp ({[ pi'' | D ]} Psi) (nextS (j f) s)))
        -> wfis D a (ione (j f))
  | wfis_jr :
      forall (D : idict) (a : assert) r,
        (forall psi s, a psi s ->
          (exists pi'' : ocdspec,
            OCHSpec.lookup psi (S_gR s r) pi''
            /\ optp ({[ pi'' | D ]} psi) (nextS (jr r) s)))
        -> wfis D a (ione (jr r))
  | wfis_addu :
      forall (D : idict) (a : assert) rd rs rt (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (addu rd rs rt) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (addu rd rs rt) I) 
  | wfis_addiu :
      forall (D : idict) (a : assert) rd rs w (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (addiu rd rs w) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (addiu rd rs w) I) 
  | wfis_subu :
      forall (D : idict) (a : assert) rd rs rt (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (subu rd rs rt) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (subu rd rs rt) I) 
  | wfis_subiu :
      forall (D : idict) (a : assert) rd rs w (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (subiu rd rs w) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (subiu rd rs w) I) 
  | wfis_move :
      forall (D : idict) (a : assert) rd rs (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (move rd rs) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (move rd rs) I) 
  | wfis_li :
      forall (D : idict) (a : assert) rd w (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (li rd w) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (li rd w) I) 
  | wfis_lw :
      forall (D : idict) (a : assert) rt w rs (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (lw rt w rs) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (lw rt w rs) I)
  | wfis_sw :
      forall (D : idict) (a : assert) rt w rs (I : instrseq),
        (exists a',
            (forall Psi s, a Psi s ->
              optp (a' Psi) (nextS (sw rt w rs) s))
            /\ wfis D a' I)
        -> wfis D a (iseq (sw rt w rs) I) 
  | wfis_blt :
      forall (D : idict) (a : assert) rs rt f (I : instrseq),
        (exists a' : assert, 
          (forall psi s, a psi s ->
            if blt_nat (S_gR s rs) (S_gR s rt) 
              then 
                exists pi'' : ocdspec,
                  OCHSpec.lookup psi f pi''
                  /\ optp ({[ pi'' | D ]} psi) (nextS (blt rs rt f) s)
              else optp (a' psi) (nextS (blt rs rt f) s))
          /\ wfis D a' I)
        -> wfis D a (iseq (blt rs rt f) I) 
  | wfis_bgtz :
      forall (D : idict) (a : assert) rs f (I : instrseq),
        (exists a' : assert, 
          (forall psi s, a psi s ->
            if blt_nat O (S_gR s rs) 
              then 
                exists pi'' : ocdspec,
                  OCHSpec.lookup psi f pi''
                  /\ optp ({[ pi'' | D ]} psi) (nextS (bgtz rs f) s)
              else optp (a' psi) (nextS (bgtz rs f) s))
          /\ wfis D a' I)
        -> wfis D a (iseq (bgtz rs f) I)
.*)

Lemma wfis_weak_D :
  forall (D D': idict) (a : assert) (is : instrseq),
    wfis D a is
    -> DSpec.sub D D'
    -> wfis D' a is.
Proof.
  intros D D' a is HDis Hsub.
  gen_clear a HDis.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
  simpl; intros a Himp; trivial.
  (* case i = j f *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct s as [hp rf pc].
  destruct Himp as [pi'' [Hin Hpi'']].
  exists pi''; split_l.
  apply (biginterp_enlarge_D Hsub Psi _ Hpi'').
  (* case i = jal f fret *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct s as [hp rf pc].
  destruct Himp as [pi'' [Hin Hpi'']].
  exists pi''; split_l.
  apply (biginterp_enlarge_D Hsub Psi _ Hpi'').
  (* case i = jr r *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct s as [hp rf pc].
  destruct Himp as [pi'' [Hin Hpi'']].
  exists pi''; split_l.
  apply (biginterp_enlarge_D Hsub Psi _ Hpi'').
  (* case iseq *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
  unfold wfis; fold wfis; intros a Himp; trivial.
  (* case i = addu rd rs rt *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = addiu rd rs w *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = subu rd rs rt *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = subiu rd rs w *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = move rd rs *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = li rd w *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = lw rd w rs *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = sw rd w rs *)
  intros Psi s Has.
  substH Himp with (Himp Psi s Has).
  destruct Himp as [a' [Himp HDis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = blt rs rt f *)
  intros Psi s Hs.
  substH Himp with (Himp Psi s Hs).
  destruct s as [hp rf pc].
  unfold nextS in Himp |- * .
  decbltnat (gR rf rs) (gR rf rt).
    destruct Himp as [pi'' [Hin Hpi'']].
    exists pi''; split_l.
    apply (biginterp_enlarge_D Hsub Psi _ Hpi'').
  destruct Himp as [a' [Ha' Hwfis]].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = beq rs rt f *)
  intros Psi s Hs.
  substH Himp with (Himp Psi s Hs).
  destruct s as [hp rf pc].
  unfold nextS in Himp |- * .
  decbeqnat (gR rf rs) (gR rf rt).
    destruct Himp as [pi'' [Hin Hpi'']].
    exists pi''; split_l.
    apply (biginterp_enlarge_D Hsub Psi _ Hpi'').
  destruct Himp as [a' [Himp Ha']].
  exists a'; split_l.
  apply IHis; trivial.
  (* case i = bgtz rs f *)
  intros Psi s Hs.
  substH Himp with (Himp Psi s Hs).
  destruct s as [hp rf pc].
  unfold nextS in Himp |- * .
  decbltnat O (gR rf rs).
    destruct Himp as [pi'' [Hin Hpi'']].
    exists pi''; split_l.
    apply (biginterp_enlarge_D Hsub Psi _ Hpi'').
  destruct Himp as [a' [Himp Ha']].
  exists a'; split_l.
  apply IHis; trivial.
Qed.
Implicit Arguments wfis_weak_D [D D' a is].

Lemma wfis_weak_a :
  forall (D : idict) (a a0 : assert) (is : instrseq),
    wfis D a is
    -> (forall (psi : chspec) (s : state), a0 psi s -> a psi s)
    -> wfis D a0 is.
Proof.
  intros D a a0 is.
  destruct is.  
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
  intros Himp Himp0; trivial.
  (* case i = j f *)
  intros Psi s Ha0s.
  substH Ha0s with (Himp0 Psi s Ha0s) into Has.
  destruct  (Himp Psi s Has) 
    as [pi'' [Hin Hpi'']].
  exists pi''; split_lr.
  (* case i = jal f fret*)
  intros Psi s Ha0s.
  substH Ha0s with (Himp0 Psi s Ha0s) into Has.
  destruct  (Himp Psi s Has) 
    as [pi'' [Hin Hpi'']].
  exists pi''; split_lr.
  (* case i = jr r *)
  intros Psi s Ha0s.
  substH Ha0s with (Himp0 Psi s Ha0s) into Has.
  destruct  (Himp Psi s Has) 
    as [pi'' [Hin Hpi'']].
  exists pi''; split_lr.
  (* case iseq *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
    unfold wfis; fold wfis; 
    intros Himp Himp0; trivial.
  (* case i = addu rd rs rt *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = addiu rd rs w *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = subu rd rs rt *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = subiu rd rs w *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = move rd rs *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = li rd w *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = lw rd w rs *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = sw rd w rs *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = blt rs rt f *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = beq rs rt f *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
  (* case i = bgtz rs f *)
  intros Psi s Ha0s.
  apply Himp.
  apply Himp0; trivial.
Qed.
Implicit Arguments wfis_weak_a [D a a0 is].

Lemma wfis_weak_D_pi :
  forall (D D': idict) (pi : ocdspec) (is : instrseq),
    DSpec.sub D D'
    -> describe_pi D pi
    -> wfis D {[pi | D]} is
    -> wfis D' {[pi | D']} is.
Proof.
  intros D D' pi is HsubD Hdespi HwfisD.
  assert (H0 := biginterp_contract_D D D' pi HsubD Hdespi).
  apply wfis_weak_a with (a := {[pi | D ]}).
  apply (wfis_weak_D HwfisD HsubD).
  assumption.
Qed.
Implicit Arguments wfis_weak_D_pi [D D' pi is].

Lemma wfis_weak_a_lift :
  forall (D : idict) (psi0 : chspec ) (a a0 : assert)
    (is : instrseq),
    wfis D <<a/psi0>> is
    -> a0 ===> a
    -> wfis D <<a0/psi0>> is.
Proof.
  intros D psi0 a a0 is HaI Himp.
  assert (H : lift psi0 a0 ===> lift psi0 a).
  unfold lift.
  unfold aimp in *.
  intros psi s [Hsub Ha].
  split_l.
  apply (Himp psi s Ha).
  unfold aimp in H.
  apply (wfis_weak_a HaI H).
Qed.
Implicit Arguments wfis_weak_a_lift [D psi0 a a0 is].

Lemma wfis_weak_D_pi_lift :
  forall (D D': idict) (psi0 : chspec )(pi : ocdspec) 
    (is : instrseq),
    DSpec.sub D D'
    -> describe_pi D pi
    -> wfis D <<{[pi | D]} / psi0>> is
    -> wfis D' <<{[pi | D']} / psi0>> is.
Proof.
  intros D D' psi0 pi is HsubD Hdespi HisD.
  assert (H0 : wfis D' (lift psi0 {[ pi | D ]}) is).
  apply wfis_weak_D with (D := D);
    assumption.
  assert (H1 :  {[ pi | D' ]}  ===> {[ pi | D ]}).
  unfold aimp.
  apply (biginterp_contract_D D D' pi HsubD Hdespi).
  apply (wfis_weak_a_lift H0 H1).
Qed.
Implicit Arguments wfis_weak_D_pi_lift [D D' psi0 pi is].

Lemma wfis_weaken_lift_Psi :
  forall (D : idict) (Psi Psi': chspec) (a : assert) (is : instrseq),
    wfis D << a / Psi >> is 
    -> OCHSpec.sub Psi Psi'
    -> wfis D << a / Psi'>> is.
Proof.
  intros D Psi Psi' a is His Hsub.
  apply wfis_weak_a with (a := << a / Psi >> ); trivial.
  intros Psi0 s.
  unfold lift.
  intros [Hsub' Has].
  split_r.
  apply OCHSpec.sub_trans with Psi'; trivial.
Qed.

Definition wfch (D : idict) (Psi Psi' : chspec) (C : codeheap) :=
  (describe D Psi /\ describe D Psi') /\
  (forall (f : label) (pi : ocdspec),
    OCHSpec.lookup Psi' f pi
    -> (exists I : _ , codeheap.getis C f = Some I
      /\ wfis D (lift Psi {[ pi | D ]}) I)).

Definition wfst (a : assert) (Psi : chspec) (S : state) :=
  a Psi S.

Definition wfmach (D : idict) (Psi: chspec) (P : machine):=
  let (C, S) := P in
    let (h, rf, pc) := S in
      (exists Psi', wfch D Psi Psi' C /\ OCHSpec.sub Psi Psi')
      /\ (exists a : assert, 
        wfst a Psi S 
        /\ (exists I : instrseq, 
          (codeheap.getis C pc = Some I) 
          /\ wfis D a I)).

Lemma wfch_Psi_domsub_ch :
  forall (D : idict) (Psi Psi' : chspec) (ch : codeheap),
    wfch D Psi Psi' ch
    -> forall (f : label) (pi : ocdspec), 
      OCHSpec.lookup Psi' f pi -> CHeap.indom ch f.
Proof.
  intros D Psi Psi' C [_ HC] f pi Hin.
  substH HC with (HC f pi Hin).
  destruct HC as [I [HC _]].
  apply CHeap.get_indom.
  apply (getis_ex_get C f I HC).
Qed.

Lemma link :
  forall (D1 D2 : idict) (Psi1 Psi2 : chspec) 
    (C1 C2 : codeheap) (Psi1' Psi2' : chspec),
    wfch D1 Psi1 Psi1' C1
    -> wfch D2 Psi2 Psi2' C2
    -> DSpec.compat D1 D2
    -> CHeap.compat C1 C2
    -> wfch (DSpec.merge D1 D2) (OCHSpec.merge Psi1 Psi2)
    (OCHSpec.merge Psi1' Psi2') (CHeap.merge C1 C2).
Proof.
  intros D1 D2 Psi1 Psi2 C1 C2 Psi1' Psi2'.
  intros HwfC1 HwfC2 HDc HCc.
  generalize HwfC1 HwfC2.
  destruct HwfC1 as [[Hd1psi Hd1psi'] HC1].
  destruct HwfC2 as [[Hd2psi Hd2psi'] HC2].
  intros HwfC1 HwfC2.
  split.
  (* case describe *)
  split; apply describe_merge; trivial.
  (* case I *)
  intros f pi Hin.
  substH HC1 with (HC1 f pi).
  substH HC2 with (HC2 f pi).
  substH Hin with (OCHSpec.lookup_merge_elim Hin).
  destruct Hin as [Hin1 | Hin2].
  (* case (f, pi) in Psi1' *)
  destruct (HC1 Hin1) as [I [HC1I HwfI]].
  exists I.
  split.
  apply getis_compat_merge_intro; 
    [left | idtac]; assumption.
  apply wfis_weak_a with 
    (a := lift Psi1 {[ pi | DSpec.merge D1 D2 ]}).
  apply (wfis_weak_D_pi_lift (D := D1) 
    (D' := DSpec.merge D1 D2) (psi0 := Psi1)
    (pi:=pi) (is:=I)).
  apply (DSpec.sub_merge_l HDc).
  apply (describe_psi_describe_pi D1 Psi1' f pi Hd1psi'
    Hin1).
  apply HwfI.
  intros Psi S.
  unfold lift.
  intros [H1 H2].
  split. apply (OCHSpec.sub_merge_elim_l H1).
  apply H2.
  (* case (f, pi) in Psi2' *)
  destruct (HC2 Hin2) as [I [HC2I HwfI]].
  exists I.
  split.
  apply getis_compat_merge_intro; 
    [right | idtac]; assumption.
  apply wfis_weak_a with 
    (a := lift Psi2 {[ pi | DSpec.merge D1 D2 ]}).
  apply (wfis_weak_D_pi_lift (D := D2) 
    (D' := DSpec.merge D1 D2) (psi0 := Psi2)
    (pi:=pi) (is:=I)).
  apply (DSpec.sub_merge_r HDc).
  apply (describe_psi_describe_pi D2 Psi2' f pi Hd2psi'
    Hin2).
  apply HwfI.
  intros Psi S.
  unfold lift.
  intros [H1 H2].
  split. apply (OCHSpec.sub_merge_elim_r H1).
  apply H2.
Qed.

Lemma link_2 :
  forall (D : idict) (Psi : chspec) 
    (C : codeheap) (Psi' : chspec) f pi I,
    wfch D Psi Psi' C
    -> describe_pi D pi
    -> getis C f = Some I
    -> wfis D << {[ pi | D]} / Psi >> I
    -> wfch D Psi (OCHSpec.merge (OCHSpec.sig_map f pi) Psi') C.
Proof.
  intros D Psi C Psi' f pi I.
  intros HwfC Hpi HI HwfI .
  unfold wfch.
  split.
    destruct HwfC as [[H1 H2] HwfC].
    split_l.
    apply describe_merge_2; trivial.
    unfold describe, OCHSpec.sig_map.
    unfold describe_pi in Hpi.
    unfold OCHSpec.lookup.
    destruct pi as [id theta].
    simpl.
    intros.
    destruct H as [H | Hf]; [idtac | destruct Hf].
    injection H.
    intros; subst.
    apply Hpi.
  destruct HwfC as [[_ _] HwfC].
  unfold OCHSpec.lookup, OCHSpec.sig_map in * .
  destruct pi as [id theta].
  simpl.
  intros f0 pi0 [Heq | Hin].
    injection Heq; intros; subst.
    exists I.
    split_lr.
  substH HwfC with (HwfC f0 pi0 Hin).
  trivial.
Qed.

Lemma wfch_weaken_D : forall D D' Psi Psi' C,
  DSpec.sub D D'
  -> wfch D Psi Psi' C
  -> wfch D' Psi Psi' C.
Proof.
  intros D D' Psi Psi' C HDsub HC.
  unfold wfch.
  destruct HC as [[H1 H2] HC].
  split.
    split; eapply describe_sub; eauto.
  intros f pi Hpi.
  destruct (HC f pi Hpi) as [I [HgetI HwfI]].
  exists I.
  split_l.
  apply wfis_weak_D_pi_lift with D; trivial.
  apply describe_psi_describe_pi with Psi' f; trivial.
Qed.


(* ************************************************
 *
 *   Soundness Theoerem 
 *
 *   Progress + Preservation
 *
 *)

Theorem Soundness :
  forall D Psi C S,
    wfmach D Psi (C, S) ->
    (exists S' : _, ((C, S) ~~> (C, S'))
      /\ wfmach D Psi (C, S')).
Proof.
  intros D Psi C S.
  destruct S as [H R pc].
  intros [HC [a [HS [I [HCI HI]]]]].
  generalize HC.
  destruct HC as [Psi' [HC Hsub]].
  intro HwfC.
  destruct I.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r ];
    unfold wfis in HI; fold wfis in HI;
    try contradiction.
  (* case i = j f *)
  destruct (HI Psi <.H, R, pc.> HS) 
    as [pi'' [Hin HS']].
  exists <.H, R, f.> .
  unfold step, nextS.
  rewrite (getis_ione_get _ _ _ HCI).
  split_l.
  split_l.
  exists <<{[pi'' | D]} / Psi>> .
  simpl in HS'.
  unfold wfst, lift.
  split_l.
  split_r.
  apply OCHSpec.sub_refl.
  unfold wfch in HC.
  destruct HC as [[_ _ ] HC].
  assert (Hin' := OCHSpec.lookup_sub_lookup Hin Hsub).
  destruct (HC f pi'' Hin') as [I'' [HI''in HI'']].
  exists I''.
  split_lr.
  (* case i = jal f fret *)
  destruct (HI Psi <.H, R, pc.> HS) 
    as [pi'' [Hin HS']].
  exists <.H, sR R ra fret, f.> .
  unfold step, nextS.
  rewrite (getis_ione_get _ _ _ HCI).
  split_l.
  split_l.
  exists <<{[pi'' | D]} / Psi>> .
  simpl in HS'.
  unfold wfst, lift.
  split_l.
  split_r.
  apply OCHSpec.sub_refl.
  unfold wfch in HC.
  destruct HC as [[_ _ ] HC].
  assert (Hin' := OCHSpec.lookup_sub_lookup Hin Hsub).
  destruct (HC f pi'' Hin') as [I'' [HI''in HI'']].
  exists I''.
  split_lr.
  (* case i = jr r *)
  destruct (HI Psi <.H, R, pc.> HS) 
    as [pi'' [Hin HS']].
  exists <.H, R, gR R r.> .
  unfold step, nextS.
  rewrite (getis_ione_get _ _ _ HCI).
  split_l.
  split_l.
  exists <<{[pi'' | D]} / Psi>> .
  simpl in HS'.
  unfold wfst, lift.
  split_l.
  split_r.
  apply OCHSpec.sub_refl.
  unfold wfch in HC.
  destruct HC as [[_ _] HC].
  assert (Hin' := OCHSpec.lookup_sub_lookup Hin Hsub).
  destruct (HC (gR R r) pi'' Hin') as [I'' [HI''in HI'']].
  exists I''.
  split_lr.
  (* case i = iseq i is *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r ]; 
    unfold wfis in HI; fold wfis in HI;
    try contradiction HI.
  (* case i = addu rd rs rt *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  exists <.H, sR R rd (gR R rs + gR R rt), pc+1.> .
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = addiu rd rs w *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  simpl in * .
  exists <.H, sR R rd (gR R rs + w), pc+1.> .
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  split_l. 
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = subu rd rs rt *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  exists <.H, sR R rd (gR R rs - gR R rt), pc+1.> .
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = subiu rd rs w *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  simpl in * .
  exists <.H, sR R rd (gR R rs - w), pc+1.> .
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = move rd rs *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  simpl in * .
  exists <.H, sR R rd (gR R rs), pc+1.> .
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = li rd w *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  simpl in * .
  exists <.H, sR R rd w, pc+1.> .
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = lw rd w rs *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  simpl in Himp.
  destruct (gH H (gR R rs + w)) as [w' | Hnin];
    [idtac | destruct Himp].
  exists <.H, sR R rt w', pc+1.> .
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = st rt w rs *)
  substH HI with (HI Psi _ HS).
  destruct HI as [a' [Himp HI]]. 
  unfold step, nextS.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  simpl in Himp.
  destruct (Heap.get_dec H (gR R rs + w)) 
    as [[w' Hsome] | Hnone];
      [idtac |
        rewrite Hnone in Himp; 
          destruct Himp].
  rewrite Hsome in Himp.
  rewrite Hsome.
  set (H' := sH H (gR R rs + w) (gR R rt)).
  exists <.H', R, pc+1.> .
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = blt rs rt f *)
  substH HI with (HI Psi _ HS).
  unfold step, nextS.
  unfold nextS in HI.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  decbltnat (gR R rs) (gR R rt).
  (* case rs < rt *)
  destruct HI as [pi'' [Hin HS']].
  exists <.H, R, f.> .
  split_l.
  split_l.
  exists <<{[pi'' | D]} / Psi>> .
  simpl in HS'.
  unfold wfst, lift.
  split_l.
  split_r.
  apply OCHSpec.sub_refl.
  unfold wfch in HC.
  destruct HC as [[_ _] HC].
  assert (Hin' := OCHSpec.lookup_sub_lookup Hin Hsub).
  destruct (HC f pi'' Hin') as [I'' [HI''in HI'']].
  exists I''.
  split_lr.
  (* case rs >= rt *)
  destruct HI as [a' [Ha' HI]].
  exists <.H, R, pc+1.> .
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = beq rs rt f *)
  substH HI with (HI Psi _ HS).
  unfold step, nextS.
  unfold nextS in HI.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  decbeqnat (gR R rs) (gR R rt).
  (* case rs = rt *)
  destruct HI as [pi'' [Hin HS']].
  exists <.H, R, f.> .
  split_l.
  split_l.
  exists <<{[pi'' | D]} / Psi>> .
  simpl in HS'.
  unfold wfst, lift.
  split_l.
  split_r.
  apply OCHSpec.sub_refl.
  unfold wfch in HC.
  destruct HC as [[_ _] HC].
  assert (Hin' := OCHSpec.lookup_sub_lookup Hin Hsub).
  destruct (HC f pi'' Hin') as [I'' [HI''in HI'']].
  exists I''.
  split_lr.
  (* case rs <> rt *)
  destruct HI as [a' [Himp HI]].
  exists <.H, R, pc+1.> .
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
  (* case i = bgtz rs f *)
  substH HI with (HI Psi _ HS).
  unfold step, nextS.
  unfold nextS in HI.
  destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His].
  rewrite Hi.
  decbltnat O (gR R rs).
  (* case O < rt *)
  destruct HI as [pi'' [Hin HS']].
  exists <.H, R, f.> .
  split_l.
  split_l.
  exists <<{[pi'' | D]} / Psi>> .
  simpl in HS'.
  unfold wfst, lift.
  split_l.
  split_r.
  apply OCHSpec.sub_refl.
  unfold wfch in HC.
  destruct HC as [[_ _] HC].
  assert (Hin' := OCHSpec.lookup_sub_lookup Hin Hsub).
  destruct (HC f pi'' Hin') as [I'' [HI''in HI'']].
  exists I''.
  split_lr.
  (* case O = rs *)
  destruct HI as [a' [Himp HI]].
  exists <.H, R, pc+1.> .
  split_l.
  split_l.
  exists a'.
  split_l.
  exists I.
  split_lr.
Qed.

Theorem Type_Safety :
  forall n D Psi C S,
    wfmach D Psi (C, S) ->
    (exists S' : _, (stepn n (C, S) = Some (C, S'))
      /\ wfmach D Psi (C, S')).
Proof.
  intros n D Psi C.
  induction n; intros S H.
    exists S; split_lr.
  unfold stepn; fold stepn.
  destruct (IHn S H) as [S' [Hs HS']].
  destruct (Soundness _ _ _ _ HS') 
    as [S'' [Hs'' HS'']].
  rewrite Hs.
  exists S''.
  split_lr.
Qed.

Theorem Correctness :
  forall n D Psi C S,
    wfmach D Psi (C, S) ->
    (exists S' : _, (stepn n (C, S) = Some (C, S'))
      /\ wfmach D Psi (C, S')
      /\ let (H', R', pc') := S' in
        match codeheap.get C pc' with
           | Some (j f) => 
             exists pi, 
               OCHSpec.lookup Psi f pi
               /\ {[pi | D]} Psi <.H',R',f.>
           | Some (jal f fret) => 
             exists pi, 
               OCHSpec.lookup Psi f pi
               /\ {[pi | D]} Psi <.H',sR R' ra fret,f.>
           | Some (jr r) =>
             exists pi, 
               OCHSpec.lookup Psi (gR R' r) pi
               /\ {[pi | D]} Psi <.H',R',(gR R' r).>
           | Some (blt rs rt f) =>
             match blt_nat (gR R' rs) (gR R' rt) with
               | true => 
                   exists pi,
                     OCHSpec.lookup Psi f pi
                     /\ {[pi | D]} Psi <.H',R',f.>
               | false => True
             end
           | Some (beq rs rt f) =>
             match beq_nat (gR R' rs) (gR R' rt) with
               | true => 
                   exists pi,
                     OCHSpec.lookup Psi f pi
                     /\ {[pi | D]} Psi <.H',R',f.>
               | false => True
             end
           | Some (bgtz rs f) =>
             match blt_nat O (gR R' rs) with
               | true => 
                   exists pi,
                     OCHSpec.lookup Psi f pi
                     /\ {[pi | D]} Psi <.H',R',f.>
               | false => True
             end
          | _ => True
         end).
Proof.
  intros n D Psi C S HS.
  destruct (Type_Safety n D Psi C S HS) as [S' [Hstep HS']].
  exists S'.
  split_l.
  split_l.
  destruct S' as [H' R' pc'].
  set (S' := <.H',R',pc'.>).
  destruct (Soundness D Psi C S' HS')
    as [S'' [Hstep' HS'']].
  unfold wfmach in HS'.
  destruct HS' as [HC' [a' [HS' [I' [HCI' HI']]]]].
  destruct I'.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r ];
    unfold wfis in HI'; fold wfis in HI';
    try contradiction HI'.
  (* case i = j f *)
  rewrite (getis_ione_get _ _ _ HCI').
  destruct (HI' Psi S' HS') 
    as [pi'' [Hin Hpi'']].
  simpl in Hpi''.
  exists pi''; split_lr.
  (* case i = jal f fret *)
  rewrite (getis_ione_get _ _ _ HCI').
  destruct (HI' Psi S' HS') 
    as [pi'' [Hin Hpi'']].
  simpl in Hpi''.
  exists pi''; split_lr.
  (* case i = jr r *)
  rewrite (getis_ione_get _ _ _ HCI').
  destruct (HI' Psi S' HS') 
    as [pi'' [Hin Hpi'']].
  simpl in Hpi''.
  exists pi''; split_lr.
  (* case i = iseq i is *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r ]; 
    unfold wfis in HI'; fold wfis in HI';
    try contradiction HI';
    destruct (getis_iseq_get _ _ _ _ HCI') 
      as [Hi His]; rewrite Hi;
    trivial.
  (* case i = blt rs rt f *)
  substH HI' with (HI' Psi _ HS').
  unfold nextS in HI'.
  decbltnat (gR R' rs) (gR R' rt); trivial.
  (* case i = beq rs rt f *)
  substH HI' with (HI' Psi _ HS').
  unfold nextS in HI'.
  decbeqnat (gR R' rs) (gR R' rt); trivial.
  (* case i = bgtz rs f *)
  substH HI' with (HI' Psi _ HS').
  unfold nextS in HI'.
  decbltnat O (gR R' rs); trivial.
Qed.
