(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Separation Logic                                            *)
(*   basic tactics                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: seplogic2.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Coq.Lists.List.
Require Import machine.
Require Import seplogic.
Require Import extac.

Open Local Scope seplogic_scope.

(* Tactic Notation "introsep" simple_intropattern (pat) := *)
(*   match goal with  *)
(*     | |- ?h ||- ?A ** ?B => *)
(*         apply (@sconj_intro *)

Ltac sep_trivial :=
  match goal with
    | |- ?h ||- top =>
        apply top_intro; trivial
    | _ => trivial
  end.

Ltac sep_ex x :=
  match goal with
    | |- _ ||- lex _ => 
        apply lex_intro1; exists x
    | |- _ ||- sconj (lex _) _ => 
        apply lex_intro_left; apply lex_intro1;
        exists x
    | |- _ ||- sconj _ (lex _) => 
        apply lex_intro_right; apply lex_intro1;
        exists x

    | _ => fail 1 "fail"
  end.

Ltac sep_ex_pos t n :=
  match eval compute in n with
    | O => 
      match goal with 
        | |- _ ||- sconj (lex _) _ => 
          apply lex_intro_left; apply lex_intro1; 
          exists t
        | _ => fail 1 "fail"
      end
    | S _ => assocr; apply sconj_sym; assocr; sep_ex_pos t (pred n)
  end.

Tactic Notation "sep_exists" constr (t) :=
  sep_ex t.

Tactic Notation "sep_exists" constr (t) "at" constr (n) :=
  sep_ex_pos t n.


Ltac sep_split :=
  match goal with
    | |- ?h ||- ! ?A ** ?B => apply bang_intro_left
    | |- ?h ||- ?B ** ! ?A => apply bang_intro_right
    | _ => fail 1 "fail"
  end.

Ltac sep_split_2 ha hb :=
  match goal with
    | |- ?h ||- ?A ** ?B =>
        apply sconj_intro with (ha := ha) (hb := hb)
    | _ => fail 1 "fail"
  end.

Tactic Notation "sep_split" :=
  sep_split.

Tactic Notation "sep_split" "into" constr (ha) constr (hb) :=
  sep_split_2 ha hb.

Definition id (A : Type) (x : A) := x.
Implicit Arguments id [A].

Ltac sep_des H :=
  match type of (H) with
    | ?h ||- ! ?A ** ?B => 
           destruct (bang_elim_left H)
    | ?h ||- ?B ** ! ?A => 
           destruct (bang_elim_right H)

    | ?h ||- ?A ** ?B =>
           destruct (id H)

    | _ => fail 1 "fail"
  end.

Tactic Notation "sep_destruct" hyp (H):=
  dessep H.

Tactic Notation "sep_destruct_clear" hyp (H):=
  dessep H; clear H.

Tactic Notation "sep_destruct" hyp (H) "as" simple_intropattern (pat) :=
  match type of (H) with
    | ?h ||- ! ?A ** ?B => 
           destruct (bang_elim_left H) as pat
    | ?h ||- ?B ** ! ?A => 
           destruct (bang_elim_right H) as pat

    | ?h ||- (lex ?A) =>
           destruct (lex_elim H) as pat

    | ?h ||- ?A ** ?B =>
           destruct (id H) as pat

    | _ => fail 1 "fail"
  end.

Tactic Notation "sep_destruct_clear" hyp (H) "as" simple_intropattern (pat) :=
  match type of (H) with
    | ?h ||- ! ?A ** ?B => 
           destruct (bang_elim_left H) as pat; clear H
    | ?h ||- ?B ** ! ?A => 
           destruct (bang_elim_right H) as pat; clear H

    | ?h ||- (lex ?A) =>
           destruct (lex_elim H) as pat; clear H

    | ?h ||- ?A ** ?B =>
           destruct (id H) as pat; clear H

    | _ => fail 1 "fail"
  end.
