(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SCAP                                                        *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: scap.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Require Export optionT.
Require Import MAPT.
Require Import ListB_MAPT.
Require Import machine.

Require Import extac.

Open Local Scope tm_scope.

Definition stpred := state -> Prop.

Definition acpred := state -> state -> Prop.

Inductive cdspec : Type := 
  | mk_cdspec : stpred -> acpred -> cdspec.

Notation " [ p & g ] " := (mk_cdspec p g) 
  (at level 0) : scap_scope.

Open Local Scope scap_scope.

Module CdSpec <: ONE_TYPE.
  Definition B := cdspec.
End CdSpec.

Module CHSpecDef := ListMapT CdSpec.
Module CHSpec := MapEx CHSpecDef.
(* psi : chspec *)
Definition chspec := CHSpec.map.

Fixpoint wfis (psi : chspec) (p : stpred) (g : acpred) (is : instrseq)
  {struct is} : Prop :=
  match is with
    | ione i =>
      match i with
        | j f =>
            (forall s, p s
              -> exists p'', exists g'' : acpred,
                CHSpec.get psi f = SomeT [p'' & g'']
                /\ optp p'' (nextS i s)
                /\ forall s',
                  optg g'' (nextS i s) s' -> g s s')
        | jal f fret =>
            (forall s, p s
              -> exists p'', exists g'',
              CHSpec.get psi f = SomeT (mk_cdspec p'' g'')
              /\ optp p'' (nextS i s)
              /\ forall s',
                optg g'' (nextS i s) s'
                -> exists p', exists g',
                  CHSpec.get psi (S_pc s') = SomeT (mk_cdspec p' g')
                  /\ p' s'
                  /\ forall s'', g' s' s'' -> g s s'')
        | jr r =>
            match (beq_reg r r31) with
              | true =>
                (forall s, p s -> optp (g s) (nextS i s))
              | false =>
                (forall s, p s
                  -> exists p'', exists g'',
                    CHSpec.get psi (rget s r) = SomeT (mk_cdspec p'' g'')
                    /\ optp p'' (nextS i s)
                    /\ forall s',
                      optg g'' (nextS i s) s' -> g s s')
            end
        | _ => False
          
      end
    | iseq i is' =>
      match i with
        | j f => False
        | jal f fret => False
        | jr r => False

        | blt rs rt f as i =>
          (forall s, p s -> if blt_nat (S_gR s rs) (S_gR s rt)
            then 
              (exists p'', exists g'',
                CHSpec.get psi f = SomeT [p'' & g'']
                /\ optp p'' (nextS i s)
                /\ forall s', 
                  optg g'' (nextS i s) s' -> g s s')
            else 
              (exists p', exists g',
                optp p' (nextS i s)
                /\ (forall s', optg g' (nextS i s) s' -> g s s')
                /\ wfis psi p' g' is'))
              
        | beq rs rt f as i =>
          (forall s, p s -> if beq_nat (S_gR s rs) (S_gR s rt)
            then 
              (exists p'', exists g'',
                CHSpec.get psi f = SomeT [p'' & g'']
                /\ optp p'' (nextS i s)
                /\ (forall s', optg g'' (nextS i s) s' -> g s s'))
            else 
              (exists p', exists g',
                optp p' (nextS i s)
                /\ (forall s', optg g' (nextS i s) s' -> g s s')
                /\  wfis psi p' g' is'))

        | bgtz rs f as i =>
          (forall s, p s -> if blt_nat (O) (S_gR s rs)
            then 
              (exists p'', exists g'',
                CHSpec.get psi f = SomeT [p'' & g'']
                /\ optp p'' (nextS i s)
                /\ (forall s', optg g'' (nextS i s) s' -> g s s'))
            else 
              (exists p', exists g',
                optp p' (nextS i s)
                /\ (forall s', optg g' (nextS i s) s' -> g s s')
                /\  wfis psi p' g' is'))
            
        | _ as i =>
          (forall s, p s
            -> exists p', exists g',
              optp p' (nextS i s)
              /\ (forall s', optg g' (nextS i s) s' -> g s s')
              /\  wfis psi p' g' is')
      end
  end.

Lemma wfis_weakening : 
  forall is psi (p p' : stpred) (g g' : acpred),
    (forall s, p s -> p' s /\ forall s', g' s s' -> g s s') 
    -> wfis psi p' g' is 
    -> wfis psi p g is.
Proof.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f
      | rs rt f | rs f
      | f | f fret | r]; 
    unfold wfis; fold wfis;
    intros psi p p0 g g0 Hw His;
      try inversion_clear His.
  (* case i = j f *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p'' [g'' [Hf [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').
  (* case i = jal f fret *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  destruct Hg'' as [p' [g' [Hf' [Hp' Hg']]]].
  exists p'; exists g'.
  split_l.
  split_l.
  intros s'' H.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = jr r *)
  decbeqreg r ra.
    intros s Hp.
    substH Hw with (Hw s Hp).
    destruct Hw as [Hw1 Hw2].
    substH His with (His s Hw1).
    destruct (nextS (jr ra) s); [idtac|destruct His].
    apply Hw2; trivial.
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p'' [g'' [Hf [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').

  (* case is = iseq i is *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f
      | rs rt f | rs f
      | f | f fret | r]; 
    unfold wfis; fold wfis;
    intros psi p p0 g g0 Hw His;
    try inversion_clear His.
  (* case i = addu rd rs rt *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = addiu rd rs w *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = subu rd rs rt *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = subiu rd rs w *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = move rd rs *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = li rd w *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = lw rt w rs *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = sw rt w rs *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = blt rs rt f *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  decbltnat (S_gR s rs) (S_gR s rt).
  (* rs < rt *)
  destruct His as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').
  (* rs >= rt*)
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = beq rs rt f *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  decbeqnat (S_gR s rs) (S_gR s rt).
  (* rs = rt *)
  destruct His as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').
  (* rs <> rt*)
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
  (* case i = bgtz rs f *)
  intros s Hp.
  substH Hw with (Hw s Hp).
  destruct Hw as [Hw1 Hw2].
  substH His with (His s Hw1).
  decbltnat (O) (S_gR s rs).
  (* O < rs *)
  destruct His as [p'' [g'' [Hin'' [Hp'' Hg'']]]].
  exists p''; exists g''.
  split_l.
  split_l.
  intros s' Hg''s'.
  substH Hg'' with (Hg'' s' Hg''s').
  apply (Hw2 s' Hg'').
  (* 0 = rs *)
  destruct His as [p' [g' [Hp' [Hg' His]]]].
  exists p'; exists g'.
  split_l.
  split_r.
  intros s' Hg's'.
  apply Hw2.
  apply Hg'; trivial.
Qed.

Definition wfch (psi psi' : chspec) (ch : codeheap) :=
  forall f p g,
    CHSpec.lookup psi' f (mk_cdspec p g)
    -> (exists is : _, codeheap.getis ch f = Some is
      /\ wfis psi p g is).
