(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SCAP                                                        *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: scap_ck.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Require Import machine.

Require Export optionT.
Require Import MAPT.
Require Import ListB_MAPT.
Require Export ListT.

Require Import scap.

Require Import extac.

Open Local Scope tm_scope.
Open Local Scope scap_scope.

Fixpoint wfch_prove (psi : chspec) (psi' : chspec) (ch : codeheap)
  {struct psi'} : Prop :=
  match psi' with 
    | nilT => True
    | consT (pairT f (mk_cdspec p g)) psi' =>
      match (codeheap.getis ch f) with
        | Some is => wfis psi p g is 
        | None => False
      end
      /\ wfch_prove psi psi' ch
  end.

Lemma wfch_prove_sound_0 :
  forall psi psi' ch,
    wfch_prove psi psi' ch
    -> forall (f : label) p g,
      CHSpec.lookup psi' f (mk_cdspec p g)
      -> (exists is,
        codeheap.getis ch f = Some is
        /\ wfis psi p g is).
Proof.
  intros psi psi'; generalize psi; clear psi.
  induction psi'.
    simpl.
    unfold wfch.
    intros.
    inversion_clear H0.
  destruct a as [f' [p' g']].
  intros psi ch Hck f p g Hin.
  simpl in Hck.
  unfold CHSpec.lookup, CHSpec.get, CHSpecDef.get in Hin.
  unfold CHSpecDef.beq in Hin. 
  fold CHSpecDef.get in Hin.
  destruct Hck as [Hck Hpsi'].
  decbeqnat f f'.
    inj_hyp Hin; intros; subst.
    substbnat f'.
    unfold get.
    destruct (codeheap.getis ch f) as [is | ]; try contradiction.
    exists is.
    split_lr.
  destruct (codeheap.getis ch f');
    try contradiction.
  apply IHpsi'; trivial.
Qed.

Theorem wfch_prove_sound :
  forall psi psi' ch,
    wfch_prove psi psi' ch
    -> wfch psi psi' ch.
Proof.
  intros psi psi' ch Hck.
  intros f a Hin.
  apply (wfch_prove_sound_0 psi psi'
    ch Hck f a Hin).
Qed.

(* 
  a simple vc generator
*)

Fixpoint vc (psi : chspec) (is : instrseq) 
  {struct is} : optionT cdspec :=
  match is with

    | ione (j f) => 
      match (CHSpec.get psi f) with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => p' (pcset s f))
            (fun s => fun s'' => g' (pcset s f) s''))
        | _ => NoneT
      end

    | ione (jal f fret) => 
      match (CHSpec.get psi f) with
        | SomeT (mk_cdspec p'' g'') =>
          SomeT (mk_cdspec
            (fun s => p'' (pcset (rset s ra fret) f)
            /\ (forall s', exists p', exists g',
              (CHSpec.get psi (S_pc s') = SomeT [p' & g']) /\
              (g'' (pcset (rset s ra fret) f) s' ->
              p' s')))
            (fun s => fun s'' => 
              (exists s', exists p', exists g',
              (CHSpec.get psi (S_pc s') = SomeT [p' & g']) /\
              g'' (pcset (rset s ra fret) f) s' /\
              g' s' s'')))
        | _ => NoneT
      end

    | ione (jr r) => 
      match (beq_reg r r31) with
        | true =>
          SomeT (mk_cdspec 
            (fun s => True)
            (fun s => fun s' => 
              optp (fun s'' => s'' = s') (nextS2 (jr r) s)))
        | false => 
          SomeT (mk_cdspec
            (fun s => 
              match (CHSpec.get psi (rget s r)) with
                | SomeT (mk_cdspec p' g') => 
                    optp (fun s' => p' s') (nextS2 (jr r) s)
                | NoneT => False
              end)
            (fun s s' =>
              match (CHSpec.get psi (rget s r)) with
                | SomeT (mk_cdspec p' g') => 
                    optg g' (nextS2 (jr r) s) s'
                | NoneT => False
              end))
      end
      
    | iseq (blt rs rt f) is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              if blt_nat (rget s rs) (rget s rt)
                then 
                  (optpT (fun bsp : cdspec => let (p'', _) := bsp in 
                    (p'' (pcset s f)))
                    (CHSpec.get psi f))
                else (p' (pcinc s)))
            (fun s => fun s' =>
              if blt_nat (rget s rs) (rget s rt) 
                then
                  (opt_predT (CHSpec.get psi f) (fun bsp =>
                    let (_, g'') := bsp in 
                      (g'' (pcset s f) s')))
                else (g' (pcinc s) s')))
        | NoneT => NoneT
      end

    | iseq (beq rs rt f) is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              if beq_nat (rget s rs) (rget s rt)
                then 
                  (optpT (fun bsp : cdspec => let (p'', _) := bsp in 
                    (p'' (pcset s f)))
                    (CHSpec.get psi f))
                else (p' (pcinc s)))
            (fun s => fun s' =>
              if beq_nat (rget s rs) (rget s rt) 
                then
                  (opt_predT (CHSpec.get psi f) (fun bsp =>
                    let (_, g'') := bsp in 
                      (g'' (pcset s f) s')))
                else (g' (pcinc s) s')))
        | NoneT => NoneT
      end

    | iseq (bgtz rs f) is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              if blt_nat O (rget s rs)
                then 
                  (optpT (fun bsp : cdspec => let (p'', _) := bsp in 
                    (p'' (pcset s f)))
                    (CHSpec.get psi f))
                else (p' (pcinc s)))
            (fun s => fun s' =>
              if blt_nat O (rget s rs) 
                then
                  (opt_predT (CHSpec.get psi f) (fun bsp =>
                    let (_, g'') := bsp in 
                      (g'' (pcset s f) s')))
                else (g' (pcinc s) s')))
        | NoneT => NoneT
      end
      
    | iseq (j f) is' => NoneT

    | iseq (jal f fret) is' => NoneT

    | iseq (jr r) is' => NoneT

    | iseq i is' =>
      match vc psi is' with
        | SomeT (mk_cdspec p' g') =>
          SomeT (mk_cdspec
            (fun s => 
              optp (fun s' => p' s') (nextS2 i s))
            (fun s => fun s'' => 
              optp (fun s' => g' s' s'') (nextS2 i s)))
        | NoneT => NoneT
      end
    | _ => NoneT
  end.

Theorem vc_sound : forall psi is p g,
  (vc psi is) = SomeT (mk_cdspec p g) ->
  wfis psi p g is.
Proof.
  intros psi.
  induction is.
  intros p g.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt | rs f
      | f | f fret | r ]; 
    simpl; intro H; try discriminate.
  (* case i = j f *)
  intros s Hp.
  destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
    [idtac | discriminate H].
  exists p''; exists g''.
  split_l.
  injection H; intros; subst p; subst g; clear H.
  destruct s as [h rf pc].
  split_lr.
  (* case i = jal f fret *)
  intros s Hp.
  destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
    [idtac | discriminate H].
  exists p''; exists g''.
  split_l.
  injection H; intros; subst p; subst g; clear H.
  destruct Hp as [Hp'' Hp].
  destruct s as [h rf pc].
  split_l.
  intros [h' rf' pc'] Hg''.
  substH Hp with (Hp (mkst h' rf' pc')).
  destruct Hp as [p' [g' [Heq Hp']]].
  exists p'; exists g'.
  split_l.
  simplst.
  substH Hp' with (Hp' Hg'').
  split_l.
  intros s'' Hg'.
  exists (mkst h' rf' pc').
  exists p'; exists g'.
  split_l.
  split_lr.
  (* case i = jr r *)
  decbeqreg r ra.
    intros [h rf pc] Hps.
    injection H; intros; subst p; subst g; clear H.
    trivial.
  intros s Hp.
  injection H; intros; subst p; subst g; clear H.
  destruct (CHSpec.get psi (rget s r)) 
    as [[p'' g''] | ]; try contradiction.
  exists p''; exists g''.
  split_l.
  destruct s as [hp rf pc].
  simplst.
  split_l.
  trivial.
  (* case is = iseq i is' *)
  intros p g.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt | rs f
      | f | f fret | r ]; 
    unfold wfis, vc; fold wfis; fold vc;
    intro H; try discriminate.
  (* case i = addu rd rs rt *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = addiu rd rs w *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = subu rd rs rt *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = subiu rd rs w *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = move rd rs *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = li rd w *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = lw rt w rs *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = sw rt w rs *)
  intros [hp rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  exists p'; exists g'.
  unfold nextS; simplst.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = blt rs rt f *)
  intros [h rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  unfold nextS; simplst.
  decbltnat (gR rf rs) (gR rf rt).
    destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
      try contradiction Hp.
    exists p''; exists g''.
    split_l.
    split_lr.
  exists p'; exists g'.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = beq rs rt f *)
  intros [h rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  unfold nextS; simplst.
  decbeqnat (gR rf rs) (gR rf rt).
    destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
      try contradiction Hp.
    exists p''; exists g''.
    split_l.
    split_lr.
  exists p'; exists g'.
  split_l.
  split_l.
  apply IHis; trivial.
  (* case i = bgtz rs rt f *)
  intros [h rf pc] Hp.
  destruct (vc psi is) as [[p' g'] | ]; 
    try discriminate H.
  injection H; intros; subst p; subst g; clear H.
  unfold nextS; simplst.
  decbltnat O (gR rf rs); simpl in Hb; 
    rewrite Hb in Hp; rewrite Hb; clear Hb.
    destruct (CHSpec.get psi f) as [[p'' g''] | ]; 
      try contradiction Hp.
    exists p''; exists g''.
    split_l.
    split_lr.
  exists p'; exists g'.
  split_l.
  split_l.
  apply IHis; trivial.
Qed.

Ltac simplpg :=
  match goal with
    | [ |- exists p, exists g, 
      SomeT (mk_cdspec ?p0 ?g0) = SomeT (mk_cdspec p g)] =>
    exists p0; exists g0; reflexivity
  end.

Ltac simplvc :=
  match goal with
    | [ |- wfis ?Psi ?pre ?guar ?is ] =>
      let E := fresh "E" in let p := fresh "p" in let g := fresh "g" in
        let E1 := fresh "E" in let E2 := fresh "E" in
          (assert (E : exists p : stpred, exists g : acpred, 
            vc Psi is = SomeT (mk_cdspec p g));
          [simpl; simplpg | destruct E as [p [g E]]];
          apply wfis_weakening with (p' := p) (g' := g);
            [ injection E; intros E1 E2; subst p; subst g; clear E 
              |
              apply vc_sound; simpl; auto])
  end.

Ltac prove_wfch :=  
  match goal with
    | |- wfch ?psi ?psi' ?C =>
      apply wfch_prove_sound; prove_wfch
    | |- wfch_prove ?psi ?psi ?C =>
      unfold psi at 2, wfch_prove; prove_wfch
    | |- wfch_prove ?psi ?psi' ?C =>
      unfold psi', wfch_prove; prove_wfch
    | |- ?a /\ ?b => split; prove_wfch
    | |- context [(getis ?C ?f)] =>
      (let v := (eval compute in (getis C f))
        in change (getis C f) with v);
      cbv iota beta
    | |- True => trivial
    | _ => idtac
  end.
