(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* SCAP over OCAP                                              *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: scap_ocap.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Require Import machine.
Require Import scap.
Require Import ocap.

Require Import extac.
Require Import deptac.

Open Local Scope tm_scope.
Open Local Scope ocap_scope.
Open Local Scope scap_scope.

Module Type SCAPPARAM.
Parameter rho_scap : nat.
  
Parameter D0 : idict.

Axiom rho_scap_not_indom_D0 :
  ~ DSpec.indom D0 rho_scap.
End SCAPPARAM.

Module ScapOcap (Params : SCAPPARAM).

Import Params.

Notation Lscap := scap.cdspec.
Notation mk_Lscap := scap.mk_cdspec.
Notation scap_chspec := scap.chspec.
Notation scap_wfis := scap.wfis.
Notation ocap_wfis := ocap.wfis.
Notation scap_wfch := scap.wfch.
Notation ocap_wfch := ocap.wfch.

Module ScapCHSpec := scap.CHSpec.

(* *********************************************************** *)

Definition scap_chspec_lift (psi : scap_chspec) (OPsi : chspec) : Prop :=
  (forall f p g,
    ScapCHSpec.lookup psi f [p & g] 
    -> OCHSpec.lookup OPsi f <* rho_scap, [p & g] *>)
  /\ (forall f pi, 
    OCHSpec.lookup OPsi f pi
    -> exists p, exists g, 
      pi = <* rho_scap, [p & g] *>
      /\ ScapCHSpec.lookup psi f [p & g]).

(* *********************************************************** *)
Fixpoint scapWFST (n : nat) (g : acpred) (s : state)
  (psi : chspec) {struct n} : Prop :=
  match n with
    | O => forall s', g s s' ->
      exists pi : ocdspec,
        OCHSpec.lookup psi (S_pc s') pi
        /\ {[ pi | D0 ]} psi s'
    | S n => forall s' : state, g s s' ->
      exists p' : stpred, exists g' : acpred,
        OCHSpec.lookup psi (S_pc s')
            (mk_pi rho_scap (xcdspec_pack [p' & g']))
       /\ p' s'
       /\ scapWFST n g' s' psi
  end.

Definition itpscap (theta : Lscap) : assert := 
  let (p, g) := theta in
    fun (psi : chspec) (s : state)  =>
      p s /\ 
      exists n, scapWFST n g s psi.

Notation "[| theta |]" := (itpscap theta) 
  (at level 0) : ocap_scope.

Notation "[| p & g |]" := (itpscap (mk_Lscap p g))
  (at level 0) : ocap_scope.

Definition D_scap : idict := fun id =>
  if (beq_nat id rho_scap) then SomeT <@ itpscap @> else
    NoneT.

Lemma D_scap_disj_D0 :
  DSpec.disj D_scap D0.
Proof.
  apply DSpec.indom_disj.
  intro a.
  split; intros H1 H2.
  unfold DSpec.indom, D_scap in H1.
  destruct H1 as [b H1].
  unfold DSpec.get, DSpecDef.get in H1.
  decbeqnat a rho_scap.
  substbnat a.
  destruct (rho_scap_not_indom_D0 H2).
  discriminate H1.
  unfold DSpec.indom, D_scap in H2.
  destruct H2 as [b H2].
  unfold DSpec.get, DSpecDef.get in H2.
  decbeqnat a rho_scap.
  substbnat a.
  destruct (rho_scap_not_indom_D0 H1).
  discriminate H2.
Qed.

Definition D1 := DSpec.merge D0 D_scap.

Lemma D0_sub_D1 : DSpec.sub D0 D1.
Proof.
  unfold D1.
  apply DSpec.sub_merge_l.
  apply DSpec.disj_compat.
  apply DSpec.disj_sym.
  apply D_scap_disj_D0.
Qed.

Lemma D_scap_sub_D1 : DSpec.sub D_scap D1.
Proof.
  unfold D1.
  apply DSpec.sub_merge_r.
  apply DSpec.disj_compat.
  apply DSpec.disj_sym.
  apply D_scap_disj_D0.
Qed.

Lemma stack_strengthen :
  forall psi n (g g' : acpred) s s',
    scapWFST n g s psi
    -> (forall s'', g' s' s'' -> g s s'')
    -> scapWFST n g' s' psi.
Proof.
  intros.
  induction n.
  unfold scapWFST in *.
  intros s'' Hg'.
  substH H0 with (H0 s'' Hg').
  substH H with (H s'' H0).
  apply H.
  simpl in *.
  intros s'' Hg'.
  substH H0 with (H0 s'' Hg').
  substH H with (H s'' H0).
  apply H.
Qed.

Lemma scap_j : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (f : label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s
    -> exists p'', exists g'',
      OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (j f) s)
      /\ (forall s', optg g'' (nextS (j f) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [p & g] *> | D]} / OPsi>> 
  (ione (j f)).
Proof.
  intros D OPsi p g f Hd Himp.
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *> .
  destruct Himp as [Hin Himp].
  split_l by
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [hp rf pc].
  set (s := <.hp, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
Qed.

Lemma scap_jal : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (f fret: label),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s 
    -> exists p'', exists g'', 
      OCHSpec.lookup OPsi f <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (jal f fret) s)
      /\ (forall s', 
        optg g'' (nextS (jal f fret) s) s'
        -> exists p', exists g',
          OCHSpec.lookup OPsi (S_pc s') <* rho_scap, [p' & g']*>
          /\ p' s'
          /\ (forall s'', g' s' s'' -> g s s'')))
  -> wfis D << {[ <* rho_scap, [p & g] *> | D ]} / OPsi>> 
  (ione (jal f fret)).
Proof.
  intros D OPsi p g f fret Hd Himp.
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *> .
  destruct Himp as [Hin Himp].
  split_l by
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [hp rf pc].
  set (s := <.hp, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  unfold nextS in Himp.
  destruct Himp as [Hp'' Himp].
  simpl.
  split_l.

  destruct Hk as [n Hk].
  exists (S n).
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  destruct Himp as [p''' [g''' Himp]].
  exists p'''; exists g'''.
  destruct Himp as [Hin' [Hp''' Hg''']].
  split_l by
    (apply (OCHSpec.lookup_sub_lookup Hin' Hsub)).
  split_l.
  apply (stack_strengthen Psi n g g''' s s'
    Hk Hg''').
Qed.

Lemma scap_jr : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred) (r : reg),
  DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s
    -> exists p'', exists g'',
      OCHSpec.lookup OPsi (S_gR s r) <* rho_scap, [p'' & g''] *>
      /\ optp p'' (nextS (jr r) s)
      /\ (forall s', 
        optg g'' (nextS (jr r) s) s' -> g s s'))
  -> wfis D << {[ <* rho_scap, [ p & g] *> | D ]} /OPsi>> 
  (ione (jr r)).
Proof.
  intros D OPsi p g r Hd Himp.
  unfold lift, itpscap, BigInterp.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Himp with (Himp s Hp).
  destruct Himp as [p'' [g'' Himp]].
  exists <* rho_scap, [p'' & g''] *> .
  destruct Himp as [Hin Himp].
  split_l by
    (apply (OCHSpec.lookup_sub_lookup Hin Hsub)).
  destruct s as [hp rf pc].
  set (s := <.hp, rf, pc.>).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  destruct Himp as [Hp'' Himp].
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Himp with (Himp s' Hg'').
  substH Hk with (Hk s' Himp).
  trivial.
Qed.

Lemma scap_ret : forall (D : idict) (OPsi : chspec)
  (p : stpred) (g : acpred),
  DSpec.sub D0 D
  -> DSpec.lookup D rho_scap <@itpscap@>
  -> (forall s, p s -> optp (g s) (nextS (jr ra) s))
  -> wfis D << {[ <* rho_scap, [ p & g] *> | D]} /OPsi>> 
  (ione (jr ra)).
Proof.
  intros D OPsi p g HDsub Hd Hs.
  unfold lift, itpscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hs with (Hs s Hp) into Hg.
  destruct Hk as [n Hk].
  destruct n.
  (* case n = 0 *)
  simpl in * .
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  set (s' := <.h, rf, gR rf ra.>).
  substH Hk with (Hk s' Hg).
  destruct Hk as [pi [Hfin H]].
  simpl in * .
  exists pi.
  split_l.
  apply (biginterp_enlarge_D HDsub Psi s' H).
  (* case n > 0 *)
  simpl in * .
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  set (s' := <.h, rf, gR rf ra.>).
  substH Hk with (Hk s' Hg).
  destruct Hk as [p' [g' Hk]].
  destruct Hk as [Hin [Hp's' Hk]].
  exists <* rho_scap, [p' & g'] *> .
  simpl in * .
  split_l.
  exists itpscap.
  split_l.
  unfold itpscap.
  split_l.
  exists n.
  assumption.
Qed.

Lemma scap_addu : forall (D : idict) (OPsi : chspec) (psi : scap_chspec)
  (p : stpred) (g : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (addu rd rs rt) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (addu rd rs rt) is).
Proof.
  intros D OPsi psi p g rd rs rt is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap.wfis in Hscap; fold scap.wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct s as [hp rf pc].
  unfold nextS in Hscap |- * .
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_addiu : 
  forall (D : idict) (OPsi : chspec) (psi : scap_chspec)
    (p : stpred) (g : acpred) 
    (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (addiu rd rs w) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (addiu rd rs w) is).
Proof.
  intros D OPsi psi p g rd rs w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct s as [hp rf pc].
  unfold nextS in Hscap |- * .
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_subu : forall (D : idict) (OPsi : chspec) 
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (subu rd rs rt) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (subu rd rs rt) is).
Proof.
  intros D OPsi psi p g rd rs rt is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct s as [hp rf pc].
  unfold nextS in Hscap |- * .
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_subiu : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (subiu rd rs w) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (subiu rd rs w) is).
Proof.
  intros D OPsi psi p g rd rs w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct s as [hp rf pc].
  unfold nextS in Hscap |- * .
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_move : forall (D : idict) (OPsi : chspec) 
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rd rs rt : reg) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (move rd rs) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (move rd rs) is).
Proof.
  intros D OPsi psi p g rd rs w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct s as [hp rf pc].
  unfold nextS in Hscap |- * .
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_li : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rd rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (li rd w) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (li rd w) is).
Proof.
  intros D OPsi psi p g rd rs rt is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct s as [hp rf pc].
  unfold nextS in Hscap |- * .
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_lw : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rt rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (lw rt w rs) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (lw rt w rs) is).
Proof.
  intros D OPsi psi p g rt rs w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  destruct (nextS (lw rt w rs) s) as [s0 | ];
    [idtac | tauto].
  destruct s as [hp rf pc].
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_sw : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rt rs : reg) w (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (sw rt w rs) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (sw rt w rs) is).
Proof.
  intros D OPsi psi p g rt rs w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi s [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap s Hp).
  destruct Hscap as [p' [g' Hscap]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  destruct Hscap as [Hp' [Hg' His]].
  split_r by (apply IHis; trivial).
  destruct (nextS (sw rt w rs) s) as [s0 | ];
    [idtac | tauto].
  destruct s as [hp rf pc].
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l.
  split_l.
  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_blt : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rs rt : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (blt rs rt f) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (blt rs rt f) is).
Proof.
  intros D OPsi psi p g rs rt w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi [hp rf pc] [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap _ Hp).
  unfold nextS in Hscap |- * .
  decbltnat (gR rf rs) (gR rf rt).
  (* case rs < rt *)
  destruct Hscap as [p'' [g'' Hscap]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Hscap as [Hin [Hp'' Hscap]].
  split.
    substH Hin with ((proj1 Hlift) _ _ _ Hin).
    apply (OCHSpec.lookup_sub_lookup Hin Hsub).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  simpl.
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Hscap with (Hscap s' Hg'').
  substH Hk with (Hk s' Hscap).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Hscap with (Hscap s' Hg'').
  substH Hk with (Hk s' Hscap).
  trivial.
  
  (* case rs >= rt *)
  destruct Hscap as [p' [g' [Hp' [Hg' His]]]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l; split_l.

  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_beq : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rs rt : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (beq rs rt f) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (beq rs rt f) is).
Proof.
  intros D OPsi psi p g rs rt w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi [hp rf pc] [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap _ Hp).
  unfold nextS in Hscap |- * .
  decbeqnat (gR rf rs) (gR rf rt).
  (* case rs = rt *)
  destruct Hscap as [p'' [g'' Hscap]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Hscap as [Hin [Hp'' Hscap]].
  split.
    substH Hin with ((proj1 Hlift) _ _ _ Hin).
    apply (OCHSpec.lookup_sub_lookup Hin Hsub).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  simpl.
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Hscap with (Hscap s' Hg'').
  substH Hk with (Hk s' Hscap).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Hscap with (Hscap s' Hg'').
  substH Hk with (Hk s' Hscap).
  trivial.
  
  (* case rs <> rt *)
  destruct Hscap as [p' [g' [Hp' [Hg' His]]]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l; split_l.

  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_bgtz : forall (D : idict) (OPsi : chspec)
  (psi : scap_chspec) (p : stpred) (g : acpred) 
  (rs : reg) (f : label) (is : instrseq),
  DSpec.lookup D rho_scap <@itpscap@>
  -> scap_chspec_lift psi OPsi
  -> (forall (p : stpred) (g : acpred),
    scap_wfis psi p g is ->
    wfis D << {[<* rho_scap, [p & g] *> | D]} / OPsi >> is)
  -> scap_wfis psi p g (iseq (bgtz rs f) is)
  -> wfis D << {[ <* rho_scap, [ p & g ] *> | D ]} / OPsi>> 
  (iseq (bgtz rs f) is).
Proof.
  intros D OPsi psi p g rs w is Hd Hlift IHis Hscap.
  unfold lift, itpscap.
  unfold wfis; fold wfis.
  unfold scap_wfis in Hscap; fold scap_wfis in Hscap.
  intros Psi [hp rf pc] [Hsub [itp [Hd' Hitp]]].
  assert (Hitpeq := DSpec.lookup_unique Hd Hd').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hitp.
  clear Hd' Hitpeq.
  destruct Hitp as [Hp Hk].
  substH Hscap with (Hscap _ Hp).
  unfold nextS in Hscap |- * .
  decbltnat O (gR rf rs).
  (* case O < rs *)
  destruct Hscap as [p'' [g'' Hscap]].
  exists (<* rho_scap, [p'' & g''] *>).
  destruct Hscap as [Hin [Hp'' Hscap]].
  split.
    substH Hin with ((proj1 Hlift) _ _ _ Hin).
    apply (OCHSpec.lookup_sub_lookup Hin Hsub).
  exists itpscap.
  unfold itpscap at 2.
  split_l.
  simpl.
  split_l.

  destruct Hk as [n Hk].
  exists n.
  destruct n.
  simpl in * .
  intros s' Hg''.
  substH Hscap with (Hscap s' Hg'').
  substH Hk with (Hk s' Hscap).
  trivial.
  simpl in * .
  intros s' Hg''.
  substH Hscap with (Hscap s' Hg'').
  substH Hk with (Hk s' Hscap).
  trivial.
  
  (* case O >= rs *)
  destruct Hscap as [p' [g' [Hp' [Hg' His]]]].
  exists (<< {[ <* rho_scap, [p' & g']*> | D ]} / OPsi>>).
  split_r by (apply IHis; trivial).
  unfold lift, itpscap, BigInterp.
  split_l.
  exists itpscap; simpl; split_l; split_l.

  destruct Hk as [n Hk].
  exists n.
  induction n; unfold scapWFST; fold scapWFST.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
  intros s' Hg's0.
  substH Hk with (Hk s' (Hg' s' Hg's0)).
  trivial.
Qed.

Lemma scap_wfis_ocap_wfis :
  forall D OPsi psi is p g,
    DSpec.sub D0 D
    -> DSpec.lookup D rho_scap <@itpscap@>
    -> scap_chspec_lift psi OPsi
    -> scap_wfis psi p g is
    -> ocap_wfis D << {[ <* rho_scap, [p & g] *> | D]} / OPsi>> is.
Proof.
  intros D OPsi psi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r]; 
    unfold scap_wfis; fold scap_wfis;
    unfold ocap_wfis; fold ocap_wfis;
    intros p g Hd Hin Hlift Hscap;
      try inversion_clear Hscap.

  apply scap_j; trivial.
  intros s Hp.
  destruct (Hscap s Hp) as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_r.
  apply ((proj1 Hlift) _ _ _ Hin').

  apply scap_jal; trivial.
  intros s Hp.
  substH Hscap with (Hscap s Hp). 
  destruct Hscap as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_l with ((proj1 Hlift) _ _ _ Hin').
  destruct H' as [Hp'' H].
  split_l.
  intros s' Hg''.
  substH H with (H s' Hg'').
  destruct H as [p''' [g''' [Hin''' [Hp''' Hg''']]]].
  exists p'''; exists g'''.
  split_l with ((proj1 Hlift) _ _ _ Hin''').
  split_lr.

  decbeqreg r ra.
    apply scap_ret; trivial.
  apply scap_jr; trivial.
  intros s Hp.
  destruct (Hscap s Hp) as [p'' [g'' [Hin' H']]].
  exists p''; exists g''.
  split_r.
  apply ((proj1 Hlift) _ _ _ Hin').

  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r ]; 
    unfold scap_wfis; fold scap_wfis;
    intros p g Hd Hin Hlift Hscap;
      try inversion_clear Hscap.
  
  apply scap_addu with psi; auto.
  apply scap_addiu with psi; auto.
  apply scap_subu with psi; auto.
  apply scap_subiu with psi; auto.
  apply scap_move with psi; auto.
  apply scap_li with psi; auto.
  apply scap_lw with psi; auto.
  apply scap_sw with psi; auto.
  apply scap_blt with psi; auto.
  apply scap_beq with psi; auto.
  apply scap_bgtz with psi; auto.
Qed.

Lemma D_scap_describe_lift_psi :
  forall OPsi psi,
    scap_chspec_lift psi OPsi
    -> describe D_scap OPsi.
Proof.
  intros OPsi psi.
  unfold scap_chspec_lift.
  intros H f rho lty theta Hin.
  destruct H as [H1 H2].
  destruct (H2 f <*rho,theta*> Hin) as [p [g [Hpi Hin']]].
  injection Hpi.
  intros Heq1 Heq2 Heq3.
  subst.
  exists itpscap.
  apply DSpec.get_lookup.
  unfold D_scap, DSpec.get, DSpecDef.get.
  rewbnat (beq_nat rho_scap rho_scap = true).
  trivial.
Qed.

Lemma scap_wfch_ocap_wfch :
  forall ch psi psi' OPsi OPsi',
    scap_chspec_lift psi OPsi
    -> scap_chspec_lift psi' OPsi'
    -> scap_wfch psi psi' ch
    -> ocap_wfch (DSpec.merge D0 D_scap) OPsi OPsi' ch.
Proof.
  intros ch psi psi' OPsi OPsi'.
  intros Hpsi Hpsi' Hch.
  unfold wfch.
  split.
    split.
      assert (H : D_scap ## OPsi).
      eapply D_scap_describe_lift_psi; eauto.
      apply describe_sub with D_scap; trivial.
      apply D_scap_sub_D1.
    assert (H : D_scap ## OPsi').
    eapply D_scap_describe_lift_psi; eauto.
    apply describe_sub with D_scap; trivial.
    apply D_scap_sub_D1.
  intros f pi Hin.
  unfold scap_chspec_lift in Hpsi'.
  unfold wfch in Hch.
  destruct Hpsi' as [_ Hpsi'].
  destruct (Hpsi' f pi Hin) as [p [g [Hpi Hin']]].
  subst pi.
  
  destruct (Hch f p g (CHSpec.lookup_get Hin')) as [is [Hgis Hwfis]].
  exists is; split_l.
  apply scap_wfis_ocap_wfis with psi.
  apply D0_sub_D1.
  apply DSpec.lookup_merge_intro.
  right.
  apply DSpec.get_lookup.
  unfold D_scap, DSpec.get, DSpecDef.get.
  rewbnat (beq_nat rho_scap rho_scap = true).
  trivial.
  apply DSpec.disj_compat.
  apply DSpec.disj_sym.
  apply D_scap_disj_D0.
  trivial.
  trivial.
Qed.

Theorem scap_sound :
  forall ch psi psi' OPsi OPsi',
    scap_chspec_lift psi OPsi
    -> scap_chspec_lift psi' OPsi'
    -> scap_wfch psi psi' ch
    -> ocap_wfch (DSpec.merge D0 D_scap) OPsi OPsi' ch.
Proof scap_wfch_ocap_wfch.

End ScapOcap.
