(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-0, a monomorphic version                                *)
(***************************************************************)

(* $Id: tal0_ck.v,v 1.1.1.1 2006/12/07 16:43:19 ssg05 Exp $ *)

Require Import tal0.
Require Import machine.
Require Import extac.

(* *********************************************************** *)

(* type checker *)

Fixpoint typechk (psi : chspec) (a : codeTy) (is : instrseq) 
  {struct is} : Prop :=
  match is with 
    | ione (j f) => 
        match (CHSpec.get psi f) with
          | Some a'' => codeTy_sub a a''
          | _ => False
        end

    | ione (jr r) => 
        match (RfTy.sel a r) with
          | Some (valty_code a'') => codeTy_sub a a''
          | _ => False
        end

    | iseq (addu rd rs rt) is' => 
        match beq_reg rd r0 with
          | false => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some valty_int, Some valty_int) => 
                typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | true => False 
        end

    | iseq (addiu rd rs w) is' => 
        match beq_reg rd r0 with 
          | false =>
            match RfTy.sel a rs with
              | Some valty_int => 
                typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | true => False
        end

    | iseq (subu rd rs rt) is' => 
        match beq_reg rd zero with
          | false => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some valty_int, Some valty_int) => 
                   typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (subiu rd rs w) is' => 
        match beq_reg rd zero with
          | false => 
            match RfTy.sel a rs with
              | Some valty_int => 
                   typechk psi (RfTy.upd a rd valty_int) is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (move rd rs) is' =>
        match beq_reg rd r0 with
          | false =>
            match RfTy.sel a rs with 
              | Some t => 
                typechk psi (RfTy.upd a rd t) is'
              | _ => False
            end
          | true => False
        end

    | iseq (li rd w) is' =>
        match beq_reg rd r0 with
          | false =>
            match (CHSpec.get psi w) with 
              | Some a'' => 
                typechk psi (RfTy.upd a rd (valty_code a'')) is'
              | _ => False
            end
          | true => False
        end

    | iseq (lw rt w rs) is' => 
        match beq_reg rt zero with
          | false => 
            match RfTy.sel a rs with
              | Some (valty_tup o) =>
                match (TupTy.get o w) with 
                  | Some (init, t) => 
                    typechk psi (RfTy.upd a rt t) is'
                  | _ => False
                end
              | _ => False
            end
          | _ => False 
        end

    | iseq (sw rt w rs) is' => 
        match beq_reg rs zero with
          | false => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some (valty_tup o), Some t) =>
                match (TupTy.get o w) with 
                  | Some (init, t') => 
                    t' = t 
                    /\ typechk psi (RfTy.upd a rs 
                      (valty_tup (TupTy.upd o w (init, t)))) is'
                  | Some (junk, t') => 
                    t' = t 
                    /\ typechk psi (RfTy.upd a rs 
                      (valty_tup (TupTy.upd o w (init, t)))) is'
                  | _ => False
                end
              | _ => False
            end
          | _ => False 
        end

    | iseq (blt rs rt f) is' => 
        match CHSpec.get psi f with
          | Some a'' => 
            match (RfTy.sel a rs, RfTy.sel a rt) with
              | (Some valty_int, Some valty_int) => 
                  codeTy_sub a a''
                  /\ typechk psi a is'
              | _ => False
            end
          | _ => False 
        end

    | iseq (bgtz rs f) is' => 
        match CHSpec.get psi f with
          | Some a'' => 
            match RfTy.sel a rs with
              | Some valty_int => 
                  codeTy_sub a a''
                  /\ typechk psi a is'
              | _ => False
            end
          | _ => False 
        end

    | _ => False
  end.

Theorem type_checker_sound : forall psi is a,
  typechk psi a is -> wfis psi a is.
Proof.
  intros psi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    simpl; intros a H; 
      try contradiction.
  
  (* case i = j f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in H].
  eapply wfis_j; eauto.
  apply CHSpec.get_lookup; trivial.

  (* case i = jr r *)
  assert (Hr : exists t, 
    RfTy.sel a r = Some t);
    [destruct (RfTy.sel a r) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hr as [t Hr];
        rewrite Hr in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_jr with a''; trivial;
    try constructor; trivial.

  (* case is = iseq i is' *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs f
      | f | r]; 
    simpl; intros a H; try contradiction.

  (* case i = addu rd rs rt *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].

  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_addu; try constructor; trivial.
  eapply IHis; eauto.
  (* case i = addiu rd rs w *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_addiu; try constructor; trivial.
  eapply IHis; eauto.
  (* case i = subu rd rs rt *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_subu; try constructor; trivial.
  eapply IHis; eauto.
  (* case i = subiu rd rs w *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  apply wfis_subiu; try constructor; trivial.
  eapply IHis; eauto.
  (* case i = move rd rs *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  apply wfis_move with t; try constructor; trivial.
  eapply IHis; eauto.
  (* case i = li rd w *)
  assert (Hrdnr0 : beq_reg rd zero = false);
    [destruct (beq_reg rd zero); 
      [contradiction | trivial]
      | rewrite Hrdnr0 in H].
  assert (Hw : exists a'', 
    CHSpec.get psi w = Some a'');
    [destruct (CHSpec.get psi w) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hw as [a'' Hw];
        rewrite Hw in H].
  apply wfis_li with a''; 
    try constructor; trivial.
  eapply IHis; eauto.
  (* case i = lw rt w rs *)
  assert (Hrtnr0 : beq_reg rt zero = false);
    [destruct (beq_reg rt zero); 
      [contradiction | trivial]
      | rewrite Hrtnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | a'' ]; try contradiction.
  assert (Hw : exists ct, 
    TupTy.get o i = Some ct);
    [destruct (TupTy.get o i) as [ct |] ;
      try contradiction;
        exists ct; trivial
      | destruct Hw as [ct Hw];
        rewrite Hw in H].
  destruct ct as [f t]; destruct f.
  apply wfis_lw with o t; 
    try constructor; trivial.
  apply TupTy.get_lookup; trivial.
  eapply IHis; eauto.
  contradiction.
  (* case i = sw rt w rs *)
  assert (Hrsnr0 : beq_reg rs zero = false);
    [destruct (beq_reg rs zero); 
      [contradiction | trivial]
      | rewrite Hrsnr0 in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | ax ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  assert (Hw : exists ct, 
    TupTy.get o i = Some ct);
    [destruct (TupTy.get o i) as [ct |] ;
    try contradiction;
      exists ct; trivial
      | destruct Hw as [ct Hw];
        rewrite Hw in H].
  destruct ct as [f t']; destruct f.

  destruct H as [Heq H].
  subst t'.
  apply wfis_sw with o (init, t) t; 
    try constructor; trivial.
  apply TupTy.get_lookup; trivial.
  eapply IHis; eauto.

  destruct H as [Heq H].
  subst t'.
  apply wfis_sw with o (junk, t) t; 
    try constructor; trivial.
  apply TupTy.get_lookup; trivial.
  eapply IHis; eauto.
  (* case i = blt rs rt f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | ax ]; try contradiction.
  assert (Hrt : exists t, 
    RfTy.sel a rt = Some t);
    [destruct (RfTy.sel a rt) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in H].
  destruct t as [ | | o | ax ]; try contradiction.
  destruct H as [Hsub H].
  apply wfis_blt with a''; trivial;
    try constructor; trivial.
  eapply IHis; eauto.
  (* case i = bgtz rs rt f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try contradiction;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in H].
  assert (Hrs : exists t, 
    RfTy.sel a rs = Some t);
    [destruct (RfTy.sel a rs) as [t |] ;
      try contradiction;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in H].
  destruct t as [ | | o | ax ]; try contradiction.
  destruct H as [Hsub H].
  apply wfis_bgtz with a''; trivial;
    try constructor; trivial.
  eapply IHis; eauto.
Qed.

Ltac typecheck :=
  match goal with 
    | |- wfis ?psi ?a ?is =>
      apply type_checker_sound;
        compute;
          simpl;
            match goal with
              | |- False => fail 1 "Ill-formed type."
              | _ => idtac
            end
    | _ => fail 1 "The goal is not a TAL instrseq judgement."
  end.

(* *********************************************************** *)

(*
Ltac psi_lookup label :=
  match goal with
    | H : CHSpec.lookup ?psi ?f ?a
      |- _ =>
      decbeqnat f label;
      [ substH H with (@CHSpec.lookup_get psi f a H);
        unfold CHSpec.get, CHSpec.get, CHSpecDef.get in H;
        unfold psi in H;
        bnat_rew_all;
        beq_nat_subst f;
        simpl in H;
          (match type of H with 
             | None = Some a  => discriminate H
             | Some ?b = Some a =>  inj_hyp H; subst a
           end)
        | idtac]
    | _ => idtac
  end.

Ltac psi_lookup_not :=
  match goal with
    | H : CHSpec.lookup ?psi ?f ?a |- _ =>
      substH H with (@CHSpec.lookup_get psi f a H);
      unfold CHSpec.get, CHSpec.get, CHSpecDef.get in H;      
      unfold psi in H; bnat_rew_all; simpl in H;
        (match type of H with
           | None = Some a => discriminate H
           | _ => fail
         end)
    | _ => fail
  end.
*)
(* *********************************************************** *)
