(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Rule.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Ty.
Require Import TyEnv.
Require Import Varmap.
Require Import Subst.

Import TyEnv.

(* **************************************************)
(* type ok                                          *)

Inductive valTyok : tenv -> valTy -> Prop :=
  | valTyok_var : forall (e : tenv) (n : nat),
                    TyEnv.ine e n
                    -> valTyok e (valTy_var n)
  | valTyok_int : forall (e : tenv),
                    valTyok e valTy_int
  | valTyok_tup : forall (e : tenv) (o : tupTy),
                    tupTyok e o 
                    -> valTyok e (valTy_tup o)
  | valTyok_code : 
             forall (e e' : tenv) (Y : rfTy),
               rfTyok (TyEnv.app e e') Y
               -> valTyok e (valTy_code e' Y)
with rfTyok : tenv -> rfTy -> Prop :=
  | rfTyok_nil : 
             forall (e : tenv),         
               rfTyok e rfTy_nil
  | rfTyok_cons :
             forall (e : tenv) r t Y,
               valTyok e t
               -> rfTyok e Y
               -> rfTyok e (rfTy_cons (r, t) Y)
with tupTyok : tenv -> tupTy -> Prop :=
  | tupTyok_nil :
             forall (e : tenv),
               tupTyok e tupTy_nil
  | tupTyok_cons :
             forall (e : tenv) f t o,
               valTyok e t
               -> tupTyok e o
               -> tupTyok e (tupTy_cons (f, t) o).

Definition codeTyok (e0 : tenv) (a : codeTy) : Prop :=
  let (e, Y) := a in
    valTyok e0 (valTy_code e Y).


(* **************************************************)
(* value type list ok                               *)

Inductive valTyLok : tenv -> varmap -> Prop :=
  | valTyLok_vO : forall e,
                 valTyLok e nil
  | valTyLok_vS : 
               forall e n t tl,
                 valTyok e t
                 -> valTyLok e tl
                 -> valTyLok e (cons (n, t) tl).

(* **************************************************)
(* cell subtyping                                   *)

Inductive ceTy_sub : forall (ct ct' : cellTy), Prop :=
  | ceTy_sub_refl : forall (ct : cellTy),
                       ceTy_sub ct ct
  | ceTy_sub_init : forall (t : valTy), 
                       ceTy_sub (init, t) (junk, t).

(* **************************************************)
(* regfile subtyping                                *)

Inductive rfTy_sub : forall (e0 : tenv) (Y Y' : rfTy), Prop :=
  | rfTy_sub_def :
      forall e0 (Y Y' : rfTy),
        rfTyok e0 Y
        -> rfTyok e0 Y'
        -> (forall r : reg,
             RfTy.indom Y' r -> RfTy.sel Y' r = RfTy.sel Y r)
        -> rfTy_sub e0 Y Y'.

(* **************************************************)
(* code subtyping                                   *)

Definition codeTy_sub (e0 : tenv) (a a' : codeTy) : Prop :=
  let (e, Y) := a in
  let (e', Y') := a' in
    exists tl : varmap,
      codeTyok e0 (e, Y) 
      /\ codeTyok e0 (e', Y')
      /\ valTyLok (TyEnv.app e0 e) tl 
      /\ rfTy_sub (TyEnv.app e0 e) Y (instYL (shiftY Y' e e' ) tl e' 0).

Lemma codeTy_sub_rfTyok_left :
  forall e0 e Y a,
    codeTy_sub e0 (e, Y) a
    -> rfTyok (app e0 e) Y.
Proof.
  intros e0 e Y a Hsub.
  destruct a.
  simpl in Hsub.
  destruct Hsub as [tl [H Hx]].
  inversion_clear H; trivial.
Qed.

(*
Lemma codeTy_sub_emp_rfTyok_left :
  forall e Y a,
    codeTy_sub TyEnv.emp (e, Y) a
    -> rfTyok e Y.
Proof.
  intros e Y a Hsub.
  destruct a.
  simpl in Hsub.
  destruct Hsub as [tl [H Hx]].
  inversion_clear H; trivial.
Qed.
*)
(* **************************************************)
(* value typing                                     *)

Inductive valok : chspec -> dhspec -> word -> valTy -> Prop :=
  | valok_var : forall (Psi : chspec) (Phi : dhspec) (w : word)
                  (n : nat),
                  valok Psi Phi w (valTy_var n)
  | valok_lab : 
              forall (Psi : chspec) (Phi : dhspec) 
                (f : word) (e : tenv) (Y : rfTy) (a : codeTy),
                CHSpec.lookup Psi f a
                -> codeTy_sub TyEnv.emp (e, Y) a
                -> valok Psi Phi f (valTy_code e Y)
  | valok_int : forall (Psi : chspec) (Phi : dhspec) (w : word),
              valok Psi Phi w valTy_int
  | valok_tup : forall (Psi : chspec) (Phi : dhspec) 
                    (l : label) (tup : tupTy),
              (forall i, 
                TupTy.indom tup i
                -> exists ct, exists ct',
                  TupTy.lookup tup i ct
                  /\ DHSpec.lookup Phi (l+i) ct'
                  /\ ceTy_sub ct' ct)
              -> valok Psi Phi l (valTy_tup tup).

Inductive valok2 : chspec -> word -> valTy -> Prop :=
  | valok2_lab : 
              forall (Psi : chspec) f e Y,
                CHSpec.lookup Psi f (e, Y)
                -> codeTyok TyEnv.emp (e, Y)
                -> valok2 Psi f (valTy_code e Y)
  | valok2_int : forall (Psi : chspec) w,
              valok2 Psi w valTy_int.

(*
Lemma valok2_valok : forall psi phi t w,
  valok2 psi w t -> valok psi phi w t.
Proof.
  intros psi phi t w H.
  inversion_clear H.
  apply valok_lab with (e, Y); trivial.
  apply valok_int; trivial.
Qed.
  *)
            
Inductive regok : tenv -> rfTy -> reg -> valTy -> Prop :=
  | regok_reg : forall (e : tenv) (Y : rfTy) (r : reg) (t : valTy),
              RfTy.sel Y r = Some t
              -> valTyok e t
              -> regok e Y r t.

Inductive ceok : chspec -> dhspec -> word -> cellTy -> Prop :=
  | ceok_junk : forall Psi Phi v (t : valTy),
(*                    valTyok TyEnv.emp t -> *)
                   ceok Psi Phi v (junk, t)
  | ceok_init : forall Psi Phi v t,
                   valok Psi Phi v t ->
(*                    valTyok TyEnv.emp t -> *)
                   ceok Psi Phi v (init, t).

(* **************************************************)
(* instruction sequence typing                      *)

Fixpoint wfis (Psi : chspec) (a : codeTy) 
  (is : instrseq) {struct is} : Prop :=
  let (e, Gamma) := a in
  match is with

    | ione (j f) => 
          exists a'', CHSpec.lookup Psi f a''
            /\ codeTy_sub emp (e, Gamma) a''

    | ione (jal f fret) => 
          exists a', CHSpec.lookup Psi fret a' 
            /\ codeTyok emp a'
            /\ let (e', Y') := a' in
              exists a'', CHSpec.lookup Psi f a''
                /\ codeTy_sub emp 
                (e, RfTy.upd Gamma ra (valTy_code e' Y')) a''

    | ione (jr r) => 
          exists a'' : codeTy,
            let (e'', Gamma'') := a'' in
              regok e Gamma r (valTy_code e'' Gamma'')
              /\ codeTy_sub e (emp, Gamma) (e'', Gamma'')

    | ione i => False
      
    | iseq (addu rd rs rt) is =>
         beq_reg rd r0 = false
         /\ regok e Gamma rs valTy_int
         /\ regok e Gamma rt valTy_int
         /\ wfis Psi (e, RfTy.upd Gamma rd valTy_int) is

    | iseq (addiu rd rs w) is =>
         beq_reg rd r0 = false
         /\ regok e Gamma rs valTy_int
         /\ wfis Psi (e, RfTy.upd Gamma rd valTy_int) is
         
    | iseq (subu rd rs rt) is =>
         beq_reg rd r0 = false
         /\ regok e Gamma rs valTy_int
         /\ regok e Gamma rt valTy_int
         /\ wfis Psi (e, RfTy.upd Gamma rd valTy_int) is

    | iseq (subiu rd rs w) is =>
         beq_reg rd r0 = false
         /\ regok e Gamma rs valTy_int
         /\ wfis Psi (e, RfTy.upd Gamma rd valTy_int) is
         
    | iseq (move rd rs) is =>
         exists t,
           beq_reg rd r0 = false
           /\ regok e Gamma rs t
           /\ wfis Psi (e, RfTy.upd Gamma rd t) is

    | iseq (li rd w) is =>
         beq_reg rd r0 = false
         /\ exists t, valok2 Psi w t
           /\ wfis Psi (e, RfTy.upd Gamma rd t) is

    | iseq (lw rt w rs) is =>
          exists o, exists t,
            beq_reg rt r0 = false
            /\ regok e Gamma rs (valTy_tup o)
            /\ TupTy.lookup o w (init, t)
            /\ wfis Psi (e, RfTy.upd Gamma rt t) is
            
    | iseq (sw rt w rs) is =>
           exists o, exists t, exists flag,
             beq_reg rs r0 = false
             /\ regok e Gamma rt t
             /\ regok e Gamma rs (valTy_tup o)
             /\ TupTy.lookup o w (flag, t)
             /\ ceTy_sub (init, t) (flag, t)
             /\ wfis Psi (e, RfTy.upd Gamma rs 
               (valTy_tup (TupTy.upd o w (init, t)))) is

    | iseq (blt rs rt f) is =>
           exists a'', CHSpec.lookup Psi f a''
            /\ regok e Gamma rs valTy_int
            /\ regok e Gamma rt valTy_int
            /\ codeTy_sub emp (e, Gamma) a''
            /\ wfis Psi (e, Gamma) is
             
    | iseq (beq rs rt f) is =>
           exists a'', CHSpec.lookup Psi f a''
            /\ regok e Gamma rs valTy_int
            /\ regok e Gamma rt valTy_int
            /\ codeTy_sub emp (e, Gamma) a''
            /\ wfis Psi (e, Gamma) is
             
    | iseq (bgtz rs f) is =>
           exists a'', CHSpec.lookup Psi f a''
            /\ regok e Gamma rs valTy_int
            /\ codeTy_sub emp (e, Gamma) a''
            /\ wfis Psi (e, Gamma) is

    | iseq _ is => False

  end.

(* **************************************************)
(* codeheap typing                                  *)

Definition wfpsi (Psi : chspec) : Prop :=
  forall f a, 
    CHSpec.lookup Psi f a
    -> codeTyok emp a.

Definition wfch (Psi Psi' : chspec) (C : codeheap) : Prop :=
  forall (f : label) (a : codeTy),
    CHSpec.lookup Psi' f a
    -> (exists I : _ , 
        codeheap.getis C f = Some I
        /\ wfis Psi a I).

(* **************************************************)
(* regfile typing                                   *)

Definition wfrf Psi Phi Gamma R : Prop :=
  forall r t, 
    RfTy.sel Gamma r = Some t
    -> valok Psi Phi (Regfile.get R r) t.

(* **************************************************)
(* dataheap typing                                  *)

Definition wfdh Psi Phi Phi' H : Prop :=
  forall l ct, 
    DHSpec.lookup Phi' l ct
    -> exists v,
        Heap.lookup H l v
        /\ ceok Psi Phi v ct.

(* **************************************************)
(* state typing                                     *)

Definition wfst Psi (a : codeTy) (S : state) : Prop :=
  match S with
    | mkst H R pc =>
      exists Phi : dhspec,
        wfdh Psi Phi Phi H 
        /\ exists Y', 
          codeTy_sub emp (emp, Y') a
          /\ wfrf Psi Phi Y' R
  end.

(* **************************************************)
(* program typing                                   *)
(*
Definition wfcodety (Psi : chspec) (a : codeTy) (pc : pcreg) : Prop :=
  CHSpec.get Psi pc = None 
  \/ exists a',
       CHSpec.get Psi pc = Some a'
       /\ codeTy_sub emp a a'.

Definition wfpc (Psi : chspec) (C : codeheap) (a : codeTy) (pc : pcreg) : Prop :=
  exists I : instrseq, 
    codeheap.getis C pc = Some I
    /\ wfis Psi a I pc.

Definition wfmach (Psi : chspec) (M : machine) : Prop := 
  let (C, S) := M in
  let (H, R, pc) := S in
    exists a,
      wfpsi Psi
      /\ wfch Psi Psi C
      /\ wfcodety Psi a pc
      /\ wfst Psi a H R pc
      /\ wfpc Psi C a pc.

Definition safe_onestep (Psi : chspec) (M : machine) : Prop :=
  exists M' : _, 
    step M = Some M'
    /\ wfmach Psi M'.

Definition safe (Psi : chspec) (M : machine) : Prop :=
  forall n,
    exists M' : _, 
      stepn n M = Some M'
      /\ wfmach Psi M'.
*)