(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Rule_lemmas.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Ty.
Require Import TyEnv.
Require Import Varmap.

Require Import Subst.
Require Import Subst_lemmas.

Require Import Rule.
Require Import Tyok_lemmas.
Require Import SubTy_lemmas.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Import TyEnv.

(***************************************************************)

Lemma wfpsi_codeTyok :
  forall Psi f a,
    wfpsi Psi
    -> CHSpec.get Psi f = Some a
    -> codeTyok emp a.
Proof.
  intros Psi f a HPsi Hf.
  unfold wfpsi in HPsi.
  apply HPsi with f.
  apply CHSpec.get_lookup; trivial.
Qed.

(*
Lemma wfpsi_lookup_wfcodety : forall Psi f a,
  wfpsi Psi
  -> CHSpec.lookup Psi f a
  -> wfcodety Psi a f.
Proof.
  intros Psi f a HPsi Hf.
  right.
  exists a.
  split.
    apply CHSpec.lookup_get.
    trivial.
  apply codeTy_sub_refl.
  unfold wfpsi in HPsi.
  eapply HPsi; eauto.
Qed. *)

Lemma wfst_weakening : forall Psi a a' H R pc,
  codeTy_sub emp a a'
  -> wfst Psi a (mkst H R pc)
  -> wfst Psi a' (mkst H R pc).
Proof.
  intros Psi [Delta Gamma] [Delta' Gamma'] H R pc.
  unfold wfst; intros Hsub Ha.
  destruct Ha as [Phi [HH [Gamma'' [Hsub' HR]]]].
  exists Phi.
  split_l.
  exists Gamma''.
  split_r.
  eapply codeTy_sub_trans; eauto.
Qed.

(***************************************************************)

Lemma reg_typing : 
  forall Psi Phi Gamma R r t,
    wfrf Psi Phi Gamma R
    -> regok TyEnv.emp Gamma r t
    -> valok Psi Phi (gR R r) t.
Proof.
  intros Psi Phi Gamma R r t HR Hr.
  destruct Hr as [Gamma r t H].
  unfold wfrf in HR.
  auto.
Qed.

Lemma regfile_update :
  forall Psi Phi Gamma R r w t,
    beq_reg r zero = false
    -> wfrf Psi Phi Gamma R
    -> valok Psi Phi w t
    -> wfrf Psi Phi (RfTy.upd Gamma r t) (sR R r w).
Proof.
  intros Psi Phi Gamma R r w t Hzero HR Hw.
  unfold wfrf in * .
  intros r0 t0.
  assert (HR' := HR r0 t0).
  unfold Regfile.set, Regfile.get in * .
  decbeqreg r0 r.
    intros Ht.
    rewrite (RfTy.sel_upd_req Gamma r r t (beq_refl _)) in Ht.
    injection Ht; intro; subst t0.
    trivial.
  intro Ht.
  rewrite (RfTy.sel_upd_rneq Gamma r0 r t Hb) in Ht.
  decbeqreg zero r0; auto.
Qed.
Implicit Arguments regfile_update
  [Psi Phi Gamma R r w t].

Lemma regfile_update_reg :
  forall Psi Phi Delta Gamma R rd rs t,
    beq_reg rd zero = false
    -> wfrf Psi Phi Gamma R
    -> regok Delta Gamma rs t
    -> wfrf Psi Phi (RfTy.upd Gamma rd t) (sR R rd (gR R rs)).
Proof.
  intros Psi Phi Delta Gamma R rd rs t Hzero HR Hrs.
  unfold wfrf in * .
  intros r0 t0.
  assert (HR' := HR r0 t0).
  unfold Regfile.set, Regfile.get in * .
  decbeqreg r0 rd.
    intros Ht.
    rewrite (RfTy.sel_upd_req Gamma rd rd t (beq_refl _)) in Ht.
    injection Ht; intro; subst t0.
    inversion_clear Hrs.
    apply HR; trivial.
  intro Ht.
  rewrite (RfTy.sel_upd_rneq Gamma r0 rd t Hb) in Ht.
  decbeqreg zero r0; auto.
Qed.
Implicit Arguments regfile_update_reg
  [Psi Phi Delta Gamma R rd rs t].

(* Lemma wfrf_weaken_Gamma :  *)
Lemma regfile_weakening :
  forall Psi Phi Delta Gamma Gamma' R,
    wfrf Psi Phi Gamma R
    -> rfTy_sub Delta Gamma Gamma'
    -> wfrf Psi Phi Gamma' R.
Proof.
  intros Psi Phi Delta Gamma Gamma' R HR Hsub.
  unfold wfrf in * .
  intros r t Hr.
  inversion_clear Hsub as [e0x Yx Y'x HY HY' Hin].
  unfold RfTy.indom in Hin.
  apply HR.
  apply sym_eq.
  assert (Hin' := Hin r).
  rewrite Hr in Hin'.
  apply Hin'.
  exists t.
  trivial.
Qed.
Implicit Arguments regfile_weakening 
   [Psi Phi Delta Gamma Gamma' R].

Lemma heapval_exists:
  forall Psi Phi H o ct addr offset,
    wfdh Psi Phi Phi H
    -> TupTy.lookup o offset ct
    -> valok Psi Phi addr (valTy_tup o)
    -> exists v, 
      Heap.lookup H (addr+offset) v.
Proof.
  intros Psi Phi H o ct0 addr offset.
  intros HH Ht Ho.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists ct0; trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct0.
  unfold wfdh in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v [Hin Hceok]].
  exists v.
  trivial.
Qed.
Implicit Arguments heapval_exists
    [Psi Phi H o ct addr offset].

Lemma heapval_typing :
  forall Psi Phi H o t addr offset v,
    wfdh Psi Phi Phi H
    -> TupTy.lookup o offset (init, t)
    -> valok Psi Phi addr (valTy_tup o)
    -> Heap.lookup H (addr+offset) v
    -> valok Psi Phi v t.
Proof.
  intros Psi Phi H o t addr offset v.
  intros HH Ht Ho Hv.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists (init, t); trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct.
  unfold wfdh in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v' [Hv' Hceok]].
  assert (He := Heap.lookup_unique Hv Hv').
  subst v'.
  inversion Hsub.
  subst ct'; subst ct.
  inversion_clear Hceok.
  trivial.
Qed.
Implicit Arguments heapval_typing 
    [Psi Phi H o t addr offset v].

(* Lemma heap_val_typing :  *)
(*
Lemma heapval_typing_old :
  forall Psi Phi H o t addr offset,
    wfdh Psi Phi H
    -> TupTy.lookup o offset (init, t)
    -> valok Psi Phi addr (valTy_tup o)
    -> exists v, Heap.lookup H (addr+offset) v
      /\ valok Psi Phi v t.
Proof.
  intros Psi Phi H o t addr offset.
  intros HH Ht Ho.
  inversion_clear Ho.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists (init, t); trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  assert (He := TupTy.lookup_unique _ _ _ _ Hct Ht).
  subst ct.
  unfold wfdh in HH.
  clear Hin.
  destruct (HH (addr+offset) ct' Hct') as
    [v [Hin Hceok]].
  exists v.
  split_l.
  inversion Hsub.
  subst ct'; subst ct.
  inversion_clear Hceok.
  trivial.
Qed.
Implicit Arguments heapval_typing 
    [ Psi Phi H o t addr offset].
*)

Lemma heap_typing_reg_tying_ex_t :
  forall Psi Phi Delta Gamma H R r o offset ct,
    wfdh Psi Phi Phi H
    -> wfrf Psi Phi Gamma R
    -> regok Delta Gamma r (valTy_tup o)
    -> TupTy.lookup o offset ct
    -> exists ct',
      DHSpec.lookup Phi (gR R r + offset) ct'
      /\ ceTy_sub ct' ct.
Proof.
  intros Psi Phi Delta Gamma H R r o offset ct.
  intros HH HR Hr Hct.
  inversion_clear Hr.
  unfold wfrf in HR.
  substH HR with (HR r (valTy_tup o) H0).
  inversion_clear HR.
  assert (Hin : TupTy.indom o offset).
  unfold TupTy.indom.
  exists ct; trivial.
  substH H1 with (H2 offset Hin).
  destruct H1 as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  exists ctx'.
  split_l.
  rewrite (TupTy.lookup_unique _ _ _ _ Hct Hctx).
  trivial.
Qed.
Implicit Arguments heap_typing_reg_tying_ex_t
    [Psi Phi Gamma H R r o offset ct].

Lemma heap_extend_valok : 
  forall Psi Phi addr ct t v,
    DHSpec.get Phi addr = None
    -> valok Psi Phi v t
    -> valok Psi (DHSpec.set Phi addr ct) v t.
Proof.
  intros Psi Phi addr ct t v.
  intros Hnin Hv.
  inversion_clear Hv.
  apply valok_var; trivial.
  apply valok_lab with a; trivial.
  apply valok_int; trivial.
  apply valok_tup.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsub]]]].
  destruct (nat_eq_dec addr (v + i)) as [Heq | Hneq].
(* case addr = v + i *)
  subst addr.
  rewrite Hctx' in Hnin.
  discriminate Hnin.
(* case addr <> v + i *)
  exists ctx. exists ctx'.
  split_l.
  split.
  apply DHSpec.get_lookup.
  rewrite (DHSpec.set_a_get_a' Phi 
    ct Hneq).
  apply (DHSpec.lookup_get Hctx').
  assumption.
Qed.
Implicit Arguments heap_extend_valok
  [Psi Phi addr ct t v].

Lemma heap_update_valok : 
  forall Psi Phi Phi' addr ct ct' t v,
    DHSpec.lookup Phi addr ct
    -> ceTy_sub ct' ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> valok Psi Phi v t
    -> valok Psi Phi' v t.
Proof.
  intros Psi Phi Phi' addr ct ct' t v.
  intros Hct Hsub Hphi' Hvalok.
  inversion_clear Hvalok.
  apply valok_var; trivial.
  apply valok_lab with a; trivial.
  apply valok_int; trivial.
  apply valok_tup.
  intros i Hindom.
  substH H with (H i Hindom).
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsubx]]]].
  destruct (nat_eq_dec addr (v + i)) as [Heq | Hneq].
(* case addr = v + i *)
  subst addr.
  exists ctx. exists ct'.
  split_l.
  split.
  apply (DHSpec.upd_lookup Hphi').
  assert (Hg := DHSpec.upd_a_get_a Hphi').
  substH Hct with (DHSpec.get_lookup
    (m:=Phi) Hct).
  assert (Heq := DHSpec.lookup_unique Hctx' Hct).
  subst ctx'.
  apply (@ceTy_sub_trans ct' ct ctx Hsub Hsubx).
(* case addr <> v + i *)
  assert (Hg := DHSpec.upd_a_get_a' Hneq Hphi').
  exists ctx. exists ctx'.
  split_l.
  split.
  apply DHSpec.get_lookup.
  rewrite Hg.
  apply (DHSpec.lookup_get Hctx').
  assumption.
Qed.
Implicit Arguments heap_update_valok
  [Psi Phi Phi' addr ct ct' t v].

Lemma heap_extend_ceok :
  forall Psi Phi addr ct ct' v,
    DHSpec.get Phi addr = None
    -> ceok Psi Phi v ct
    -> ceok Psi (DHSpec.set Phi addr ct') v ct.
Proof.
  intros Psi Phi addr ct ct' v.
  intros Hnin Hceok.
  inversion_clear Hceok.
  apply ceok_junk; trivial.
  apply ceok_init; trivial.
  destruct ct'.
  eapply heap_extend_valok; eauto.
Qed.
Implicit Arguments heap_extend_ceok
  [Psi Phi addr ct ct' v].
  
Lemma heap_update_ceok :
  forall Psi Phi Phi' addr ct ct' ct'' v,
    DHSpec.lookup Phi addr ct
    -> ceTy_sub ct' ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> ceok Psi Phi v ct''
    -> ceok Psi Phi' v ct''.
Proof.
  intros Psi Phi Phi' addr ct ct' ct0 v.
  intros Hct Hsub Hphi' Hceok.
  inversion_clear Hceok.
  apply ceok_junk; trivial.
  apply ceok_init; trivial.
  eapply heap_update_valok; eauto.
Qed.
Implicit Arguments heap_update_ceok
  [Psi Phi Phi' addr ct ct' ct'' v].
  
Lemma heap_extend_wfdh :
  forall Psi Phi H ct addr w,
    wfdh Psi Phi Phi H
    -> DHSpec.get Phi addr = None
    -> ceok Psi (DHSpec.set Phi addr ct) w ct
    -> wfdh Psi (DHSpec.set Phi addr ct) (DHSpec.set Phi addr ct)
            (Heap.set H addr w).
Proof.
  intros Psi Phi H ct addr w.
  intros HH Hnin Hw.
  unfold wfdh in * .
  intros l ct0 Hct0.
  destruct (nat_eq_dec addr l) as [Heq | Hneq].
(* case l = addr *)
  subst addr.
  exists w.
  substH Hct0 with (DHSpec.lookup_get Hct0).
  rewrite (DHSpec.set_a_get_a Phi l ct) in Hct0.
  inj_hyp Hct0.
  subst ct0.
  split_r.
  apply Heap.get_lookup.
  apply Heap.set_a_get_a.
(* case l <> addr *)
  substH Hct0 with (DHSpec.lookup_get Hct0).
  rewrite (DHSpec.set_a_get_a' Phi ct Hneq) in Hct0.
  substH Hct0 with (DHSpec.get_lookup Hct0).
  destruct (HH l ct0 Hct0) as [v [Hv Hceok]].
  exists v.
  split.
  apply Heap.get_lookup.
  rewrite <- (Heap.lookup_get Hv).
  apply (Heap.set_a_get_a' H w Hneq).
  apply (heap_extend_ceok (ct':=ct) Hnin Hceok).
Qed.
Implicit Arguments heap_extend_wfdh
    [Psi Phi H ct addr w].

Lemma heap_update_wfdh :
  forall Psi Phi Phi' H ct ct' addr w,
    wfdh Psi Phi Phi H
    -> ceTy_sub ct' ct
    -> DHSpec.lookup Phi addr ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> ceok Psi Phi' w ct'
    -> wfdh Psi Phi' Phi' (Heap.set H addr w).
Proof.
  intros Psi Phi Phi' H ct ct' addr w.
  intros HH Hsub Hct Hphi' Hw.
  unfold wfdh in * .
  intros l ct0 Hct0.
  destruct (nat_eq_dec addr l) as [Heq | Hneq].
(* case l = addr *)
  subst l.
  assert (Hg := DHSpec.get_lookup (m := Phi) Hct).
  substH Hphi' with (DHSpec.upd_lookup Hphi').
  assert (Hcteq := DHSpec.lookup_unique Hct0 Hphi').
  subst ct0.
  exists w.
  split_r.
  apply Heap.get_lookup.
  apply Heap.set_a_get_a.
(* case l <> addr *)
  assert (H0 := DHSpec.upd_a_get_a' Hneq Hphi').
  substH Hct0 with (DHSpec.lookup_get (m := Phi') Hct0).
  rewrite Hct0 in H0; clear Hct0.
  substH HH with (HH l ct0 
    (DHSpec.get_lookup (m:=Phi) (sym_eq H0))).
  destruct HH as [v [Hv Hceok]].
  exists v.
  split.
  apply Heap.get_lookup.
  rewrite <- (Heap.lookup_get Hv).
  apply (Heap.set_a_get_a' H w Hneq).
  apply (heap_update_ceok Hct Hsub Hphi' Hceok).
Qed.
Implicit Arguments heap_update_wfdh
    [Psi Phi Phi' H ct ct' addr w].

Lemma heap_extend_wfrf :
  forall Psi Phi Gamma R ct addr,
    wfrf Psi Phi Gamma R
    -> DHSpec.get Phi addr = None
    -> wfrf Psi (DHSpec.set Phi addr ct) Gamma R.
Proof.
  intros Psi Phi Gamma R ct addr.
  intros HR Hnin.
  unfold wfrf in * .
  intros r t Hr.
  apply heap_extend_valok; trivial.
  apply HR; trivial.
Qed.
Implicit Arguments heap_extend_wfrf 
    [Psi Phi Gamma R ct addr].

Lemma heap_update_wfrf :
  forall Psi Phi Phi' Gamma R ct ct' addr,
    wfrf Psi Phi Gamma R
    -> DHSpec.lookup Phi addr ct
    -> DHSpec.canupd Phi addr ct' Phi'
    -> ceTy_sub ct' ct
    -> wfrf Psi Phi' Gamma R.
Proof.
  intros Psi Phi Phi' Gamma R ct ct' addr.
  intros HR Hct Hphi' Hsub.
  unfold wfrf in * .
  intros r t Hr.
  apply heap_update_valok with (Phi:=Phi) (addr:=addr) 
    (ct := ct) (ct' := ct'); trivial.
  apply HR; trivial.
Qed.
Implicit Arguments heap_update_wfrf 
    [Psi Phi Phi' Gamma R ct ct' addr].

(* Lemma heap_update_valok_strong :  *)
Lemma heap_update_tuple_valok :
  forall Psi Phi Phi' addr offset o ct t,
    valok Psi Phi addr (valTy_tup o)
    -> TupTy.lookup o offset ct
    -> ceTy_sub (init, t) ct
    -> DHSpec.canupd Phi (addr + offset) (init, t) Phi'
    -> valok Psi Phi' addr 
           (valTy_tup (TupTy.upd o offset (init, t))).
Proof.
  intros Psi Phi Phi' addr offset tup ct t.
  intros Hvalok Htup Hsub Hphi'.
  inversion_clear Hvalok.
  apply valok_tup.
  intros i Hindom. 
  assert (Hindom' : TupTy.indom tup i).
  eapply TupTy.indom_upd_indom; eauto.
  substH H with (H i Hindom').
  destruct H as [ctx [ctx' [Hctx [Hctx' Hsubx]]]].
  destruct (nat_eq_dec i offset) as [Heq | Hneq].
  subst i.
  exists (init, t). exists (init, t).
  split_l by (apply (TupTy.lookup_upd _ _ _ (init, t) Hctx)).
  split_l by (apply (DHSpec.canupd_lookup Hphi')).
  apply ceTy_sub_refl.
  exists ctx. exists ctx'.
  split_l by 
    (apply (TupTy.lookup_upd_neq _ _ _ _ (init, t) Hneq Hctx)).
  split_r.
  assert (Hneq' : addr + i <> addr + offset).
  intro Hf; apply Hneq.
  eapply Arith.Plus.plus_reg_l; eauto.
  apply (DHSpec.canupd_lookup_neq Hneq' Hphi' Hctx').
Qed.
Implicit Arguments heap_update_tuple_valok 
   [Psi Phi Phi' addr offset o ct t].

Lemma heap_update_wfrf_reg :
  forall Psi Phi Phi' Delta Gamma R o ct ct' t r offset,
    beq_reg r zero = false
    -> wfrf Psi Phi Gamma R
    -> regok Delta Gamma r (valTy_tup o)
    -> TupTy.lookup o offset ct
    -> DHSpec.lookup Phi 
         (gR R r + offset) ct'
    -> DHSpec.canupd Phi 
         (gR R r + offset) (init, t) Phi'
    -> ceTy_sub ct' ct
    -> ceTy_sub (init, t) ct'
    -> wfrf Psi Phi' (RfTy.upd Gamma r 
         (valTy_tup (TupTy.upd o offset (init, t)))) R.
Proof.
  intros Psi Phi Phi' Delta Gamma R o ct ct' t r offset.
  intros Hnr0 HR Hr Hct Hctin Hphi' Hsub Hsub'.
  unfold wfrf in * .
  intros r0 t0 Hr0.
  decbeqreg r0 r.
(* case : r = r0 *)
  rewrite (RfTy.sel_upd_req Gamma r r 
    (valTy_tup (TupTy.upd o offset (init, t))))
    in Hr0; [idtac | solvebreg].
  inj_hyp Hr0.
  subst t0.
  assert (H : valok Psi Phi (gR R r) 
      (valTy_tup o)).
  inversion_clear Hr.
  apply (HR r (valTy_tup o) H).
  assert (Hsub'' : ceTy_sub (init, t) ct).
  eapply ceTy_sub_trans; eauto.
  apply (heap_update_tuple_valok H Hct Hsub'' Hphi').
(* case : r <> r0 *)
  rewrite (RfTy.sel_upd_rneq Gamma r0 r 
    (valTy_tup (TupTy.upd o offset (init, t))) Hb)
    in Hr0.
  apply (heap_update_valok Hctin Hsub' Hphi' 
    (HR r0 t0 Hr0)).
Qed.
Implicit Arguments heap_update_wfrf_reg
    [Psi Phi Phi' Gamma R o ct ct' t r offset].

Open Local Scope tm_scope.

Lemma regok_t_weakening_rfTy :
  forall e0 Y0 e Y r t tl,
    rfTy_sub e0 Y0 (instYL (shiftY Y e0 e) tl e 0)
    -> valTyLok e0 tl
    -> regok e Y r t
    -> regok e0 Y0 r (instL (shift t e0 e) tl e 0).
Proof.
  intros e0 Y0 e Y r t tl Hsub Htl Hr.
  inversion_clear Hr; simplenv.
  constructor.
  apply rfTy_sub_sel with e0 (instYL (shiftY Y e0 e) tl e 0);
    trivial.
  apply RfTy_sel_some_instYL.
  apply RfTy_sel_some_shiftY; trivial.
  apply valTyok_instL; trivial.
  apply valTyok_shift_2; trivial.
Qed.

Lemma regok_t_codety_weakening :
  forall e Y e0 Y0 r t,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> regok e Y r t
    -> exists t0, regok e0 Y0 r t0.
Proof.
  intros e Y e0 Y0 r t Hsub Hr.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub]]]].
  exists (instL (shift t e0 e) tl e 0).
  simplenv.
  eapply regok_t_weakening_rfTy; eauto.
Qed.

Lemma regok_int_codety_weakening :
  forall e Y e0 Y0 r,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> regok e Y r valTy_int
    -> regok e0 Y0 r valTy_int.
Proof.
  intros e Y e0 Y0 r Hsub Hr.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub]]]].
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r valTy_int tl Hsub Htl Hr).
  simpl in Hx.
  trivial.
Qed.

Lemma regok_code_weakening_rfTy :
  forall e0 Y0 e Y r e'' Y'' tl,
    rfTy_sub e0 Y0 (instYL (shiftY Y e0 e) tl e 0)
    -> valTyLok e0 tl
    -> regok e Y r (valTy_code e'' Y'')
    -> regok e0 Y0 r (valTy_code e'' (instYL (shiftY Y'' e0 (app e e'')) (shiftL tl e'' emp) e (len e''))).
Proof.
  intros e0 Y0 e Y r e'' Y'' tl Hsub Htl Hr.
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r (valTy_code e'' Y'') tl Hsub Htl Hr).
  simpl in Hx.
  trivial.
Qed.

Lemma regok_tup_weakening_rfTy :
  forall e0 Y0 e Y r o tl,
    rfTy_sub e0 Y0 (instYL (shiftY Y e0 e) tl e 0)
    -> valTyLok e0 tl
    -> regok e Y r (valTy_tup o)
    -> regok e0 Y0 r (valTy_tup 
         (instoL (shifto o e0 e) tl e 0)).
Proof.
  intros e0 Y0 e Y r o tl Hsub Htl Hr.
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r (valTy_tup o) tl Hsub Htl Hr).
  simpl in Hx.
  trivial.
Qed.

Lemma codetysub_codetysub_valTyok :
  forall e Y e0 Y0 r t,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> valTyok emp t
    -> codeTy_sub emp (e0, RfTy.upd Y0 r t) (e, RfTy.upd Y r t).
Proof.
  intros e Y e0 Y0 r t Hsub Ht.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub]]]].
  inversion_clear Ha0.
  inversion_clear Ha.
  simplenv.
  exists tl.
  split.
    constructor; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    apply valTyok_weaken_emp; trivial.
  split.
    constructor; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    apply valTyok_weaken_emp; trivial.
  split.
    simplenv; trivial.
  simplenv.
  rewrite <- RfTy_upd_shiftY; simpl.
  rewrite <- RfTy_upd_instYL; simpl.
  assert (Hx : valTyok e t).
    apply valTyok_weaken_emp; trivial.
  rewrite valTyok_shift_unchange_emp; trivial.
  rewrite valTyok_instL_unchange_emp; trivial.
  apply rfTy_sub_upd; trivial.
  apply valTyok_weaken_emp; trivial.
Qed.

Lemma codetysub_codetysub_regok_regok :
  forall e Y e0 Y0 r r' t t0,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> regok e Y r t
    -> regok e0 Y0 r t0
    -> codeTy_sub emp (e0, RfTy.upd Y0 r' t0) 
    (e, RfTy.upd Y r' t).
Proof.
  intros e Y e0 Y0 r r' t t0.
  intros Hsub Hr H0r.
  destruct (regok_t_codety_weakening Hsub Hr) as [t' H'r].
  inversion_clear H0r; inversion_clear H'r.
  rewrite H1 in H; inj_hyp H; subst t'.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub]]]].
  exists tl; simplenv.
  split.
    constructor; simplenv.
    inversion_clear Ha0; simplenv.
    apply RfTy_upd_rfTyok; trivial.
  split.
    constructor; simplenv.
    inversion_clear Ha; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    inversion_clear Hr; trivial.
  split_l.
  assert (He : t0 = (instL (shift t e0 e) tl e 0)).
    assert (H := regok_t_weakening_rfTy Hsub Htl Hr).
    inversion_clear H.
    rewrite H1 in H3; inj_hyp H3; subst t0; trivial.
  rewrite He.
  rewrite <- RfTy_upd_shiftY; simpl.
  rewrite <- RfTy_upd_instYL; simpl.
  apply rfTy_sub_upd; trivial.
  apply valTyok_instL; trivial.
  apply valTyok_shift_2; trivial.
  inversion_clear Hr; trivial.
Qed.

Lemma regok_tup_codety_weakening :
  forall e Y e0 Y0 r o,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> regok e Y r (valTy_tup o)
    -> exists o0, regok e0 Y0 r (valTy_tup o0).
Proof.
  intros e Y e0 Y0 r o Hsub Hr.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub]]]].
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r (valTy_tup o) tl Hsub Htl Hr).
  simpl in Hx.
  exists (instoL (shifto o e0 e) tl e 0).
  trivial.
Qed.

Lemma codetysub_codetysub_regok_regok_load_tup :
  forall e Y e0 Y0 r r' o i flag t o0,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> regok e Y r (valTy_tup o)
    -> regok e0 Y0 r (valTy_tup o0)
    -> TupTy.lookup o i (flag, t)
    -> exists t0, codeTy_sub emp (e0, RfTy.upd Y0 r' t0)
      (e, RfTy.upd Y r' t)
      /\ TupTy.lookup o0 i (flag, t0).
Proof.
  intros e Y e0 Y0 r r' o i flag t o0.
  intros Hsub Hr Hr0 Ho.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub]]]].
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r (valTy_tup o) tl Hsub Htl Hr).
  simpl in Hx.
  assert (He : o0 = instoL (shifto o e0 e) tl e 0).
    inversion_clear Hr0.
    inversion_clear Hx.
    rewrite H in H1.
    inj_hyp H1; subst o0; trivial.
  subst o0.
  exists (instL (shift t e0 e) tl e 0).
  split.
  exists tl; simplenv.
  split.
    constructor; simplenv.
    inversion_clear Ha0; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    apply valTyok_instL; trivial.
    apply valTyok_shift_2; trivial.
    inversion_clear Ha; simplenv.
    inversion_clear Hr.
    inversion_clear H2.
    eapply TupTy_lookup_valTyok; eauto.
  split.
    constructor; simplenv.
    inversion_clear Ha; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    inversion_clear Hr.
    inversion_clear H1.
    eapply TupTy_lookup_valTyok; eauto.
  split_l.
  rewrite <- RfTy_upd_shiftY; simpl.
  rewrite <- RfTy_upd_instYL; simpl.
  apply rfTy_sub_upd; trivial.
  apply valTyok_instL; trivial.
  apply valTyok_shift_2; trivial.
  inversion_clear Hr.
  inversion_clear H0.
  eapply TupTy_lookup_valTyok; eauto.

  apply TupTy_lookup_instoL.
  apply TupTy_lookup_shifto; trivial.
Qed.

Lemma codetysub_codetysub_regok_regok_write_tup :
  forall e Y e0 Y0 r r' r'' o i flag t t0 o0,
    codeTy_sub emp (e0, Y0) (e, Y)
    -> regok e Y r t
    -> regok e0 Y0 r t0
    -> regok e Y r' (valTy_tup o)
    -> regok e0 Y0 r' (valTy_tup o0)
    -> TupTy.lookup o i (flag, t)
    -> codeTy_sub emp 
        (e0, RfTy.upd Y0 r'' (valTy_tup (TupTy.upd o0 i (init, t0))))
        (e, RfTy.upd Y r'' (valTy_tup (TupTy.upd o i (init, t))))
      /\ TupTy.lookup o0 i (flag, t0).
Proof.
  intros e Y e0 Y0 r r' r'' o i flag t t0 o0.
  intros Hsub Hr Hr0 Hr' Hr'0 Ho.
  destruct Hsub as [tl [Ha0 [Ha [Htl Hsub0]]]].
  simplenv.
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r t tl Hsub0 Htl Hr).
  assert (He : t0 = instL (shift t e0 e) tl e 0).
    inversion_clear Hr0.
    inversion_clear Hx.
    rewrite H in H1; inj_hyp H1; subst t0; trivial.
  subst t0. clear Hx.
  assert (Hx := @regok_t_weakening_rfTy e0 Y0 e Y
    r' (valTy_tup o) tl Hsub0 Htl Hr').
  simpl in Hx.
  assert (He : o0 = instoL (shifto o e0 e) tl e 0).
    inversion_clear Hr'0.
    inversion_clear Hx.
    rewrite H in H1.
    inj_hyp H1; subst o0; trivial.
  subst o0; clear Hx.
(*   exists (instL (shift t e0 e) tl e 0). *)

  split.

  exists tl; simplenv.
  inversion_clear Ha0; simplenv.
  inversion_clear Ha; simplenv.
  inversion_clear Hr; simplenv.
  inversion_clear Hr'; simplenv.
  inversion_clear H4.
  split.
    constructor; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    constructor.
    apply TupTy_upd_tupTyok; trivial.
    apply tupTyok_instoL; trivial.
    apply tupTyok_shifto_2; trivial.
    apply valTyok_instL; trivial.
    apply valTyok_shift_2; trivial.
  split.
    constructor; simplenv.
    apply RfTy_upd_rfTyok; trivial.
    constructor.
    eapply TupTy_upd_tupTyok; eauto.
  split_l.
  rewrite <- RfTy_upd_shiftY; simpl.
  rewrite <- RfTy_upd_instYL; simpl.
  rewrite <- TupTy_upd_shifto; simpl.
  rewrite <- TupTy_upd_instoL; simpl.
  apply rfTy_sub_upd; trivial.
  constructor.
  eapply TupTy_upd_tupTyok; eauto.
  apply tupTyok_instoL; trivial.
  apply tupTyok_shifto_2; trivial.
  apply valTyok_instL; trivial.
  apply valTyok_shift_2; trivial.

  apply TupTy_lookup_instoL.
  apply TupTy_lookup_shifto.
  trivial.
Qed.

Lemma wfis_weakening :
  forall Psi  I (a a0 : codeTy),
    wfpsi Psi
    -> codeTy_sub emp a0 a
    -> wfis Psi a I
    -> wfis Psi a0 I.
Proof.
  intros Psi I.
  induction I.
  intros [e Y] [e0 Y0] HPsi Hsub0 HaI.
  simplenv.
  destruct i as
    [ rd rs rt | rd rs w
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f | r]; 
    try (simpl in HaI; 
      contradiction).
  (* case j *)
  destruct HaI as [[e'' Y''] [Hf Hsub]].
  exists (e'', Y'').
  split_l.
  eapply codeTy_sub_trans; eauto.
  (* case jal f fret *)
  destruct HaI as [[e' Y'] [Hspc [Ha'ok HaI]]].
  destruct HaI as [[e'' Y''] [Hf Hsub]].
  exists (e', Y'); split_l.
  split_l.
  exists (e'', Y''); split_l.
  apply codeTy_sub_trans with
    (e, RfTy.upd Y ra (valTy_code e' Y')); trivial.
  assert (He'Y' := HPsi fret (e',Y') Hspc).
  destruct Hsub0 as [tl [Hok [Hok' [Htl Hsub0]]]].
  exists tl.
  split.
    inversion_clear Hok.
    constructor.
    simplenv.
    apply RfTy_upd_rfTyok; trivial.
    apply valTyok_weaken_emp; trivial.
  split.
    inversion_clear Hok'.
    constructor.
    simplenv.
    apply RfTy_upd_rfTyok; trivial.
    apply valTyok_weaken_emp; trivial.
  split_l.
  simplenv.
  rewrite <- RfTy_upd_shiftY.
  rewrite <- RfTy_upd_instYL.
  rewrite valTyok_shift_unchange_emp; trivial.
  rewrite valTyok_instL_unchange_emp; trivial.
  apply rfTy_sub_upd; trivial.
  apply valTyok_weaken_emp; trivial.
  apply valTyok_weaken_emp; trivial.
  (* case jr *)
  simpl in * .
  destruct HaI as [[e'' Y''] [Hr Hsub]].
  destruct Hsub0 as [tl [Ha0 [Ha [Htl Hsub0]]]].
  simplenv.
  rewrite shiftY_emp in Hsub.
  destruct Hsub as [tl' [HY [Ha'' [Htl' Hsub']]]].
  set (Yx := instYL (shiftY Y'' e0 (app e e''))
    (shiftL tl e'' emp) e (len e'')).
  exists (e'', Yx).
  split.
  subst Yx.
  apply regok_code_weakening_rfTy with Y; trivial.
  set (tlx := (LinstL (shiftL tl' e0 e) tl e 0)).
  constructor 1 with tlx.
  split_l.
    inversion_clear Ha0; constructor; 
      simplenv; trivial.
  split_l.
    constructor.
    subst Yx.
    apply rfTyok_instYL_aux; trivial.
    assert (Hx := @rfTyok_shiftY_aux Y'' 
      (app e e'') emp e0); simplenv.
    inversion_clear Hr.
    inversion_clear H0.
    substH Hx with (Hx H1).
    trivial.
  split_l.
    simplenv.
    subst tlx.
    apply valTyLok_LinstL; trivial.
    apply valTyLok_shiftL_2; trivial.
  simplenv.
  rewrite shiftY_emp.
  assert (Hx := @rfTy_sub_shiftY emp Y 
    (instYL Y'' tl' e'' 0) e e0).
  simplenv.
  substH Hx with (Hx Hsub').
  substH Hx with (@rfTy_sub_instYL e0 e 
    (shiftY Y e0 e) 
    (shiftY (instYL Y'' tl' e'' 0) e0 e)
    tl Htl Hx).
  substH Hx with (rfTy_sub_trans Hsub0 Hx).
  subst Yx.
  rewrite <- instYL_shiftY_2_emp in Hx.
  rewrite instYL_dist in Hx.
  trivial.
  (* case iseq *)
  intros [e Y] [e0 Y0] HPsi Hsub0 HaI.
  destruct i as
    [ rd rs rt | rd rs w
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f| r]; 
    simpl in HaI; 
    try contradiction.
  (* case addu *)
  destruct HaI as [Hb [Hrs [Hrt His]]].
  unfold wfis. fold wfis.
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply regok_int_codety_weakening; eauto).
  apply (IHI (e, RfTy.upd Y rd valTy_int)
    (e0, RfTy.upd Y0 rd valTy_int)); trivial.
  apply codetysub_codetysub_regok_regok with rs; trivial.
  eapply regok_int_codety_weakening; eauto.

  (* case iseq (addiu rd rs w) is *)
  destruct HaI as [Hb [Hrs His]].
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  apply (IHI (e, RfTy.upd Y rd valTy_int)
    (e0, RfTy.upd Y0 rd valTy_int)); trivial.
  apply codetysub_codetysub_regok_regok with rs; trivial.
  eapply regok_int_codety_weakening; eauto.

  (* case iseq (subu rd rs rt) is *)
  destruct HaI as [Hb [Hrs [Hrt His]]].
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply regok_int_codety_weakening; eauto).
  apply (IHI (e, RfTy.upd Y rd valTy_int)
    (e0, RfTy.upd Y0 rd valTy_int)); trivial.
  apply codetysub_codetysub_regok_regok with rs; trivial.
  eapply regok_int_codety_weakening; eauto.

  (* case subiu *)
  destruct HaI as [Hb [Hrs His]].
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  apply (IHI (e, RfTy.upd Y rd valTy_int)
    (e0, RfTy.upd Y0 rd valTy_int)); trivial.
  apply codetysub_codetysub_regok_regok with rs; trivial.
  eapply regok_int_codety_weakening; eauto.

  (* case i = iseq (move rd rs) is*)
  destruct HaI as [t [Hb [Hrs His]]].
  destruct (regok_t_codety_weakening Hsub0 Hrs) as [t0 Hrs0].
  exists t0.
  split_l.
  split_l.
  apply (IHI (e, RfTy.upd Y rd t)
    (e0, RfTy.upd Y0 rd t0)); trivial.
  apply codetysub_codetysub_regok_regok with rs; trivial.

  (* case i = iseq (li rd w) is *)
  destruct HaI as [Hrd [t [Hw His]]].
  split_l.
  exists t; split_l.
  apply (IHI (e, RfTy.upd Y rd t)
    (e0, RfTy.upd Y0 rd t)); trivial.
  apply codetysub_codetysub_valTyok; trivial.
  destruct Hw as [Psi f e'' Y'' Hf | ].
    apply (HPsi f (e'', Y'') Hf).
  constructor.
  (* case i = lw *)
  destruct HaI as [o [t [Hb [Hrs [Ho His]]]]].
  destruct (regok_tup_codety_weakening Hsub0 Hrs) as [o0 Hrs0].
  destruct (codetysub_codetysub_regok_regok_load_tup 
    rt Hsub0 Hrs Hrs0 Ho) as [t0 [Hsub Ho0]].
  exists o0.
  exists t0.
  split_l.
  split_l.
  split_l.
  apply (IHI (e, RfTy.upd Y rt t)
    (e0, RfTy.upd Y0 rt t0)); trivial.

  (* case i = sw *)
  destruct HaI as [o [t [flag [Hb [Hrt [Hrs [Ho [Hce Ha']]]]]]]].
  destruct (regok_t_codety_weakening Hsub0 Hrt) as [t0 Hrt0].
  destruct (regok_tup_codety_weakening Hsub0 Hrs) as [o0 Hrs0].
  destruct (codetysub_codetysub_regok_regok_write_tup
    rs Hsub0 Hrt Hrt0 Hrs Hrs0 Ho) as [Hsub Ho0].
  exists o0; exists t0; exists flag.
  split_l.
  split_l.
  split_l.
  split_l.
  split_l by (inversion_clear Hce; constructor).
  apply (IHI (e, RfTy.upd Y rs (valTy_tup (TupTy.upd o i (init, t))))
    (e0, RfTy.upd Y0 rs (valTy_tup (TupTy.upd o0 i (init, t0))))); trivial.

  (* case i = blt *)
  destruct HaI as [[e'' Y''] [Hf [Hrs [Hrt [Hsub Ha']]]]].
  exists (e'', Y'').
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply codeTy_sub_trans; eauto).
  apply IHI with (e, Y); trivial.

  (* case i = beq *)
  destruct HaI as [[e'' Y''] [Hf [Hrs [Hrt [Hsub Ha']]]]].
  exists (e'', Y'').
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply codeTy_sub_trans; eauto).
  apply IHI with (e, Y); trivial.

  (* case i = bgtz rs f *)
  destruct HaI as [[e'' Y''] [Hf [Hrs [Hsub Ha']]]].
  exists (e'', Y'').
  split_l.
  split_l by (eapply regok_int_codety_weakening; eauto).
  split_l by (eapply codeTy_sub_trans; eauto).
  apply IHI with (e, Y); trivial.
Qed.

(***************************************************************)

Lemma wfst_codeval_exists :
  forall Psi e Y e'' Y'' H R pc r, 
    wfst Psi (e, Y) (mkst H R pc)
    -> regok e Y r (valTy_code e'' Y'')
    -> exists a,
      CHSpec.lookup Psi (gR R r) a.
Proof.
  intros Psi e Y e'' Y'' H R pc r.
  intros HS Hr.
  generalize HS; intro HS'.
  destruct HS' as [Phi [HH [Y' [Hsub' HR]]]].
  destruct Hsub' as [tl [Ha0 [Ha'' [Htlok Hsub'']]]]; simplenv.
  rewrite shiftY_emp in Hsub''.
  inversion_clear Hr as [ex Yx rx tx Hsel Htok].
  assert (Hsel' := RfTy_sel_some_instYL tl e 0 Hsel).
  simpl in Hsel'.
  substH Hsel' with (rfTy_sub_sel Hsub'' Hsel').
  assert (Hrok := HR r (valTy_code e'' 
    (instYL Y'' (shiftL tl e'' emp) e (len e''))) Hsel').
  inversion_clear Hrok as 
    [px hx fx ex Yx Hf | | | ].
  exists a; trivial.
Qed.


Lemma wfst_codety_reg_wfst :
  forall Psi e Y e'' Y'' H R pc r a, 
    wfst Psi (e, Y) (mkst H R pc)
    -> regok e Y r (valTy_code e'' Y'')
    -> CHSpec.lookup Psi (gR R r) a
    -> codeTy_sub e (emp, Y) (e'', Y'')
    -> wfst Psi a (mkst H R (gR R r)).
Proof.
  intros Psi e Y e'' Y'' H R pc r a.
  intros HS Hr Ha Hsub.
  generalize HS; intro HS'.
  destruct HS' as [Phi [HH [Y' [Hsub' HR]]]].
  destruct Hsub' as [tl [Ha0 [Ha'' [Htlok Hsub'']]]]; simplenv.
  rewrite shiftY_emp in Hsub''.
  
  inversion_clear Hr as [ex Yx rx tx Hsel Htok].
  assert (Hsel' := RfTy_sel_some_instYL tl e 0 Hsel).
  simpl in Hsel'.
  substH Hsel' with (rfTy_sub_sel Hsub'' Hsel').
  assert (Hrok := HR r (valTy_code e'' 
    (instYL Y'' (shiftL tl e'' emp) e (len e''))) Hsel').
  inversion_clear Hrok as 
    [px hx fx ex Yx Hf | | | ].

  unfold wfst.
  exists Phi.
  split_l.
  exists (instYL Y tl e 0).
  assert (He := CHSpec.lookup_unique Ha H0).
  subst a0.
  split.
    destruct a.
    assert (Hsub2 : codeTy_sub emp (emp, instYL Y tl e 0) 
      (e'', instYL Y'' (shiftL tl e'' emp) e (len e''))).
      apply codeTy_sub_instYL_2; trivial.
    eapply codeTy_sub_trans; eauto.
  eapply regfile_weakening; eauto.
Qed.
(* 
Lemma wfst_codeval_subty :
  forall Psi e Y e'' Y'' H R pc r,
    wfst Psi (e, Y) (mkst H R pc)
    -> regok e Y r (valTy_code e'' Y'')
    -> exists a,
      CHSpec.lookup Psi (gR R r) a
      /\ codeTy_sub emp (app e e'', Y'') a.
Proof.
  intros Psi e Y e'' Y'' H R pc r.
  intros HS Hr.
  generalize HS; intro HS'.
  destruct HS' as [Phi [HH [Y' [Hsub' HR]]]].
  destruct Hsub' as [tl [Ha0 [Ha'' [Htlok Hsub'']]]]; simplenv.
  rewrite shiftY_emp in Hsub''.
  inversion_clear Hr as [ex Yx rx tx Hsel Htok].
  assert (Hsel' := RfTy_sel_some_instYL tl e 0 Hsel).
  simpl in Hsel'.
  substH Hsel' with (rfTy_sub_sel Hsub'' Hsel').
  assert (Hrok := HR r (valTy_code e'' 
    (instYL Y'' (shiftL tl e'' emp) e (len e''))) Hsel').
  inversion_clear Hrok as 
    [px hx fx ex Yx Hf | | | ].
  exists a.
  split_l.
  unfold codeTy_sub.
  destruct a as [e0 Y0].
  simplenv.
Qed.
*)
Lemma wfst_heapval_exists :
  forall Psi e Y H R pc o flag t rs offset,
    wfst Psi (e, Y) (mkst H R pc)
    -> regok e Y rs (valTy_tup o)
    -> TupTy.lookup o offset (flag, t)
    -> exists v, 
      Heap.lookup H (gR R rs +offset) v.
Proof.
  intros Psi e Y H R pc o flag t rs offset.
  intros HS Hrs Ho.
  destruct HS as [Phi [HH [Y' [Hsub' HR]]]].
  unfold wfdh in HH.
  destruct Hsub' as [tl [Ha' [Ha [Htl Hsub'']]]]; simplenv.
  assert (Hx := regok_tup_weakening_rfTy Hsub'' Htl Hrs).
  assert (Hx0 := reg_typing HR Hx).
  rewrite shifto_emp in Hx0.
  inversion_clear Hx0.
  assert (Hin : TupTy.indom (instoL o tl e 0) offset).
    unfold TupTy.indom.
    exists (flag, (instL t tl e 0)); trivial.
    apply TupTy_lookup_instoL; trivial.
  destruct (H0 offset Hin) as [ct [ct' [Hct [ Hct' Hsub]]]].
  destruct (HH (gR R rs +offset) ct' Hct') as
    [v [Hin' Hceok]].
  exists v; trivial.
Qed.

Lemma wfst_RfTy_upd_wfst : 
  forall Psi e Y H R pc pc' rd t w,
  beq_reg rd r0 = false
  -> wfst Psi (e, Y) (mkst H R pc)
  -> valTyok e t
  -> valok2 Psi w t 
  -> wfst Psi (e, RfTy.upd Y rd t) (mkst H (sR R rd w) pc').
Proof.
  intros Psi e Y H R pc pc' rd t w.
  intros Hb HS Ht Hw.
  destruct HS as [Phi [HH [Y' [Hsub' HR]]]].
  exists Phi.
  split_l.
  generalize Hsub'; intro Hsub.
  destruct Hsub' as [tl [Ha' [Ha [Htl Hsub'']]]]; simplenv.
  inversion_clear Ha'.
  inversion_clear Ha.
  exists (RfTy.upd Y' rd (instL t tl e 0)).
  split. 
    rewrite shiftY_emp in Hsub''.
    exists tl.
    simplenv.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
      apply valTyok_instL; simplenv; trivial.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
    split_l.
    rewrite <- RfTy_upd_shiftY.
    rewrite <- RfTy_upd_instYL.
    rewrite shiftY_emp.
    rewrite shift_emp.
    apply rfTy_sub_upd; trivial.
    apply valTyok_instL; simplenv; trivial.
  eapply regfile_update; eauto.
  destruct t as [n | | | e0 Y0 ];
    inversion_clear Hw; simplenv; simpl.
  constructor.
  rewrite rfTyok_instYL_unchange; trivial.
  constructor 2 with (e0, Y0); trivial.
  apply codeTy_sub_refl; trivial.
  inversion_clear H3.
  simplenv; trivial.
Qed.

Lemma wfst_RfTy_upd_int_wfst : 
  forall Psi e Y H R pc pc' rd w,
  beq_reg rd r0 = false
  -> wfst Psi (e, Y) (mkst H R pc)
  -> wfst Psi (e, RfTy.upd Y rd valTy_int) 
     (mkst H (sR R rd w) pc').
Proof.
  intros Psi e Y H R pc pc' rd w.
  intros Hb HS.
  eapply wfst_RfTy_upd_wfst with pc; eauto.
  constructor.
  constructor.
Qed.

Lemma wfst_RfTy_upd_reg_wfst : 
  forall Psi e Y H R pc pc' rd t rs,
  beq_reg rd r0 = false
  -> wfst Psi (e, Y) (mkst H R pc)
  -> regok e Y rs t 
  -> wfst Psi (e, RfTy.upd Y rd t) 
       (mkst H (sR R rd (gR R rs)) pc').
Proof.
  intros Psi e Y H R pc pc' rd t rs.
  intros Hb HS Hrs.
  destruct HS as [Phi [HH [Y' [Hsub' HR]]]].
  exists Phi.
  split_l.
  generalize Hsub'; intro Hsub.
  destruct Hsub' as [tl [Ha' [Ha [Htl Hsub'']]]]; simplenv.
  inversion_clear Ha'.
  inversion_clear Ha.
  exists (RfTy.upd Y' rd (instL t tl e 0)).
  split. 
    rewrite shiftY_emp in Hsub''.
    inversion_clear Hrs.
    exists tl.
    simplenv.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
      apply valTyok_instL; simplenv; trivial.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
    split_l.
    rewrite <- RfTy_upd_shiftY.
    rewrite <- RfTy_upd_instYL.
    rewrite shiftY_emp.
    rewrite shift_emp.
    apply rfTy_sub_upd; trivial.
    apply valTyok_instL; simplenv; trivial.
  assert (Hx := regok_t_weakening_rfTy Hsub'' Htl Hrs).
  rewrite shift_emp in Hx.
  eapply regfile_update_reg; eauto.
Qed.

Lemma wfst_RfTy_upd_reg_load_tup_wfst : 
  forall Psi e Y H R pc pc' i rd t rs o v,
    beq_reg rd r0 = false
    -> wfst Psi (e, Y) (mkst H R pc)
    -> regok e Y rs (valTy_tup o)
    -> TupTy.lookup o i (init, t)
    -> Heap.lookup H ((gR R rs) + i) v
    -> wfst Psi (e, RfTy.upd Y rd t) (mkst H (sR R rd v) pc').
Proof.
  intros Psi e Y H R pc pc' i rd t rs o v.
  intros Hb HS Hrs Ho Hv.
  destruct HS as [Phi [HH [Y' [Hsub' HR]]]].
  exists Phi.
  split_l.
  generalize Hsub'; intro Hsub.
  destruct Hsub' as [tl [Ha' [Ha [Htl Hsub'']]]]; simplenv.
  inversion_clear Ha'.
  inversion_clear Ha.
  exists (RfTy.upd Y' rd (instL t tl e 0)).
  split. 
    rewrite shiftY_emp in Hsub''.
    inversion_clear Hrs.
    exists tl.
    simplenv.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
      apply valTyok_instL; simplenv; trivial.
      inversion_clear H3.
      eapply TupTy_lookup_valTyok; eauto.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
      inversion_clear H3.
      eapply TupTy_lookup_valTyok; eauto.
    split_l.
    rewrite <- RfTy_upd_shiftY.
    rewrite <- RfTy_upd_instYL.
    rewrite shiftY_emp.
    rewrite shift_emp.
    apply rfTy_sub_upd; trivial.
    apply valTyok_instL; simplenv; trivial.
    inversion_clear H3.
    eapply TupTy_lookup_valTyok; eauto.
  assert (Hx := regok_t_weakening_rfTy Hsub'' Htl Hrs).
  rewrite shift_emp in Hx.
  eapply regfile_update; eauto.
  apply (@heapval_typing Psi Phi H (instoL o tl e 0) 
    (instL t tl e 0) (gR R rs) i v HH); trivial.
    apply TupTy_lookup_instoL; trivial.
  apply reg_typing with Y'; trivial.
Qed.

Lemma wfst_RfTy_upd_reg_write_tup_wfst : 
  forall Psi e Y H R pc pc' i flag t rs o rt,
    beq_reg rs r0 = false
    -> wfst Psi (e, Y) (mkst H R pc)
    -> regok e Y rs (valTy_tup o)
    -> regok e Y rt t
    -> TupTy.lookup o i (flag, t)
    -> wfst Psi (e, RfTy.upd Y rs (valTy_tup (TupTy.upd o i (init, t)))) 
        (mkst (sH H (gR R rs + i) (gR R rt)) R pc').
Proof.
  intros Psi e Y H R pc pc' i flag t rs o rt.
  intros Hb HS Hrs Hrt Ho.
  destruct HS as [Phi [HH [Y' [Hsub' HR]]]].
  generalize Hsub'; intro Hsub.
  destruct Hsub' as [tl [Ha' [Ha [Htl Hsub'']]]]; simplenv.
  inversion_clear Ha'.
  inversion_clear Ha.
  assert (Hrs' := regok_tup_weakening_rfTy Hsub'' Htl Hrs).
  assert (Ho' : TupTy.lookup 
    (instoL (shifto o emp e) tl e 0) i 
    (flag, instL (shift t emp e) tl e 0)).
    apply TupTy_lookup_instoL; trivial.
    apply TupTy_lookup_shifto; trivial.
  destruct (heap_typing_reg_tying_ex_t emp
    HH HR Hrs' Ho') as [ct' [Hct'in Hsub']].
  set (addr := (gR R rs + i)) in * .
  set (Phi' := DHSpec.set Phi addr (init, (instL t tl e 0))).
  simplenv.
  rewrite shiftY_emp in Hsub''.
  rewrite shifto_emp in Ho'.
  rewrite shifto_emp in Hrs'.
  rewrite shift_emp in Ho'.
  rewrite shift_emp in Hsub'.
  exists Phi'.
  assert (Hx := DHSpec.lookup_canupd (m:=Phi) (a:=addr) 
    (b:= ct') (init, (instL t tl e 0)) Hct'in).
  subst Phi'.
  set (Phi' := DHSpec.set Phi addr (init, (instL t tl e 0))) in * .
  assert (Hsubx : ceTy_sub (init, instL t tl e 0) ct').
    inversion_clear Hsub'.
    destruct flag; constructor.
    constructor.
  split.
    refine (heap_update_wfdh HH Hsubx Hct'in Hx _).
    apply ceok_init.
    eapply reg_typing; eauto.
    eapply heap_update_wfrf; eauto.
    rewrite <- (shift_emp t e).
    apply regok_t_weakening_rfTy with Y; trivial.
    rewrite shiftY_emp; trivial.
  exists (RfTy.upd Y' rs (valTy_tup (TupTy.upd 
    (instoL o tl e 0) i (init, (instL t tl e 0))))).
  split. 
    exists tl.
    inversion_clear Hrs.
    inversion_clear H3.
    simplenv.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
      constructor.
      apply TupTy_upd_tupTyok; trivial.
      apply tupTyok_instoL; simplenv; trivial.
      trivial.
      apply valTyok_instL; simplenv; trivial.
      eapply TupTy_lookup_valTyok; eauto.
    split.
      constructor; simplenv.
      apply RfTy_upd_rfTyok; trivial.
      constructor.
      apply TupTy_upd_tupTyok; trivial.
      eapply TupTy_lookup_valTyok; eauto.
    split_l.
    rewrite <- RfTy_upd_shiftY.
    rewrite <- RfTy_upd_instYL.
    rewrite shiftY_emp.
    rewrite shift_emp.
    simpl.
    rewrite <- TupTy_upd_instoL.
    apply rfTy_sub_upd; trivial.
    constructor.
    apply TupTy_upd_tupTyok; trivial.
    apply tupTyok_instoL; simplenv; trivial.
    apply valTyok_instL; simplenv; trivial.
    eapply TupTy_lookup_valTyok; eauto.
  eapply heap_update_wfrf_reg; eauto.
Qed.

(***************************************************************)

Unset Implicit Arguments.
