(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Soundness.v,v 1.1.1.1 2006/12/07 16:43:20 ssg05 Exp $ *)

Require Import Ty.
Require Import TyEnv.
Require Import Varmap.

Require Import Subst.
Require Import Subst_lemmas.

Require Import Rule.
Require Import Tyok_lemmas.
Require Import SubTy_lemmas.
Require Import Rule_lemmas.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Import TyEnv.

Open Local Scope tm_scope.

(***************************************************************)

Definition wfpc (Psi : chspec) (C : codeheap) (a : codeTy) (pc : pcreg) : Prop :=
  exists I : instrseq, 
    codeheap.getis C pc = Some I
    /\ wfis Psi a I.

Definition wfmach (Psi : chspec) (M : machine) : Prop := 
  let (C, S) := M in
  let (H, R, pc) := S in
    exists a,
      wfpsi Psi
      /\ wfch Psi Psi C
      /\ wfst Psi a S
      /\ wfpc Psi C a pc.

Definition safe_onestep (Psi : chspec) (M : machine) : Prop :=
  exists M' : _, 
    step M = Some M'
    /\ wfmach Psi M'.

Definition safe (Psi : chspec) (M : machine) : Prop :=
  forall n,
    exists M' : _, 
      stepn n M = Some M'
      /\ wfmach Psi M'.

(***************************************************************)

Lemma soundness_onestep :
  forall Psi M,
    wfmach Psi M
    -> safe_onestep Psi M.
Proof.
  intros Psi M HM.
  destruct M as [C [H R pc]].
  destruct HM as [[Delta Gamma] [HPsi [HC [HS Hpc]]]].
  destruct Hpc as [I [HCI HI]].
  destruct I.
  destruct i as
    [ rd rs rt | rd rs w
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f | r ]; 
    unfold wfis in HI; fold wfis in HI;
    try contradiction;
    simpl.
  (* case i = j f *)
  destruct HI as [a'' [Hf Hsub]].
  exists (C, <.H, R, f.>).
  unfold step.
  rewrite (getis_ione_get _ _ _ HCI).
  split_l.
  exists a''.
  split_l.
  split_l.
  split_l by 
    (apply wfst_weakening with (Delta, Gamma); trivial).
  apply (HC f a'' Hf).
  (* case i = jal f fret *)
  destruct HI as [[e' Y'] [Hfret [Ha'ok HI]]].
  destruct HI as [a'' [Hf Hsub]].
  exists (C, <. H, set R ra fret, f .>).
  unfold step.
  rewrite (getis_ione_get _ _ _ HCI).
  unfold nextS.
  split_l.
  exists a''.
  split_l.
  split_l.
  split_l by 
    (apply wfst_weakening 
      with (Delta, RfTy.upd Gamma ra (valTy_code e' Y')); 
      trivial).
  apply wfst_RfTy_upd_wfst with f; trivial.
  apply valTyok_weaken_emp.
  apply (HPsi fret (e', Y') Hfret).
  apply valok2_lab; trivial.
  apply (HC f a'' Hf).
  (* case jr r *)
  destruct HI as [[Delta'' Gamma''] [Hr Hsub]].
  exists (C, <.H, R, gR R r.>).
  unfold step.
  rewrite (getis_ione_get _ _ _ HCI).
  split_l.
  destruct (wfst_codeval_exists HS Hr) as [a''' Ha'''].
  exists a'''.
  split_l.
  split_l.
  split_l by (eapply wfst_codety_reg_wfst; eauto).
  apply (HC (gR R r) a''' Ha''').

 (* case iseq *)
  destruct i as
    [ rd rs rt | rd rs w
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f | r ];
    unfold wfis in HI; fold wfis in HI;
    try contradiction;
    destruct (getis_iseq_get _ _ _ _ HCI) as [Hi His];
    unfold step; simpl.
  (* case addu *)
  destruct HI as [Hb [Hrs [Hrt HI]]].
  exists (C, <.H, sR R rd (gR R rs + gR R rt), pc + 1 .>).
  unfold step.
  rewrite Hi.
  split_l.
  exists (Delta, RfTy.upd Gamma rd valTy_int).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_int_wfst with (pc := pc); eauto.
  unfold wfpc.
  exists I.
  split_lr.
  (* case addiu *)
  destruct HI as [Hb [Hrd Ha']].
  exists (C, <.H, sR R rd (gR R rs + w), pc+1.>).
  unfold step.
  rewrite Hi.
  split_l.
  exists (Delta, RfTy.upd Gamma rd valTy_int).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_int_wfst with (pc := pc); eauto.
  exists I.
  split_lr.
  (* case subu *)
  destruct HI as [Hb [Hrs [Hrt Ha']]].
  exists (C, <.H, sR R rd (gR R rs - gR R rt), pc+1.>).
  unfold step.
  rewrite Hi.
  split_l.
  exists (Delta, RfTy.upd Gamma rd valTy_int).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_int_wfst with (pc := pc); eauto.
  exists I.
  split_lr.
  (* case subiu *)
  destruct HI as [Hb [Hrd Ha']].
  exists (C, <.H, sR R rd (gR R rs - w), pc+1.>).
  unfold step.
  rewrite Hi.
  split_l.
  exists (Delta, RfTy.upd Gamma rd valTy_int).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_int_wfst with (pc := pc); eauto.
  exists I; split_lr.
  (* case move *)
  destruct HI as [t [Hb [Hrd Ha']]].
  exists (C, <.H, sR R rd (gR R rs), pc+1.>).
  unfold step.
  rewrite Hi.
  split_l.
  exists (Delta, RfTy.upd Gamma rd t).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_reg_wfst with (pc := pc); eauto.
  exists I; split_lr.
  (* case li *)
  destruct HI as [Hb [t [Hw HI]]].
  exists (C, <.H, sR R rd w, pc+1.>).
  unfold step.
  rewrite Hi.
  split_l.
  exists (Delta, RfTy.upd Gamma rd t).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_wfst with (pc := pc); eauto.
    apply valTyok_weaken_emp.
    inversion_clear Hw; trivial.
    constructor.
  exists I; split_lr.
  (* case lw *)
  destruct HI as [o [t [Hb [Hrs [Ho HI]]]]].
  destruct (wfst_heapval_exists HS Hrs Ho) as [v Hv].
  exists (C, <.H, sR R rt v, pc+1.>).
  unfold step.
  rewrite Hi.
  unfold nextS.
  rewrite (Heap.lookup_get Hv).
  split_l.
  exists (Delta, RfTy.upd Gamma rt t).
  split_l.
  split_l.
  split.
    eapply wfst_RfTy_upd_reg_load_tup_wfst with (pc := pc); eauto.
  exists I; split_lr.
  (* case sw *)
  destruct HI as [o [t [flag [Hb [Hrt [Hrs [Ho [Hce HI]]]]]]]].
  destruct (wfst_heapval_exists HS Hrs Ho) as [v Hv].
  set (H' := sH H (gR R rs + i) (gR R rt)).
  exists (C, <.H', R, pc+1.>).
  unfold step.
  rewrite Hi.
  unfold nextS.
  rewrite (Heap.lookup_get Hv).
  split_l.
  exists (Delta, RfTy.upd Gamma rs (valTy_tup 
    (TupTy.upd o i (init, t)))).
  split_l.
  split_l.
  split.
    subst H'.
    eapply wfst_RfTy_upd_reg_write_tup_wfst with (pc := pc); eauto.
  exists I; split_lr.
  (* case blt *)
  destruct HI as [[Delta'' Gamma''] [Hf [Hrs [Hrt [Hsub HI]]]]].
  decbltnat (gR R rs) (gR R rt).

  exists (C, <.H, R, f.>).
  unfold step, nextS.
  rewrite Hi.
  simplbnat.
  split_l.
  exists (Delta'', Gamma'').
  split_l.
  split_l.
  split_l by 
    (apply wfst_weakening with (Delta, Gamma); trivial).
  apply (HC f (Delta'', Gamma'') Hf).

  exists (C, <.H, R, pc+1.>).
  unfold step, nextS.
  rewrite Hi.
  simplbnat.
  split_l.
  exists (Delta, Gamma).
  split_l.
  split_l.
  split_l.
  exists I; split_lr.
  (* case beq *)
  destruct HI as [[Delta'' Gamma''] [Hf [Hrs [Hrt [Hsub HI]]]]].
  decbeqnat (gR R rs) (gR R rt).

  exists (C, <.H, R, f.>).
  unfold step, nextS.
  rewrite Hi.
  simplbnat.
  split_l.
  exists (Delta'', Gamma'').
  split_l.
  split_l.
  split_l by 
    (apply wfst_weakening with (Delta, Gamma); trivial).
  apply (HC f (Delta'', Gamma'') Hf).

  exists (C, <.H, R, pc+1.>).
  unfold step, nextS.
  rewrite Hi.
  simplbnat.
  split_l.
  exists (Delta, Gamma).
  split_l.
  split_l.
  split_l.
  exists I; split_lr.
  (* case bgtz *)
  destruct HI as [[Delta'' Gamma''] [Hf [Hrs [Hsub HI]]]].
  decbltnat 0 (gR R rs).

  exists (C, <.H, R, f.>).
  unfold step, nextS.
  rewrite Hi.
  simplbnat.
  split_l.
  exists (Delta'', Gamma'').
  split_l.
  split_l.
  split_l by 
    (apply wfst_weakening with (Delta, Gamma); trivial).
  apply (HC f (Delta'', Gamma'') Hf).
  
  exists (C, <.H, R, pc+1.>).
  unfold step, nextS.
  rewrite Hi.
  simplbnat.
  split_l.
  exists (Delta, Gamma).
  split_l.
  split_l.
  split_l.
  exists I; split_lr.
Qed.

Theorem Soundness :
  forall Psi M,
    wfmach Psi M
    -> safe Psi M.
Proof.
  intros Psi M HM n.
  induction n; simpl.
    exists M; split_lr.
  destruct IHn as [M' [Hstep HM']].
  destruct (soundness_onestep HM') 
    as [M'' [Hstep' HM'']].
  exists M''.
  rewrite Hstep.
  split_lr.
Qed.

(***************************************************************)

Unset Implicit Arguments.
