(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: SubTy_lemmas.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Ty.
Require Import TyEnv.
Require Import Varmap.
Require Import Rule.
Require Import Subst.
Require Import Subst_lemmas.
Require Import Tyok_lemmas.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

Import TyEnv.

(***************************************************************)
(* cellTy Subtyping Lemmas                                     *)

Lemma ceTy_sub_trans : forall ct ct' ct'',
  ceTy_sub ct ct' 
  -> ceTy_sub ct' ct''
  -> ceTy_sub ct ct''.
Proof.
  intros.
  inversion H0.
  inversion H.
  subst ct'.
  subst ct''.
  apply ceTy_sub_refl.
  subst ct''.
  rewrite <- H4.
  apply ceTy_sub_init.
  inversion H.
  subst ct'.
  apply ceTy_sub_init.
  subst ct'.
  discriminate.
Qed.

Lemma ceTy_sub_init_ct : forall t ct ct',
  ceTy_sub ct' ct
  -> ceTy_sub (init, t) ct
  -> ceTy_sub (init, t) ct'.
Proof.
  intros; destruct ct'; destruct ct.
  inversion_clear H; inversion_clear H0;
    try apply ceTy_sub_init;
      try apply ceTy_sub_refl.
Qed.

Lemma ceTy_sub_shift : forall f1 f2 t1 t2 e' e,
  ceTy_sub (f1, t1) (f2, t2)
  -> ceTy_sub (f1, shift t1 e' e) (f2, shift t2 e' e).
Proof.
  intros.
  inversion_clear H.
  apply ceTy_sub_refl.
  apply ceTy_sub_init.
Qed.

Lemma ceTy_sub_instL : forall f1 t1 f2 t2 tl e,
  ceTy_sub (f1, t1) (f2, t2)
  -> ceTy_sub (f1, instL t1 tl e 0) (f2, instL t2 tl e 0).
Proof.
  intros.
  inversion_clear H.
  apply ceTy_sub_refl.
  apply ceTy_sub_init.
Qed.


(***************************************************************)
(* RfTy Subtyping Lemmas *)

Lemma rfTy_sub_sel :
  forall e Y Y' r t,
    rfTy_sub e Y Y'
    -> RfTy.sel Y' r = Some t
    -> RfTy.sel Y r = Some t.
Proof.
  intros e Y Y' r t Hsub Hsel.
  inversion_clear Hsub.
  rewrite <- H1; trivial.
  exists t; trivial.
Qed.

Lemma rfTy_sub_sel_2 :
  forall e Y Y' r t,
    rfTy_sub e Y Y'
    -> RfTy.sel Y' r = Some t
    -> RfTy.sel Y r = RfTy.sel Y' r.
Proof.
  intros e Y Y' r t Hsub Hsel.
  inversion_clear Hsub.
  rewrite <- H1; trivial.
  exists t; trivial.
Qed.

Lemma rfTy_sub_nil :
  forall e Y,
    rfTyok e Y
    -> rfTy_sub e Y rfTy_nil.
Proof.
  intros e Y H.
  constructor; trivial.
  constructor.
  intros.
  destruct (RfTy.indom_nil r H0).
Qed.

Lemma rfTy_sub_cons :
  forall e Y Y' r t,
    valTyok e t
    -> rfTy_sub e Y Y'
    -> rfTy_sub e (rfTy_cons (r, t) Y) (rfTy_cons (r, t) Y').
Proof.
  induction Y; destruct Y'; try destruct p;
    simpl; intros; trivial.
  constructor; constructor; trivial.
  constructor.
  constructor.
  inversion_clear H0.
  assert (Hx : RfTy.indom (rfTy_cons (r, v) Y') r).
  unfold RfTy.indom.
  exists v; simpl; simplbreg; trivial.
  substH H3 with (H3 r Hx).
  simpl in H3.
  simplbreg.
  discriminate.
  inversion_clear H0.
  constructor.
  constructor; trivial.
  constructor; trivial.
  intros.
  unfold RfTy.indom in H0.
  simpl.
  simpl in H0.
  decbeqreg r1 r0; trivial.
  destruct H0; discriminate.
  destruct p0.
  inversion_clear H0.
  constructor.
  constructor; trivial.
  constructor; trivial.
  unfold RfTy.indom in * .
  simpl in * .
  intros r2.
  assert (H4 := H3 r2).
  intros.
  decbeqreg r2 r0; trivial.
  decbeqreg r2 r1; trivial.
  apply (H4 H0).
  apply (H4 H0).
Qed.

Lemma rfTy_sub_upd : 
  forall e Y Y' r t,
    valTyok e t
    -> rfTy_sub e Y Y'
    -> rfTy_sub e (RfTy.upd Y r t) (RfTy.upd Y' r t).
Proof.
  intros e Y Y' r t Ht Hsub.
  inversion_clear Hsub.
  constructor.
  apply RfTy_upd_rfTyok; trivial.
  apply RfTy_upd_rfTyok; trivial.
  intros r0 [t0 Hsel].
  decbeqreg r r0.
  rewrite RfTy.sel_upd_req; simplbreg; trivial.
  rewrite RfTy.sel_upd_req; simplbreg; trivial.
  rewrite RfTy.sel_upd_rneq; simplbreg; trivial.
  rewrite RfTy.sel_upd_rneq; simplbreg; trivial.
  apply H1.
  rewrite RfTy.sel_upd_rneq in Hsel; simplbreg; trivial.
  
  exists t0; trivial.
Qed.

Lemma rfTy_sub_sort_r :
  forall Y Y' Y'' e,
    rfTy_sub e Y Y'
    -> RfTy.sort Y Y' = Some Y''
    -> rfTy_sub e Y'' Y'.
Proof.
  intros.
  inversion_clear H.
  constructor; trivial.
  apply RfTy_sort_rfTyok with Y Y'; trivial.
  intros r Hx.
  substH H3 with (H3 r Hx).
  rewrite H3.
  eapply RfTy.sort_sel; eauto.
Qed.
  
Lemma rfTy_sub_sort :
  forall Y Y' Y'' e,
    rfTyok e Y
    -> RfTy.sort Y Y' = Some Y''
    -> rfTy_sub e Y Y''.
Proof.
  intros.
  constructor; trivial.
  apply RfTy_sort_rfTyok with Y Y'; trivial.
  intros r Hx.
  rewrite <- (RfTy.sort_sel_2 Y Y' Y'' r Hx H0). 
  trivial.
Qed.

Lemma rfTy_sub_refl : 
  forall e Y, rfTyok e Y -> rfTy_sub e Y Y.
Proof.
  intros.
  apply rfTy_sub_def; trivial.
Qed.

Lemma rfTy_sub_trans : 
  forall e0 Y Y' Y'',
    rfTy_sub e0 Y Y'
    -> rfTy_sub e0 Y' Y''
    -> rfTy_sub e0 Y Y''.
Proof.
  intros e0 Y Y' Y'' H12 H23.
  inversion_clear H12 as [ex Yx Y'x HY HY' Hin].
  inversion_clear H23 as [ex Y'x Y''x _ HY'' Hin'].
  apply rfTy_sub_def; trivial.
  intros r HinY''.
  inversion HinY''.
  rewrite (Hin' r HinY'') in H.
  assert (H2 : RfTy.indom Y' r).
  unfold RfTy.indom.
  exists x; trivial.
  rewrite (Hin' r HinY'').
  apply (Hin r H2).
Qed.

Lemma rfTy_sub_shiftY : 
  forall e0 Y Y' e e',
    rfTy_sub (app e0 e) Y Y'
    -> rfTy_sub (app e0 (app e' e)) (shiftY Y e' e) (shiftY Y' e' e).
Proof.
  intros e0 Y Y' e e' Hsub.
  inversion_clear Hsub as [e0x Yx Y'x HY HY' Hind].
  apply rfTy_sub_def.
  apply rfTyok_shiftY_aux; trivial.
  apply rfTyok_shiftY_aux; trivial.
  intros r Hin.
  assert (Hin' := @RfTy_indom_shiftY_rev Y' r e e' Hin).
  substH Hind with (Hind r Hin').
  apply RfTy_sel_shiftY; trivial.
Qed.

Lemma rfTy_sub_shiftY_rev : 
  forall e0 Y Y' e e',
    rfTy_sub (app e0 (app e' e)) (shiftY Y e' e) (shiftY Y' e' e)
    -> rfTy_sub (app e0 e) Y Y'.
Proof.
  intros e0 Y Y' e e' Hsub.
  inversion_clear Hsub as [e0x Yx Y'x HY HY' Hind].
  apply rfTy_sub_def.
  eapply rfTyok_shiftY_rev_aux; eauto.
  eapply rfTyok_shiftY_rev_aux; eauto.

  intros r Hin.
  assert (Hin' := @RfTy_indom_shiftY Y' r e e' Hin).
  substH Hind with (Hind r Hin').
  eapply RfTy_sel_shiftY_rev; eauto.
Qed.

Lemma rfTy_sub_instYL :
  forall e0 e Y Y' tl,
    valTyLok e0 tl
    -> rfTy_sub (app e0 e) Y Y'
    -> rfTy_sub e0 (instYL Y tl e 0) (instYL Y' tl e 0).
Proof.
  intros e0 e Y Y' tl' Htlok Hsub.
  inversion_clear Hsub as [e0x Yx Y'x HY HY' Hin].
  apply rfTy_sub_def.
  eapply rfTyok_instYL; eauto.
  eapply rfTyok_instYL; eauto.
  intros r HinY'.
  assert (HinY'0 := @RfTy_indom_instYL_rev Y' 
    r tl' e 0 HinY').
  assert (H := Hin r HinY'0).
  apply RfTy_sel_instYL; trivial.
Qed.


(***************************************************************)
(* codeTy Subtyping Lemmas                                     *)
(*
Lemma codeTy_sub_upd : 
  forall e0 e Y e' Y' tl r t,
    valTyLok (app e0 e) tl
    -> 
    -> codeTy_sub e0 (e, Y) (e', Y')
    -> codeTy_sub e0 (e, RfTy.upd Y r t) 
                     (e', RfTy.upd Y' r t).
Proof.

Qed.
*)
Lemma codeTy_sub_refl :
  forall e0 a, codeTyok e0 a -> codeTy_sub e0 a a.
Proof.
  intros e0 [e Y] Haok.
  exists (id_tl (len e)); trivial.
  split_l.
  split_l.
  split_l by (apply valTyLok_weaken).
  apply valTyLok_idtl.
  inversion_clear Haok.
  rewrite instYL_idtl.
  apply rfTy_sub_refl.
  trivial.
Qed.

Lemma codeTy_sub_trans : forall e0 a a' a'',
  codeTy_sub e0 a a'
  -> codeTy_sub e0 a' a''
  -> codeTy_sub e0 a a''.
Proof.
  intros e0 [e Y] [e' Y'] [e'' Y''] H12 H23.
  destruct H12 as [tl' [Ha [Ha' [Htl'ok Hsub12]]]].
  inversion_clear H23 as [tl'' [_ [Ha'' [Htl''ok Hsub23]]]].
  assert (Hsub23' := @rfTy_sub_shiftY _ Y' 
    (instYL (shiftY Y'' e' e'') tl'' e'' 0) e' e Hsub23).
  rewrite <- TyEnv.app_assoc_l in Hsub23'.
  assert (Hsub23'' := @rfTy_sub_instYL 
    (app e0 e) e' 
    (shiftY Y' e e')
    (shiftY (instYL (shiftY Y'' e' e'') tl'' e'' 0) e e')
    tl' Htl'ok Hsub23').
  assert (Hsub13 := rfTy_sub_trans Hsub12 Hsub23'').
  clear Hsub12 Hsub23 Hsub23' Hsub23''.
  rewrite shiftY_instYL_shiftY in Hsub13; trivial.
  rewrite (instYL_dist 
    (shiftY Y'' (app e e') e'')
    tl' (shiftL tl'' e e')
    e' e'') in Hsub13.
  rewrite <- shiftY_1s2_1s3 in Hsub13.
  rewrite instYL_shiftY_unchange in Hsub13; trivial.
  exists (LinstL (shiftL tl'' e e') tl' e' 0).
  split_l.
  split_l.
  split_r. 
  apply valTyLok_LinstL; trivial.
  simplenv.
  apply valTyLok_shiftL_aux; trivial.
Qed.

(* TODO : interesting result *)
(*
Lemma codeTy_sub_lift : forall e0 e Y e' Y',
  codeTy_sub e0 (e, Y) (e', Y')
  -> codeTy_sub emp (app e0 e, Y) (app e0 e', Y').
Proof.
  intros e0 e Y e' Y' H.
  inversion_clear H.
  constructor 1 with tl.
  
  inversion_clear H0; constructor; simplenv; trivial.

  inversion_clear H1; constructor; simplenv; trivial.

  simplenv; trivial.

  simplenv; trivial.

Admitted. *)

Lemma codeTy_sub_instYL : 
  forall e0 e e' Y Y' tl, 
    valTyLok e0 tl
    -> codeTy_sub e0 (e, Y) (e', Y')
    -> codeTy_sub e0 (emp, instYL Y tl e 0)
      (e', Y').
Proof.
  intros e0 e e' Y Y' tl Htlok Hsub.
  destruct Hsub as [tl' [Ha [Ha' [Htl'ok Hsub']]]].
  assert (Hsub := rfTy_sub_instYL Htlok Hsub').
  rewrite (instYL_dist (shiftY Y' e e') 
    tl tl' e e') in Hsub; trivial.
  rewrite instYL_shiftY_unchange in Hsub; trivial.
  exists (LinstL tl' tl e 0); simplenv.
  split_l by (eapply codeTyok_instYL; eauto).
  split_l.
  split_l by (eapply valTyLok_LinstL; eauto).
  rewrite shiftY_emp; trivial.
Qed.

Lemma codeTy_sub_instYL_2 : 
  forall e0 e e' Y Y' tl, 
    valTyLok e0 tl
    -> codeTy_sub (app e0 e) (emp, Y) (e', Y')
    -> codeTy_sub e0 (emp, instYL Y tl e 0)
      (e', instYL Y' (shiftL tl e' emp) e (len e')).
Proof.
  intros e0 e e' Y Y' tl Htl Hsub.
  destruct Hsub as [tl' [Ha [Ha' [Htl'ok Hsub']]]].
  simplenv.
  assert (Hsub := rfTy_sub_instYL Htl Hsub').
  constructor 1 with (LinstL tl' tl e 0).
  split_l.
    constructor; simplenv.
    inversion_clear Ha; simplenv.
    apply rfTyok_instYL; trivial.
  split_l.
    constructor.
    inversion_clear Ha'; simplenv.
    apply rfTyok_instYL_aux; trivial.
  split_l.
    simplenv.
    apply valTyLok_LinstL; trivial.
  simplenv.
  rewrite shiftY_emp.
  rewrite <-instYL_dist.
  rewrite shiftY_emp in Hsub.
  trivial.
Qed.

Unset Implicit Arguments.
