(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Subst.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Ty.
Require Import Varmap.
Require Import TyEnv.

Fixpoint shift (t : valTy) (ed ek : tenv) {struct t}: valTy :=
  match t with 
    | valTy_var m => if blt_nat m (TyEnv.len ek)
                       then valTy_var m
                       else valTy_var (m + (TyEnv.len ed))
    | valTy_int => t
    | valTy_tup o => valTy_tup (shifto o ed ek)
    | valTy_code e Y =>
          valTy_code e (shiftY Y ed (TyEnv.app ek e))
  end
with shiftY (Y : rfTy) (ed ek : tenv) {struct Y} : rfTy :=
  match Y with
    | rfTy_nil => rfTy_nil
    | rfTy_cons (r, t) m' => rfTy_cons (r, (shift t ed ek))
         (shiftY m' ed ek)
  end
with shifto (o : tupTy) (ed ek : tenv) {struct o} : tupTy :=
  match o with
    | tupTy_nil => tupTy_nil
    | tupTy_cons (f, t) o' => tupTy_cons (f, (shift t ed ek))
         (shifto o' ed ek)
  end.

Fixpoint rshift (t : valTy) (ed ek : tenv) {struct t}: valTy :=
  match t with 
    | valTy_var m => if blt_nat m (TyEnv.len ek)
                       then valTy_var m
                       else valTy_var (m - (TyEnv.len ed))
    | valTy_int => t
    | valTy_tup o => valTy_tup (rshifto o ed ek)
    | valTy_code e Y =>
          valTy_code e (rshiftY Y ed (TyEnv.app ek e))
  end
with rshiftY (Y : rfTy) (ed ek : tenv) {struct Y} : rfTy :=
  match Y with
    | rfTy_nil => rfTy_nil
    | rfTy_cons (r, t) m' => rfTy_cons (r, (rshift t ed ek))
         (rshiftY m' ed ek)
  end
with rshifto (o : tupTy) (ed ek : tenv) {struct o} : tupTy :=
  match o with
    | tupTy_nil => tupTy_nil
    | tupTy_cons (f, t) o' => tupTy_cons (f, (rshift t ed ek))
         (rshifto o' ed ek)
  end.

Fixpoint shiftL (tl : varmap) (ed ek : tenv) 
  {struct tl}: varmap :=
  match tl with
    | nil => nil
    | cons (n, t) tl => cons (n, shift t ed ek) (shiftL tl ed ek) 
  end.

Fixpoint instL (t : valTy) (tl : varmap) (e : tenv) (n : nat)
  {struct t} : valTy :=
  match t with 
    | valTy_var m => 
        if blt_nat m n 
          then valTy_var m
          else 
            if blt_nat m (TyEnv.len e + n)
              then Varmap.get2 tl (valTy_int) (m - n)
              else valTy_var (m - TyEnv.len e)
    | valTy_int => t
    | valTy_tup o => valTy_tup (instoL o tl e n)
    | valTy_code e0 Y =>
        valTy_code e0 (instYL Y (shiftL tl e0 TyEnv.emp)
          e (n + TyEnv.len e0))
  end
  with instYL (Y : rfTy) (tl : varmap) (e : tenv) (n : nat) 
  {struct Y} : rfTy :=
  match Y with
    | rfTy_nil => rfTy_nil
    | rfTy_cons (r, t) Y => 
          rfTy_cons (r, (instL t tl e n)) (instYL Y tl e n)
  end
  with instoL (o : tupTy) (tl : varmap) (e : tenv) (n : nat) 
  {struct o} : tupTy :=
  match o with
    | tupTy_nil => tupTy_nil
    | tupTy_cons (f, t) o => 
          tupTy_cons (f, (instL t tl e n)) (instoL o tl e n)
  end.

Fixpoint LinstL (tl tl': varmap) (e : tenv) (n : nat)
  {struct tl} : varmap :=
  match tl with 
    | nil => nil
    | cons (n0, t) tl => 
        cons (n0, instL t tl' e n) (LinstL tl tl' e n)
  end.

Fixpoint bcpactL (t : valTy) (tl : varmap) (e : tenv) (n : nat)
  {struct t} : bool :=
  match t with 
    | valTy_var m => 
        if blt_nat m n 
          then true
          else 
            if blt_nat m (TyEnv.len e + n) 
              then bindom tl (m - n)
              else true
    | valTy_int => true
    | valTy_tup o => bcpactoL o tl e n
    | valTy_code e0 Y =>
        bcpactYL Y (shiftL tl e0 TyEnv.emp) e (n + TyEnv.len e0)
  end
  with bcpactYL (Y : rfTy) (tl : varmap) (e : tenv) (n : nat) 
  {struct Y} : bool :=
  match Y with
    | rfTy_nil => true
    | rfTy_cons (r, t) Y => andb
          (bcpactL t tl e n) (bcpactYL Y tl e n)
  end
  with bcpactoL (o : tupTy) (tl : varmap) (e : tenv) (n : nat) 
  {struct o} : bool :=
  match o with
    | tupTy_nil => true
    | tupTy_cons (f, t) o => andb
          (bcpactL t tl e n) (bcpactoL o tl e n)
  end.

Fixpoint bLcpactL (tl tl': varmap) (e : tenv) (n : nat)
  {struct tl} : bool :=
  match tl with 
    | nil => true
    | cons (n0, t) tl => andb
      (bcpactL t tl' e n) (bLcpactL tl tl' e n)
  end.
