(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Subst_lemmas.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Varmap.
Require Import Ty.
Require Import TyEnv.
Require Import Subst.

Require Import extac.

Set Implicit Arguments.
Unset Strict Implicit.

(***************************************************************)
(* shift, shiftL lemmas                                        *)

Lemma shift_emp :
  forall t e,
    shift t emp e = t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e,
        shiftY Y emp e
        = Y)
    (P1 := fun o =>
      forall e,
        shifto o emp e
        = o);
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e.
  simpl.
  simplenv; simplpm.
  decbltnat n0 (len e); trivial.
  (* t = int *)
  intros e.
  simpl;  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e.
  simpl.
  rewrite (IHo e).
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e.
  simpl.
  rewrite (IHY (app e e0)).
  trivial.
  (* Y = e0 nil *)
  intros e.
  simpl;  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e.
  simpl.
  rewrite (IHt e).
  rewrite (IHY e).
  trivial.
  (* o = nil *)
  intros e.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e.
  simpl.
  rewrite (IHt e).
  rewrite (IHo e).
  trivial.
Qed.

Lemma shiftY_emp : forall Y e, 
  shiftY Y emp e = Y.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY.
  rewrite shift_emp; trivial.
Qed.

Lemma shifto_emp : forall o e, 
  shifto o emp e = o.
Proof.
  induction o; try destruct p as [f t];
    simpl; intros; trivial.
  rewrite IHo.
  rewrite shift_emp; trivial.
Qed.

Lemma inj_shift : forall t t' e e', 
  (shift t e e') = (shift t' e e') 
  -> t = t'.
Proof.
  intros t.
  induction t using valTy_ind3 with
    (P0 := fun Y =>
      forall Y' e e',
        shiftY Y e e' = shiftY Y' e e' -> Y = Y')
    (P1 := fun o =>
      forall o' e e',
        shifto o e e' = shifto o' e e' -> o = o');
    intro t'; destruct t'; simpl; intros.
  decbltnat n0 (len e'); decbltnat n (len e').
  trivial.
  injection H; clear H; intros; subst.
  desbnat; elimtype False; omega.
  injection H; clear H; intros; subst.
  desbnat; elimtype False; omega.
  injection H; clear H; intros; subst.
  assert (n0=n); [omega | subst]; trivial.

  decbltnat n (len e'); discriminate.
  decbltnat n (len e'); discriminate.
  decbltnat n (len e'); discriminate.
  decbltnat n (len e'); discriminate.
  trivial.
  discriminate.
  discriminate.
  decbltnat n (len e'); discriminate.
  discriminate.
  injection H; clear H; intros; subst.
  rewrite (IHt t e e' H); trivial.
  discriminate.
  decbltnat n (len e'); discriminate.
  discriminate.
  discriminate.
  injection H; clear H; intros; subst.
  rewrite (IHt r e1 (app e' e0) H); trivial.
  
  trivial.
  destruct p; discriminate.
  discriminate.
  destruct p.
  injection H; clear H; intros; subst.
  rewrite (IHt v e e' H0).
  rewrite (IHt0 t' e e' H); trivial.

  trivial.
  destruct p; discriminate.
  discriminate.
  destruct p.
  injection H; clear H; intros; subst.
  rewrite (IHt v e e' H0).
  rewrite (IHt0 t' e e' H); trivial.
Qed.

(***************************************************************)

(*
 *                     |--------||
 *                         1     0       
 *            |========|--------||
 *                2        1     0      
 *   |========|########|--------||
 *        2       3        1     0      
*)

Lemma shift_1s2_1s3 :
  forall t e e' e'', 
    shift (shift t e' e ) e'' e = shift t (app e' e'') e.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e' e'',
        shiftY (shiftY Y e' e) e'' e
        = shiftY Y (app e' e'') e)
    (P1 := fun o => 
      forall e e' e'',
        shifto (shifto o e' e) e'' e
        = shifto o (app e' e'') e);
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' e''.
  simpl; simplenv.
  decbltnat n0 (len e); simpl; simplbnat; trivial.
  rewbnat (blt_nat (n0 + len e') (len e) = false).
  simplpm; trivial.
  (* t = int *)
  intros e e' e''.
  simpl;  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e e' e''.
  simpl.
  rewrite (IHo e e' e'').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' e''.
  simpl; simplenv.
  rewrite (IHY (app e e0) e' e'').
  trivial.
  (* Y = e0 nil *)
  intros e e' e''.
  simpl;  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e' e''.
  simpl.
  rewrite (IHt e e' e'').
  rewrite (IHY e e' e'').
  trivial.
  (* o = nil *)
  intros e e' e''.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e' e''.
  simpl.
  rewrite (IHt e e' e'').
  rewrite (IHo e e' e'').
  trivial.
Qed.

Lemma shiftY_1s2_1s3 :
  forall Y e e' e'', 
    shiftY (shiftY Y e' e) e'' e
    = shiftY Y (app e' e'') e.
Proof.
  induction Y; try destruct p as [r t]; 
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite shift_1s2_1s3; trivial.
Qed.

(*
 *           |--------||         
 *                1    0           
 *  |--------|========||         
 *      1         2    0            
*)

Lemma shift_0s1_0s2 :
  forall t e e', 
    shift (shift t e emp ) e' emp
    = shift t (app e e') emp.
Proof.
  intros t e e'.
  rewrite (shift_1s2_1s3 t emp e e').
  trivial.
Qed.

Lemma shiftY_0s1_0s2 :
  forall Y e e', 
    shiftY (shiftY Y e emp ) e' emp
    = shiftY Y (app e e') emp.
Proof.
  induction Y; try destruct p as [r t]; 
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite shift_0s1_0s2; trivial.
Qed.

(***************************************************************)

(*
 *                     |--------||
 *                         1     0       
 *            |========|--------||
 *                2        1     0      
 *   |########|========|--------||
 *        3       2        1     0      
*)

Lemma shift_1s2_2s3 :
  forall t e e' e'', 
    shift (shift t e' e) e'' (app e' e)
    = shift t (app e'' e') e.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e' e'',
        shiftY (shiftY Y e' e) e'' (app e' e)
        = shiftY Y (app e'' e') e)
    (P1 := fun o => 
      forall e e' e'',
        shifto (shifto o e' e) e'' (app e' e)
        = shifto o (app e'' e') e);
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' e''.
  simpl; simplenv.
  decbltnat n0 (len e); simpl; simplenv; simplpm.
    rewbnat (blt_nat n0 (len e' + len e)= true).
    trivial.
  rewbnat (blt_nat (n0 + len e') (len e' + len e) = false). 
  rep_arith (n0 + len e' + len e'') 
    with (n0 + (len e'' + len e')).
  trivial.
  (* t = int *)
  intros e e' e''.
  simpl;  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e e' e''.
  simpl.
  rewrite (IHo e e' e'').
  trivial.
  (* t = coe'e e0 Y *)
  intros e0 Y IHY.
  intros e e' e''.
  simpl; simplenv.
  rewrite (IHY (app e e0) e' e'').
  trivial.
  (* Y = e0 nil *)
  intros e e' e''.
  simpl;  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e' e''.
  simpl.
  rewrite (IHt e e' e'').
  rewrite (IHY e e' e'').
  trivial.
  (* o = nil *)
  intros e e' e''.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e' e''.
  simpl.
  rewrite (IHt e e' e'').
  rewrite (IHo e e' e'').
  trivial.
Qed.

Lemma shiftY_1s2_2s3 :
  forall Y e e' e'', 
    shiftY (shiftY Y e' e) e'' (app e' e)
    = shiftY Y (app e'' e') e.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY.
  rewrite shift_1s2_2s3; trivial.
Qed.

(*
 *                     |--------||
 *                         1     0       
 *            |========|--------||
 *                2        1     0      
*)

Lemma shift_0s1_1s2 :
  forall t e e', 
    shift (shift t e emp) e' e
    = shift t (app e' e) emp .
Proof.
  intros t e e'.
  assert (H := shift_1s2_2s3 t emp e e').
  simplenv; simplpm.
  trivial.
Qed.

Lemma shiftY_0s1_1s2 :
  forall Y e e', 
    shiftY (shiftY Y e emp) e' e
    = shiftY Y (app e' e) emp.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite shift_0s1_1s2; trivial.
Qed.

(***************************************************************)

Lemma shift_1s2_0s3_aux :
  forall t e e' e'' e''', 
    shift (shift t e' (app e e''')) e'' e'''
    = shift (shift t e'' e''') e' (app (app e e'') e''').
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e' e'' e''',
        shiftY (shiftY Y e' (app e e''')) e'' e'''
        = shiftY (shiftY Y e'' e''') e' (app (app e e'') e'''))
    (P1 := fun o => 
      forall e e' e'' e''',
        shifto (shifto o e' (app e e''')) e'' e'''
        = shifto (shifto o e'' e''') e' (app (app e e'') e'''));
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' e'' e'''.
  simpl; simplenv.
  decbltnat n0 (len e + len e'''); simpl; simplenv; simplpm.
    decbltnat n0 (len e'''); simpl; simplenv; simplpm.
      rewbnat (blt_nat n0 (len e + (len e'' + len e''')) = true).
      trivial.
    rewbnat (blt_nat (n0 + len e'') (len e + (len e'' + len e''')) = true).
    trivial.
  rewbnat (blt_nat (n0 + len e') (len e''') = false).
  rewbnat (blt_nat n0 (len e''') = false).
  simpl; simplenv; simplpm.
  rewbnat (blt_nat (n0 + len e'') (len e + (len e'' + len e''')) = false).
  rep_arith (n0 + len e' + len e'') 
    with (n0 + len e'' + len e').
  trivial.
  (* t = int *)
  intros e e' e'' e'''.
  simpl;  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e e' e'' e'''.
  simpl.
  rewrite (IHo e e' e'' e''').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' e'' e'''.
  simpl; simplenv.
  rewrite (IHY e e' e'' (app e''' e0)).
  simplenv.
  trivial.
  (* Y = e0 nil *)
  intros e e' e'' e'''.
  simpl;  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e' e'' e'''.
  simpl.
  rewrite (IHt e e' e'' e''').
  rewrite (IHY e e' e'' e''').
  trivial.
  (* o = nil *)
  intros e e' e'' e'''.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e' e'' e'''.
  simpl.
  rewrite (IHt e e' e'' e''').
  rewrite (IHo e e' e'' e''').
  trivial.
Qed.

Lemma shift_1s2_0s3 :
  forall t e e' e'', 
    shift (shift t e' e) e'' emp
    = shift (shift t e'' emp) e' (app e e'').
Proof.
  intros t e e' e''.
  assert (H := shift_1s2_0s3_aux t e e' e'' emp).
  simplenv.
  trivial.
Qed.

(* ******************************************** *)

Lemma rshift_emp :
  forall t e,
    rshift t emp e = t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e,
        rshiftY Y emp e
        = Y)
    (P1 := fun o =>
      forall e,
        rshifto o emp e
        = o);
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e.
  simpl.
  simplenv; simplpm.
  decbltnat n0 (len e); trivial.
  (* t = int *)
  intros e.
  simpl;  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e.
  simpl.
  rewrite (IHo e).
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e.
  simpl.
  rewrite (IHY (app e e0)).
  trivial.
  (* Y = e0 nil *)
  intros e.
  simpl;  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e.
  simpl.
  rewrite (IHt e).
  rewrite (IHY e).
  trivial.
  (* o = nil *)
  intros e.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e.
  simpl.
  rewrite (IHt e).
  rewrite (IHo e).
  trivial.
Qed.

Lemma rshiftY_emp :
  forall Y e,
    rshiftY Y emp e = Y.
Proof.
  induction Y; try destruct p; simpl; intros;
    trivial.
  rewrite IHY.
  rewrite rshift_emp.
  trivial.
Qed.

Lemma rshifto_emp :
  forall o e,
    rshifto o emp e = o.
Proof.
  induction o; try destruct p; simpl; intros;
    trivial.
  rewrite IHo.
  rewrite rshift_emp.
  trivial.
Qed.

Lemma shift_rshift : 
  forall t e e',
    valTy_boute2 t e' e = true
    -> shift (rshift t e' e) e' e = t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e',
        rfTy_boute2 Y e' e = true
        -> shiftY (rshiftY Y e' e) e' e = Y)
    (P1 := fun o => 
      forall e e',
        tupTy_boute2 o e' e = true
        -> shifto (rshifto o e' e) e' e = o);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' Hout.
  unfold TyEnv.bine2, negb, andb, ifb in Hout.
  decbltnat n0 (len e' + len e);
    decbltnat n0 (len e); 
    simplbnat; simpl; simplbnat; trivial.
  rewbnat (blt_nat (n0 - len e') (len e) = false).
  repbnat (n0 - len e' + len e') with n0; trivial.
  (* t = int *)
  intros e e' Hout.
  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e e' Hout.
  simpl.
  rewrite (IHo e e' Hout).
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' Hout.
  rewrite (IHY (app e e0) e' Hout); trivial.
  (* Y = e0 nil *)
  intros e e' Hout.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e' Hout.
  destruct (andb_prop _ _ Hout) as [Hout1 Hout2].
  rewrite (IHt e e' Hout1).
  rewrite (IHY e e' Hout2).
  trivial.
  (* o = nil *)
  intros e e' Hout.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e' Hout.
  destruct (andb_prop _ _ Hout) as [Hout1 Hout2].
  rewrite (IHt e e' Hout1).
  rewrite (IHo e e' Hout2).
  trivial.
Qed.

(* ******************************************** *)

Lemma shiftL_emp :
  forall tl e,
    shiftL tl emp e = tl.
Proof.
  intros; induction tl; try destruct a; simpl; auto.
  rewrite IHtl.
  rewrite (shift_emp v e).
  trivial.
Qed.

Lemma shiftL_shiftL_merge :
  forall tl e e' e'',
    shiftL (shiftL tl e' e) e'' e
    = shiftL tl (app e' e'') e.
Proof.
  induction tl; try destruct a; simpl; intros; trivial.
  rewrite IHtl.
  rewrite shift_1s2_1s3; trivial.
Qed.

Lemma shiftL_shiftL_merge_cont :
  forall tl e e' e'',
    shiftL (shiftL tl e' e) e'' (app e' e) 
    = shiftL tl (app e'' e') e.
Proof.
  induction tl; try destruct a; simpl; intros; trivial.
  rewrite shift_1s2_2s3.
  rewrite IHtl; trivial.
Qed.

Lemma shiftL_shiftL_merge_cont_emp :
  forall tl e' e'',
    shiftL (shiftL tl e' emp) e'' e'
    = shiftL tl (app e'' e') emp.
Proof.
  induction tl; try destruct a;simpl; intros; trivial.
  rewrite shift_0s1_1s2.
  rewrite IHtl; trivial.
Qed.

Lemma shiftL_shiftL_comm :
  forall tl e e' e'',
    shiftL (shiftL tl e' e) e'' emp
    = shiftL (shiftL tl e'' emp) e' (app e e'').
Proof.
  intros.
  induction tl.
  simpl.
  trivial.
  destruct a.
  simpl.
  rewrite <- shift_1s2_0s3.
  rewrite IHtl.
  trivial.
Qed.

(***************************************************************)

Lemma beq_valTy_shift : 
  forall t t' e e',
    beq_valTy (shift t e e') (shift t' e e')
    = beq_valTy t t'.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall Y' e e',
        beq_rfTy (shiftY Y e e') (shiftY Y' e e')
        = beq_rfTy Y Y')
    (P1 := fun o =>
      forall o' e e',
        beq_tupTy (shifto o e e') (shifto o' e e')
        = beq_tupTy o o');
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros t' e e'; destruct t'; try discriminate;
    simpl.
  decbltnat n (len e');
  decbltnat n0 (len e'); simpl; trivial.
  rewbnat (beq_nat n0 n = false); solvebnat.
  rewbnat (beq_nat n0 n = false); solvebnat.
  decbeqnat n0 n; solvebnat.
  decbltnat n0 (len e'); simpl; trivial.
  decbltnat n0 (len e'); simpl; trivial.
  decbltnat n0 (len e'); simpl; trivial.
  (* t = int *)
  intros t' e e'; destruct t'; 
    try discriminate;
    simpl; try decbltnat n (len e'); 
      simpl; trivial.
  (* t = tup o *)
  intros o IHo .
  intros t' e e'; destruct t'; 
    try discriminate;
    simpl; try decbltnat n (len e'); 
      simpl; trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros t' e e'; destruct t'; 
    try discriminate;
    simpl; trivial.
  try decbltnat n (len e'); 
      simpl; trivial.
  unfold beq_env.
  decbeqnat e0 e1; trivial.
  substbnat e1.
  apply IHY.
  (* Y = e0 nil *)
  intros Y' e e'; destruct Y'; 
    try destruct p;
    try discriminate;
    simpl; trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros Y' e e'; destruct Y'; 
    try destruct p;
    try discriminate;
    simpl; trivial.
  decbeqreg r r0; trivial.
  rewrite IHt.
  rewrite IHY.
  trivial.
  (* o = nil *)
  intros o' e e'; destruct o'; 
    try destruct p;
    try discriminate;
    simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros o' e e'; destruct o'; 
    try destruct p;
    try destruct f;
    try destruct f0;
    try discriminate;
    simpl; trivial.
  rewrite IHt.
  rewrite IHo.
  trivial.
  rewrite IHt.
  rewrite IHo.
  trivial.
Qed.

Lemma varmap_get2_shiftL : 
  forall tl e e' ax ax' i,
    bindom tl i = true
    -> get2 (shiftL tl e e') ax i
       = shift (get2 tl ax' i) e e'.
Proof.
  induction tl; try destruct a; 
    simpl; simplbnat; simplpm;
        intros e e' ax ax' i H; trivial.
  discriminate.
  decbeqnat i n; trivial.
  simpl in H.
  apply IHtl; trivial.
Qed.

Lemma varmap_get2_shiftL_2 : 
  forall tl e e' ax i,
    bindom tl i = true
    -> get2 (shiftL tl e e') ax i
       = shift (get2 tl ax i) e e'.
Proof.
  induction tl; try destruct a; 
    simpl; simplbnat; simplpm;
        intros e e' ax i H; trivial.
  discriminate.
  decbeqnat i n; trivial.
  simpl in H.
  apply IHtl; trivial.
Qed.

Lemma varmap_get2_shiftL_3 : 
  forall tl e e' i,
    get2 (shiftL tl e e') valTy_int i
    = shift (get2 tl valTy_int i) e e'.
Proof.
  induction tl; try destruct a; 
    simpl; simplbnat; simplpm;
        intros e e' i; trivial.
  decbeqnat i n; trivial.
Qed.

Lemma varmap_bindom_shiftL :
  forall tl e e' i,
    Varmap.bindom (shiftL tl e e') i
    = Varmap.bindom tl i.
Proof.
  induction tl; try destruct a; simpl; intros; trivial.
  rewrite IHtl; trivial.
Qed.

Lemma varmap_bsubset_shiftL :
  forall n tl tl' e e0 e'',
    Varmap.bsubset (shiftL tl e0 e) 
      (shiftL tl' e0 e) n e'' 
    = Varmap.bsubset tl tl' n e''.
Proof.
  induction n; simpl; 
    intros; trivial.
  rewrite varmap_bindom_shiftL.
  rewrite varmap_bindom_shiftL.
  rewrite IHn.
  destruct (Varmap.bindom_dec tl (n + len e''))
    as [Hin | Hin]; rewrite Hin; trivial.
  destruct (Varmap.bindom_dec tl' (n + len e''))
    as [Hin' | Hin']; rewrite Hin'; trivial.
  rewrite varmap_get2_shiftL_2; trivial.
  rewrite varmap_get2_shiftL_2; trivial.
  rewrite beq_valTy_shift; trivial.
Qed.


(***************************************************************)
(* bcpactL, bLcpactL lemmas                                    *)

Lemma bcpactL_shiftL : 
  forall t tl ed n e e',
    bcpactL t (shiftL tl e' e) ed n
    = bcpactL t tl ed n.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl ed n e e',
        bcpactYL Y (shiftL tl e' e) ed n
        = bcpactYL Y tl ed n)
    (P1 := fun o => 
      forall tl ed n e e',
        bcpactoL o (shiftL tl e' e) ed n
        = bcpactoL o tl ed n);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros tl ed n e e'.
  decbltnat n0 n; trivial.
  decbltnat n0 (len ed + n); trivial.
  rewrite varmap_bindom_shiftL; trivial.
  (* t = int *)
  intros tl ed n e e'.
  trivial.
  (* t = tup o *)
  intros o IHo.
  intros tl ed n e e'.
  rewrite (IHo tl ed n e e').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros tl ed n e e'.
  simpl; simplenv.
  rewrite shiftL_shiftL_comm.
  rewrite IHY; trivial.
  (* Y = e0 nil *)
  intros tl ed n e e'.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros tl ed n e e'.
  rewrite (IHt tl ed n e e').
  rewrite (IHY tl ed n e e').
  trivial.
  (* o = nil *)
  intros tl ed n e e'.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl ed n e e'.
  rewrite (IHt tl ed n e e').
  rewrite (IHo tl ed n e e').
  trivial.
Qed.

Lemma bcpactYL_shiftL : 
  forall Y tl ed n e e',
    bcpactYL Y (shiftL tl e' e) ed n
    = bcpactYL Y tl ed n.
Proof.
  induction Y; try destruct p as [r t]; simpl; 
    intros; trivial.
  rewrite bcpactL_shiftL.
  rewrite IHY; trivial.
Qed.

Lemma bcpactoL_shiftL : 
  forall o tl ed n e e',
    bcpactoL o (shiftL tl e' e) ed n
    = bcpactoL o tl ed n.
Proof.
  induction o; try destruct p as [f t]; simpl; 
    intros; trivial.
  rewrite bcpactL_shiftL.
  rewrite IHo; trivial.
Qed.

Lemma bcpactL_shift_left : 
  forall t tl e e' e'',
    bcpactL (shift t e'' e') tl e (len e'' + len e')
    = bcpactL t tl e (len e').
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl e e' e'',
        bcpactYL (shiftY Y e'' e') tl e (len e'' + len e')
        = bcpactYL Y tl e (len e'))
    (P1 := fun o => 
      forall tl e e' e'',
        bcpactoL (shifto o e'' e') tl e (len e'' + len e')
        = bcpactoL o tl e (len e'));
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros tl e e' e''.
  decbltnat n0 (len e'); simpl; trivial.
    rewbnat (blt_nat n0 (len e'' + len e') = true).
    trivial.
  rewbnat (blt_nat (n0 + len e'') (len e'' + len e') = false).
  decbltnat n0 (len e + len e'); trivial.
    rewbnat (blt_nat (n0 + len e'') (len e + (len e'' + len e')) = true).
    simplpm; trivial.
  rewbnat (blt_nat (n0 + len e'') (len e + (len e'' + len e')) = false).
  trivial.
  (* t = int *)
  intros tl e e' e''.
  trivial.
  (* t = tup o *)
  intros o IHo.
  intros tl e e' e''.
  rewrite (IHo tl e e' e'').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros tl e e' e''.
  rewrite bcpactYL_shiftL.
  rewrite bcpactYL_shiftL.
  assert (H := IHY tl e (app e' e0) e'').
  simplenv. simplpm.
  apply H.
  (* Y = e0 nil *)
  intros tl e e' e''.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros tl e e' e''.
  rewrite (IHt tl e e' e'').
  rewrite (IHY tl e e' e'').
  trivial.
  (* o = nil *)
  intros tl e e' e''.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl e e' e''.
  rewrite (IHt tl e e' e'').
  rewrite (IHo tl e e' e'').
  trivial.
Qed.

Lemma bLcpactL_shiftL :
  forall tl tl' e e' e'' n,
    bLcpactL tl (shiftL tl' e' e'') e n
    = bLcpactL tl tl' e n.
Proof.
  induction tl; simpl; intros; trivial.
  destruct a as [n0 t].
  simpl.
  rewrite bcpactL_shiftL.
  rewrite IHtl; trivial.
Qed.

Lemma bLcpactL_shiftL_left :
  forall tl tl' e e' e'',
    bLcpactL (shiftL tl e'' e') tl' e (len e'' + len e')
    = bLcpactL tl tl' e (len e').
Proof.
  induction tl; simpl; intros; trivial.
  destruct a as [n0 t].
  simpl.
  rewrite bcpactL_shift_left.
  rewrite IHtl; trivial.
Qed.

Lemma bcpactL_shift_left_emp :
  forall t tl e e',
    bcpactL (shift t e' emp) tl e (len e')
    = bcpactL t tl e 0.
Proof.
  intros t tl e e'.
  assert (H := bcpactL_shift_left t tl e emp e').
  simplenv; simplpm.
  trivial.
Qed.

Lemma bLcpactL_shiftL_left_emp :
  forall tl tl' e e',
    bLcpactL (shiftL tl e' emp) tl' e (len e')
    = bLcpactL tl tl' e 0.
Proof.
  induction tl; simpl; intros; trivial.
  destruct a as [n0 t].
  simpl.
  assert (H := bcpactL_shift_left t tl' e emp e').
  simplenv; simplpm.
  rewrite IHtl; trivial.
  rewrite H; trivial.
Qed.

(***************************************************************)
(* instL, LinstL lemmas                                        *)

Lemma instL_shift_unchange :
  forall t tl e e',
    instL (shift t e e') tl e (len e') 
    = t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl e e',
        instYL (shiftY Y e e') tl e (len e') = Y)
    (P1 := fun o =>
      forall tl e e',
        instoL (shifto o e e') tl e (len e') = o);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros tl e e'.
  decbltnat n0 (len e'); 
    simpl; simplbnat; trivial.
  rewbnat (blt_nat (n0 + len e) (len e') = false).
  rewbnat (blt_nat (n0 + len e) (len e + len e') = false).
  simplpm; trivial.
  (* t = int *)
  intros tl e e'.
  trivial.
  (* t = tup o *)
  intros o IHo .
  intros tl e e'.
  rewrite (IHo tl e e').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros tl e e'.
  assert (H := IHY (shiftL tl e0 emp) 
    e (TyEnv.app e' e0)).
  simplenv.
  rewrite H; trivial.
  (* Y = e0 nil *)
  intros tl e e'.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros tl e e'.
  rewrite (IHt tl e e').
  rewrite (IHY tl e e').
  trivial.
  (* o = nil *)
  intros tl e e'.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl e e'.
  rewrite (IHt tl e e').
  rewrite (IHo tl e e').
  trivial.
Qed.

Lemma instYL_shiftY_unchange :
  forall Y tl e e',
    instYL (shiftY Y e e') tl e (len e') 
    = Y.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite instL_shift_unchange; trivial.
Qed.

(***************************************************************)

Lemma varmap_get2_LinstL :
  forall tl' tl e n i,
    Varmap.get2 (LinstL tl' tl e n) valTy_int i 
    = instL (Varmap.get2 tl' valTy_int i) tl e n.
Proof.
  induction tl' as [| (n', t') tl' IHtl']; simpl.
  intros tl e n i.
   trivial.
  intros tl e n i.
  decbeqnat i n'; trivial.
Qed.

(***************************************************************)

Lemma instL_shift :
  forall t tl' e e' ed,
    instL (shift t e' e) (shiftL tl' e' e) ed (len e' + len e)
    = shift (instL t tl' ed (len e)) e' e.
Proof.
  intro t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl' e e' ed,
        instYL (shiftY Y e' e) 
          (shiftL tl' e' e) ed (len e' + len e)
        = shiftY (instYL Y tl' ed (len e)) e' e)
    (P1 := fun o => 
      forall tl' e e' ed,
        instoL (shifto o e' e) 
          (shiftL tl' e' e) ed (len e' + len e)
        = shifto (instoL o tl' ed (len e)) e' e);
    clear t.
  (* t = var n0 *)
  intros n0.
  simpl.
  intros tl' e e' ed.
  decbltnat n0 (len e);
    simpl; simplpm; simplbnat; trivial.
  rewbnat (blt_nat (n0 + len e') (len e' + len e) = false).
  decbltnat n0 (len ed + len e); 
    simpl; simplpm; simplbnat; trivial.
  rewbnat (blt_nat (n0 + len e') (len ed + (len e' + len e)) = true).
  rewrite varmap_get2_shiftL_3; trivial.
  rewbnat (blt_nat (n0 + len e') (len ed + (len e' + len e)) = false).
  rewbnat (blt_nat (n0 - len ed) (len e) = false).
  repbnat (n0 + len e' - len ed)
    with (n0 - len ed + len e').
  trivial.
  (* t = int *)
  intros tl' e e' ed.
  simpl; trivial.
  (* t = tup o *)
  intros o IHo.
  intros tl' e e' ed.
  simpl.
  simpl in IHo.
  rewrite (IHo tl' e e' ed); trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY; simpl.
  intros tl' e e' ed.
  rewrite shiftL_shiftL_comm.
  assert (Hx := IHY (shiftL tl' e0 emp) (app e e0) e' ed).
  simplenv; simplpm.
  rewrite Hx; trivial.
  (* Y = e0 nil *)
  intros tl' e e' ed.
  simpl; trivial.
  (* Y = e0 (r, t) :: Y *)
  intros r t Y IHt IHY; simpl.
  intros tl' e e' ed.
  rewrite (IHt tl' e e' ed).
  rewrite (IHY tl' e e' ed).
  trivial.
  (* o = nil *)
  intros tl' e e' ed.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo; simpl.
  intros tl' e e' ed.
  rewrite (IHt tl' e e' ed).
  rewrite (IHo tl' e e' ed).
  trivial.
Qed.

Lemma instL_shift_emp :
  forall t tl' e e',
    instL (shift t e emp) (shiftL tl' e emp) e' (len e)
    = shift (instL t tl' e' 0) e emp.
Proof.
  intros t tl' e e'.
  assert (H := @instL_shift t tl' emp e e').
  simplenv; simplpm; trivial.
Qed.

Lemma instYL_shiftY_emp :
  forall Y tl' e e',
    instYL (shiftY Y e emp) (shiftL tl' e emp) e' (len e)
    = shiftY (instYL Y tl' e' 0) e emp.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite instL_shift_emp; trivial.
Qed.

Lemma shiftL_LinstL :
  forall tl' tl e ed,
    shiftL (LinstL tl' tl ed 0) e emp
    = LinstL (shiftL tl' e emp) (shiftL tl e emp) ed (len e).
Proof.
  intros tl'.
  induction tl' as [|(n', t') tl' IHtl']; simpl.
  intros tl e ed.
    trivial.
  intros tl e ed.
    rewrite (IHtl' tl e); trivial.
    rewrite instL_shift_emp; trivial.
Qed.

(* 

          |--e'--|--e---|--ex--|        
   |..e''.|--e'--|--e---|--ex--|
   |..e''.|--e'--|++e+++|--ex--| 
          |..e''.|--e'--|--ex--|

be equal to 

          |--e'--|--e---|--ex--|        
          |--e'--|++e+++|--ex--|
                 |--e'--|--ex--|
          |..e''.|--e'--|--ex--|
*)

Lemma instL_shift_2 :
  forall t tl ex e e' e'',
    instL (shift t e'' (app e' (app e ex)))
          (shiftL (shiftL tl ex emp) e'' (app e' ex))
          e (len ex)
    = shift (instL t (shiftL tl ex emp) e (len ex)) e'' (app e' ex).
Proof.
  intro t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl ex e e' e'',
        instYL (shiftY Y e'' (app e' (app e ex))) 
               (shiftL (shiftL tl ex emp) e'' (app e' ex))
               e (len ex) 
        = shiftY (instYL Y (shiftL tl ex emp) e (len ex)) e'' (app e' ex))
    (P1 := fun o => 
      forall tl ex e e' e'',
        instoL (shifto o e'' (app e' (app e ex))) 
               (shiftL (shiftL tl ex emp) e'' (app e' ex))
               e (len ex) 
        = shifto (instoL o (shiftL tl ex emp) e (len ex)) e'' (app e' ex));
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros tl' ex e e' e''.
  simplenv.
  decbltnat n0 (len ex);
    simpl; simplpm; simplbnat; simplenv; trivial.
    rewbnat (blt_nat n0 (len e' + len ex) = true).
    rewbnat (blt_nat n0 (len e' + (len e + len ex)) = true).
    simpl.
    simplbnat; trivial.
  decbltnat n0 (len e + len ex); 
    simpl; simplpm; simplbnat; simplenv; trivial.
    set (tlx := shiftL tl' ex emp).
    rewrite varmap_get2_shiftL_3; trivial.
  decbltnat n0 (len e' + (len e + len ex));
    simpl; simplpm; simplbnat; simplenv; trivial.
    rewbnat (blt_nat (n0 - len e) (len e' + len ex) = true).
    trivial.
  rewbnat (blt_nat (n0 + len e'') (len ex) = false).
  rewbnat (blt_nat (n0 + len e'') (len e + len ex) = false).
  rewbnat (blt_nat (n0 - len e) (len e' + len ex) = false).
  repbnat (n0 + len e'' - len e)
    with (n0 - len e + len e'').
  trivial.
  (* t = int *)
  intros tl ex e e' e''.
  trivial.
  (* t = tup o *)
  intros o IHo.
  intros tl ex e e' e''.
  simpl in IHo.
  rewrite (IHo tl ex e e' e''); trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY; simpl.
  intros tl ex e e' e''.
  assert (Hx := IHY tl (app ex e0) e e' e'').
  simplenv; simplpm.
  rewrite shiftL_shiftL_merge.
  rewrite shiftL_shiftL_comm.
  rewrite shiftL_shiftL_merge.
  simplenv.
  rewrite Hx; trivial.
  (* Y = e0 nil *)
  intros tl ex e e' e''.
  trivial.
  (* Y = e0 (r, t) :: Y *)
  intros r t Y IHt IHY.
  intros tl ex e e' e''.
  rewrite (IHt tl ex e e' e'').
  rewrite (IHY tl ex e e' e'').
  trivial.
  (* o = nil *)
  intros tl ex e e' e''.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl ex e e' e''.
  rewrite (IHt tl ex e e' e'').
  rewrite (IHo tl ex e e' e'').
  trivial.
Qed.

Lemma instL_shift_2_emp :
  forall t tl e e' e'',
    instL (shift t e'' (app e' e)) (shiftL tl e'' e') e 0
    = shift (instL t tl e 0) e'' e'.
Proof.
  intros t tl e e' e''.
  assert (Hx := instL_shift_2 t (shiftL tl emp emp) emp e e' e'').
  simplenv; simplpm.
  rewrite shiftL_emp in Hx.
  rewrite shiftL_emp in Hx.
  trivial.
Qed.

Lemma instYL_shiftY_2_emp :
  forall Y tl e e' e'',
    instYL (shiftY Y e'' (app e' e)) (shiftL tl e'' e') e 0
    = shiftY (instYL Y tl e 0) e'' e'.
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  rewrite IHY.
  rewrite instL_shift_2_emp.
  trivial.
Qed.

Lemma instoL_shifto_2_emp :
  forall o tl e e' e'',
    instoL (shifto o e'' (app e' e)) (shiftL tl e'' e') e 0
    = shifto (instoL o tl e 0) e'' e'.
Proof.
  induction o; try destruct p; simpl; intros; trivial.
  rewrite IHo.
  rewrite instL_shift_2_emp.
  trivial.
Qed.

(* 

          |..e'..|--e''-|--e'''|        
          |..e'..|++e''+|--e'''|        
                 |..e'..|--e'''|        
          |..e...|..e'..|--e'''|        

be equal to 

   |..e...|..e'..|--e''-|--e'''|        
   |..e...|..e'..|++e''+|--e'''|        
          |..e...|..e'..|--e'''|        

*)

Lemma shift_instL_shift_aux : forall t tl e e' e'' e''',
  shift (instL (shift t e' (app e'' e'''))
               (shiftL tl e''' emp) 
               e''
               (len e''')) e (app e' e''')
  = instL (shift t (app e e') (app e'' e'''))
          (shiftL (shiftL tl e''' emp) e (app e' e'''))
          e''
          (len e''').
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl e e' e'' e''',
        shiftY (instYL (shiftY Y e' (app e'' e''')) 
                     (shiftL tl e''' emp)
                     e''
                     (len e''')) e (app e' e''')
      = instYL (shiftY Y (app e e') (app e'' e'''))
              (shiftL (shiftL tl e''' emp) e (app e' e'''))
              e''
              (len e'''))
    (P1 := fun o =>
      forall tl e e' e'' e''',
        shifto (instoL (shifto o e' (app e'' e'''))
                     (shiftL tl e''' emp)
                     e''
                     (len e''')) e (app e' e''')
      = instoL (shifto o (app e e') (app e'' e'''))
              (shiftL (shiftL tl e''' emp) e (app e' e'''))
              e''
              (len e'''));
    clear t.
  (* t = var n0 *)
  intros n0; simpl.
  intros tl e e' e'' e'''.
  simplenv; simplpm; simplbnat.
  decbltnat n0 (len e'' + len e'''); 
    simpl; simplbnat; trivial.
    decbltnat n0 (len e'''); 
      simpl; simplbnat; simplenv; trivial.
      rewbnat (blt_nat n0 (len e' + len e''') = true).
      trivial.
     rewrite (@varmap_get2_shiftL_3 (shiftL tl e''' emp)
      e (app e' e''') (n0 - len e''')).
    trivial.
  rewbnat (blt_nat (n0 + len e') (len e''') = false).
  rewbnat (blt_nat (n0 + (len e + len e')) (len e''') = false).
  rewbnat (blt_nat (n0 + len e') (len e'' + len e''') = false).
  rewbnat (blt_nat (n0 + (len e + len e')) (len e'' + len e''') = false).
  simpl; simplenv.
  rewbnat (blt_nat (n0 + len e' - len e'') (len e' + len e''') = false).
  repbnat (n0 + len e' - len e'' + len e)
    with (n0 + (len e + len e') - len e'').
  trivial.
  (* t = int *)
  intros tl e e' e'' e'''; simpl.
  trivial.
  (* t = tup o *)
  intros o IHo; simpl.
  intros tl e e' e'' e'''.
  rewrite (IHo tl e e' e'' e''').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY; simpl.
  intros tl e e' e'' e'''.
  assert (H := IHY tl e e' e'' (app e''' e0)).
  simplenv.
  rewrite shiftL_shiftL_merge.
  rewrite shiftL_shiftL_comm.
  rewrite shiftL_shiftL_merge.
  simplenv.
  rewrite H; trivial.
  (* Y = e0 nil *)
  intros tl e e' e'' e'''; simpl.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY; simpl.
  intros tl e e' e'' e'''.
  rewrite (IHt tl e e' e'' e''').
  rewrite (IHY tl e e' e'' e''').
  trivial.
  (* o = nil *)
  intros tl e e' e'' e'''.
  simpl; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo; simpl.
  intros tl e e' e'' e'''.
  rewrite (IHt tl e e' e'' e''').
  rewrite (IHo tl e e' e'' e''').
  trivial.
Qed.

Lemma shift_instL_shift : forall t tl e e' e'',
  shift (instL (shift t e' e'') tl e'' 0) e e'
  = instL (shift t (app e e') e'') 
          (shiftL tl e e')
          e''
          0.
Proof.
  intros t tl e e' e''.
  assert (H := @shift_instL_shift_aux t tl e e' e'' emp).
  simplenv.
  rewrite shiftL_emp in H.
  auto.
Qed.

Lemma shiftY_instYL_shiftY : forall Y tl e e' e'',
  shiftY (instYL (shiftY Y e' e'') tl e'' 0) e e'
  = instYL (shiftY Y (app e e') e'') 
          (shiftL tl e e') 
          e''
          0.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite shift_instL_shift; trivial.
Qed.

(***************************************************************)

Lemma instL_dist_aux : 
  forall t tl tl' (e e' e'' : tenv),
    instL (instL t (shiftL tl' e'' emp) e' (len e''))
             (shiftL tl e'' emp)
             e
             (len e'')
     = instL (instL t 
                (shiftL (shiftL tl e'' emp) e' e'') e
                (len e' + len e''))
             (shiftL (LinstL tl' tl e 0) e'' emp)
             e'
             (len e'').
Proof.
  intro t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl tl' e e' e'',
        instYL (instYL Y
                  (shiftL tl' e'' emp) e' (len e''))
                  (shiftL tl e'' emp)
                  e
                  (len e'')
         = instYL (instYL Y 
                     (shiftL (shiftL tl e'' emp) e' e'')
                     e
                     (len e' + len e''))
                  (shiftL (LinstL tl' tl e 0) e'' emp)
                  e'
                  (len e''))
    (P1 := fun o => 
      forall tl tl' e e' e'',
         instoL (instoL o
                (shiftL tl' e'' emp) e' (len e''))
                  (shiftL tl e'' emp)
                  e
                  (len e'')
         = instoL (instoL o 
                     (shiftL (shiftL tl e'' emp) e' e'')
                     e
                     (len e' + len e''))
                  (shiftL (LinstL tl' tl e 0) e'' emp)
                  e'
                  (len e''));
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0; simpl.
  intros tl tl' e e' e''.
  decbltnat n0 (len e''); simpl; 
    simplbnat; simplpm; trivial.
  decbltnat n0 (len e' + len e''); 
    simpl; simplbnat; simplpm; trivial.
  (*  e' + e'' > n0 > e'' *)
    rewrite shiftL_LinstL; trivial.
    rewrite varmap_get2_LinstL; trivial.
  decbltnat n0 (len e + len e' + len e''); 
    simpl; simplpm; simplbnat; trivial.
    (*  e + e' + e'' > n0 > e' + e'' *)
    rewbnat (blt_nat (n0 - len e') (len e'') = false).
    rewbnat (blt_nat (n0 - len e') (len e + len e'') = true).
    rewrite (@varmap_get2_shiftL_3 (shiftL tl e'' emp)
      e' e'' (n0 - len e' - len e'')).
    set (tx := Varmap.get2 (shiftL tl e'' emp) 
      (valTy_int) (n0 - len e' - len e'')).
    rewrite instL_shift_unchange; trivial.
    (*  x + e + e' + e'' > n0 > e + e' + e'' *)
  simpl.
  rewbnat (blt_nat (n0 - len e') (len e'') = false).
  rewbnat (blt_nat (n0 - len e') (len e + len e'') = false).
  rewbnat (blt_nat (n0 - len e) (len e' + len e'') = false).
  rewbnat (blt_nat (n0 - len e) (len e'') = false).
  repbnat (n0 - len e' - len e)
    with (n0 - len e - len e'); trivial.
 (* t = int *)
  intros tl tl' e e' e''.
  trivial.
  (* t = tup o *)
  intros o IHo.
  intros tl tl' e e' e''.
  rewrite (IHo tl tl' e e' e'').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros tl tl' e e' e''.
  assert (H := IHY tl tl' e e' (app e'' e0)).
  simplenv; simplpm.
  rewrite shiftL_shiftL_merge.
  rewrite shiftL_shiftL_merge.
  rewrite shiftL_shiftL_merge.
  rewrite H; trivial.
  rewrite shiftL_shiftL_merge_cont_emp.
  rewrite shiftL_shiftL_merge_cont_emp.
  rewrite shiftL_shiftL_merge.
  simplenv; trivial.
  (* Y = nil *)
  intros tl tl' e e' e''.
  trivial.
  (* Y = (r,t) :: Y *)
  intros r t Y IHt IHY.
  intros tl tl' e e' e''.
  rewrite (IHt tl tl' e e' e'').
  rewrite (IHY tl tl' e e' e'').
  trivial.
  (* o = nil *)
  intros tl tl' e e' e''.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl tl' e e' e''.
  rewrite (IHt tl tl' e e' e'').
  rewrite (IHo tl tl' e e' e'').
  trivial.
Qed.

Lemma instL_dist : 
  forall t tl tl' (e e' : tenv),
    instL (instL t tl' e' 0) tl e 0
    = instL (instL t (shiftL tl e' emp) e (len e'))
            (LinstL tl' tl e 0) e' 0.
Proof.
  intros t tl tl' e e'.
  assert (H := @instL_dist_aux t tl tl' e e' emp). 
  simplenv; simplpm.
  rewrite shiftL_emp in H.
  rewrite shiftL_emp in H.
  rewrite shiftL_emp in H.
  simplenv; simplpm; trivial.
Qed.

Lemma instYL_dist : 
  forall Y tl tl' (e e' : tenv),
    instYL (instYL Y tl' e' 0) tl e 0
    = instYL (instYL Y (shiftL tl e' emp) e (len e'))
             (LinstL tl' tl e 0) e' 0.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite (IHY tl tl' e e').
  rewrite (@instL_dist t tl tl' e e').
  simplenv; simplpm; trivial.
Qed.

(***************************************************************)

Lemma varmap_bsubset_bcpactL_true :
  forall t tl tl' e' e'',
    Varmap.bsubset tl tl' (len e') emp = true
    -> bcpactL t tl e' (len e'') = true
    -> bcpactL t tl' e' (len e'') = true.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl tl' e' e'',
        Varmap.bsubset tl tl' (len e') emp = true
        -> bcpactYL Y tl e' (len e'') = true
        -> bcpactYL Y tl' e' (len e'') = true)
    (P1 := fun o =>
      forall tl tl' e' e'',
        Varmap.bsubset tl tl' (len e') emp = true
        -> bcpactoL o tl e' (len e'') = true
        -> bcpactoL o tl' e' (len e'') = true);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros tl tl' e' e'' Hsub Hcp.
  decbltnat n0 (len e''); trivial.
  decbltnat n0 (len e' + len e''); trivial.
  assert (H := @Varmap.bsubset_bindom (len e') 
    (n0 - len e'') tl tl' emp).
  simplenv; simplpm.
  apply H; trivial.
  solvebnat.
  (* t = int *)
  intros tl tl' e' e''.
  trivial.
  (* t = tup o *)
  intros o IHo .
  intros tl tl' e' e'' Hsub Hcp.
  rewrite (IHo tl tl' e' e'' Hsub Hcp).
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros tl tl' e' e'' Hsub Hcp.
  assert (H := IHY (shiftL tl e0 emp) (shiftL tl' e0 emp)
    e' (app e'' e0)).
  simplenv.
  rewrite H; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  (* Y = e0 nil *)
  intros tl tl' e' e'' Hsub Hcp.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros tl tl' e' e'' Hsub Hcp.
  destruct (andb_prop _ _ Hcp).
  rewrite (IHt tl tl' e' e'' Hsub H).
  rewrite (IHY tl tl' e' e'' Hsub H0).
  trivial.
  (* o = nil *)
  intros tl tl' e' e'' Hsub Hcp.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl tl' e' e'' Hsub Hcp.
  destruct (andb_prop _ _ Hcp).
  rewrite (IHt tl tl' e' e'' Hsub H).
  rewrite (IHo tl tl' e' e'' Hsub H0).
  trivial.
Qed.

Lemma varmap_bsubset_bcpactYL_true :
  forall Y tl tl' e' e'',
    Varmap.bsubset tl tl' (len e') emp = true
    -> bcpactYL Y tl e' (len e'') = true
    -> bcpactYL Y tl' e' (len e'') = true.
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  destruct (andb_prop _ _ H0).
  rewrite (IHY tl tl' e' e''); trivial.
  rewrite (@varmap_bsubset_bcpactL_true v tl); trivial.
Qed.

Lemma varmap_bsubset_bcpactoL_true :
  forall o tl tl' e' e'',
    Varmap.bsubset tl tl' (len e') emp = true
    -> bcpactoL o tl e' (len e'') = true
    -> bcpactoL o tl' e' (len e'') = true.
Proof.
  induction o; try destruct p; simpl; intros; trivial.
  destruct (andb_prop _ _ H0).
  rewrite (IHo tl tl' e' e''); trivial.
  rewrite (@varmap_bsubset_bcpactL_true v tl); trivial.
Qed.

Lemma varmap_bsubset_instL_unchange :
  forall t tl tl' e' e'',
    Varmap.bsubset tl tl' (len e') emp = true
    -> bcpactL t tl e' (len e'') = true
    -> instL t tl e' (len e'')
    = instL t tl' e' (len e'').
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall tl tl' e' e'',
        Varmap.bsubset tl tl' (len e') emp = true
        -> bcpactYL Y tl e' (len e'') = true
        -> instYL Y tl e' (len e'')
        = instYL Y tl' e' (len e''))
    (P1 := fun o =>
      forall tl tl' e' e'',
        Varmap.bsubset tl tl' (len e') emp = true
        -> bcpactoL o tl e' (len e'') = true
        -> instoL o tl e' (len e'')
        = instoL o tl' e' (len e''));
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros tl tl' e' e'' Hsub Hcp.
  decbltnat n0 (len e''); trivial.
  decbltnat n0 (len e' + len e''); trivial.
  assert (H := @Varmap.bsubset_sem (len e') 
    (n0 - len e'') tl tl' emp).
  simplenv; simplpm.
  apply H; trivial.
  solvebnat.
  (* t = int *)
  intros tl tl' e' e''.
  trivial.
  (* t = tup o *)
  intros o IHo .
  intros tl tl' e' e'' Hsub Hcp.
  rewrite (IHo tl tl' e' e'' Hsub Hcp).
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros tl tl' e' e'' Hsub Hcp.
  assert (H := IHY (shiftL tl e0 emp) (shiftL tl' e0 emp)
    e' (app e'' e0)).
  simplenv.
  rewrite H; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  (* Y = e0 nil *)
  intros tl tl' e' e'' Hsub Hcp.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros tl tl' e' e'' Hsub Hcp.
  destruct (andb_prop _ _ Hcp).
  rewrite (IHt tl tl' e' e'' Hsub H).
  rewrite (IHY tl tl' e' e'' Hsub H0).
  trivial.
  (* o = nil *)
  intros tl tl' e' e'' Hsub Hcp.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros tl tl' e' e'' Hsub Hcp.
  destruct (andb_prop _ _ Hcp).
  rewrite (IHt tl tl' e' e'' Hsub H).
  rewrite (IHo tl tl' e' e'' Hsub H0).
  trivial.
Qed.

Lemma varmap_bsubset_instYL_unchange :
  forall Y tl tl' e' e'',
    Varmap.bsubset tl tl' (len e') emp = true
    -> bcpactYL Y tl e' (len e'') = true
    -> instYL Y tl e' (len e'')
    = instYL Y tl' e' (len e'').
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  destruct (andb_prop _ _ H0).
  rewrite (IHY tl tl' e' e'' H H2).
  rewrite (varmap_bsubset_instL_unchange H H1).
  trivial.
Qed.

Lemma varmap_bsubset_instoL_unchange :
  forall o tl tl' e' e'',
    Varmap.bsubset tl tl' (len e') emp = true
    -> bcpactoL o tl e' (len e'') = true
    -> instoL o tl e' (len e'')
    = instoL o tl' e' (len e'').
Proof.
  induction o; try destruct p; simpl; intros; trivial.
  destruct (andb_prop _ _ H0).
  rewrite (IHo tl tl' e' e'' H H2).
  rewrite (varmap_bsubset_instL_unchange H H1).
  trivial.
Qed.

(***************************************************************)

Lemma RfTy_sel_some_shiftY :
  forall Y r t e e',
    RfTy.sel Y r = Some t
    -> RfTy.sel (shiftY Y e e') r
       = Some (shift t e e').
Proof.
  induction Y as [|(r0,t0) Y IHY]; simpl; intros.
    discriminate.
  decbeqreg r r0.
  inversion_clear H; trivial.
  apply IHY; trivial.
Qed.

Lemma RfTy_sel_none_shiftY :
  forall Y r e e',
    RfTy.sel Y r = None
    -> RfTy.sel (shiftY Y e e') r
       = None.
Proof.
  induction Y as [|(r0,t0) Y IHY]; simpl; intros.
    trivial.
  decbeqreg r r0.
  inversion_clear H; trivial.
  apply IHY; trivial.
Qed.

Lemma RfTy_sel_shiftY :
  forall Y Y' r e e',
    RfTy.sel Y r = RfTy.sel Y' r
    -> RfTy.sel (shiftY Y e e') r
       = RfTy.sel (shiftY Y' e e') r.
Proof.
  intros Y Y' r e e'.
  generalize (refl_equal (RfTy.sel Y' r)).
  pattern (RfTy.sel Y' r) at 2 3.
  case (RfTy.sel Y' r).
    intros t HY' HY. 
    rewrite (RfTy_sel_some_shiftY e e' HY).
    rewrite (RfTy_sel_some_shiftY e e' HY').
    trivial.
  intros HY' HY.
  rewrite (RfTy_sel_none_shiftY e e' HY'). 
  rewrite (RfTy_sel_none_shiftY e e' HY). 
  trivial.
Qed.

Lemma RfTy_sel_some_shiftY_rev :
  forall Y r t e e',
    RfTy.sel (shiftY Y e' e) r
           = Some (shift t e' e)  
    -> RfTy.sel Y r = Some t.
Proof.
  induction Y as [|(r0,t0) Y IHY]; simpl; intros.
    discriminate.
  decbeqreg r r0.
  injection H. intro H0.
  substH H0 with (inj_shift H0).
  subst t0; trivial.
  eapply IHY; eauto.
Qed.

Lemma RfTy_sel_none_shiftY_rev :
  forall Y r e e',
    RfTy.sel (shiftY Y e' e) r
       = None    
    -> RfTy.sel Y r = None.
Proof.
  induction Y as [|(r0,t0) Y IHY]; simpl; intros.
    trivial.
  decbeqreg r r0.
  inversion_clear H; trivial.
  eapply IHY; eauto.
Qed.

Lemma RfTy_sel_shiftY_rev :
  forall Y Y' r e e',
    RfTy.sel (shiftY Y e' e) r
       = RfTy.sel (shiftY Y' e' e) r    
    -> RfTy.sel Y r = RfTy.sel Y' r.
Proof.
  intros Y Y' r e e'.
  generalize (refl_equal (RfTy.sel Y' r)).
  pattern (RfTy.sel Y' r) at 2 3.
  case (RfTy.sel Y' r).
    intros t HY' HY.
    rewrite (RfTy_sel_some_shiftY e' e HY') in HY.
    rewrite (RfTy_sel_some_shiftY_rev HY).
    trivial.
  intros HY' HY.
  rewrite (RfTy_sel_none_shiftY e' e HY') in HY. 
  rewrite (RfTy_sel_none_shiftY_rev HY). 
  trivial.
Qed.

Lemma RfTy_sel_some_instYL :
  forall Y r t tl e n,
    RfTy.sel Y r = Some t
    -> RfTy.sel (instYL Y tl e n) r
       = Some (instL t tl e n).
Proof.
  induction Y as [|(r0,t0) Y IHY]; simpl; intros.
    discriminate.
  decbeqreg r r0.
  inversion_clear H; trivial.
  apply IHY; trivial.
Qed.

Lemma RfTy_sel_none_instYL :
  forall Y r tl e n,
    RfTy.sel Y r = None
    -> RfTy.sel (instYL Y tl e n) r
       = None.
Proof.
  induction Y as [|(r0,t0) Y IHY]; simpl; intros.
    trivial.
  decbeqreg r r0.
  inversion_clear H; trivial.
  apply IHY; trivial.
Qed.

Lemma RfTy_sel_instYL :
  forall Y Y' r tl e n,
    RfTy.sel Y r = RfTy.sel Y' r
    -> RfTy.sel (instYL Y tl e n) r
       = RfTy.sel (instYL Y' tl e n) r.
Proof.
  intros Y Y' r tl e n.
  generalize (refl_equal (RfTy.sel Y' r)).
  pattern (RfTy.sel Y' r) at 2 3.
  case (RfTy.sel Y' r).
    intros t HY' HY. 
    rewrite (RfTy_sel_some_instYL tl e n HY).
    rewrite (RfTy_sel_some_instYL tl e n HY').
    trivial.
  intros HY' HY.
  rewrite (RfTy_sel_none_instYL tl e n HY'). 
  rewrite (RfTy_sel_none_instYL tl e n HY). 
  trivial.
Qed.

Lemma RfTy_upd_shiftY : forall Y r t e e',
  RfTy.upd (shiftY Y e' e) r (shift t e' e)
  = shiftY (RfTy.upd Y r t) e' e.
Proof.
  intros Y r t e e'.
  induction Y; simpl; intros; auto.
  destruct p as [rx tx].
  simpl.
  decbeqreg r rx; simpl; trivial.
  rewrite IHY; trivial.
Qed.

Lemma RfTy_upd_instYL : forall Y r tl t e n,
  RfTy.upd (instYL Y tl e n) r (instL t tl e n)
  = instYL (RfTy.upd Y r t) tl e n.
Proof.
  intros Y r tl t e n.
  induction Y; simpl; intros; auto.
  destruct p as [rx tx].
  simpl.
  decbeqreg r rx; simpl; trivial.
  rewrite IHY; trivial.
Qed.

Lemma RfTy_upd_int_instYL : forall Y r tl e n,
  RfTy.upd (instYL Y tl e n) r valTy_int 
  = instYL (RfTy.upd Y r valTy_int) tl e n.
Proof.
  intros Y r tl n.
  induction Y; simpl; intros; auto.
  destruct p as [rx tx].
  simpl.
  decbeqreg r rx; simpl; trivial.
  rewbnat IHY; trivial.
Qed.

Lemma TupTy_lookup_shifto : 
  forall i o t flag e e',
    TupTy.lookup o i (flag, t)
    -> TupTy.lookup (shifto o e' e) i 
         (flag, shift t e' e).
Proof.
  double induction i o; simpl; intros; auto.
    inversion_clear H0.
    trivial.
  destruct p as [fx tx].
  auto.
Qed.

Lemma TupTy_lookup_instoL : 
  forall i o tl t flag e n,
    TupTy.lookup o i (flag, t)
    -> TupTy.lookup (instoL o tl e n) i 
         (flag, instL t tl e n).
Proof.
  double induction i o; simpl; intros; auto.
    inversion_clear H0.
    trivial.
  destruct p as [fx tx].
  auto.
Qed.

Lemma TupTy_upd_shifto : forall i o flag t e e',
  TupTy.upd (shifto o e' e) i (flag, shift t e' e)
  = shifto (TupTy.upd o i (flag, t)) e' e.
Proof.
  double induction i o; simpl; intros; auto.
  destruct p as [fx tx]; trivial.
  destruct p as [fx tx]; trivial.
  rewrite H0.
  trivial.
Qed.

Lemma TupTy_upd_instoL : forall i o flag t tl e n,
  TupTy.upd (instoL o tl e n) i (flag, instL t tl e n)
  = instoL (TupTy.upd o i (flag, t)) tl e n.
Proof.
  double induction i o; simpl; intros; auto.
  destruct p as [fx tx]; trivial.
  destruct p as [fx tx]; trivial.
  rewrite H0.
  trivial.
Qed.

Lemma RfTy_indom_instYL :
  forall Y r tl e n,
  RfTy.indom Y r
  -> RfTy.indom (instYL Y tl e n) r.
Proof.
  intros Y r tl e n Hin.
  induction Y as [| (r0, t0) Y IHY].
  simpl in Hin.
    trivial.
  unfold RfTy.indom in IHY,Hin |- * .
    simpl in * .
    decbeqreg r r0.
    exists (instL t0 tl e n); trivial.
    apply IHY.
    apply Hin.
Qed.

Lemma RfTy_indom_instYL_rev :
  forall Y r tl e n,
  RfTy.indom (instYL Y tl e n) r
  -> RfTy.indom Y r.
Proof.
  intros Y r tl e n Hin.
  induction Y as [| (r0, t0) Y IHY].
  simpl in Hin.
    trivial.
  unfold RfTy.indom in IHY,Hin |- * .
    simpl in * .
    decbeqreg r r0.
    exists t0; trivial.
    apply IHY.
    apply Hin.
Qed.

Lemma RfTy_indom_shiftY :
  forall Y r e e',
    RfTy.indom Y r
    -> RfTy.indom (shiftY Y e' e) r.
Proof.
  intros Y r e e' Hin.
  induction Y as [| (r0, t0) Y IHY].
  simpl in Hin.
    trivial.
  unfold RfTy.indom in IHY,Hin |- * .
    simpl in * .
    decbeqreg r r0.
    exists (shift t0 e' e); trivial.
    apply IHY.
    apply Hin.
Qed.

Lemma RfTy_indom_shiftY_rev :
  forall Y r e e',
  RfTy.indom (shiftY Y e' e) r
  -> RfTy.indom Y r.
Proof.
  intros Y r e e' Hin.
  induction Y as [| (r0, t0) Y IHY].
  simpl in Hin.
    trivial.
  unfold RfTy.indom in IHY,Hin |- * .
    simpl in * .
    decbeqreg r r0.
    exists t0; trivial.
    apply IHY.
    apply Hin.
Qed.

(* ****************************************************** *)
Lemma instL_idtl_aux :
  forall t e e',
    instL (shift t e (app e e')) 
          (shiftL (id_tl (len e)) e' emp) e (len e') 
    = t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y => 
      forall e e',
        instYL (shiftY Y e (app e e')) 
          (shiftL (id_tl (len e)) e' emp) e (len e') = Y)
    (P1 := fun o =>
      forall e e',
        instoL (shifto o e (app e e')) 
          (shiftL (id_tl (len e)) e' emp) e (len e') = o);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e'.
  simplenv.
  decbltnat n0 (len e + len e'); simpl;
    simplenv; simplbnat; simpl; simplpm; trivial.
    simplenv; simplbnat; simpl.
    decbltnat n0 (len e');
      simpl; simplbnat; simplpm; trivial.
    rewrite varmap_get2_shiftL_2.
    rewrite Varmap.get2_idtl.
    simpl; simplenv; simplbnat; simplpm.
    repbnat (n0 - len e' + len e') with n0.
    trivial.
    solvebnat.
    rewrite Varmap.bindom_idtl; trivial.
    solvebnat.
  rewbnat (blt_nat (n0 + len e) (len e') = false).
  rewbnat (blt_nat (n0 + len e) (len e + len e') = false).
  trivial.
  (* t = int *)
  intros e e'.
  trivial.
  (* t = tup o *)
  intros o IHo.
  intros e e'.
  rewrite (IHo e e').
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e'.
  assert(H := IHY e (app e' e0)).
  rewrite shiftL_shiftL_merge.
  simplenv.
  rewrite H.
  trivial.
  (* Y = nil *)
  intros e e'.
  trivial.
  (* Y = cons (r, t) Y *)
  intros r t Y IHt IHY.
  intros e e'.
  rewrite (IHt e e').
  rewrite (IHY e e').
  trivial.
  (* o = nil *)
  intros e e'.
  trivial.
  (* o = cons (f, t) o *)
  intros f t o IHt IHo. 
  intros e e'.
  rewrite (IHt e e').
  rewrite (IHo e e').
  trivial.
Qed.

Lemma instL_idtl :
  forall t e,
    instL (shift t e e) (id_tl (len e)) e 0 = t.
Proof.
  intros t e.
  assert (H := instL_idtl_aux t e emp).
  simplenv.
  rewrite shiftL_emp in H; trivial.
Qed.

Lemma instYL_idtl :
  forall Y e,
    instYL (shiftY Y e e) (id_tl (len e)) e 0 = Y.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  rewrite IHY; trivial.
  rewrite instL_idtl; trivial.
Qed.

Lemma varmap_bcpactL_idtl :
  forall t e e',
    bcpactL t (id_tl (len e)) e (len e') = true.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e',
        bcpactYL Y (id_tl (len e)) e (len e') 
        = true)
    (P1 := fun o =>
      forall e e',
        bcpactoL o (id_tl (len e)) e (len e') 
        = true);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e'.
  decbltnat n0 (len e'); trivial.
  decbltnat n0 (len e + len e'); trivial.
  apply Varmap.bindom_idtl; trivial.
  solvebnat.
  (* t = int *)
  intros e e'.
  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e e'.
  apply IHo.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e'.
  assert (H := IHY e (app e' e0)).
  simplenv; simplpm.
  rewrite bcpactYL_shiftL; trivial.
  (* Y = e0 nil *)
  intros e e'.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e'.
  rewrite (IHt e e').
  rewrite (IHY e e').
  trivial.
  (* o = nil *)
  intros e e'.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e'.
  rewrite (IHt e e').
  rewrite (IHo e e').
  trivial.
Qed.

Lemma varmap_bcpactYL_idtl :
  forall Y e e',
    bcpactYL Y (id_tl (len e)) e (len e') = true.
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  rewrite varmap_bcpactL_idtl.
  rewrite IHY.
  simpl; trivial.
Qed.

Unset Implicit Arguments.
