(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* - Type Cheker                                               *)
(***************************************************************)

(* $Id: TyCk.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Ty.
Require Import TyEnv.
Require Import Varmap.

Require Import Subst.
Require Import Subst_lemmas.
Require Import Unify.
Require Import Unify_lemmas.

Require Import Rule.
Require Import Tyok_lemmas.
Require Import SubTy_lemmas.
Require Import Rule_lemmas.

Require Import extac.

Import TyEnv.

Open Local Scope tm_scope.

(* *********************************************************** *)

Fixpoint valTyok_ck (e : tenv) (t : valTy) {struct t}: bool :=
  match t with
    | valTy_var n => TyEnv.bine e n
    | valTy_int => true
    | valTy_tup o => tupTyok_ck e o 
    | valTy_code e' Y => rfTyok_ck (TyEnv.app e e') Y
  end
with rfTyok_ck (e : tenv) (Y : rfTy) {struct Y} : bool :=
  match Y with
    | rfTy_nil => true
    | rfTy_cons (r, t) Y => andb (valTyok_ck e t) (rfTyok_ck e Y)
  end
with tupTyok_ck (e : tenv) (o : tupTy) {struct o}: bool :=
  match o with
    | tupTy_nil => true
    | tupTy_cons (f, t) o => andb (valTyok_ck e t) (tupTyok_ck e o)
  end.

Lemma valTyok_ck_sound :
  forall t e,
    valTyok_ck e t = true
    -> valTyok e t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e,
        rfTyok_ck e Y = true
        -> rfTyok e Y)
    (P1 := fun o =>
      forall e,
        tupTyok_ck e o = true
        -> tupTyok e o);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e H.
  constructor.
  apply TyEnv.bine_ine; trivial.
  (* t = int *)
  intros e H.
  constructor.
  (* t = tup o *)
  intros o IHo .
  intros e H.
  simpl.
  constructor; trivial.
  apply IHo; trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e H.
  simpl.
  constructor.
  apply IHY; trivial.
  (* Y = e0 nil *)
  intros e H.
  constructor.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e H.
  bool_destruct H as [Ht HY].
  constructor; auto.
  (* o = nil *)
  intros e H.
  constructor; trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e H.
  bool_destruct H as [Ht Ho].
  constructor; auto.
Qed.

Lemma rfTyok_ck_sound :
  forall Y e,
    rfTyok_ck e Y = true
    -> rfTyok e Y.
Proof.
  induction Y as [ | (r, t) Y IHY]; simpl.
    intros e H; constructor.
  intros e H.
  bool_destruct H as [H1 H2].
  constructor; auto.
  apply valTyok_ck_sound; trivial.
Qed.

Lemma tupTyok_ck_sound :
  forall o e,
    tupTyok_ck e o = true
    -> tupTyok e o.
Proof.
  induction o as [ | (r, t) o IHo]; simpl.
    intros e H; constructor.
  intros e H.
  bool_destruct H as [H1 H2].
  constructor; auto.
  apply valTyok_ck_sound; trivial.
Qed.

Definition codeTyok_ck (e : tenv) (a : codeTy) :=
  match a with
    | (e', Y') => rfTyok_ck (TyEnv.app e e') Y'
  end.

Lemma codeTyok_ck_sound :
  forall e a,
    codeTyok_ck e a = true
    -> codeTyok e a.
Proof.
  intros e [e' Y] H.
  unfold codeTyok.
  constructor.
  apply rfTyok_ck_sound; trivial.
Qed.

(* *********************************************************** *)

Definition codeTy_sub_ck (e : env) (a a' : codeTy) : bool :=
  codeTyok_ck e a && codeTyok_ck e a' &&
  match a, a' with
    | (e, Y), (e', Y') =>
      match RfTy.sort Y Y' with
        | Some Yx =>
          match unifyY Yx (shiftY Y' e e') e' emp with
            | Some tl => true
            | None => false
          end
        | None => false
      end
  end.

Lemma codeTy_sub_ck_sound :
  forall e a a',
    codeTy_sub_ck e a a' = true
    -> codeTy_sub e a a'.
Proof.
  intros e0 a a' H.
  destruct a as [e Y]; destruct a' as [e' Y'].
  unfold codeTy_sub_ck in H.
  bool_destruct H as [H' H].
  bool_destruct H' as [Ha Ha'].
  substH Ha with (codeTyok_ck_sound _ _ Ha).
  substH Ha' with (codeTyok_ck_sound _ _ Ha').
  destruct (RfTy.sort_dec Y Y') as [[Yx HY] | HY];
    rewrite HY in H; try discriminate H.
  set (Yy := shiftY Y' e e') in H.
  destruct (unifyY_dec Yx Yy e' emp) 
    as [[tl Hu] | Hu];
    rewrite Hu in H; try discriminate H.
  constructor 1 with tl; trivial.
  split.
    inversion_clear Ha.
    constructor; trivial.
  split.
    inversion_clear Ha'.
    constructor; trivial.
  split.
    eapply unifyY_valTyLok; eauto.
    simplenv.
    eapply RfTy_sort_rfTyok; eauto.
    inversion_clear Ha; trivial.
  subst Yy.
  rewrite (unifyY_sound Yx (shiftY Y' e e') e' tl Hu).
  inversion_clear Ha.
  eapply rfTy_sub_sort; eauto.
Qed.
(* *********************************************************** *)

Definition ceflag_ck (f f' : flag) : bool :=
  match f, f' with
    | junk, init => false
    | _, _ => true
  end.

Fixpoint tupok_ck (Psi : chspec) (Phi : dhspec) (l : word) (o : tupTy)
  {struct o} : bool :=
  match o with
    | tupTy_nil => true
    | tupTy_cons (f, t) o' =>
      match DHSpec.get Phi l with
        | Some (f', t') => ceflag_ck f' f && beq_valTy t t'
                        && tupok_ck Psi Phi (l + 1) o'
        | None => false
      end
  end.

Lemma tupok_ck_sound : 
  forall i o Psi Phi l,
    tupok_ck Psi Phi l o = true
    -> TupTy.indom o i
    -> exists ct, exists ct',   
      TupTy.lookup o i ct
      /\ DHSpec.lookup Phi (l+i) ct'
      /\ ceTy_sub ct' ct.
Proof.
  induction i; destruct o; simpl; intros.
    destruct H0.
    destruct (TupTy.lookup_nil _ _ H0).
    destruct p.
    destruct H0.
    simpl in H0.
    subst x.
    simplpm.
    destruct (DHSpec.get_dec Phi l) as [[[f' t'] Ht] | Ht];
      rewrite Ht in H; try discriminate.
    bool_destruct H as [H1 Ho].
    bool_destruct H1 as [Hc He].
    substH He with (beq_valTy_eq _ _ He).
    subst t'.
    exists (f, v).
    exists (f', v).
    split_l.
    split.
      apply DHSpec.get_lookup; trivial.
    destruct f'; destruct f; try constructor.
    simpl in Hc; discriminate.
    destruct H0.
    destruct (TupTy.lookup_nil _ _ H0).
  destruct p as [f t].
  destruct (DHSpec.get_dec Phi l) as [[[f' t'] Ht] | Ht];
    rewrite Ht in H; try discriminate.
  destruct (andb_prop _ _ H) as [H1 Ho].
  destruct (andb_prop _ _ H1) as [Hc He].
  substH He with (beq_valTy_eq _ _ He).
  subst t'.
  clear H H1.
  destruct H0.
  simpl in H.
  assert (Hin : TupTy.indom o i). 
    exists x; trivial.
  substH IHi with (IHi o Psi Phi (l + 1) Ho Hin).
  destruct IHi as [ct [ct' [Hct [Hct' Hce]]]].
  exists ct; exists ct'.
  split_l.
  simplpm.
  split_lr.
Qed.

Definition valok_ck (Psi : chspec) (Phi : dhspec) (w : word) (t : valTy) 
  : bool :=
  match t with 
    | valTy_var n0 => true
    | valTy_int => true
    | valTy_code e Y => 
      match CHSpec.get Psi w with
        | Some (e', Y') => codeTy_sub_ck TyEnv.emp (e, Y) (e', Y')
        | None => false
      end
    | valTy_tup o => tupok_ck Psi Phi w o
  end.

Lemma valok_ck_sound : 
  forall Psi Phi w t,
    valok_ck Psi Phi w t = true
    -> valok Psi Phi w t.
Proof.
  intros Psi Phi w t Hck.
  destruct t; try constructor.
    simpl in Hck.
    intros i H.
    apply tupok_ck_sound with Psi; trivial.
  unfold valok_ck in Hck.
  destruct (CHSpec.get_dec Psi w) as [[[e' Y'] Hs] | Hs];
    rewrite Hs in Hck;
    try discriminate.
  constructor 2 with (e', Y').
  apply CHSpec.get_lookup; trivial.
  apply codeTy_sub_ck_sound; trivial.
Qed.

Fixpoint wfpsi_ck (psi : chspec) : bool :=
  match psi with
    | nil => true
    | cons (f, (e, Y)) psi =>
        rfTy_bine Y e && wfpsi_ck psi
  end.

Lemma wfpsi_ck_sound : 
  forall psi, wfpsi_ck psi = true -> wfpsi psi.
Proof.
  induction psi; try destruct a as [f' a']; simpl; intros.
    intros f a Hin; inversion_clear Hin.
  destruct a' as [e' Y'].
  destruct (andb_prop _ _ H) as [Ha' Hpsi].
  intros f a Hin.
  unfold CHSpec.lookup, CHSpec.get, CHSpecDef.get in Hin.
  unfold CHSpecDef.beq in Hin.
  decbeqnat f f'.
  substbnat f'.
  inj_hyp Hin; subst a; trivial.
  constructor; simplenv.
  apply rfTy_bine_rfTyok; trivial.
  fold CHSpecDef.get in Hin.
  change CHSpecDef.get with CHSpec.get in Hin.
  substH Hin with (CHSpec.get_lookup Hin).
  apply (IHpsi Hpsi f a Hin).
Qed.

Fixpoint wfdh_ck (Psi : chspec) (Phi Phi' : dhspec) (H : heap) 
  {struct Phi'} : bool :=
  match Phi' with 
    | nil => true
    | cons (l, (f, t))  Phi' =>
      match Heap.get H l with
        | Some w => match f with
                      | init => valok_ck Psi Phi w t
                      | junk => true
                    end
                   && wfdh_ck Psi Phi Phi' H
        | None => false
      end
  end.

Lemma wfdh_ck_sound : 
  forall Psi Phi Phi' H,
    wfdh_ck Psi Phi Phi' H = true
    -> wfdh Psi Phi Phi' H.
Proof.
  intros Psi Phi Phi'.
  induction Phi'; simpl; intros.
    intros l ct Hin.
    inversion_clear Hin.
  destruct a as [l' [f' t']].
  intros l ct Hin.
  unfold DHSpec.lookup, DHSpec.get, DHSpecDef.get in Hin.
  unfold DHSpecDef.beq in Hin.
  decbeqnat l l'.
    inj_hyp Hin; subst ct.
    substbnat l'.
    destruct (Heap.get_dec H l) as [[w Hw] | Hw];
      rewrite Hw in H0; try discriminate.
    exists w.
    split.
      apply Heap.get_lookup; trivial.
    destruct (andb_prop _ _ H0) as [H1 H2].
    destruct f'; constructor.
    apply valok_ck_sound; trivial.
  fold DHSpecDef.get in Hin.
  change DHSpecDef.get with DHSpec.get in Hin.
  apply IHPhi'.
  destruct (Heap.get H l'); try discriminate.
  destruct (andb_prop _ _ H0) as [H1 H2].
  trivial.
  apply DHSpec.get_lookup; trivial.
Qed.

Fixpoint wfrf_ck (Psi : chspec) (Phi : dhspec) (Y : rfTy) (R : regfile) 
  {struct Y} : bool :=
  match Y with 
    | rfTy_nil => true
    | rfTy_cons (r, t) Y' => valok_ck Psi Phi (gR R r) t
                          && wfrf_ck Psi Phi Y' R
  end.

Lemma wfrf_ck_sound : 
  forall Psi Phi Y R,
    wfrf_ck Psi Phi Y R = true
    -> wfrf Psi Phi Y R.
Proof.
  intros Psi Phi Y.
  induction Y; simpl; intros R H.
    unfold wfrf; simpl; intros.
    discriminate.
  destruct p as [r t].
  unfold wfrf; simpl; intros.
  destruct (andb_prop _ _ H) as [H1 H2].
  decbeqreg r0 r.
    inj_hyp H0; subst t0; trivial.
    apply valok_ck_sound; trivial.
  apply IHY; trivial.
Qed.

Definition wfst_ck (Psi : chspec) (Phi : dhspec) 
  (a : codeTy) (S : state) : bool :=
  match a, S with
    | (e, Y), mkst H R pc =>
      wfdh_ck Psi Phi Phi H 
      && wfrf_ck Psi Phi (instYL Y Varmap.emp e 0) R
      && rfTy_bine Y e
  end.

Lemma wfst_ck_sound :
  forall Psi Phi a S,
    wfst_ck Psi Phi a S = true
    -> wfst Psi a S.
Proof.
  intros Psi Phi [e Y] [H R pc] Hck.
  exists Phi.
  destruct (andb_prop _ _ Hck) as [Hck' Hok].
  destruct (andb_prop _ _ Hck') as [Hdh Hrf].
  clear Hck Hck'.
  split.
    apply wfdh_ck_sound; trivial.
  exists (instYL Y Varmap.emp e 0).
  split.
    apply codeTy_sub_instYL.
    constructor.
    apply codeTy_sub_refl; trivial.
    substH Hok with (rfTy_bine_rfTyok Hok).
    constructor; simplenv; trivial.
  apply wfrf_ck_sound; trivial.
Qed.

(***************************************************************)

Fixpoint wfis_ck (psi : chspec) (a : codeTy) (is : instrseq) 
  {struct is} : bool :=
  codeTyok_ck emp a
  &&
  match a with
    | (e, Y) => 
      match is with
        | ione (j f) =>
          match (CHSpec.get psi f) with
            | Some a'' => codeTy_sub_ck TyEnv.emp (e, Y) a''
            | _ => false
          end
          
        | ione (jal f fret) =>
          match (CHSpec.get psi fret), (CHSpec.get psi f) with
            | Some (e', Y'), Some a'' => 
                codeTyok_ck TyEnv.emp (e', Y') 
                && codeTy_sub_ck TyEnv.emp (e, RfTy.upd Y ra (valTy_code e' Y')) a''
            | _, _ => false
          end
          
        | ione (jr r) =>
          match (RfTy.sel Y r) with
            | Some (valTy_code e'' Y'') => 
                codeTy_sub_ck e (emp, Y) (e'', Y'')
            | _ => false
          end
          
        | iseq (addu rd rs rt) is' =>
          match (beq_reg rd r0) with
            | false => 
              match (RfTy.sel Y rs, RfTy.sel Y rt) with
                | (Some valTy_int, Some valTy_int) =>
                  let Y' := (RfTy.upd Y rd valTy_int) in
                    wfis_ck psi (e, Y') is'
                | _ => false
              end
            | true => false
          end
          
        | iseq (addiu rd rs w) is' => 
          match (beq_reg rd r0) with
            | false => 
              match RfTy.sel Y rs with
                | Some valTy_int => 
                  let Y' := (RfTy.upd Y rd valTy_int) in
                    wfis_ck psi (e, Y') is'
                | _ => false
              end
            | true => false
          end

        | iseq (subu rd rs rt) is' => 
          match beq_reg rd r0 with
            | false => 
              match (RfTy.sel Y rs, RfTy.sel Y rt) with
                | (Some valTy_int, Some valTy_int) =>
                  let Y' := (RfTy.upd Y rd valTy_int) in
                    wfis_ck psi (e, Y') is'
                | _ => false
              end
            | _ => false
          end
          
        | iseq (subiu rd rs w) is' => 
          match beq_reg rd r0 with
            | false => 
              match RfTy.sel Y rs with
                | Some valTy_int =>
                  let Y' := (RfTy.upd Y rd valTy_int) in
                    wfis_ck psi (e, Y') is'
                | _ => false
              end
            | _ => false
          end
          
        | iseq (move rd rs) is' =>
          match (beq_reg rd r0) with
            | false => 
              match RfTy.sel Y rs with 
                | Some t => 
                  let Y' := (RfTy.upd Y rd t) in
                    wfis_ck psi (e, Y') is'
                | _ => false
              end
            | _ => false 
          end

        | iseq (li rd w) is' =>
          match (beq_reg rd r0) with
            | false =>
              match (CHSpec.get psi w) with
                | Some (e'', Y'') =>
                  codeTyok_ck TyEnv.emp (e'', Y'')
                  &&
                  let Y' := (RfTy.upd Y rd (valTy_code e'' Y'')) in
                    wfis_ck psi (e, Y') is'
                | _ => false
              end
              || 
              (let Y' := (RfTy.upd Y rd valTy_int) in
                wfis_ck psi (e, Y') is')
            | _ => false 
          end
          
        | iseq (lw rt w rs) is' => 
          match beq_reg rt r0 with
            | false => 
              match RfTy.sel Y rs with
                | Some (valTy_tup o) =>
                  match (TupTy.get o w) with 
                    | Some (init, t) =>
                      let Y' := (RfTy.upd Y rt t) in
                        wfis_ck psi (e, Y') is'
                    | _ => false
                  end
                | _ => false
              end
            | _ => false 
          end

        | iseq (sw rt w rs) is' => 
          match (beq_reg rs r0) with
            | false => 
              match (RfTy.sel Y rs, RfTy.sel Y rt) with
                | (Some (valTy_tup o), Some t) =>
                  match (TupTy.get o w) with 
                    | Some (f, t') => 
                      (beq_valTy t' t) && 
                      let Y' := RfTy.upd Y rs 
                        (valTy_tup (TupTy.upd o w (init, t))) in
                        wfis_ck psi (e, Y') is'
                    | _ => false
                  end
                | _ => false
              end
            | _ => false 
          end
          
        | iseq (blt rs rt f) is' => 
          match CHSpec.get psi f with
            | Some a'' => 
              match (RfTy.sel Y rs, RfTy.sel Y rt) with
                | (Some valTy_int, Some valTy_int) => 
                  (codeTy_sub_ck TyEnv.emp (e, Y) a'')
                  && wfis_ck psi (e, Y) is'
                | _ => false
              end
            | _ => false 
          end
          
        | iseq (beq rs rt f) is' => 
          match CHSpec.get psi f with
            | Some a'' => 
              match (RfTy.sel Y rs, RfTy.sel Y rt) with
                | (Some valTy_int, Some valTy_int) => 
                  (codeTy_sub_ck TyEnv.emp (e, Y) a'')
                  && wfis_ck psi (e, Y) is'
                | _ => false
              end
            | _ => false 
          end
          
        | iseq (bgtz rs f) is' => 
          match CHSpec.get psi f with
            | Some a'' => 
              match RfTy.sel Y rs with
                | Some valTy_int => 
                  codeTy_sub_ck TyEnv.emp (e, Y) a''
                  && wfis_ck psi (e, Y) is'
                | _ => false
            end
            | _ => false 
          end
        | _ => false
      end
  end.

Theorem wfis_ck_sound : forall psi is a,
  wfis_ck psi a is = true
  -> wfis psi a is.
Proof.
  intros psi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r]; 
    unfold wfis_ck; fold wfis_ck;
    intros [e Y] Hck;
    bool_destruct Hck as [Hok Hck];
    try (simpl in Hck; discriminate).
  (* case i = j f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try discriminate;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in Hck].
  substH Hf with (CHSpec.get_lookup Hf).
  destruct a'' as [e'' Y''].
  unfold wfis.
  exists (e'', Y'').
  split_lr.
  apply codeTy_sub_ck_sound; trivial.
  (* case i = jal f fret *)
  assert (Hfret : exists a', 
    CHSpec.get psi fret = Some a');
    [destruct (CHSpec.get psi fret) as [a' |] ;
      try discriminate;
        exists a'; trivial
      | destruct Hfret as [a' Hfret];
        rewrite Hfret in Hck].
  destruct a' as [e' Y'].
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try discriminate;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in Hck].
  substH Hfret with (CHSpec.get_lookup Hfret).
  substH Hf with (CHSpec.get_lookup Hf).
  bool_destruct Hck as [Ha'ok Hck].
  destruct a'' as [e'' Y''].
  unfold wfis.
  exists (e', Y'); split_l. 
  split_l by (apply codeTyok_ck_sound; trivial).
  exists (e'', Y''); split_l.
  apply codeTy_sub_ck_sound; trivial.
  (* case i = jr r *)
  assert (Hr : exists t, 
    RfTy.sel Y r = Some t);
    [destruct (RfTy.sel Y r) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hr as [t Hr];
        rewrite Hr in Hck].
  destruct t as [ | | o | e'' Y'' ]; 
    try discriminate.
  unfold wfis.
  exists (e'', Y'').
  split.
    trivial;
    try constructor; trivial.
    eapply RfTy_sel_valTyok; eauto.
    substH Hok with (codeTyok_ck_sound _ _ Hok).
    inversion_clear Hok.
    simplenv; trivial.
  apply codeTy_sub_ck_sound; trivial.
  (* case is = iseq i is' *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f | r]; 
    unfold wfis_ck; fold wfis_ck;
    intros [e Y] Hck;
    bool_destruct Hck as [Hok Hck];
    try (simpl in Hck; discriminate).
  (* case i = addu rd rs rt *)
  decbeqreg rd r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
    try discriminate;
      exists t; trivial
    | destruct Hrs as [t Hrs];
      rewrite Hrs in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  assert (Hrt : exists t, 
    RfTy.sel Y rt = Some t);
    [destruct (RfTy.sel Y rt) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  unfold wfis; fold wfis.
  split_l.
  split_l by (constructor; trivial; constructor).
  split_l by (constructor; trivial; constructor).
  eapply IHis; eauto.
  (* case i = addiu rd rs w *)
  decbeqreg rd r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  split_l.
  split_l by (constructor; trivial; constructor).
  eapply IHis; eauto.
  (* case i = subu rd rs rt *)
  decbeqreg rd r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  assert (Hrt : exists t, 
    RfTy.sel Y rt = Some t);
    [destruct (RfTy.sel Y rt) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  split_l.
  split_l by (constructor; trivial; constructor).
  split_l by (constructor; trivial; constructor).
  eapply IHis; eauto.
  (* case i = subiu rd rs w *)
  decbeqreg rd r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  split_l.
  split_l by (constructor; trivial; constructor).
  eapply IHis; eauto.
  (* case i = move rd rs *)
  decbeqreg rd r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  unfold wfis; fold wfis.
  exists t.
  split_l.
  split.
    constructor; trivial.
    eapply RfTy_sel_valTyok; eauto.
    substH Hok with (codeTyok_ck_sound _ _ Hok).
    inversion_clear Hok.
    simplenv; trivial.
  eapply IHis; eauto.
  (* case i = li rd w *)
  decbeqreg rd r0.
  bool_destruct Hck as [Hck | Hck].
    assert (Hw : exists a'',CHSpec.get psi w = Some a'').
      destruct (CHSpec.get psi w) as [a'' |];
        try discriminate;
        exists a''; trivial.
    destruct Hw as [a'' Hw];
      rewrite Hw in Hck.
    destruct a'' as [e'' Y''].
    bool_destruct Hck as [Hok' Hck].
    unfold wfis.
    split_l.
    exists (valTy_code e'' Y'').
    split.
      constructor; trivial.
      apply codeTyok_ck_sound; trivial.
    fold wfis.
    eapply IHis; eauto.
  split_l.
  exists valTy_int.
  split_l by (constructor; trivial).
  eapply IHis; eauto.
  (* case i = lw rt, w, rs *)
  decbeqreg rt r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | e'' Y'']; try discriminate.
  assert (Hw : exists ct, 
    TupTy.get o w = Some ct);
    [destruct (TupTy.get o w) as [ct |] ;
      try discriminate;
        exists ct; trivial
      | destruct Hw as [ct Hw];
        rewrite Hw in Hck].
  destruct ct as [f t]; destruct f; try discriminate.
  exists o; exists t.
  split_l.
  split.
    constructor; trivial.
    eapply RfTy_sel_valTyok; eauto.
    substH Hok with (codeTyok_ck_sound _ _ Hok).
    inversion_clear Hok; simplenv; trivial.
  split_l by (apply TupTy.get_lookup; trivial).
  eapply IHis; eauto.
  (* case i = sw rt w rs *)
  decbeqreg rs r0.
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | ax ]; try discriminate.
  assert (Hrt : exists t, 
    RfTy.sel Y rt = Some t);
    [destruct (RfTy.sel Y rt) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in Hck].
  assert (Hw : exists ct, 
    TupTy.get o w = Some ct);
    [destruct (TupTy.get o w) as [ct |] ;
    try discriminate;
      exists ct; trivial
      | destruct Hw as [ct Hw];
        rewrite Hw in Hck].
  destruct ct as [f t'](* ; destruct f *).
  bool_destruct Hck as [Heq Hck].
  substH Heq with (beq_valTy_eq _ _ Heq).
  subst t'.
  exists o; exists t; exists f.
  split_l.
  assert (HrfTyok : rfTyok e Y).
    substH Hok with (codeTyok_ck_sound _ _ Hok).
    inversion_clear Hok; simplenv; trivial.
  split.
    constructor; trivial.
    eapply RfTy_sel_valTyok; eauto.
  split.
    constructor; trivial.
    eapply RfTy_sel_valTyok; eauto.
  split.
    apply TupTy.get_lookup; trivial.
  split.
    destruct f; constructor; trivial.
  eapply IHis; eauto.
  (* case i = blt rs rt f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try discriminate;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in Hck].
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | ax ]; try discriminate.
  assert (Hrt : exists t, 
    RfTy.sel Y rt = Some t);
    [destruct (RfTy.sel Y rt) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in Hck].
  destruct t as [ | | o | ax ]; try discriminate.
  bool_destruct Hck as [Hsub Hck].
  substH Hf with (CHSpec.get_lookup Hf).
  destruct a'' as [e'' Y''].
  exists (e'', Y'').
  split_l.
  split_l by (constructor; trivial; constructor).
  split_l by (constructor; trivial; constructor).
  split_l by (apply codeTy_sub_ck_sound; trivial).
  eapply IHis; eauto.
  (* case i = beq rs rt f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try discriminate;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in Hck].
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | ax ]; try discriminate.
  assert (Hrt : exists t, 
    RfTy.sel Y rt = Some t);
    [destruct (RfTy.sel Y rt) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrt as [t Hrt];
        rewrite Hrt in Hck].
  destruct t as [ | | o | ax ]; try discriminate.
  bool_destruct Hck as [Hsub Hck].
  substH Hf with (CHSpec.get_lookup Hf).
  destruct a'' as [e'' Y''].
  exists (e'', Y'').
  split_l.
  split_l by (constructor; trivial; constructor).
  split_l by (constructor; trivial; constructor).
  split_l by (apply codeTy_sub_ck_sound; trivial).
  eapply IHis; eauto.
  (* case i = bgtz rs f *)
  assert (Hf : exists a'', 
    CHSpec.get psi f = Some a'');
    [destruct (CHSpec.get psi f) as [a'' |] ;
      try discriminate;
        exists a''; trivial
      | destruct Hf as [a'' Hf];
        rewrite Hf in Hck].
  assert (Hrs : exists t, 
    RfTy.sel Y rs = Some t);
    [destruct (RfTy.sel Y rs) as [t |] ;
      try discriminate;
        exists t; trivial
      | destruct Hrs as [t Hrs];
        rewrite Hrs in Hck].
  destruct t as [ | | o | ax ]; try discriminate.
  bool_destruct Hck as [Hsub Hck].
  substH Hf with (CHSpec.get_lookup Hf).
  destruct a'' as [e'' Y''].
  exists (e'', Y'').
  split_l.
  split_l by (constructor; trivial; constructor).
  split_l by (apply codeTy_sub_ck_sound; trivial).
  eapply IHis; eauto.
Qed.

Fixpoint wfch_ck (psi : chspec) (psi' : chspec) (ch : codeheap)
  {struct psi'} : bool :=
  match psi' with 
    | nil => true
    | cons (f, (e, Y)) psi' =>
      match (getis ch f) with
        | Some is => 
            match wfis_ck psi (e, Y) is with 
              | true => wfch_ck psi psi' ch
              | false => false
            end
        | None => false
      end
  end.

Lemma wfch_ck_sound_0 : 
  forall psi psi' ch,
    wfpsi_ck psi' = true
    -> wfch_ck psi psi' ch = true
    -> forall (f : label) (a : codeTy),
      CHSpec.lookup psi' f a
      -> (exists is,
        getis ch f = Some is
        /\ wfis psi a is).
Proof.
  intros psi psi'; generalize psi; clear psi.
  induction psi'.
    intros.
    inversion_clear H1.
  destruct a as [f' [e' Y']].
  intros psi ch Hpsi' Hwfch f a Hin.
  simpl in Hwfch.
  unfold CHSpec.lookup, CHSpec.get, CHSpecDef.get in Hin.
  unfold CHSpecDef.beq in Hin. 
  fold CHSpecDef.get in Hin.
  decbeqnat f f'.
    inj_hyp Hin; intros; subst.
    substbnat f'.
    unfold get.
    destruct (getis ch f); try discriminate.
    exists i.
    split_l.
    generalize (refl_equal (wfis_ck psi (e', Y') i)).
    pattern i at 1.
    elim (wfis_ck psi (e', Y') i);
      intros H1; rewrite H1 in Hwfch;
      try discriminate.
    apply wfis_ck_sound; trivial.
  unfold wfpsi_ck in Hpsi'; fold wfpsi_ck in Hpsi'.
  bool_destruct Hpsi' as [Ha' Hpsi'].
  apply IHpsi'; trivial.
  destruct (getis ch f'); try discriminate.
  destruct (wfis_ck psi (e', Y') i); try discriminate.
  trivial.
Qed.


Theorem wfch_ck_sound :
  forall psi psi' ch,
    wfpsi_ck psi' = true
    -> wfch_ck psi psi' ch = true
    -> wfch psi psi' ch.
Proof.
  intros psi psi' ch Hpsi' Hwfch.
  intros f a Hin.
  assert (H := wfch_ck_sound_0 psi psi'
    ch Hpsi' Hwfch f a Hin).
  destruct H as [is [Hch2 His]].
  exists is; split_lr.
Qed.

Theorem wfch_ck_sound_2 :
  forall psi ch,
    wfpsi_ck psi = true
    -> wfch_ck psi psi ch = true
    -> wfch psi psi ch.
Proof.
  intros psi ch Hpsi Hwfch.
  intros f a Hin.
  assert (H := wfch_ck_sound_0 psi psi 
    ch Hpsi Hwfch f a Hin).
  destruct H as [is [Hch His]].
  exists is; split_lr.
Qed.

(* *********************************************************** *)
(*
Ltac typeck_is :=
  match goal with 
    | |- wfis ?psi ?a ?is =>
      apply wfis_ck_sound;
        compute;
          simpl;
            match goal with
              | |- false = true => fail 1 "Ill-formed type."
              | _ => reflexivity
            end
    | _ => fail 1 "The goal is not a TAL instrseq judgement."
  end.*)

Ltac tyck_ch ch :=
  match goal with 
    | |- wfch ?psi ?psi' ?ch =>
      apply wfch_ck_sound;
        [compute; reflexivity 
          | compute; reflexivity];
        fail 1 "Ill-formed type."
    | _ => fail 1 "The goal is not a TAL instrseq judgement."
  end.

(* *********************************************************** *)
