(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Tyok_lemmas.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Ty.
Require Import Varmap.
Require Import TyEnv.
Require Import Rule.
Require Import Subst.
Require Import Subst_lemmas.

Require Import extac.


(* Lemmas *)

Set Implicit Arguments.
Unset Strict Implicit.

Import TyEnv.

Lemma valTy_bine_valTyok :
  forall t e,
    valTy_bine t e = true 
    -> valTyok e t.
Proof.
  intros t.
  elim t using valTy_ind3 with 
    (P0 := fun (Y : rfTy) =>
      forall e, 
        rfTy_bine Y e = true
        -> rfTyok e Y) 
    (P1 := fun (o : tupTy) =>
      forall e, 
        tupTy_bine o e = true
        -> tupTyok e o);
    simpl;
    clear t.
(* case t := valty_var n *)
  intros n0. 
  intros e Hine.
  constructor.
  unfold ine.
  trivial.
(* case t := valTy_int *)
  intros e Hine.
  constructor.
(* case t := valTy_tup o *)
  intros o IHo.
  intros e Hine.
  constructor.
  eapply IHo; eauto.
(* case t := valTy_code e Y *)
  intros e0 Y IHY.
  intros e Hine.
  constructor.
  eapply IHY; trivial.
(* case Y := rfty_nil *)
  intros e Hine.
  constructor.
(* case Y := rfTy_cons (r, t) Y *)
  intros r t Y IHt IHY. 
  intros e Hine.
  destruct (andb_prop _ _ Hine).
  constructor.
  eapply IHt; eauto.
  eapply IHY; eauto.
(* case o := tupty_nil *)
  intros e Hine.
  constructor.
(* case o := tupTy_cons (f, t) o *)
  intros f t o IHt IHo. 
  intros e Hine.
  destruct (andb_prop _ _ Hine).
  constructor.
  eapply IHt; eauto.
  eapply IHo; eauto.
Qed.

Lemma rfTy_bine_rfTyok :
  forall Y e,
    rfTy_bine Y e = true 
    -> rfTyok e Y.
Proof.
  induction Y; try destruct p as [r t]; simpl;
    intros; constructor.
  destruct (andb_prop _ _ H) as [Hine HY].
  apply valTy_bine_valTyok; trivial.
  destruct (andb_prop _ _ H) as [Hine HY].
  apply IHY; trivial.
Qed.

Lemma tupTy_bine_tupTyok :
  forall o e,
    tupTy_bine o e = true 
    -> tupTyok e o.
Proof.
  induction o; try destruct p as [f t]; simpl;
    intros; constructor.
  destruct (andb_prop _ _ H) as [Hine Ho].
  apply valTy_bine_valTyok; trivial.
  destruct (andb_prop _ _ H) as [Hine Ho].
  apply IHo; trivial.
Qed.

Lemma rfTy_bine_codeTyok :
  forall Y e,
    rfTy_bine Y e = true 
    -> codeTyok emp (e, Y).
Proof.
  intros.
  constructor; simplenv.
  apply rfTy_bine_rfTyok; trivial.
Qed.

(***************************************************************)

Lemma valTyok_weaken :
  forall t e e',
    valTyok e t
    -> valTyok (app e' e) t.
Proof.
  intros t.
  elim t using valTy_ind3 with 
    (P0 := fun (rf : rfTy) =>
      forall e e', 
        rfTyok e rf 
        -> rfTyok (app e' e) rf) 
    (P1 := fun (o : tupTy) =>
      forall e e', 
        tupTyok e o
        -> tupTyok (app e' e) o);
    clear t.
(* case t := valty_var n *)
  intros n e e' H.
  apply valTyok_var.
  inversion_clear H.
  apply ine_weaken_l; trivial.
(* case t := valTy_int *)
  intros e e' H.
  apply valTyok_int.
(* case t := valTy_tup o *)
  intros o IH e e' H.
  apply valTyok_tup.
  inversion_clear H.
  eapply IH; eauto.
(* case t := valTy_code e Y *)
  intros e0 Y IH.
  intros e e' H.
  apply valTyok_code.
  inversion_clear H.
  assert(Hx := IH (app e e0) e' H0).
  simplenv; trivial.
(* case Y := rfty_nil *)
  intros e e' H.
  apply rfTyok_nil.
(* case Y := rfTy_cons (r, t) Y *)
  intros r t Y IH IH0 e e' H.
  inversion_clear H.
  apply rfTyok_cons.
  eapply IH; eauto.
  eapply IH0; eauto.
(* case o := tupty_nil *)
  intros e e' H.
  apply tupTyok_nil.
(* case o := tupTy_cons (f, t) o *)
  intros f t o IH IH0 e e' H.
  inversion_clear H.
  apply tupTyok_cons.
  eapply IH; eauto.
  eapply IH0; eauto.
Qed.

Lemma tupTyok_weaken :
  forall o e e',
    tupTyok e o
    -> tupTyok (app e' e) o.
Proof.
  intros o e e' H.
  induction o.
  apply tupTyok_nil.
  destruct p.
  inversion_clear H.
  apply tupTyok_cons.
  apply valTyok_weaken; trivial.
  auto.
Qed.

Lemma rfTyok_weaken :
  forall Y e e',
    rfTyok e Y
    -> rfTyok (app e' e) Y.
Proof.
  intros Y e e' H.
  induction Y.
  apply rfTyok_nil.
  destruct p.
  inversion_clear H.
  apply rfTyok_cons.
  apply valTyok_weaken; trivial.
  auto.
Qed.

Lemma valTyok_weaken_emp :
  forall t e,
    valTyok emp t
    -> valTyok e t.
Proof.
  intros t e Ht.
  assert (H := @valTyok_weaken t emp e).
  simplenv.
  auto.
Qed.

Lemma rfTyok_weaken_emp :
  forall Y e,
    rfTyok emp Y
    -> rfTyok e Y.
Proof.
  induction Y; try destruct p; simpl; intros e H; 
    constructor;
    inversion_clear H; auto.
  apply valTyok_weaken_emp; trivial.
Qed.

Lemma tupTyok_weaken_emp :
  forall o e,
    tupTyok emp o
    -> tupTyok e o.
Proof.
  induction o; try destruct p; simpl; intros e H; 
    constructor;
    inversion_clear H; auto.
  apply valTyok_weaken_emp; trivial.
Qed.

Lemma valTyLok_weaken :
  forall tl e e',
    valTyLok e tl
    -> valTyLok (TyEnv.app e' e) tl.
Proof.
  induction tl; simpl; intros.
  apply valTyLok_vO.
  inversion_clear H.
  apply valTyLok_vS.
  eapply valTyok_weaken; eauto. 
  apply IHtl; auto.
Qed.

(***************************************************************)

Lemma valTyok_shift_aux :
  forall t e e' e'',
    valTyok (app e' e) t
    -> valTyok (app e' (app e'' e)) 
               (shift t e'' e).
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y => 
      forall e e' e'',
        rfTyok (app e' e) Y 
        -> rfTyok (app e' (app e'' e)) (shiftY Y e'' e))
    (P1 := fun o =>
      forall e e' e'',
        tupTyok (app e' e) o
        -> tupTyok (app e' (app e'' e)) (shifto o e'' e));
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' e'' H.
  inversion_clear H.
  simpl.
  simplenv.
  decbltnat n0 (len e); trivial.
    apply valTyok_var.
    apply TyEnv.blt_ine.
    inv_ine H0.
    simplenv.
    solvebnat.
  apply valTyok_var.
  apply TyEnv.blt_ine.
  inv_ine H0.
  simplenv.
  solvebnat.
  (* t = int *)
  intros e e' e'' H.
  simpl.
  apply valTyok_int.
  (* t = tup o *)
  intros o IHo.
  intros e e' e'' H.
  simpl.
  apply valTyok_tup.
  inversion_clear H.
  apply IHo; trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' e'' H.
  inversion_clear H.
  simpl.
  apply valTyok_code.
  assert (Hx := IHY (app e e0) e' e'').
  simplenv.
  auto.
  (* Y = nil *)
  intros e e' e'' H.
  simpl;  apply rfTyok_nil.
  (* Y = cons (r, t) Y *)
  intros r t Y IHt IHY.
  intros e e' e'' H.
  simpl.
  inversion_clear H.
  apply rfTyok_cons.
  apply IHt; trivial.
  apply IHY; trivial.
  (* o = nil *)
  intros e e' e'' H.
  apply tupTyok_nil.
  (* o = cons (f, t) o *)
  intros f t o IHt IHo. 
  intros e e' e'' H.
  simpl.
  inversion_clear H.
  apply tupTyok_cons.
  apply IHt; trivial.
  apply IHo; trivial.
Qed.

Lemma rfTyok_shiftY_aux :
  forall Y e e' e'',
    rfTyok (app e' e) Y
    -> rfTyok (app e' (app e'' e)) 
               (shiftY Y e'' e).
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  apply rfTyok_nil.
  inversion_clear H.
  apply rfTyok_cons.
  apply valTyok_shift_aux; trivial.
  apply IHY; trivial.
Qed.

Lemma valTyLok_shiftL_aux :
  forall tl e e' e'',
    valTyLok (app e e'') tl
    -> valTyLok (app e (app e' e'')) (shiftL tl e' e'').
Proof.
  induction tl.
  intros e e' e'' H.
    simpl.
    apply valTyLok_vO.
  intros e e' e'' H.
    simpl.
    inversion_clear H.
    apply valTyLok_vS.
    apply valTyok_shift_aux.
    trivial.
    apply IHtl; trivial.
Qed.

Lemma valTyok_shift : 
  forall t e e',
    valTyok e t
    -> valTyok (app e e') (shift t e' emp).
Proof.
  intros.
  assert (Hx := @valTyok_shift_aux t emp e e').
  simplenv.
  apply (Hx H).
Qed.

Lemma valTyLok_shiftL :
  forall tl e e',
    valTyLok e tl
    -> valTyLok (app e e') (shiftL tl e' emp).
Proof.
  induction tl.
  intros e e' H.
    simpl.
    apply valTyLok_vO.
  intros e e' H.
    simpl.
    inversion_clear H.
    apply valTyLok_vS.
    apply valTyok_shift.
    trivial.
    apply IHtl; trivial.
Qed.

Lemma valTyok_shift_2 : 
  forall t e e',
    valTyok e t
    -> valTyok (app e' e) (shift t e' e).
Proof.
  intros.
  assert (Hx := @valTyok_shift_aux t e emp e').
  simplenv.
  apply (Hx H).
Qed.

Lemma rfTyok_shiftY_2 : 
  forall Y e e',
    rfTyok e Y
    -> rfTyok (app e' e) (shiftY Y e' e).
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  constructor.
  inversion_clear H.
  constructor.
  apply valTyok_shift_2; trivial.
  apply IHY; trivial.
Qed.

Lemma tupTyok_shifto_2 : 
  forall o e e',
    tupTyok e o
    -> tupTyok (app e' e) (shifto o e' e).
Proof.
  induction o; try destruct p; simpl; intros; trivial.
  constructor.
  inversion_clear H.
  constructor.
  apply valTyok_shift_2; trivial.
  apply IHo; trivial.
Qed.

Lemma valTyLok_shiftL_2 :
  forall tl e e',
    valTyLok e tl
    -> valTyLok (app e' e) (shiftL tl e' e).
Proof.
  induction tl.
  intros e e' H.
    simpl.
    apply valTyLok_vO.
  intros e e' H.
    simpl.
    inversion_clear H.
    apply valTyLok_vS.
    apply valTyok_shift_2.
    trivial.
    apply IHtl; trivial.
Qed.

Lemma valTyok_shift_rev_aux :
  forall t e e' e'',
    valTyok (app e' (app e'' e)) 
               (shift t e'' e)
    -> valTyok (app e' e) t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y => 
      forall e e' e'',
        rfTyok (app e' (app e'' e)) (shiftY Y e'' e)
        -> rfTyok (app e' e) Y)
    (P1 := fun o =>
      forall e e' e'',
        tupTyok (app e' (app e'' e)) (shifto o e'' e)
        -> tupTyok (app e' e) o);
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' e'' H.
  simpl in H.
  decbltnat n0 (len e); trivial.
    apply valTyok_var.
    apply TyEnv.blt_ine.
    simplenv.
    solvebnat.
  inversion_clear H.
  apply valTyok_var. 
  inv_ine H0.
  apply TyEnv.blt_ine.
  simplenv.
  solvebnat.
  (* t = int *)
  intros e e' e'' H.
  simpl.
  apply valTyok_int.
  (* t = tup o *)
  intros o IHo.
  intros e e' e'' H.
  simpl.
  apply valTyok_tup.
  inversion_clear H.
  eapply IHo; eauto.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' e'' H.
  inversion_clear H.
  simpl.
  apply valTyok_code.
  assert (Hx := IHY (app e e0) e' e'').
  simplenv.
  auto.
  (* Y = nil *)
  intros e e' e'' H.
  simpl;  apply rfTyok_nil.
  (* Y = cons (r, t) Y *)
  intros r t Y IHt IHY.
  intros e e' e'' H.
  simpl.
  inversion_clear H.
  apply rfTyok_cons.
  eapply IHt; eauto.
  eapply IHY; eauto.
  (* o = nil *)
  intros e e' e'' H.
  apply tupTyok_nil.
  (* o = cons (f, t) o *)
  intros f t o IHt IHo. 
  intros e e' e'' H.
  simpl.
  inversion_clear H.
  apply tupTyok_cons.
  eapply IHt; eauto.
  eapply IHo; eauto.
Qed.

Lemma rfTyok_shiftY_rev_aux :
  forall Y e e' e'',
    rfTyok (app e' (app e'' e)) (shiftY Y e'' e) 
    -> rfTyok (app e' e) Y.
Proof.
  induction Y; try destruct p as [r t];
    simpl; intros; trivial.
  apply rfTyok_nil.
  inversion_clear H.
  apply rfTyok_cons.
  eapply valTyok_shift_rev_aux; eauto.
  eapply IHY; eauto.
Qed.

Lemma valTyok_rshift_aux :
  forall t e e' e'',
    valTy_boute2 t e'' e = true
    -> valTyok (app e' (app e'' e)) t
    -> valTyok (app e' e) (rshift t e'' e).
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y => 
      forall e e' e'',
        rfTy_boute2 Y e'' e = true
        -> rfTyok (app e' (app e'' e)) Y
        -> rfTyok (app e' e) (rshiftY Y e'' e))
    (P1 := fun o =>
      forall e e' e'',
        tupTy_boute2 o e'' e = true
        -> tupTyok (app e' (app e'' e)) o
        -> tupTyok (app e' e) (rshifto o e'' e));
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' e'' He H.
  unfold bine2 in He.
  decbltnat n0 (len e); trivial.
    apply valTyok_var.
    apply TyEnv.blt_ine.
    simplenv.
    solvebnat.
  decbltnat n0 (len e'' + len e); try discriminate.
  inversion_clear H.
  apply valTyok_var. 
  inv_ine H0.
  apply TyEnv.blt_ine.
  simplenv.
  substH H0 with (ine_blt _ _ H0).
  unfold bine2 in He.
  solvebnat.
  (* t = int *)
  intros e e' e'' He H.
  apply valTyok_int.
  (* t = tup o *)
  intros o IHo.
  intros e e' e'' He H.
  apply valTyok_tup.
  inversion_clear H.
  eapply IHo; eauto.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' e'' He H.
  inversion_clear H.
  simpl.
  apply valTyok_code.
  assert (Hx := IHY (app e e0) e' e'').
  simplenv.
  auto.
  (* Y = nil *)
  intros e e' e'' He H.
  apply rfTyok_nil.
  (* Y = cons (r, t) Y *)
  intros r t Y IHt IHY.
  intros e e' e'' He H.
  inversion_clear H. 
  destruct (andb_prop _ _ He).
  apply rfTyok_cons.
  eapply IHt; eauto.
  eapply IHY; eauto.
  (* o = nil *)
  intros e e' e'' He H.
  apply tupTyok_nil.
  (* o = cons (f, t) o *)
  intros f t o IHt IHo. 
  intros e e' e'' He H.
  inversion_clear H.
  destruct (andb_prop _ _ He).
  apply tupTyok_cons.
  eapply IHt; eauto.
  eapply IHo; eauto.
Qed.

Lemma rfTyok_rshiftY_aux :
  forall Y e e' e'',
    rfTy_boute2 Y e'' e = true
    -> rfTyok (app e' (app e'' e)) Y
    -> rfTyok (app e' e) (rshiftY Y e'' e).
Proof.
  induction Y; try destruct p; simpl; intros.
  constructor.
  destruct (andb_prop _ _ H).
  inversion_clear H0.
  constructor.
  apply valTyok_rshift_aux; trivial.
  apply IHY; trivial.
Qed.

Lemma tupTyok_rshifto_aux :
  forall o e e' e'',
    tupTy_boute2 o e'' e = true
    -> tupTyok (app e' (app e'' e)) o
    -> tupTyok (app e' e) (rshifto o e'' e).
Proof.
  induction o; try destruct p; 
    try destruct f; simpl; intros.
  constructor.
  destruct (andb_prop _ _ H).
  inversion_clear H0.
  constructor.
  apply valTyok_rshift_aux; trivial.
  apply IHo; trivial.
  destruct (andb_prop _ _ H).
  inversion_clear H0.
  constructor.
  apply valTyok_rshift_aux; trivial.
  apply IHo; trivial.
Qed.

Lemma valTyok_rshift :
  forall t e' e'',
    valTy_boute2 t e'' emp = true
    -> valTyok (app e' e'') t
    -> valTyok e' (rshift t e'' emp).
Proof.
  intros t e' e'' He H.
  assert (Hx := @valTyok_rshift_aux t emp e' e'' He).
  simplenv.
  auto.
Qed.

Lemma rfTyok_rshiftY :
  forall Y e' e'',
    rfTy_boute2 Y e'' emp = true
    -> rfTyok (app e' e'') Y
    -> rfTyok e' (rshiftY Y e'' emp).
Proof.
  induction Y; try destruct p; simpl; intros.
  constructor.
  destruct (andb_prop _ _ H).
  inversion_clear H0.
  constructor.
  apply valTyok_rshift; trivial.
  apply IHY; trivial.
Qed.

Lemma tupTyok_rshifto :
  forall o e' e'',
    tupTy_boute2 o e'' emp = true
    -> tupTyok (app e' e'') o
    -> tupTyok e' (rshifto o e'' emp).
Proof.
  induction o; try destruct p; 
    try destruct f; simpl; intros.
  constructor.
  destruct (andb_prop _ _ H).
  inversion_clear H0.
  constructor.
  apply valTyok_rshift; trivial.
  apply IHo; trivial.
  destruct (andb_prop _ _ H).
  inversion_clear H0.
  constructor.
  apply valTyok_rshift; trivial.
  apply IHo; trivial.
Qed.

(***************************************************************)

Lemma valTyok_varmap_get2 :
  forall tl e ax i,
    valTyLok e tl
    -> Varmap.bindom tl i = true
    -> valTyok e (Varmap.get2 tl ax i).
Proof.
  induction tl; simpl;  intros e ax i H1 H2.
  inversion_clear H2.
  destruct a.
  inversion_clear H1.
  decbeqnat n i; trivial.
  apply IHtl; trivial.
Qed.

Lemma valTyok_varmap_get2_2 :
  forall tl e i,
    valTyLok e tl
    -> valTyok e (Varmap.get2 tl valTy_int i).
Proof.
  induction tl; simpl;  intros e i H1.
  constructor.
  destruct a.
  inversion_clear H1.
  decbeqnat n i; trivial.
  apply IHtl; trivial.
Qed.

Lemma valTyLok_list_app :
  forall tl tl' e,
    valTyLok e tl
    -> valTyLok e tl'
    -> valTyLok e (List.app tl tl').
Proof.
  induction tl; destruct tl'; 
    try destruct p; try destruct p0; 
    simpl; intros e H1 H2;
    inversion_clear H1; inversion_clear H2;
    try constructor;
    trivial.
  rewrite <- app_nil_end; trivial.
  apply IHtl; trivial.
  constructor; trivial.
Qed.

(* ******************************************************* *)

Lemma valTyok_shift_unchange :
  forall t e e' e'',
    valTyok e t
    -> shift t e'' (app e' e) = t.
Proof.
  intro t.
  elim t using valTy_ind3 with 
    (P0 := fun Y =>
      forall e e' e'',
        rfTyok e Y
        -> shiftY Y e'' (app e' e) = Y)
    (P1 := fun o =>
      forall e e' e'',
        tupTyok e o
        -> shifto o e'' (app e' e) = o);
    simpl;
    clear t.
  (* t = valty_var n0 *)
  intros n0.
  intros e e' e'' Ht.
  decbltnat n0 (len e); simpl; simplenv; simplpm; trivial.
    rewbnat (blt_nat n0 (len e' + len e) = true).
    trivial.
  inversion_clear Ht.
  inv_ine H.
  simplbnat.
  (* case t := valty_int *)
  intros e e' e'' Ht.
  trivial.
  (* case t := valty_tup o *)
  intros o IHo.
  intros e e' e'' Ht.
  inversion_clear Ht.
  rewrite IHo; trivial.
  (* case t := valty_code e0 Y *)
  intros e0 Y IHY.
  intros e e' e'' Ht.
  inversion_clear Ht.
  simplenv.
  rewrite (IHY (app e e0) e' e''); trivial.
  (* case Y := e0 nil *)
  intros e e' e'' Ht.
  trivial.
  (* case Y := e0 (r, t) Y *)
  intros r t Y IHt IHY.
  intros e e' e'' Ht.
  inversion_clear Ht.
  rewrite IHt; trivial.
  rewrite IHY; trivial.
  (* case o := nil *)
  intros e e' e'' Ht.
  trivial.
  (* case o := (f, t) o *)
  intros f t o IHt IHo.
  intros e e' e'' Ht.
  inversion_clear Ht.
  rewrite IHt; trivial.
  rewrite IHo; trivial.
Qed.

Lemma rfTyok_shiftY_unchange :
  forall Y e e' e'',
    rfTyok e Y
    -> shiftY Y e'' (app e' e) = Y.
Proof.
  induction Y; try destruct p; simpl; intros; trivial.
  inversion_clear H.
  rewrite valTyok_shift_unchange; trivial.
  rewrite IHY; trivial.
Qed.

Lemma tupTyok_shifto_unchange :
  forall o e e' e'',
    tupTyok e o
    -> shifto o e'' (app e' e) = o.
Proof.
  induction o; try destruct p; simpl; intros; trivial.
  inversion_clear H.
  rewrite valTyok_shift_unchange; trivial.
  rewrite IHo; trivial.
Qed.

Lemma valTyok_shift_unchange_emp :
  forall t e e'',
    valTyok e t
    -> shift t e'' e = t.
Proof.
  intros t e e'' Ht.
  assert (H := @valTyok_shift_unchange t e emp e'' Ht).
  simplenv; trivial.
Qed.

Lemma rfTyok_shiftY_unchange_emp :
  forall Y e e'',
    rfTyok e Y
    -> shiftY Y e'' e = Y.
Proof.
  intros Y e e'' Ht.
  assert (H := @rfTyok_shiftY_unchange Y e emp e'' Ht).
  simplenv; trivial.
Qed.

Lemma tupTyok_shifto_unchange_emp :
  forall o e e'',
    tupTyok e o
    -> shifto o e'' e = o.
Proof.
  intros o e e'' Ht.
  assert (H := @tupTyok_shifto_unchange o e emp e'' Ht).
  simplenv; trivial.
Qed.

Lemma valTyok_instL_aux :
  forall t e e' e'' tl,
    valTyok (app e (app e' e'')) t
    -> valTyLok e tl
    -> valTyok (app e e'') 
         (instL t (shiftL tl e'' emp) e' (len e'')).
Proof.
  intro t.
  elim t using valTy_ind3 with 
    (P0 := fun Y =>
      forall e e' e'' tl,
        rfTyok (app e (app e' e'')) Y
        -> valTyLok e tl
        -> rfTyok (app e e'')
             (instYL Y (shiftL tl e'' emp) e' (len e'')))
    (P1 := fun o =>
      forall e e' e'' tl,
        tupTyok (app e (app e' e'')) o
        -> valTyLok e tl
        -> tupTyok (app e e'') 
             (instoL o (shiftL tl e'' emp) e' (len e'')));
    simpl;
    clear t.
  (* t = valty_var n0 *)
  intros n0.
  intros e e' e'' tl Ht Htlok.
  decbltnat n0 (len e''); simpl; simplenv; simplpm; trivial.
    apply valTyok_var.
    apply TyEnv.blt_ine; trivial.
    simplenv.
    solvebnat.
  decbltnat n0 (len e' + len e''); 
    simpl; simplenv; simplpm; trivial.
    apply valTyok_varmap_get2_2.
    apply valTyLok_shiftL; trivial.
  apply valTyok_var.
  inversion_clear Ht.
  assert (H1 := TyEnv.ine_blt _ _ H).
  apply TyEnv.blt_ine; trivial.
  simplenv.
  solvebnat.
  (* case t := valty_int *)
  intros e e' e'' tl Ht Htlok.
  apply valTyok_int.
  (* case t := valty_tup o *)
  intros o IHo.
  intros e e' e'' tl Ht Htlok.
  inversion_clear Ht.
  apply valTyok_tup.
  eapply IHo; eauto.
  (* case t := valty_code e0 Y *)
  intros e0 Y IHY.
  intros e e' e'' tl Ht Htlok.
  simpl.
  inversion_clear Ht.
  apply valTyok_code.
  simpl.
  assert (H1 := IHY e e' (app e'' e0) tl).
  rewrite shiftL_shiftL_merge.
  simplenv.
  apply H1; trivial.
  (* case Y := e0 nil *)
  intros e e' e'' tl Ht Htlok.
  apply rfTyok_nil.
  (* case Y := e0 (r, t) Y *)
  intros r t Y IHt IHY.
  intros e e' e'' tl Ht Htlok.
  inversion_clear Ht.
  apply rfTyok_cons.
  eapply IHt; eauto.
  eapply IHY; eauto.
  (* case o := nil *)
  intros e e' e'' tl Ht Htlok.
  apply tupTyok_nil.
  (* case o := (f, t) o *)
  intros f t o IHt IHo.
  intros e e' e'' tl Ht Htlok.
  inversion_clear Ht.
  apply tupTyok_cons.
  eapply IHt; eauto.
  eapply IHo; eauto.
Qed.

Lemma valTyok_instL :
  forall t e e' tl,
    valTyok (app e e') t
    -> valTyLok e tl
    -> valTyok e (instL t tl e' 0).
Proof.
  intros t e e' tl.
  assert (H := @valTyok_instL_aux t e e' emp tl).
  simplenv.
  rewrite shiftL_emp in H.
  apply H.
Qed.

Lemma valTyLok_LinstL :
  forall tl e tl' e',
    valTyLok (app e e') tl
    -> valTyLok e tl'
    -> valTyLok e (LinstL tl tl' e' 0).
Proof.
  induction tl; simpl; intros.
    apply valTyLok_vO.
  destruct a.
  inversion_clear H.
  apply valTyLok_vS.
  apply valTyok_instL; trivial.
  apply IHtl; trivial.
Qed.

Lemma rfTyok_instYL_aux :
  forall Y e e' e'' tl,
    rfTyok (app e (app e' e'')) Y
    -> valTyLok e tl
    -> rfTyok (app e e'') 
         (instYL Y (shiftL tl e'' emp) e' (len e'')).
Proof.
  intros Y e e' e'' tl HY Htl.
  induction Y as [|(r,t) Y IHY].
  simpl.
    apply rfTyok_nil.
  simpl.
    inversion_clear HY.
    apply rfTyok_cons.
    assert (H1 := @valTyok_instL_aux t e e' e'' tl).
    apply H1; trivial.
    trivial.
  apply IHY; trivial.
Qed.

Lemma rfTyok_instYL :
  forall Y e e' tl,
    rfTyok (app e e') Y
    -> valTyLok e tl
    -> rfTyok e (instYL Y tl e' 0).
Proof.
  intros Y e e' tl HY Htl'.
  induction Y as [|(r,t) Y IHY].
  simpl.
    apply rfTyok_nil.
  simpl.
    inversion_clear HY.
    apply rfTyok_cons.
    assert (H1 := @valTyok_instL t e e' tl).
    apply H1; trivial.
    trivial.
  apply IHY; trivial.
Qed.

Lemma tupTyok_instoL_aux :
  forall o e e' e'' tl,
    tupTyok (app e (app e' e'')) o
    -> valTyLok e tl
    -> tupTyok (app e e'') 
         (instoL o (shiftL tl e'' emp) e' (len e'')).
Proof.
  intros o e e' e'' tl Ho Htl.
  induction o as [|(f,t) o IHo].
  simpl.
    apply tupTyok_nil.
  simpl.
    inversion_clear Ho.
    apply tupTyok_cons.
    assert (H1 := @valTyok_instL_aux t e e' e'' tl).
    apply H1; trivial.
    trivial.
  apply IHo; trivial.
Qed.

Lemma tupTyok_instoL :
  forall o e e' tl,
    tupTyok (app e e') o
    -> valTyLok e tl
    -> tupTyok e (instoL o tl e' 0).
Proof.
  intros o e e' tl Ho Htl'.
  induction o as [|(f,t) o IHo].
  simpl.
    apply tupTyok_nil.
  simpl.
    inversion_clear Ho.
    apply tupTyok_cons.
    assert (H1 := @valTyok_instL t e e' tl).
    apply H1; trivial.
    trivial.
  apply IHo; trivial.
Qed.

Lemma codeTyok_instYL : 
  forall e0 e Y tl,
    valTyLok e0 tl
    -> codeTyok e0 (e, Y)
    -> codeTyok e0 (emp, instYL Y tl e 0).
Proof.
  unfold codeTyok; simpl.
  intros e0 e Y tl Htlok Hsub.
  inversion_clear Hsub.
  apply valTyok_code.
  simplenv.
  eapply rfTyok_instYL; eauto.
Qed.

Lemma valTyok_instL_unchange :
  forall t e e' tl,
    valTyok e t
    -> instL t tl e' (len e) 
    = t.
Proof.
  intros t.
  elim t using valTy_ind3 with
    (P0 := fun Y =>
      forall e e' tl,
        rfTyok e Y
        -> instYL Y tl e' (len e)
        = Y)
    (P1 := fun o => 
      forall e e' tl,
        tupTyok e o
        -> instoL o tl e' (len e)
        = o);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros e e' tl Ht.
  inversion_clear Ht.
  inv_ine H.
  rewrite Hblt; trivial.
  (* t = int *)
  intros e e' tl Ht.
  simpl;  trivial.
  (* t = tup o *)
  intros o IHo .
  intros e e' tl Ht.
  inversion_clear Ht.
  rewrite (IHo e e' tl H).
  trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros e e' tl Ht.
  inversion_clear Ht.
  rewrite <- (TyEnv.len_app e e0).
  rewrite (IHY (TyEnv.app e e0) e'); trivial. 
  (* Y = e0 nil *)
  intros e e' tl Ht.
  trivial.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros e e' tl Ht.
  inversion_clear Ht.
  rewrite (IHt e e' tl H).
  rewrite (IHY e e' tl H0).
  trivial.
  (* o = nil *)
  intros e e' tl Ht.
  trivial.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros e e' tl Ht.
  inversion_clear Ht.
  rewrite (IHt e e' tl H).
  rewrite (IHo e e' tl H0).
  trivial.
Qed.

Lemma rfTyok_instYL_unchange :
  forall Y e e' tl,
    rfTyok e Y
    -> instYL Y tl e' (len e) = Y.
Proof.
  intros Y e e' tl HY.
  induction Y; simpl; intros; trivial.
  destruct p.
  inversion_clear HY.
  rewrite valTyok_instL_unchange; trivial.
  rewrite (IHY H0); trivial.
Qed.

Lemma tupTyok_instoL_unchange :
  forall o e e' tl,
    tupTyok e o
    -> instoL o tl e' (len e) = o.
Proof.
  intros o e e' tl Ho.
  induction o; simpl; intros; trivial.
  destruct p.
  inversion_clear Ho.
  rewrite valTyok_instL_unchange; trivial.
  rewrite (IHo H0); trivial.
Qed.

Lemma valTyok_instL_unchange_emp :
  forall t e tl,
    valTyok emp t
    -> instL t tl e 0 
    = t.
Proof.
  intros t e tl Ht.
  assert (H := @valTyok_instL_unchange t emp e tl Ht).
    simplenv; trivial.
Qed.

Lemma rfTyok_instYL_unchange_emp :
  forall Y e tl,
    rfTyok emp Y   
    -> instYL Y tl e 0 = Y.
Proof.
  intros Y e tl Ht.
  assert (H := @rfTyok_instYL_unchange Y emp e tl Ht).
    simplenv; trivial.
Qed.

Lemma tupTyok_instoL_unchange_emp :
  forall o e tl,
    tupTyok emp o
    -> instoL o tl e 0 = o.
Proof.
  intros o e tl Ht.
  assert (H := @tupTyok_instoL_unchange o emp e tl Ht).
    simplenv; trivial.
Qed.

Lemma RfTy_sel_valTyok :
  forall e Y r t,
    rfTyok e Y
    -> RfTy.sel Y r = Some t
    -> valTyok e t.
Proof.
  induction Y; try destruct p; simpl; 
    intros; trivial.
  discriminate.
  inversion_clear H.
  decbeqreg r0 r.
  inj_hyp H0; subst v; trivial.
  eapply IHY; eauto.
Qed.

Lemma RfTy_sel_tupTyok :
  forall e Y r o,
    rfTyok e Y
    -> RfTy.sel Y r = Some (valTy_tup o)
    -> tupTyok e o.
Proof.
  intros e Y r o H H0.
  assert (H1 := RfTy_sel_valTyok H H0).
  inversion_clear H1; trivial.
Qed.

Lemma RfTy_upd_rfTyok :
  forall e Y r t,
    rfTyok e Y
    -> valTyok e t
    -> rfTyok e (RfTy.upd Y r t).
Proof.
  induction Y; try destruct p; simpl; 
    intros; trivial.
  constructor; trivial.
  inversion_clear H.
  decbeqreg r0 r.
  constructor; trivial.
  constructor; trivial.
  apply IHY; trivial.
Qed.

Lemma TupTy_get_valTyok :
  forall e i o f t,
    tupTyok e o
    -> TupTy.get o i = Some (f, t)
    -> valTyok e t.
Proof.
  induction i; destruct o; try destruct p; simpl; 
    intros; trivial.
  discriminate.
  injection H0; intros; subst.
  inversion_clear H.
  trivial.
  discriminate.
  inversion_clear H.
  eapply IHi; eauto.
Qed.

Lemma TupTy_lookup_valTyok :
  forall e i o f t,
    tupTyok e o
    -> TupTy.lookup o i (f, t)
    -> valTyok e t.
Proof.
  induction i; destruct o; try destruct p; simpl; 
    intros; trivial.
  inversion_clear H.
  contradiction.
  injection H0; intros; subst.
  inversion_clear H.
  trivial.
  contradiction.
  inversion_clear H.
  eapply IHi; eauto.
Qed.

Lemma TupTy_upd_tupTyok :
  forall e i o f t,
    tupTyok e o
    -> valTyok e t
    -> tupTyok e (TupTy.upd o i (f, t)).
Proof.
  induction i; destruct o; try destruct p; simpl; 
    intros; trivial.
  inversion_clear H.
  constructor; trivial.
  
  inversion_clear H.
  constructor; trivial.
  apply IHi; trivial.
Qed.

Lemma RfTy_sort_rfTyok : 
  forall e Y Y' Y'',
    rfTyok e Y 
    -> RfTy.sort Y Y' = Some Y''
    -> rfTyok e Y''.
Proof.
  intros e Y Y'; generalize e Y; clear e Y.
  induction Y'; try destruct p; simpl; intros; trivial.
  inj_hyp H0; subst Y''.
  constructor.
  destruct (RfTy.sel_dec Y r) as [[t Hs] | Hs];
    rewrite Hs in H0; try discriminate.
  destruct (RfTy.sort_dec Y Y') as [[Yx Hx] | Hx];
    rewrite Hx in H0; try discriminate.
  inj_hyp H0; subst Y''.
  constructor.
  eapply RfTy_sel_valTyok; eauto.
  eapply IHY'; eauto.
Qed.

(* ****************************************************** *)

Lemma valTyLok_idtl_aux : forall n e,
  n <= len e ->
    valTyLok e (id_tl n).
Proof.
  induction n.
  intros; apply valTyLok_vO.
  intros. simpl. 
  apply valTyLok_vS.
  apply valTyok_var.
  apply blt_ine.
  solvebnat.
  apply IHn.
  omega.
Qed.

Lemma valTyLok_idtl : forall e,
    valTyLok e (id_tl (len e)).
Proof.
  intros e.
  apply valTyLok_idtl_aux. 
  omega.
Qed.
