(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Unify.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Ty.
Require Import Varmap.
Require Import TyEnv.
Require Import Subst.

Fixpoint unify (t t' : valTy) (e' e'' : tenv)
  {struct t'} : option varmap :=
  match t' with 
    | valTy_var n0 =>
      if (negb (blt_nat n0 (len e' + len e'')))
        then 
          (if beq_valTy t (valTy_var (n0 - len e'))
            then Some nil
            else None)
        else 
          if andb (blt_nat n0 (len e' + len e'')) 
            (negb (blt_nat n0 (len e'')))
            then 
              if (valTy_boute2 t e'' emp) 
                then 
                  let t0 := rshift t e'' emp in
                    Some (((n0 - len e''), t0) :: nil)
                else 
                  None
            else
              if beq_valTy t (valTy_var n0)
                then Some nil
                else None
    | valTy_int =>
      match t with
        | valTy_int => Some nil
        | _ => None
      end
    | valTy_tup o' =>
      match t with
        | valTy_tup o => unifyo o o' e' e''
        | _ => None
      end
    | valTy_code e0' Y' =>
      match t with
        | valTy_code e0 Y => 
          if (beq_env e0 e0') 
            then unifyY Y Y' e' (app e'' e0')
            else None
        | _ => None
      end
  end
  with unifyY (Y Y' : rfTy) (e' e'' : tenv)
  {struct Y'} : option varmap :=
  match Y', Y with
    | rfTy_nil, rfTy_nil => Some nil
    | rfTy_cons (r', t') Y', rfTy_cons (r, t) Y =>
        if beq_reg r' r
          then
            let st := unify t t' e' e'' in
            let sY := unifyY Y Y' e' e'' in
              match st, sY with
                | Some tl_t, Some tl_Y => 
                  if Varmap.bcompat tl_t tl_Y (len e') emp
                    then Some (List.app tl_t tl_Y)
                    else None
                | _, _ => None
              end
          else
            None
    | _, _ => None
  end
  with unifyo (o o' : tupTy) (e' e'' : tenv)
  {struct o'} : option varmap :=
  match o', o with
    | tupTy_nil, tupTy_nil => Some nil
    | tupTy_cons (init, t') o', tupTy_cons (init, t) o =>
       let st := unify t t' e' e'' in
       let so := unifyo o o' e' e'' in
         match st, so with
           | Some tl_t, Some tl_o => 
             if Varmap.bcompat tl_t tl_o (len e') emp
               then Some (List.app tl_t tl_o)
               else None
           | _, _ => None
         end
    | tupTy_cons (junk, t') o', tupTy_cons (junk, t) o =>
       let st := unify t t' e' e'' in
       let so := unifyo o o' e' e'' in
         match st, so with
           | Some tl_t, Some tl_o => 
             if Varmap.bcompat tl_t tl_o (len e') emp
               then Some (List.app tl_t tl_o)
               else None
           | _, _ => None
         end
    | _, _ => None
  end.
