(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Unify_lemmas.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Ty.
Require Import Varmap.
Require Import TyEnv.
Require Import Rule.
Require Import Subst.
Require Import Unify.
Require Import Subst_lemmas.
Require Import Tyok_lemmas.

Require Import extac.

Import TyEnv.

Lemma unify_dec :
  forall t t' e' e'', 
  (exists tl, unify t t' e' e'' = Some tl)
  \/ (unify t t' e' e'' = None).
Proof.
  intros t t'; generalize t; clear t.
  elim t' using valTy_ind3 with
    (P0 := fun Y' =>
      forall Y e' e'',
      (exists tl, unifyY Y Y' e' e'' = Some tl)
      \/ (unifyY Y Y' e' e'' = None))
    (P1 := fun o' =>
      forall o e' e'',
      (exists tl, unifyo o o' e' e'' = Some tl)
      \/ (unifyo o o' e' e'' = None));
    simpl;
    clear t'.
  (* t = var n0 *)
  intros n0.
  intros t e' e''; destruct t; simpl.
  unfold bine2; simplenv; simplpm;
  rewbnat (blt_nat n 0 = false);
  decbeqnat n n0; simpl;
  decbeqnat n (n0 - len e');simpl;
  decbltnat n0 (len e'');simpl;
  decbltnat n (len e'');simpl;
  decbltnat n0 (len e' + len e'');simpl;
   try (right; trivial; fail);
   try (left; exists (nil : varmap); trivial; fail).
  left; exists ((n0 - len e'', valTy_var (n - len e'')) :: nil);
    trivial.
  left; exists ((n0 - len e'', valTy_var (n - len e'')) :: nil);
    trivial.
  left; exists ((n0 - len e'', valTy_var (n - len e'')) :: nil);
    trivial.
  left; exists ((n0 - len e'', valTy_var (n - len e'')) :: nil);
    trivial.

  decbltnat n0 (len e' + len e'');simpl;
    decbltnat n0 (len e'');simpl;
    try (right; trivial; fail);
    try (left; exists ((n0 - len e'', valTy_int) :: nil);
      trivial).
      
  decbltnat n0 (len e' + len e'');simpl;
    decbltnat n0 (len e'');simpl;
      destruct (tupTy_boute2 t e'' emp);
        try (right; trivial; fail).
  left; exists ((n0 - len e'', valTy_tup (rshifto t e'' emp)) :: nil);
    trivial.

  decbltnat n0 (len e' + len e'');simpl;
    decbltnat n0 (len e'');simpl;
      destruct (rfTy_boute2 r e'' e);
        try (right; trivial; fail).
  left; exists ((n0 - len e'', 
    valTy_code e (rshiftY r e'' e)) :: nil);
    trivial.
  (* t = int *)
  intros t e' e''; destruct t; simpl;
    try (right; trivial; fail).
  left; exists (nil : varmap); trivial.
  (* t = tup o *)
  intros o IHo.
  intros t e' e''; destruct t; simpl;
    try (right; trivial; fail).
  destruct (IHo t e' e'').
  left; trivial.
  right; trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros t e' e''; destruct t; simpl;
    try (right; trivial; fail).
  unfold beq_env.
  decbeqnat e e0.
  substbnat e.
  destruct (IHY r e' (app e'' e0)).
  left; trivial.
  right; trivial.
  right; trivial.
  (* Y = e0 nil *)
  intros t e' e''; destruct t; simpl;
    try (right; trivial; fail).
  left; exists (nil : varmap); trivial.
  (* Y = e0 (r, t)::Y *)
  intros r' t' Y' IHt IHY.
  intros t e' e''; destruct t; 
    try destruct p; simpl;
    try (right; trivial; fail).
  decbeqreg r' r.
  destruct (IHt v e' e'') as [[tl Hs] | Hs];
    destruct (IHY t e' e'') as [[tl' Hs'] | Hs'];
      rewrite Hs; try rewrite Hs'.
  destruct (Varmap.bcompat tl tl' (len e') emp).
  left; exists (List.app tl tl'); trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  right; trivial.
  (* o = nil *)
  intros t e' e''; destruct t; 
    try destruct p; simpl;
    try (right; trivial; fail).
  left; exists (nil : varmap); trivial.
  (* o = (f, t) :: o *)
  intros f' t' o' IHt IHo.
  intros t e' e''; destruct t; 
    try destruct p; simpl;
    try destruct f; try destruct f';
    try (right; trivial; fail).
  destruct (IHt v e' e'') as [ [tl Hs] | Hs];
    destruct (IHo t e' e'') as [ [tl' Hs'] | Hs'];
    rewrite Hs; try rewrite Hs';
    try destruct (Varmap.bcompat tl tl' (len e') emp);
    try (right; trivial; fail).
  left; exists (List.app tl tl'); trivial.
  destruct (IHt v e' e'') as [ [tl Hs] | Hs];
    destruct (IHo t e' e'') as [ [tl' Hs'] | Hs'];
    rewrite Hs; try rewrite Hs';
    try destruct (Varmap.bcompat tl tl' (len e') emp);
    try (right; trivial; fail).
  left; exists (List.app tl tl'); trivial.
Qed.

Lemma unifyY_dec :
  forall Y Y' e' e'', 
  (exists tl, unifyY Y Y' e' e'' = Some tl)
  \/ (unifyY Y Y' e' e'' = None).
Proof.
  induction Y; destruct Y'; 
    try destruct p;
    try destruct p0;
    simpl; intros; trivial;
    try (right; trivial; fail).
  left; exists (nil : varmap); trivial.
  decbeqreg r0 r; try (right; trivial; fail).
  destruct (unify_dec v v0 e' e'') as [ [tl Hs] | Hs];
    destruct (IHY Y' e' e'') as [ [tl' Hs'] | Hs'];
    rewrite Hs; try rewrite Hs';
    try destruct (Varmap.bcompat tl tl' (len e') emp);
    try (right; trivial; fail).
  left; exists (List.app tl tl'); trivial.
Qed.

Lemma unifyo_dec :
  forall o o' e' e'', 
  (exists tl, unifyo o o' e' e'' = Some tl)
  \/ (unifyo o o' e' e'' = None).
Proof.
  induction o; destruct o'; 
    try destruct p;
    try destruct p0; 
    try destruct f;
    try destruct f0;
    simpl; intros; trivial;
    try (right; trivial; fail).
  left; exists (nil : varmap); trivial.
  destruct (unify_dec v v0 e' e'') as [ [tl Hs] | Hs];
    destruct (IHo o' e' e'') as [ [tl' Hs'] | Hs'];
    rewrite Hs; try rewrite Hs';
    try destruct (Varmap.bcompat tl tl' (len e') emp);
    try (right; trivial; fail).
  left; exists (List.app tl tl'); trivial.
  destruct (unify_dec v v0 e' e'') as [ [tl Hs] | Hs];
    destruct (IHo o' e' e'') as [ [tl' Hs'] | Hs'];
    rewrite Hs; try rewrite Hs';
    try destruct (Varmap.bcompat tl tl' (len e') emp);
    try (right; trivial; fail).
  left; exists (List.app tl tl'); trivial.
Qed.

Lemma unify_bcpactL :
  forall t t' e' e'' tl,
    unify t t' e' e'' = Some tl
    -> bcpactL t' tl e' (len e'') = true.
Proof.
  intros t t'; generalize t; clear t.
  elim t' using valTy_ind3 with
    (P0 := fun Y' =>
      forall Y e' e'' tl,
        unifyY Y Y' e' e'' = Some tl
        -> bcpactYL Y' tl e' (len e'') = true)
    (P1 := fun o' =>
      forall o e' e'' tl,
        unifyo o o' e' e'' = Some tl
        -> bcpactoL o' tl e' (len e'') = true);
    simpl;
    clear t'.
  (* t = var n0 *)
  intros n0.
  intros t e' e'' tl; destruct t; simpl.
  unfold bine2; simplenv; simplpm.
  rewbnat (blt_nat n 0 = false).
  decbeqnat n n0; simpl;
  decbeqnat n (n0 - len e');simpl;
  decbltnat n0 (len e'');simpl;
  decbltnat n (len e'');simpl;
  decbltnat n0 (len e' + len e'');simpl;
  decbltnat n (len e' + len e'');simpl;
    try (trivial; fail);
    try (intro Hf; discriminate Hf; fail);
    try (discribnat; fail);
    try (try substbnat n0; intro Hx; inj_hyp Hx; 
      subst tl; simpl; simplbnat; split; trivial; fail).

  decbltnat n0 (len e' + len e'');simpl;
    decbltnat n0 (len e'');simpl;
    try (intro Hf; discriminate; fail);
    try (intro H; inj_hyp H; subst tl; simpl; 
      simplbnat; trivial).
      
  decbltnat n0 (len e' + len e'');simpl;
    decbltnat n0 (len e'');simpl;
      destruct (tupTy_boute2 t e'' emp);
      try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
      simplbnat; trivial).

  decbltnat n0 (len e' + len e'');simpl;
    decbltnat n0 (len e'');simpl;
      destruct (rfTy_boute2 r e'' e);
      try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
      simplbnat; trivial).
  (* t = int *)
  intros t e' e'' tl; destruct t; simpl;
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  (* t = tup o *)
  intros o IHo.
  intros t e' e'' tl; destruct t; simpl;
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  apply IHo; trivial.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros t e' e'' tl; destruct t; simpl;
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  unfold beq_env.
  decbeqnat e e0.
  substbnat e.
  rewbnat bcpactYL_shiftL.
  apply (IHY r e' (app e'' e0) tl).
  intros; discriminate.
  (* Y = e0 nil *)
  intros t e' e'' tl; destruct t; simpl;
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  (* Y = e0 (r, t)::Y *)
  intros r' t' Y' IHt IHY.
  intros t e' e'' tl; destruct t; 
    try destruct p; simpl;
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  decbeqreg r' r.
  destruct (unify_dec v t' e' e'') as [[tl1 Hu] | Hu];
    destruct (unifyY_dec t Y' e' e'') as [[tl2 HY] | HY];
      rewrite Hu; try rewrite HY;
      intros; try discriminate.
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp) as
    [Hc | Hc]; rewrite Hc in H; try discriminate.
  rewrite (@varmap_bsubset_bcpactL_true t' tl1 tl e' e'').
  rewrite (@varmap_bsubset_bcpactYL_true Y' tl2 tl e' e'').
  trivial.
  inj_hyp H; subst tl.
  apply Varmap.List_app_bsubset_r; trivial.
  apply IHY with t; trivial.
  inj_hyp H; subst tl.
  apply Varmap.List_app_bsubset_l; trivial.
  apply IHt with v; trivial.
  intro Hf; discriminate.
  (* o = nil *)
  intros t e' e'' tl; destruct t; 
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  (* o = (f, t) :: o *)
  intros f' t' o' IHt IHo.
  intros t e' e'' tl; destruct t; 
    try (intro Hf; discriminate; fail);
      try (intro H; inj_hyp H; subst tl; simpl; 
        simplbnat; trivial).
  destruct f'; intro Hf; discriminate Hf.
  destruct f'; try destruct p; try destruct f;
    try (intro Hf; discriminate Hf).

  destruct (unify_dec v t' e' e'') as [[tl1 Hu] | Hu];
    destruct (unifyo_dec t o' e' e'') as [[tl2 Ho] | Ho];
      rewrite Hu; try rewrite Ho;
      intros; try discriminate.
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp) as
    [Hc | Hc]; rewrite Hc in H; try discriminate.
  rewrite (@varmap_bsubset_bcpactL_true t' tl1 tl e' e'').
  rewrite (@varmap_bsubset_bcpactoL_true o' tl2 tl e' e'').
  trivial.
  inj_hyp H; subst tl.
  apply Varmap.List_app_bsubset_r; trivial.
  apply IHo with t; trivial.
  inj_hyp H; subst tl.
  apply Varmap.List_app_bsubset_l; trivial.
  apply IHt with v; trivial.

  destruct (unify_dec v t' e' e'') as [[tl1 Hu] | Hu];
    destruct (unifyo_dec t o' e' e'') as [[tl2 Ho] | Ho];
      rewrite Hu; try rewrite Ho;
      intros; try discriminate.
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp) as
    [Hc | Hc]; rewrite Hc in H; try discriminate.
  rewrite (@varmap_bsubset_bcpactL_true t' tl1 tl e' e'').
  rewrite (@varmap_bsubset_bcpactoL_true o' tl2 tl e' e'').
  trivial.
  inj_hyp H; subst tl.
  apply Varmap.List_app_bsubset_r; trivial.
  apply IHo with t; trivial.
  inj_hyp H; subst tl.
  apply Varmap.List_app_bsubset_l; trivial.
  apply IHt with v; trivial.
Qed.

Lemma unifyY_bcpactYL :
  forall Y Y' e' e'' tl,
    unifyY Y Y' e' e'' = Some tl
    -> bcpactYL Y' tl e' (len e'') = true.
Proof.
  intros Y Y'; generalize Y; clear Y.
  induction Y'; try destruct p; simpl; intros; trivial.
  destruct Y; try destruct p; try discriminate.
  decbeqreg r r0; try discriminate.
  destruct (unify_dec v0 v e' e'') as [[tl1 Hu] | Hu];
    destruct (unifyY_dec Y Y' e' e'') as [[tl2 HY] | HY];
      rewrite Hu in H; try rewrite HY in H;
      intros; try discriminate.
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp) as
    [Hc | Hc]; rewrite Hc in H; try discriminate.
  inj_hyp H; subst tl.
  rewrite (@varmap_bsubset_bcpactL_true v tl1).
  rewrite (@varmap_bsubset_bcpactYL_true Y' tl2). 
  trivial.
  apply Varmap.List_app_bsubset_r; trivial.
  apply IHY' with Y; trivial.
  apply Varmap.List_app_bsubset_l; trivial.
  apply unify_bcpactL with v0; trivial.
Qed.

Lemma unifyo_bcpactL :
  forall o o' e' e'' tl,
    unifyo o o' e' e'' = Some tl
    -> bcpactoL o' tl e' (len e'') = true.
Proof.
  intros o o'; generalize o; clear o.
  induction o'; try destruct p; simpl; intros; trivial.
  destruct o; try destruct p; try discriminate.
  destruct f; try discriminate.
  destruct f; try destruct f0; try discriminate;
    simpl.

  destruct (unify_dec v0 v e' e'') as [[tl1 Hu] | Hu];
    destruct (unifyo_dec o o' e' e'') as [[tl2 Ho] | Ho];
      rewrite Hu in H; try rewrite Ho in H;
      intros; try discriminate.
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp) as
    [Hc | Hc]; rewrite Hc in H; try discriminate.
  inj_hyp H; subst tl.
  rewrite (@varmap_bsubset_bcpactL_true v tl1).
  rewrite (@varmap_bsubset_bcpactoL_true o' tl2). 
  trivial.
  apply Varmap.List_app_bsubset_r; trivial.
  apply IHo' with o; trivial.
  apply Varmap.List_app_bsubset_l; trivial.
  apply unify_bcpactL with v0; trivial.

  destruct (unify_dec v0 v e' e'') as [[tl1 Hu] | Hu];
    destruct (unifyo_dec o o' e' e'') as [[tl2 Ho] | Ho];
      rewrite Hu in H; try rewrite Ho in H;
      intros; try discriminate.
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp) as
    [Hc | Hc]; rewrite Hc in H; try discriminate.
  inj_hyp H; subst tl.
  rewrite (@varmap_bsubset_bcpactL_true v tl1).
  rewrite (@varmap_bsubset_bcpactoL_true o' tl2). 
  trivial.
  apply Varmap.List_app_bsubset_r; trivial.
  apply IHo' with o; trivial.
  apply Varmap.List_app_bsubset_l; trivial.
  apply unify_bcpactL with v0; trivial.
Qed.

Lemma unify_sound_aux : forall t t' e' e'' tl,
  unify t t' e' e'' = Some tl
  -> beq_valTy t (instL t' (shiftL tl e'' emp) e' (len e'')) 
    = true.
Proof.
  intros t t'; generalize t; clear t.
  elim t' using valTy_ind3 with 
    (P0 := fun (Y' : rfTy) =>
      forall Y e' e'' tl, 
        unifyY Y Y' e' e'' = Some tl
        -> beq_rfTy Y (instYL Y' (shiftL tl e'' emp) e' (len e''))
        = true)
    (P1 := fun (o' : tupTy) =>
      forall o e' e'' tl,
        unifyo o o' e' e'' = Some tl
        -> beq_tupTy o (instoL o' (shiftL tl e'' emp) e' (len e''))
        = true);
    simpl;
    clear t'.
  (* case t := valty_var n0 *)
  intros n0 t e' e'' tl Htl.
  decbltnat n0 (len e' + len e''); simpl in * .
    decbltnat n0 (len e''); simpl in * .
      destruct (beq_valTy t (valTy_var n0)).
      trivial.
      simpl in Htl.
      discriminate.
      destruct (valTy_boute2_dec t e'' emp) as [Ht | Hf].
      rewrite Ht in Htl.
      inj_hyp Htl.
      rewrite <- Htl.
      simpl.
      rewbnat (beq_nat (n0 - len e'') (n0 - len e'') = true).
      rewrite shift_rshift; trivial.
      apply beq_valTy_refl.
    rewrite Hf in Htl.
    discriminate.
  rewbnat (blt_nat n0 (len e'') = false).
  destruct (beq_valTy_dec t 
    (valTy_var (n0 - len e'))) as [Ht | Hf].
  trivial.
  rewbnat Hf in Htl; discriminate.
  (* case t' := valTy_int *)
  intros.
  destruct t; simpl in * ; try discriminate; trivial.
  (* case t' := valTy_tup o *)
  intros o'. intros.
  destruct t as [n | | o | e Y ]; simpl in * ; try discriminate; trivial.
  eapply H; eauto.
  (* case t' := valTy_code e0' Y' *)
  intros e0' Y' IH. intros.
  destruct t as [n | | o | e0 Y ]; simpl in * ; 
    try discriminate; trivial.
  destruct (beq_env e0 e0').
  assert (H0 := IH Y e' (app e'' e0') tl H).
  simplenv.
  rewrite shiftL_shiftL_merge.
  trivial.
  simpl in H; discriminate.
(* case Y := rfty_nil *)
  intros.
  destruct Y; simpl; trivial.
  simpl in H; discriminate.
(* case Y := rfTy_cons (r, t) Y *)
  intros r' t' Y' IHt' IHY'. intros.
  destruct Y as [| (r, t) Y]; simpl.
  discriminate.
  decbeqreg r r'.
  destruct (unify_dec t t' e' e'') 
    as [[tl1 Ht] | Ht]; rewrite Ht in H.
  destruct (unifyY_dec Y Y' e' e'') 
    as [[tl2 Ht'] | Ht' ]; rewrite Ht' in H.
  assert (Hcp1 := unify_bcpactL t t' e' e'' tl1 Ht).
  assert (Hcp2 := unifyY_bcpactYL Y Y' e' e'' tl2 Ht').
  substH IHt' with (IHt' t e' e'' tl1 Ht).
  substH IHY' with (IHY' Y e' e'' tl2 Ht').
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp)
    as [Hct | Hct]; rewrite Hct in H; 
    try discriminate.
  inj_hyp H.
  rewrite (@varmap_bsubset_instL_unchange t' 
    (shiftL tl1 e'' emp) (shiftL tl e'' emp) e' e'')
    in IHt'.
  rewrite IHt'.
  rewrite (@varmap_bsubset_instYL_unchange Y' 
    (shiftL tl2 e'' emp) (shiftL tl e'' emp) e' e'')
    in IHY'.
  rewrite IHY'.
  simpl; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  subst tl.
  apply List_app_bsubset_r; trivial.
  rewrite bcpactYL_shiftL; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  subst tl.
  apply List_app_bsubset_l; trivial.
  rewrite bcpactL_shiftL; trivial.
  discriminate.
  discriminate.
  discriminate.
  (* case o := tupty_nil *)
  intros o e' e'' tl; destruct o; 
    intros; trivial.
  discriminate.
  (* case o := tupTy_cons (f, t) o *)
  intros f' t' o' IHt' IHo'.
  intros o e' e'' tl Htl; destruct f'; 
    destruct o; 
    try destruct p;
    try destruct f;
    try discriminate.
  destruct (unify_dec v t' e' e'') 
    as [[tl1 Ht] | Ht]; rewrite Ht in Htl.
  destruct (unifyo_dec o o' e' e'') 
    as [[tl2 Ht'] | Ht' ]; rewrite Ht' in Htl.
  assert (Hcp1 := unify_bcpactL v t' e' e'' tl1 Ht).
  assert (Hcp2 := unifyo_bcpactL o o' e' e'' tl2 Ht').
  substH IHt' with (IHt' v e' e'' tl1 Ht).
  substH IHo' with (IHo' o e' e'' tl2 Ht').
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp)
    as [Hct | Hct]; rewrite Hct in Htl; 
    try discriminate.
  inj_hyp Htl.
  simpl.
  rewrite (@varmap_bsubset_instL_unchange t' 
    (shiftL tl1 e'' emp) (shiftL tl e'' emp) e' e'')
    in IHt'.
  rewrite IHt'.
  rewrite (@varmap_bsubset_instoL_unchange o' 
    (shiftL tl2 e'' emp) (shiftL tl e'' emp) e' e'')
    in IHo'.
  rewrite IHo'.
  simpl; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  subst tl.
  apply List_app_bsubset_r; trivial.
  rewrite bcpactoL_shiftL; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  subst tl.
  apply List_app_bsubset_l; trivial.
  rewrite bcpactL_shiftL; trivial.
  discriminate.
  discriminate.

  destruct (unify_dec v t' e' e'') 
    as [[tl1 Ht] | Ht]; rewrite Ht in Htl.
  destruct (unifyo_dec o o' e' e'') 
    as [[tl2 Ht'] | Ht' ]; rewrite Ht' in Htl.
  assert (Hcp1 := unify_bcpactL v t' e' e'' tl1 Ht).
  assert (Hcp2 := unifyo_bcpactL o o' e' e'' tl2 Ht').
  substH IHt' with (IHt' v e' e'' tl1 Ht).
  substH IHo' with (IHo' o e' e'' tl2 Ht').
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp)
    as [Hct | Hct]; rewrite Hct in Htl; 
    try discriminate.
  inj_hyp Htl.
  simpl.
  rewrite (@varmap_bsubset_instL_unchange t' 
    (shiftL tl1 e'' emp) (shiftL tl e'' emp) e' e'')
    in IHt'.
  rewrite IHt'.
  rewrite (@varmap_bsubset_instoL_unchange o' 
    (shiftL tl2 e'' emp) (shiftL tl e'' emp) e' e'')
    in IHo'.
  rewrite IHo'.
  simpl; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  subst tl.
  apply List_app_bsubset_r; trivial.
  rewrite bcpactoL_shiftL; trivial.
  rewrite varmap_bsubset_shiftL; trivial.
  subst tl.
  apply List_app_bsubset_l; trivial.
  rewrite bcpactL_shiftL; trivial.
  discriminate.
  discriminate.
Qed.

Lemma unify_sound : forall t t' e' tl,
  unify t t' e' emp = Some tl
  -> instL t' tl e' 0 = t.
Proof.
  intros t t' e' tl Hu.
  apply sym_eq.
  apply beq_valTy_eq.
  assert (H := unify_sound_aux t t' e' emp tl Hu).
  simplenv.
  rewrite shiftL_emp in H; trivial.
Qed.

Lemma unifyY_sound : forall Y Y' e' tl,
  unifyY Y Y' e' emp = Some tl
  -> instYL Y' tl e' 0 = Y.
Proof.
  intros Y Y' e' tl H. 
  apply sym_eq.
  apply beq_rfTy_eq.
  generalize Y e' tl H; clear Y e' tl H.
  induction Y'; destruct Y; 
    try destruct p; try destruct p0; 
    simpl; intros; 
    try discriminate; trivial.
  decbeqreg r r0; try discriminate.
  destruct (unify_dec v0 v e' emp) 
    as [[tl1 Ht] | Ht]; rewrite Ht in H.
  destruct (unifyY_dec Y Y' e' emp) 
    as [[tl2 Ht'] | Ht' ]; rewrite Ht' in H.
  assert (Hcp1 := unify_bcpactL v0 v e' emp tl1 Ht).
  assert (Hcp2 := unifyY_bcpactYL Y Y' e' emp tl2 Ht').
  assert (IHt' := unify_sound v0 v e' tl1 Ht).
  substH IHY' with (IHY' Y e' tl2 Ht').
  destruct (Varmap.bcompat_dec tl1 tl2 (len e') emp)
    as [Hct | Hct]; rewrite Hct in H; 
    try discriminate.
  inj_hyp H.
  assert (H1 := @varmap_bsubset_instL_unchange v
    tl1 tl e' emp).
  simplenv.
  assert (H2 := List_app_bsubset_l tl1 tl2 (len e') emp Hct); trivial.
  rewrite H in H2.
  rewrite <- (H1 H2 Hcp1).
  clear H1 H2.
  assert (H1 := @varmap_bsubset_instYL_unchange Y' 
    tl2 tl e' emp).
  simplenv.
  assert (H2 := List_app_bsubset_r tl1 tl2 (len e') emp Hct); trivial.
  rewrite H in H2.
  rewrite <- (H1 H2 Hcp2).
  clear H1 H2.
  rewrite IHt'.
  rewrite beq_valTy_refl.
  simpl.
  trivial.
  discriminate.
  discriminate.
Qed.

Lemma unify_valTyLok : 
  forall t t' tl e e' e'',
    valTyok (app e e'') t
    -> unify t t' e' e'' = Some tl
    -> valTyLok e tl.
Proof.
  intro t.
  elim t using valTy_ind3 with 
    (P0 := fun (Y : rfTy) =>
      forall Y' tl e e' e'', 
        rfTyok (app e e'') Y
        -> unifyY Y Y' e' e'' = Some tl
        -> valTyLok e tl)
    (P1 := fun (o : tupTy) =>
      forall o' tl e e' e'',
        tupTyok (app e e'') o
        -> unifyo o o' e' e'' = Some tl
        -> valTyLok e tl);
    simpl;
    clear t.
  (* t = var n0 *)
  intros n0.
  intros t' tl e e' e''; destruct t'; simpl.
  unfold bine2; simplenv; simpl; 
    simplpm; simplbnat; simpl; intros.
  decbltnat n (len e' + len e''); simpl in H0.
  decbltnat n (len e''); simpl in H0.
  decbeqnat n0 n; try discriminate.
  inj_hyp H0; subst tl; simpl.
  constructor; trivial.
  decbltnat n0 (len e''); simpl in H0;
    try discriminate.
  inj_hyp H0; subst tl; simpl.
  constructor.
  inversion_clear H.
  substH H0 with (ine_blt _ _ H0).
  simplenv.
  constructor.
  apply blt_ine.
  simplenv.
  solvebnat.
  constructor.
  decbeqnat n0 (n - len e'); try discriminate.
  inj_hyp H0; subst tl; constructor.
  intros; discriminate.
  intros; discriminate.
  intros; discriminate.
  (* t = int *)
  intros t' tl e e' e''; destruct t'; simpl;
  simplenv; simplbnat; simplpm; simpl; intros;
    try discriminate.
  decbltnat n (len e' + len e''); simpl in H0.
  decbltnat n (len e''); simpl in H0.
  discriminate.
  inj_hyp H0; subst tl.
  constructor; constructor.
  discriminate.
  inj_hyp H0; subst tl; constructor.
  (* t = tup o *)
  intros o IHo .
  intros t' tl e e' e''; destruct t'; simpl;
  simplenv; simplbnat; simplpm; simpl; intros;
    try discriminate.
  decbltnat n (len e' + len e''); simpl in H0.
  decbltnat n (len e''); simpl in H0.
  discriminate.
  destruct (tupTy_boute2_dec o e'' emp) as [He | He];
    rewrite He in H0;
    try discriminate.
  inj_hyp H0; subst tl.
  constructor; constructor.
  inversion_clear H.
  apply tupTyok_rshifto; trivial.
  discriminate.
  inversion_clear H.
  eapply IHo; eauto.
  (* t = code e0 Y *)
  intros e0 Y IHY.
  intros t' tl e e' e''; destruct t'; simpl;
  simplenv; simplbnat; simplpm; simpl; intros;
    try discriminate.
  decbltnat n (len e' + len e''); simpl in H0.
  decbltnat n (len e''); simpl in H0; 
    try discriminate.
  destruct (rfTy_boute2_dec Y e'' e0) as [He | He];
    rewrite He in H0;
    try discriminate.
  inj_hyp H0; subst tl.
  constructor; constructor.
  inversion_clear H.
  simplenv.
  apply rfTyok_rshiftY_aux; trivial.
  discriminate.
  unfold beq_env in H0.
  decbeqnat e0 e1.
  substbnat e1.
  inversion_clear H; simplenv.
  eapply IHY; eauto.
  discriminate.
  (* Y = e0 nil *)
  intros Y' tl e e' e'' H Htl.
  destruct Y'; try destruct p; simpl in * .
  inj_hyp Htl; subst tl.
  constructor.
  discriminate.
  (* Y = e0 (r, t)::Y *)
  intros r t Y IHt IHY.
  intros Y' tl e e' e'' H Htl.
  destruct Y'; try destruct p; simpl in * .
  discriminate.
  decbeqreg r0 r.
  destruct (unify_dec t v e' e'') as [[tl1 Ht] | Ht];
    rewrite Ht in Htl;
    destruct (unifyY_dec Y Y' e' e'') as [[tl2 HY]| HY];
    try rewrite HY in Htl;
    try destruct (bcompat_dec tl1 tl2 (len e') emp) as [Hc | Hc];
    try rewrite Hc in Htl;
    try discriminate.
  inj_hyp Htl; subst tl.
  inversion_clear H.
  apply valTyLok_list_app.
  eapply IHt; eauto.
  eapply IHY; eauto.
  discriminate.
  (* o = nil *)
  intros o' tl e e' e'' H Htl;
  destruct o'; try destruct p; try destruct f;
    simpl in * .
  inj_hyp Htl; subst tl.
  constructor.
  discriminate.
  discriminate.
  (* o = (f, t) :: o *)
  intros f t o IHt IHo.
  intros o' tl e e' e'' H Htl.
  destruct o'; try destruct p; 
    try destruct f; try destruct f0; simpl in * .
  discriminate.
  discriminate.
  destruct (unify_dec t v e' e'') as [[tl1 Ht] | Ht];
    rewrite Ht in Htl;
    destruct (unifyo_dec o o' e' e'') as [[tl2 Ho]| Ho];
    try rewrite Ho in Htl;
    try destruct (bcompat_dec tl1 tl2 (len e') emp) as [Hc | Hc];
    try rewrite Hc in Htl;
    try discriminate.
  inj_hyp Htl; subst tl.
  inversion_clear H.
  apply valTyLok_list_app.
  eapply IHt; eauto.
  eapply IHo; eauto.
  discriminate.
  discriminate.

  destruct (unify_dec t v e' e'') as [[tl1 Ht] | Ht];
    rewrite Ht in Htl;
    destruct (unifyo_dec o o' e' e'') as [[tl2 Ho]| Ho];
    try rewrite Ho in Htl;
    try destruct (bcompat_dec tl1 tl2 (len e') emp) as [Hc | Hc];
    try rewrite Hc in Htl;
    try discriminate.
  inj_hyp Htl; subst tl.
  inversion_clear H.
  apply valTyLok_list_app.
  eapply IHt; eauto.
  eapply IHo; eauto.
Qed.

Lemma unifyY_valTyLok : 
  forall Y Y' tl e e' e'',
    rfTyok (app e e'') Y
    -> unifyY Y Y' e' e'' = Some tl
    -> valTyLok e tl.
Proof.
  induction Y; destruct Y'; 
    try destruct p; try destruct p0;
    simpl; intros tl e e' e'' Hok Hu.
  inj_hyp Hu; subst tl; constructor.
  discriminate.
  discriminate.
  decbeqreg r0 r; try discriminate.
  destruct (unify_dec v v0 e' e'') as [[tl1 Hv] | Hv];
    rewrite Hv in Hu;
    destruct (unifyY_dec Y Y' e' e'') as [[tl2 HY]| HY];
    try rewrite HY in Hu;
    try destruct (bcompat_dec tl1 tl2 (len e') emp) as [Hc | Hc];
    try rewrite Hc in Hu;
    try discriminate.
  inj_hyp Hu; subst tl.
  inversion_clear Hok.
  apply valTyLok_list_app.
  eapply unify_valTyLok; eauto.
  eapply IHY; eauto.
Qed.
