(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Typed Assembly Language                                     *)
(*                                                             *)
(* tal-c3                                                      *)
(***************************************************************)

(* $Id: Varmap.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import Coq.Lists.List.
Require Import Ty.
Require Import TyEnv.

Require Import extac.

Definition varmap := list (nat * valTy).

Definition emp := (nil : varmap).

Fixpoint bindom (tl : varmap) (n : nat) {struct tl} : bool :=
  match tl with
    | nil => false
    | cons (n', t') tl' => 
      if beq_nat n' n then true
        else (bindom tl' n)
  end.

Fixpoint get2 (tl : varmap) (ax : valTy) (n : nat)
  {struct tl} : valTy :=
  match tl with
    | nil => ax
    | cons (n', t') tl' =>
      if beq_nat n n' then t' else get2 tl' ax n
  end.

Fixpoint bcompat (tl tl' : varmap) (n : nat) (e : tenv)
  {struct n} : bool :=
  match n with
    | O => true
    | S n => if 
      match bindom tl (n + len e), bindom tl' (n + len e) with
        | true, true =>
          beq_valTy (get2 tl (valTy_int) (n + len e)) 
                    (get2 tl' (valTy_int) (n + len e))
        | _, _ => true
      end
      then 
        bcompat tl tl' n e
      else
        false
  end.

Fixpoint bsubset (tl tl' : varmap) (n : nat) 
  (e : tenv) {struct n} : bool :=
  match n with 
    | O => true
    | S n => 
      andb
      match bindom tl (n + len e), bindom tl' (n + len e) with
        | true, true => 
            beq_valTy (get2 tl (valTy_int) (n + len e)) 
               (get2 tl' (valTy_int) (n + len e))
        | true, false => false
        | false, true => true
        | _, _ => true
      end
      (bsubset tl tl' n e)
  end.

Lemma bindom_dec : forall tl n,
  {bindom tl n = true} 
  + {bindom tl n = false}.
Proof.
  induction tl; simpl; intros; trivial.
  right; trivial.
  destruct a.
  decbeqnat n0 n.
  left; trivial.
  apply IHtl.
Qed.

Lemma bcompat_sem : 
  forall tl tl' n e,
    bcompat tl tl' n e = true 
    -> forall i,
      blt_nat i n = true
      -> bindom tl (i + len e) = true
      -> bindom tl' (i + len e) = true
      -> get2 tl valTy_int (i + len e)
      = get2 tl' valTy_int (i + len e).
Proof.
  intros tl tl' n; generalize tl tl'; clear tl tl'.
  induction n; simpl; intros; trivial.
  discribnat.
  decbeqnat n i.
  substbnat i.
  rewrite H1 in H.
  rewrite H2 in H.
  apply beq_valTy_eq.
  destruct (beq_valTy (get2 tl valTy_int (n + len e))
    (get2 tl' valTy_int (n + len e))); 
  try discriminate; trivial.
  apply IHn; trivial.
  destruct (bindom tl (n + len e));
    destruct (bindom tl' (n + len e));
    destruct (beq_valTy (get2 tl valTy_int (n + len e))
      (get2 tl' valTy_int (n + len e)));
    destruct (bcompat tl tl' n e);
    try discriminate; trivial.
  solvebnat.
Qed.

Lemma bcompat_nil_r :
  forall tl n e,
    bcompat tl nil n e = true.
Proof.
  induction n; simpl; intros; trivial.
  destruct (bindom tl (n + len e)); trivial.
Qed.

Lemma bcompat_nil_l :
  forall tl n e,
    bcompat nil tl n e = true.
Proof.
  induction n; simpl; intros; trivial.
Qed.

Lemma bcompat_sym : 
  forall n tl tl' e,
    bcompat tl tl' n e
    = bcompat tl' tl n e.
Proof.
  induction n; simpl; intros; trivial.
  destruct (bindom tl (n + len e));
    destruct (bindom tl' (n + len e)); 
      simpl; trivial.
  destruct (beq_valTy_dec (get2 tl (valTy_int) (n + len e))
    (get2 tl' (valTy_int) (n + len e))).
  rewrite H.
  rewrite beq_valTy_sym.
  rewrite H.
  trivial.
  rewrite H.
  rewrite beq_valTy_sym.
  rewrite H.
  trivial.
Qed.

Lemma bsubset_bindom :
  forall n i tl tl' e'',
    bsubset tl tl' n e'' = true
    -> blt_nat i n = true
    -> bindom tl (i + len e'')= true
    -> bindom tl' (i + len e'')= true.
Proof.
  intros n i.
  induction n; intros.
  simplbnat.
  decbeqnat n i.
  substbnat i.
  simpl in H.
  rewrite H1 in H.
  destruct (bindom tl' (n + len e'')); 
    try discriminate.
  trivial.
  simpl in H.
  destruct (bindom tl' (n + len e'')); 
    try discriminate.
  destruct (bindom tl (n + len e'')); 
    try discriminate.
  destruct (andb_prop _ _ H).
  apply IHn with tl; trivial.
  solvebnat.
  apply IHn with tl; trivial.
  solvebnat.
  destruct (andb_prop _ _ H).
  apply IHn with tl; trivial.
  solvebnat.
Qed.

Lemma bsubset_sem :
  forall n i tl tl' e'',
    bsubset tl tl' n e'' = true
    -> blt_nat i n = true
    -> bindom tl (i + len e'')= true
    -> get2 tl (valTy_int) (i + len e'') 
    = get2 tl' (valTy_int) (i + len e'').
Proof.
  intros n i.
  induction n; intros.
  simplbnat.
  decbeqnat n i.
  substbnat i.
  simpl in H.
  rewrite H1 in H.
  destruct (bindom tl' (n + len e'')); 
    try discriminate.
  destruct (andb_prop _ _ H).
  apply beq_valTy_eq.
  trivial.
  simpl in H.
  destruct (andb_prop _ _ H).
  apply IHn; trivial.
  solvebnat.
Qed.

Lemma bcompat_dec :
  forall tl tl' n e,
    bcompat tl tl' n e = true
    \/ bcompat tl tl' n e = false.
Proof.
  intros tl tl' n; generalize tl tl'; clear tl tl'.
  induction n; simpl; intros; trivial.
  left; trivial.
  destruct (bindom_dec tl (n + len e)) as [Htl | Htl];
    destruct (bindom_dec tl' (n + len e)) as [Htl' | Htl'];
    rewrite Htl; try rewrite Htl'.
  destruct (beq_valTy (get2 tl valTy_int (n + len e))
         (get2 tl' valTy_int (n + len e))).
  apply IHn.
  right; trivial.
  apply IHn.
  apply IHn.
  apply IHn.
Qed.

Lemma List_app_bindom_l :
  forall n tl tl',
    bindom tl n = true
    -> bindom (List.app tl tl') n = true.
Proof.
  induction tl; try destruct a; simpl; intros; trivial.
  discriminate.
  decbeqnat n0 n; trivial.
  apply IHtl; trivial.
Qed.

Lemma List_app_bindom_r :
  forall n tl tl',
    bindom tl' n = true
    -> bindom (List.app tl tl') n = true.
Proof.
  intros n tl tl'; generalize tl; clear tl.
  induction tl'; try destruct a; simpl; intros; trivial.
  discriminate.
  decbeqnat n0 n.
  substbnat n0.
  induction tl; simpl; simplbnat; trivial.
  destruct a.
  decbeqnat n0 n; trivial.
Lemma tl_eq_0 : 
  forall tl tl' (n : nat) (t : valTy),
    List.app tl ((n, t) :: tl')
    = List.app (List.app tl ((n, t) :: nil)) tl'.
Proof.
  induction tl; simpl; intros; trivial.
  rewrite IHtl.
  trivial.
Qed.
  rewrite tl_eq_0.
  apply IHtl'; trivial.
Qed.

Lemma List_app_bindom :
  forall tl tl' n,
    bindom (List.app tl tl') n
    = match bindom tl n, bindom tl' n with
        | false, false => false
        | _, _ => true
      end.
Proof.
  induction tl; try destruct a; simpl; intros; trivial.
  destruct (bindom tl' n); trivial.
  decbeqnat n n0; trivial.
Qed.

Lemma bindom_false_get2 :
  forall tl ax n,
    bindom tl n = false 
    -> get2 tl ax n = ax.
Proof.
  induction tl; simpl; intros; trivial.
  destruct a.
  decbeqnat n0 n; trivial.
  apply IHtl; trivial.
Qed.

Lemma List_app_get2 :
  forall tl tl' ax n,
    get2 (List.app tl tl') ax n 
    = match bindom tl n, bindom tl' n with
        | true, _ => get2 tl ax n
        | false, true => get2 tl' ax n
        | false, false => ax
      end.
Proof.
  induction tl; simpl; intros; trivial.
  destruct (bindom_dec tl' n) as [Hb | Hb];
    rewrite Hb; trivial.
  apply bindom_false_get2; trivial.
  destruct a.
  decbeqnat n0 n; trivial.
Qed.

Lemma List_app_bcompat_l : 
  forall tl tl' n e,
    bcompat tl tl' n e = true
    -> bcompat tl (List.app tl tl') n e = true.
Proof.
  intros tl tl' n; generalize tl tl'; clear tl tl'.
  induction n; simpl; intros; trivial.
  rewrite List_app_get2.
  destruct (bindom_dec tl (n + len e)) as [Hin | Hin];
    rewrite Hin; rewrite Hin in H;
    try (rewrite List_app_bindom_l; trivial).
  destruct (bindom_dec tl' (n + len e)) as [Hin' | Hin'];
    rewrite Hin' in H.
  rewrite beq_valTy_refl.
  apply IHn.
  destruct (beq_valTy (get2 tl valTy_int (n + len e))
    (get2 tl' valTy_int (n + len e))); 
    try discriminate.
  trivial.
  rewrite beq_valTy_refl.
  apply IHn; trivial.
  apply IHn; trivial.
Qed.

Lemma List_app_bsubset_l :
  forall tl tl' n e,
    bcompat tl tl' n e = true
    -> bsubset tl (List.app tl tl') n e = true.
Proof.
  intros tl tl' n; generalize tl tl'; clear tl tl'.
  induction n; simpl; intros tl tl' e; trivial.
  rewrite List_app_get2.
  rewrite List_app_bindom.
  destruct (bindom_dec tl (n + len e)) as [Htl | Htl];
    destruct (bindom_dec tl' (n + len e)) as [Htl' | Htl'];
    rewrite Htl; try rewrite Htl';
    destruct (beq_valTy (get2 tl valTy_int (n + len e))
      (get2 tl' valTy_int (n + len e)));
    try rewrite beq_valTy_refl;
    intro H;
    try discriminate;
    simpl; apply IHn; trivial.
Qed.

Lemma List_app_bsubset_r :
  forall tl tl' n e,
    bcompat tl tl' n e = true
    -> bsubset tl' (List.app tl tl') n e = true.
Proof.
  intros tl tl' n; generalize tl tl'; clear tl tl'.
  induction n; simpl; intros tl tl' e; trivial.
  rewrite List_app_get2.
  rewrite List_app_bindom.
  destruct (bindom_dec tl (n + len e)) as [Htl | Htl];
    destruct (bindom_dec tl' (n + len e)) as [Htl' | Htl'];
    rewrite Htl; try rewrite Htl';
    destruct (beq_valTy_dec (get2 tl valTy_int (n + len e))
      (get2 tl' valTy_int (n + len e))) as [He | He];
    try rewrite He; 
    try rewrite beq_valTy_refl;
    intro H;
    try discriminate;
    try (rewrite beq_valTy_sym; rewrite He);
    simpl; apply IHn; trivial.
Qed.

(***************************************************************)

Fixpoint id_tl (n : nat) : varmap :=
  match n with
    | O => nil
    | S n => cons (n, valTy_var n) (id_tl n)
  end.

Lemma get2_idtl :
  forall n n0 ax,
    blt_nat n0 n = true
    -> get2 (id_tl n) ax n0 = valTy_var n0.
Proof.
  induction n; simpl; intros.
  simplbnat.
  decbeqnat n0 n.
  substbnat n0; trivial.
  apply IHn.
  solvebnat.
Qed.

Lemma bindom_idtl :
  forall n0 n,
    blt_nat n0 n = true
    -> bindom (id_tl n) n0 = true.
Proof.
  induction n; simpl; intros.
  simplbnat.
  decbeqnat n n0; trivial.
  apply IHn.
  solvebnat.
Qed.
