(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* TAL over OCAP                                               *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: talp_ocap.v,v 1.1.1.1 2006/12/07 16:43:21 ssg05 Exp $ *)

Require Import machine.
Require Talp.
Require Import ocap.

Require Import extac.
Require Import deptac.

Open Local Scope tm_scope.
Open Local Scope ocap_scope.
Open Local Scope tal_scope.

Module Type TALPARAM.

  Parameter rho_tal : nat.
  Parameter open_r : assert.
  Axiom open_r_only_assert_heap :
    forall Psi H R R' pc pc',
      open_r Psi <.H, R, pc.> 
      -> open_r Psi <.H, R', pc'.>.
End TALPARAM.

Module TalOcap (Params : TALPARAM).

Import Params.

Inductive Ltal : Type := 
  | mk_Ltal : Ty.codeTy -> Ltal.

(* Notation Ltal := tal0.codeTy. *)
(* Notation mk_Ltal := tal0.valty_code. *)
Notation tal_chspec := Ty.chspec.
Notation tal_wfis := Rule.wfis.
Notation ocap_wfis := ocap.wfis.
Notation tal_wfch := Rule.wfch.
Notation ocap_wfch := ocap.wfch.

Module TalCHSpec := Ty.CHSpec.

(*************************************************** *)

Definition tal_chspec_lift (psi : tal_chspec) (OPsi : chspec) : Prop :=
  (forall f (a : Ty.codeTy),
    TalCHSpec.lookup psi f a
    -> OCHSpec.lookup OPsi f <* rho_tal, (mk_Ltal a) *>)
  /\ (forall f pi, 
    OCHSpec.lookup OPsi f pi
    -> exists a, 
      pi = <* rho_tal, mk_Ltal a *>
      /\ TalCHSpec.lookup psi f a).

Definition tal_chspec_sink (OPsi : chspec) (psi : tal_chspec) : Prop :=
  forall f (a : Ty.codeTy),
    (OCHSpec.lookup OPsi f <* rho_tal, (mk_Ltal a) *>
    -> TalCHSpec.lookup psi f a)
    /\ (TalCHSpec.lookup psi f a
      -> OCHSpec.lookup OPsi f <* rho_tal, (mk_Ltal a) *>) .

Lemma lift_lookup : forall psi OPsi f a,
  tal_chspec_lift psi OPsi
  -> TalCHSpec.lookup psi f a
  -> OCHSpec.lookup OPsi f <*rho_tal, mk_Ltal a*> .
Proof.
  intros psi OPsi f a.
  unfold tal_chspec_lift.
  intros [H1 H2] Hin.
  apply H1; trivial.
Qed.

Lemma sink_lookup : forall psi OPsi f a,
  tal_chspec_sink OPsi psi
  -> TalCHSpec.lookup psi f a
  -> OCHSpec.lookup OPsi f <*rho_tal, mk_Ltal a*> .
Proof.
  intros psi OPsi f a.
  unfold tal_chspec_sink.
  intros H H1.
  substH H with (H f a).
  destruct H.
  auto.
Qed.


Lemma lift_sub : forall psi psi' OPsi Psi f a,
  tal_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> TalCHSpec.lookup psi' f a
  -> tal_chspec_sink Psi psi'
  -> OCHSpec.lookup Psi f <*rho_tal, mk_Ltal a*>.
Proof.
  intros psi psi' OPsi Psi f a.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f a) as [Hin1 Hin2].
  auto.
Qed.

Lemma sink_sub : forall psi psi' OPsi Psi f a,
  tal_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> TalCHSpec.lookup psi' f a
  -> tal_chspec_sink Psi psi'
  -> OCHSpec.lookup Psi f <*rho_tal, mk_Ltal a*>.
Proof.
  intros psi psi' OPsi Psi f a.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f a) as [Hin1 Hin2].
  auto.
Qed.

Lemma sink_sub_2 : forall psi psi' OPsi Psi f a,
  tal_chspec_lift psi OPsi
  -> OCHSpec.sub OPsi Psi
  -> TalCHSpec.lookup psi f a
  -> tal_chspec_sink Psi psi'
  -> TalCHSpec.lookup psi' f a.
Proof.
  intros psi psi' OPsi Psi f a.
  intros Hlift Hsub Hf Hsink.
  destruct (Hsink f a) as [Hin1 Hin2].
  destruct Hlift.
  auto.
Qed.

(*************************************************** *)

Definition itptal (theta : Ltal) : assert := 
  match theta with 
    | mk_Ltal a =>
  fun (OPsi : chspec) (s : state)  =>
    match s with 
      | <.H,R,pc.> =>
        exists H1, exists H2,
          Heap.join H1 H2 H
          /\ (exists psi, 
            tal_chspec_sink OPsi psi
            /\ Rule.wfst psi a (mkst H1 R pc))
          /\ open_r OPsi <.H2,R,pc.>
    end
  end.

Definition D_tal : idict := fun id =>
  if (beq_nat id rho_tal) then SomeT <@ itptal @> else
    NoneT.

(*************************************************** *)

Lemma tal_wfis_ocap_wfis_2 :
  forall D OPsi is psi a,
    DSpec.lookup D rho_tal <@itptal@>
    -> tal_chspec_lift psi OPsi
    -> tal_wfis psi a is
    -> ocap_wfis D <<{[ <*rho_tal, mk_Ltal a *> | D ]} / OPsi>> is.
Proof.
  intros D OPsi is.
  induction is.
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt w rs | rt w rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r ]; 
    intros psi [e Y] HD Hlift Htal;
      generalize Hlift; intros [Hlift1 Hlift2];
    try (simpl in Htal; contradiction).

  (* case is = one, i = j *)
  destruct Htal as [[e'' Y''] [Hf Hsub]].
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  exists <* rho_tal, (mk_Ltal (e'', Y'')) *> .
  split_l.
  apply (OCHSpec.lookup_sub_lookup 
    (Hlift1 f (e'', Y'') Hf) Hpsub).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  exists itptal.
  split_l.
  unfold itptal.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  apply Rule_lemmas.wfst_weakening with (e, Y); trivial.

  (* case is = one, i = jal f fret*)
  destruct Htal as [[e' Y'] [Hfret [Ha'ok Htal]]].
  destruct Htal as [[e'' Y''] [Hf Hsub]].
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  exists <* rho_tal, (mk_Ltal (e'', Y'')) *> .
  split.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f (e'', Y'') Hf) Hpsub).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  simpl.
  exists itptal.
  split_l.
  unfold itptal.
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  apply Rule_lemmas.wfst_weakening 
    with (e, Ty.RfTy.upd Y ra (Ty.valTy_code e' Y')); trivial.
  apply Rule_lemmas.wfst_RfTy_upd_wfst with pc; trivial.
    apply Tyok_lemmas.valTyok_weaken_emp; trivial.
  constructor; trivial.
  eapply sink_sub_2; eauto.
 
  (* case is = ione, i = jr *)
  destruct Htal as [[e'' Y''] [Hr Hsub]].
  unfold lift, itptal, BigInterp.
  intros Psi s [Hpsub Hwfs].
  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  destruct (Rule_lemmas.wfst_codeval_exists HS Hr) as [a''' Ha'''].
  exists <* rho_tal, (mk_Ltal a''') *> .
  split.
    eapply sink_sub; eauto. 
  simpl.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_codety_reg_wfst; eauto.

  (* case is = iseq *)
  destruct i as
    [ rd rs rt | rd rs w 
      | rd rs rt | rd rs w
      | rd rs | rd w
      | rt i rs | rt i rs
      | rs rt f | rs rt f | rs f
      | f | f fret | r];
    unfold wfis; fold wfis;
    intros psi [e Y] HD Hlift Htal;
      generalize Hlift; intros [Hlift1 Hlift2];
    try (simpl in Htal; contradiction);
    unfold Rule.wfis in Htal; fold Rule.wfis in Htal.

  (* case wfis_addu *)
  destruct Htal as [Hrdnr0 [Hrs [Hrt His]]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal, mk_Ltal (e, Ty.RfTy.upd Y rd Ty.valTy_int) *> | D]}
           / OPsi >> .
  split_r by (eapply IHis; eauto).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).
  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_int_wfst with (pc := pc); eauto.

  (* case is = iseq, i = addiu  *)
  destruct Htal as [Hrdnr0 [Hrs His]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal, 
    mk_Ltal (e, Ty.RfTy.upd Y rd Ty.valTy_int) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]]. 
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_int_wfst with (pc := pc); eauto.

  (* case is = iseq, i = subu  *)
  destruct Htal as [Hrdnr0 [Hrs [Hrt His]]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal, 
    mk_Ltal (e, Ty.RfTy.upd Y rd Ty.valTy_int) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto). 
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]]. 
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_int_wfst with (pc := pc); eauto.

  (* case is = iseq, i = subiu  *)
  destruct Htal as [Hrdnr0 [Hrs His]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal,
    mk_Ltal (e, Ty.RfTy.upd Y rd Ty.valTy_int) *> | D ]} 
    / OPsi >> .
  split_r by (eapply IHis; eauto). 
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]]. 
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_int_wfst with (pc := pc); eauto.

  (* case is = iseq, i = move  *)
  destruct Htal as [t [Hrdnr0 [Hrs His]]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal,
    mk_Ltal (e, Ty.RfTy.upd Y rd t) *> | D ]} 
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_reg_wfst with (pc := pc); eauto.

  (* case is = iseq, i = li  *)
  destruct Htal as [Hrdnr0 [t [Ht His]]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal, 
    mk_Ltal (e, Ty.RfTy.upd Y rd t) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  destruct t; try inversion_clear Ht.
    eapply Rule_lemmas.wfst_RfTy_upd_wfst with (pc := pc); eauto.
      constructor.
    constructor.
  eapply Rule_lemmas.wfst_RfTy_upd_wfst with (pc := pc); eauto.
    apply Tyok_lemmas.valTyok_weaken_emp; trivial.
  constructor; trivial.
  eapply sink_sub_2; eauto.

  (* case is = iseq, i = lw *)
  destruct Htal as [o [t [Hb [Hrs [Ho His]]]]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal,
    mk_Ltal (e, Ty.RfTy.upd Y rt t) *> | D ]} 
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  destruct (Rule_lemmas.wfst_heapval_exists HS Hrs Ho) as [v Hv].
  assert (Hv' := Heap.join_lookup_lookup_l H12 Hv).
  substH Hv' with (Heap.lookup_get Hv').
  rewrite Hv'.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_reg_load_tup_wfst with (pc := pc); eauto.

  (* case is = iseq, i = sw *)
  destruct Htal as [o [t [flag [Hb [Hrt [Hrs [Ho [Hce Ha']]]]]]]].
  intros Psi s [Hpsub Hwfs].
  exists << {[<* rho_tal,
    mk_Ltal (e, Ty.RfTy.upd Y rs
      (Ty.valTy_tup 
        (Ty.TupTy.upd o i (Ty.init, t)))) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  destruct s as [h rf pc].
  set (s := <.h, rf, pc.>).

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].
  destruct (Rule_lemmas.wfst_heapval_exists HS Hrs Ho) as [v Hv].
  unfold lift, itptal, BigInterp.
  subst s; unfold nextS, itptal, lift.
  rewrite (Heap.join_lookup_lookup_l H12 Hv).
  split_l.
  exists itptal.
  split_l.
  exists (Heap.set H1 (gR rf rs + i) (gR rf rt)).
  exists H2.
  split_l.
  eapply Heap.join_set_l; eauto.
  eapply Heap.lookup_indom; eauto.
  split_r by
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  eapply Rule_lemmas.wfst_RfTy_upd_reg_write_tup_wfst with (pc := pc); eauto.

  (* case is = iseq, i = blt *)
  destruct Htal as [[e'' Y''] [Hf [Hrs [Hrt [Hsub His]]]]].
  intros Psi [h rf pc] [Hpsub Hwfs].

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].

  decbltnat (gR rf rs) (gR rf rt).
    exists <* rho_tal, (mk_Ltal (e'', Y'')) *> .
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f (e'', Y'') Hf) Hpsub).
    unfold nextS, itptal.
    simplbnat.
    exists itptal.
    split_l.
    unfold itptal, proj2_xcdspec.
    exists H1; exists H2.
    split_l.
    split_r by
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    apply Rule_lemmas.wfst_weakening with (e, Y); trivial.
  exists << {[<* rho_tal, 
    mk_Ltal (e, Y) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  unfold nextS, itptal, lift.
  simplbnat.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  trivial.

  (* case is = iseq, i = beq *)
  destruct Htal as [[e'' Y''] [Hf [Hrs [Hrt [Hsub His]]]]].
  intros Psi [h rf pc] [Hpsub Hwfs].

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].

  decbeqnat (gR rf rs) (gR rf rt).
    exists <* rho_tal, (mk_Ltal (e'', Y'')) *> .
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f (e'', Y'') Hf) Hpsub).
    unfold nextS, itptal.
    simplbnat.
    exists itptal.
    split_l.
    unfold itptal, proj2_xcdspec.
    exists H1; exists H2.
    split_l.
    split_r by
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    apply Rule_lemmas.wfst_weakening with (e, Y); trivial.
  exists << {[<* rho_tal, 
    mk_Ltal (e, Y) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  unfold nextS, itptal, lift.
  simplbnat.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  trivial.

  (* case is = iseq, i = bgtz *)
  destruct Htal as [[e'' Y''] [Hf [Hrs [Hsub His]]]].
  intros Psi [h rf pc] [Hpsub Hwfs].

  destruct Hwfs as [itp [HD' Hwfs]].
  assert (Hitpeq := DSpec.lookup_unique HD HD').
  inj_hyp Hitpeq.
  deprewrite <- Hitpeq in Hwfs.
  clear HD' Hitpeq.

  destruct Hwfs as [H1 [ H2 [H12 [HS Hopenr]]]].
  destruct HS as [psi' [Hsink HS]].

  decbltnat O (gR rf rs).
    exists <* rho_tal, (mk_Ltal (e'', Y'')) *> .
    split_l.
    apply (OCHSpec.lookup_sub_lookup 
      (Hlift1 f (e'', Y'') Hf) Hpsub).
    unfold nextS, itptal.
    simplbnat.
    exists itptal.
    split_l.
    unfold itptal, proj2_xcdspec.
    exists H1; exists H2.
    split_l.
    split_r by
      (eapply open_r_only_assert_heap; eauto).
    exists psi'.
    split_l.
    apply Rule_lemmas.wfst_weakening with (e, Y); trivial.
  exists << {[<* rho_tal, 
    mk_Ltal (e, Y) *> | D ]}
    / OPsi >> .
  split_r by (eapply IHis; eauto).
  unfold nextS, itptal, lift.
  simplbnat.
  split_l.
  exists itptal.
  split_l.
  exists H1; exists H2.
  split_l.
  split_r by 
    (eapply open_r_only_assert_heap; eauto).
  exists psi'.
  split_l.
  trivial.
Qed.

Lemma tal_wfis_ocap_wfis :
  forall D OPsi is psi a,
    DSpec.lookup D rho_tal <@itptal@>
    -> tal_chspec_lift psi OPsi
    -> tal_wfis psi a is
    -> ocap_wfis D <<(itptal (mk_Ltal a)) / OPsi>> is.
Proof.
  intros D OPsi is psi a HD HLift His.
  apply wfis_weak_a with <<{[ <*rho_tal, mk_Ltal a *> | D ]} / OPsi>> .
  eapply tal_wfis_ocap_wfis_2; eauto.
  
  unfold lift, BigInterp.
  intros psi0 [h rf pc] [Hsub Hitp].
  split_l.
  exists itptal.
  split_l.
  unfold proj2_xcdspec.
  trivial.
Qed.

Lemma D_tal_describe_lift_psi :
  forall OPsi psi,
    tal_chspec_lift psi OPsi
    -> describe D_tal OPsi.
Proof.
  intros OPsi psi.
  unfold tal_chspec_lift.
  intros H f rho lty theta Hin.
  destruct H as [H1 H2].
  destruct (H2 f <*rho,theta*> Hin) as [Y [Hpi Hin']].
  injection Hpi.
  intros Heq1 Heq2 Heq3.
  subst.
  exists itptal.
  apply DSpec.get_lookup.
  unfold D_tal, DSpec.get, DSpecDef.get.
  rewbnat (beq_nat rho_tal rho_tal = true).
  trivial.
Qed.

Lemma tal_wfch_ocap_wfch :
  forall ch psi psi' OPsi OPsi',
    tal_chspec_lift psi OPsi
    -> tal_chspec_lift psi' OPsi'
    -> tal_wfch psi psi' ch
    -> ocap_wfch D_tal OPsi OPsi' ch.
Proof.
  intros ch psi psi' OPsi OPsi'.
  intros Hpsi Hpsi' Hch.
  unfold wfch.
  split.
    split.
    eapply D_tal_describe_lift_psi; eauto.
    eapply D_tal_describe_lift_psi; eauto.
  intros f pi Hin.
  unfold tal_chspec_lift in Hpsi'.
  unfold Rule.wfch in Hch.
  destruct Hpsi' as [_ Hpsi'].
  destruct (Hpsi' f pi Hin) as [Y [Hpi Hin']].
  subst pi.
  
  destruct (Hch f Y (TalCHSpec.lookup_get Hin')) as [is [Hgis Hwfis]].
  exists is; split_l.
  apply tal_wfis_ocap_wfis_2 with psi.
  apply DSpec.get_lookup.
  unfold D_tal, DSpec.get, DSpecDef.get.
  rewbnat (beq_nat rho_tal rho_tal = true).
  trivial.
  trivial.
  trivial.
Qed.

Theorem tal_sound :
  forall ch psi psi' OPsi OPsi',
    tal_chspec_lift psi OPsi
    -> tal_chspec_lift psi' OPsi'
    -> tal_wfch psi psi' ch
    -> ocap_wfch D_tal OPsi OPsi' ch.
Proof tal_wfch_ocap_wfch.

End TalOcap.
