(***************************************************************)
(* Coq code for PCAP                                           *)
(*                                                             *)
(* Target Machine                                              *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: machine.v,v 1.1.1.1 2006/12/07 16:43:17 ssg05 Exp $ *)

Require Export Word.
Require Export Reg.
Require Export Regfile.
Require Export Instr.
Require Export heap.
Require Export codeheap.

(* *********************************************************** *)

Notation Local "'do' x <-- o ; f" :=
  (match o with Some y => (fun x => f) y | None => None end)
  (at level 200, x ident) : tm_scope.

Notation Local "'ret' x" := (Some x) 
  (at level 200) : tm_scope.

(* *********************************************************** *)

Notation pcreg := (label) (only parsing).

(* *********************************************************** *)

Inductive state : Set :=
  | mkst : heap -> regfile -> pcreg -> state.

Notation " <. h , rf , pc .> " := (mkst h rf pc)
  (at level 0, rf at level 40, pc at level 50) : tm_scope.

Open Local Scope tm_scope.

Notation gR := (Regfile.get).

Notation sR := (Regfile.set).

Notation gH := (Heap.get).

Notation sH := (Heap.set).

Notation S_R := 
  (fun (s : state) =>
    match s with <.h, rf, pc.> => rf end).

Notation S_H := 
  (fun (s : state) =>
    match s with <.h, rf, pc.> => h end).

Notation S_pc := 
  (fun (s : state) =>
    match s with <.h, rf, pc.> => pc end).

Notation S_gR :=
  (fun (s : state) (r : reg) =>
  match s with <.h, rf, pc.> => Regfile.get rf r end).

Notation S_sR :=
  (fun (s : state) (r : reg) (w : word) =>
    match s with <.h, rf, pc.> => <.h, Regfile.set rf r w, pc.>
    end).

Notation S_gH :=
  (fun (s : state) (l : label) =>
  match s with <.h, rf, pc.> => Heap.get h l end).

Notation S_sH :=
  (fun (s : state) (l : label) (w : word) =>
    match s with <.h, rf, pc.> => <.Heap.set h l w, rf, pc.>
    end).

Definition rget (s : state) (r : reg) : word :=
  match s with <.h, rf, pc.> => Regfile.get rf r end.

Definition rset (s : state) (r : reg) (w : word) : state :=
  match s with <.h, rf, pc.> => <.h, Regfile.set rf r w, pc.> end.

Definition hget (s : state) (l : label) : option word :=
  match s with <.h, rf, pc.> => Heap.get h l end.

Definition hset (s : state) (l : label) (w : word) : state :=
  match s with <.h, rf, pc.> => <.Heap.set h l w, rf, pc.> end.

Definition pcinc (s : state) : state :=
  match s with <.h, rf, pc.> => <.h, rf, pc+1.> end.

Definition pcget (s : state) : pcreg :=
  match s with <.h, rf, pc.> => pc end.

Definition pcset (s : state) (f : label) : state :=
  match s with <.h, rf, pc.> => <.h, rf, f.> end.

Definition rf_s := 
  (fun (s : state) =>
    match s with <.h, rf, pc.> => rf end).

Definition hp_s := 
  (fun (s : state) =>
    match s with <.h, rf, pc.> => h end).

Definition pc_s := 
  (fun (s : state) =>
    match s with <.h, rf, pc.> => pc end).

(* *********************************************************** *)

Definition machine : Set := (codeheap * state) % type.

Implicit Types P M : machine.


(* *********************************************************** *)

Definition nextS (i : instr) (s : state) : option state :=
  match i with
    | addu rd rs rt =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs + gR rf rt) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', pc+1.>
      end
    | addiu rd rs w =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs + w) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', pc+1.>
      end
    | subu rd rs rt =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs - gR rf rt) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', pc+1.>
      end
    | subiu rd rs w =>
      match s with
        | <.h, rf, pc.> =>
          do sum <-- Some (gR rf rs - w) ;
            do rf' <-- Some (sR rf rd sum) ;
              ret <.h, rf', pc+1.>
      end
    | move rd rs =>
      match s with
        | <.h, rf, pc.> =>
          do data <-- Some (gR rf rs) ;
            do rf' <-- Some (sR rf rd data) ;
              ret <.h, rf', pc+1.>
      end
    | li rd w =>
      match s with
        | <.h, rf, pc.> =>
          do rf' <-- Some (sR rf rd w) ;
            ret <.h, rf', pc+1.>
      end
    | lw rt i rs =>
      match s with
        | <.h, rf, pc.>  =>
          do data <-- gH h (gR rf rs + i) ;
            do rf' <-- Some (sR rf rt data);
              ret <.h, rf', pc+1.>
      end
    | sw rt i rs =>
      match s with
        | <.h, rf, pc.>  =>
          do data <-- Some (gR rf rt);
            do origdata <-- gH h (gR rf rs + i);
              do h' <-- Some (sH h (gR rf rs + i) data);
                ret <.h', rf, pc+1.>
      end
    | blt rs rt f =>
      match s with
        | <.h, rf, pc.>  =>
          if blt_nat (gR rf rs) (gR rf rt)
            then
              ret <.h, rf, f.>
            else
              ret <.h, rf, pc+1.>
      end
    | beq rs rt f =>
      match s with
        | <.h, rf, pc.>  =>
          if beq_nat (gR rf rs) (gR rf rt)
            then
              ret <.h, rf, f.>
            else
              ret <.h, rf, pc+1.>
      end
    | bgtz rs f =>
      match s with
        | <.h, rf, pc.>  =>
          if blt_nat O (gR rf rs)
            then
              ret <.h, rf, f.>
            else
              ret <.h, rf, pc+1.>
      end
    | j f =>
      match s with
        | <.h, rf, pc.>  =>
          ret <.h, rf, f.>
      end
    | jal f fret =>
      match s with
        | <.h, rf, pc.>  =>
          do rf' <-- Some (sR rf ra fret);
            ret <.h, rf', f.>
      end
    | jr r =>
      match s with
        | <.h, rf, pc.> =>
          do f <-- Some (gR rf r) ;
            ret <.h, rf, f.>
      end
  end.

(* Yet another next-state function, whose 
  return-term is smaller than the above one
  for efficiency consideration *)
Definition nextS2 (i : instr) (s : state) : option state :=
  match i with
    | addu rd rs rt => 
      Some (pcinc (rset s rd (rget s rs + rget s rt)))
    | addiu rd rs w => 
      Some (pcinc (rset s rd (rget s rs + w)))
    | subu rd rs rt => 
      Some (pcinc (rset s rd (rget s rs - rget s rt)))
    | subiu rd rs w => 
      Some (pcinc (rset s rd (rget s rs - w)))
    | move rd rs => 
      Some (pcinc (rset s rd (rget s rs)))
    | li rd w => 
      Some (pcinc (rset s rd w))
    | lw rt w rs => 
      match hget s (rget s rs + w) with
        | Some v => Some (pcinc (rset s rt v))
        | None => None 
      end
    | sw rt w rs => 
      match (hget s (rget s rs + w)) with
        | Some _ => 
            Some (pcinc (hset s (rget s rs + w) (rget s rt)))
        | None => None
      end
    | blt rs rt f => 
      match blt_nat (rget s rs) (rget s rt) with
        | true => Some (pcset s f)
        | false => Some (pcinc s)
      end
    | beq rs rt f => 
      match beq_nat (rget s rs) (rget s rt) with
        | true => Some (pcset s f)
        | false => Some (pcinc s)
      end
    | bgtz rs f => 
      match blt_nat O (rget s rs) with
        | true => Some (pcset s f)
        | false => Some (pcinc s)
      end
    | j f => Some (pcset s f)
    | jal f fret => Some (pcset (rset s ra fret) f)
    | jr r => Some (pcset s (rget s r))
  end.

(* *********************************************************** *)

Definition step (P : machine) : option machine :=
  let (ch, s) := P in
    let (h, rf, pc) := s in
      do i <-- codeheap.get ch pc ;
        do s' <-- nextS i s ;
          ret (ch, s').

Fixpoint stepn (n : nat) (p : machine) {struct n} : option machine :=
  match n with
    | O => Some p
    | S n' =>
      match stepn n' p with
	| Some p' => step p'
	| None => None
      end
  end.

Notation "P ~~> P'" := (step P = Some P') 
  (at level 5) : tm_scope.

Notation optp :=
  (fun f => fun opt:option _ => 
    match opt with
      | Some t => f t
      | None => False
    end).

Notation optg :=
  (fun f => fun opt:option _ =>
    fun t' =>
      match opt with
        | Some t => f t t'
        | None => False
      end).

(* *********************************************************** *)

Notation " i ; I " := (iseq i I) 
  (at level 60, right associativity) : tm_scope.

Notation " i ;; " := (ione i)
  (at level 60) : tm_scope.

(* *********************************************************** *)

Ltac unfoldgs :=
  unfold rget, rset, hget, hset, pcinc, pcset, pcget in * .

Ltac simplst := 
  unfoldgs; Regfile.simplgs; Heap.simplgs.
