(***************************************************************)
(* Coq code for OCAP                                           *)
(*                                                             *)
(* Utilities for manipulating natural numbers                  *)
(*                                                             *)
(*                                                             *)
(***************************************************************)

(* $Id: natpm.v,v 1.1.1.1 2006/12/07 16:43:22 ssg05 Exp $ *)

Require Import Arith.
Require Import Omega.

(***************************************************************)

Lemma xpO : forall x,
  x + O = x.
Proof.
  intros; auto with arith.
Qed.

Lemma xp1 : forall x,
  x + 1 = S x.
Proof.
  intros; omega.
Qed.

Lemma pOx : forall x,
  O + x = x.
Proof.
  simpl; trivial.
Qed.

Lemma p1x : forall x,
  1 + x = S x.
Proof.
  simpl; trivial.
Qed.

Lemma xp1m1 : forall x,
  x + 1 - 1 = x.
Proof.
  intros; omega.
Qed.

Lemma xm1p1 : forall x,
  x >= 1 -> x - 1 + 1 = x.
Proof.
  intros; omega.
Qed.

Lemma xp2m2 : forall x,
  x + 2 - 2 = x.
Proof.
  intros; omega.
Qed.

Lemma xm2p2 : forall x,
  x >= 2 -> x - 2 + 2 = x.
Proof.
  intros; omega.
Qed.

Lemma xpymy : forall x y,
  x + y - y = x.
Proof.
  intros; omega.
Qed.

Lemma xp2m1 : forall x,
  x + 2 - 1 = S x.
Proof.
  intros; omega.
Qed.

Lemma xpSy : forall x y,
  x + S y = S x + y.
Proof.
  intros; omega.
Qed.

Lemma xpypz : forall x y z,
  x + (y + z) = x + y + z.
Proof.
  intros; omega.
Qed.

Lemma xmO : forall x,
  x - O = x.
Proof.
  induction x; simpl; trivial.
Qed.

Lemma xmymz : forall x y z,
  x - (y + z) = x - y - z.
Proof.
  double induction x y; simpl; intros; trivial.
Qed.

(***************************************************************)

Lemma SSxm2 : forall x,
  (S (S x)) - 2 = x.
Proof.
  intros; omega.
Qed.

Lemma SSxm1 : forall x,
  (S (S x)) - 1 = S x.
Proof.
  intros; omega.
Qed.

(* *********************************************************** *)

Ltac simplpm :=
  match goal with
    | [ |- context [(?x + O)] ] =>
      rewrite xpO; simplpm
    | [H : context [(?x + O)] |- _ ] =>
      rewrite xpO in H; simplpm

    | [ |- context [(?x + 1)]] =>
      rewrite xp1; simplpm
    | [H : context [(?x + 1)] |- _ ] =>
      rewrite xp1 in H; simplpm

    | [ |- context [(O + ?x)]] =>
      rewrite pOx; simplpm
    | [H : context [(O + ?x)] |- _ ] =>
      rewrite pOx in H; simplpm

    | [ |- context [(1 + ?x)]] =>
      rewrite p1x; simplpm
    | [H : context [(1 + ?x)] |- _ ] =>
      rewrite p1x in H; simplpm

    | [ |- context [(?x + ?y - ?y)]] =>
      rewrite xpymy; simplpm
    | [H : context [(?x + ?y - ?y)] |- _ ] =>
      rewrite xpymy in H; simplpm

    | [ |- context [(?x + 2 - 1)]] =>
      rewrite xp2m1; simplpm
    | [H : context [(?x + 2 - 1)] |- _ ] =>
      rewrite xp2m1 in H; simplpm

    | [ |- context [(?x + S ?y)]] =>
      rewrite xpSy; simplpm
    | [H : context [(?x + S ?y)] |- _ ] =>
      rewrite xpSy in H; simplpm
        
    | [ |- context [(?x + (?y + ?z))]] =>
      match goal with
        | |- context [(x + y + z)] => 
          rewrite (xpypz x y z); simplpm
        | H : context [(x + y + z)] |- _ => 
          rewrite (xpypz x y z); simplpm
      end
    | [ H : context [(?x + (?y + ?z))] |- _ ] =>
      match goal with
        | |- context [(x + y + z)] => 
          rewrite (xpypz x y z) in H; simplpm
        | H0 : context [(x + y + z)] |- _ => 
          rewrite (xpypz x y z) in H; simplpm
      end

    | [|- context [?x - O] ] =>
      rewrite (xmO x); simplpm
    | [H : context [?x - O] |- _ ] =>
      rewrite (xmO x) in H; simplpm

    | [|- context [?x - (?y + ?z)] ] =>
      rewrite (xmymz x y z); simplpm
    | [H : context [?x - (?y + ?z)] |- _ ] =>
      rewrite (xmymz x y z) in H; simplpm
       
    | [|- context [(S (S ?x) -2)] ] =>
      rewrite (SSxm2 x); simplpm
    | [H : context [S (S ?x) -2] |- _ ] =>
      rewrite (SSxm2 x) in H; simplpm

    | [|- context [S (S ?x) -1] ] =>
      rewrite (SSxm1 x); simplpm
    | [H : context [S (S ?x) -1] |- _ ] =>
      rewrite (SSxm1 x) in H; simplpm

    | _ => idtac
  end.

(* *********************************************************** *)

Lemma not_ltO : forall n, ~ (n < O).
Proof. apply lt_n_O. Qed.

Lemma lt_irrefl : forall n, ~ (n < n).
Proof. apply lt_irrefl. Qed.

Ltac simplnat :=
  match goal with 
    | H : ?n < O |- _ => 
        destruct (not_ltO n)
    | H : ?n < ?n |- _ =>
        destruct (lt_irrefl n)
  end.

(* *********************************************************** *)

Tactic Notation "rep_arith" constr (t1) "with" constr (t2) :=
  replace t1 with t2; [idtac | omega].

Tactic Notation "rep_arith" constr (t1) "with" constr (t2) "in" hyp (H) :=
  replace t1 with t2 in H; [idtac | omega].
