Require Import Map.
Require Import TM.
Require Import Assertions.
Require Import SAGL.
Require Import CSL.
Require Import Arith.
Require Import Omega.

Lemma SAGL_WFiseq_aAa :
  forall Psi A G a v I,
    SAGL_WFIseq Psi A G (sagl_spec a v) I -> presvA a A.
Proof.
  intros.
  induction I.
  inversion H. auto.
  inversion H. auto.
Qed.

Lemma SAGL_WFiseq_preca :
  forall Psi A G a v I,
    SAGL_WFIseq Psi A G (sagl_spec a v) I -> prec_a a.
Proof.
  intros.
  induction I; inversion H; auto.
Qed.

Lemma SAGL_WFthrd_iseq:
  forall Psi A G a v C R I k,
    SAGL_WFthrd Psi A G a v (C, R, I, k) ->
      SAGL_WFIseq Psi A G (sagl_spec a v) I.
Proof.
  intros.
  unfold SAGL_WFthrd in H.
  InvertAll. auto.
Qed.

Lemma SAGL_WFthrd_aAa:
  forall Psi A G a v C R I k,
    SAGL_WFthrd Psi A G a v (C, R, I, k) -> presvA a A.
Proof.
  intros.
  apply SAGL_WFiseq_aAa with Psi G v I.
  apply SAGL_WFthrd_iseq with C R k.
  auto.
Qed.

Lemma SAGL_WFthrd_preca:
  forall Psi A G a v C R I k,
    SAGL_WFthrd Psi A G a v (C, R, I, k) -> prec_a a. 
Proof.
  intros.
  apply SAGL_WFiseq_preca with Psi A G v I.
  apply SAGL_WFthrd_iseq with C R k.
  auto.
Qed.

Lemma star_av_intro:
  forall (a v : Assertion) (M_1 M_2 : Mem) (R: RegFile) (k : Tid) (L : LockMap),
    disjoint M_1 M_2 -> a (M_1, (R, k), L) -> v (M_2, (R, k), filter_L L k) -> star_av a v (merge M_1 M_2, (R, k), L).
Proof.
  intros.
  unfold star_av.
  unfold star_m.
  exists M_1. exists M_2.
  split; auto.
Qed.


Lemma preca_liftG_precG :
  forall M_1 M_2 M_1' M_2' R R' k L L' a_1 a' G,
    liftG a_1 G a' (merge M_1 M_2, (R, k), L) (merge M_1' M_2', (R', k), L') ->
    disjoint M_1 M_2 -> disjoint M_1' M_2' ->
    a_1 (M_1, (R, k), L) -> a' (M_1', (R', k), L') ->
    prec_a a_1 -> prec_a a' ->
    G (M_1, (R, k), L) (M_1', (R', k), L').
Proof.
  intros.
  unfold liftG in H.
  InvertAll.
  assert (x = M_1).
  unfold prec_a in H4.
  apply H4 with (merge M_1 M_2) R k L; auto.
  apply disjUnion_subseteq_left with x0; auto.
  apply merge_extends_map.
  assert (x1 = M_1').
  unfold prec_a in H5.
  apply H5 with (merge M_1' M_2') R' k L'; auto.
  apply disjUnion_subseteq_left with x2; auto.
  apply merge_extends_map.
  rewrite <- H9. rewrite <- H11.
  auto.
Qed.


Lemma SAGL_WFcdhp_WFiseq:
  forall Psi A G C Psi' f a'' v'' I,
    SAGL_WFcdhp Psi A G C Psi' -> 
      Psi' f = sagl_spec a'' v'' ->
      maps_to C f I ->
      SAGL_WFIseq Psi A G (sagl_spec a'' v'') I.
Proof.
  intros.
  unfold SAGL_WFcdhp in H.
  assert 
    (exists I : InstrSeq,
        maps_to C f I /\ SAGL_WFIseq Psi A G (sagl_spec a'' v'') I).
  apply H; auto.
  inversion H2.
  inversion_clear H3.
  unfold maps_to in H1. unfold maps_to in H4.
  rewrite H4 in H1. inversion H1.
  rewrite <- H6. auto.
Qed.



Lemma resplit_AB_disjBC :
  forall (Ms M_1 M_2 Ms' M_1' : Mem), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> merge Ms' M_1' = merge Ms M_1 ->
        disjUnion M_1' M_2 (merge M_1' M_2).
Proof.
  intros.
  unfold disjUnion. split; auto.
  assert (disjoint (merge Ms M_1) M_2).
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  apply disjoint_commut; auto.
  apply disjoint_commut; auto.
  rewrite <- H3 in H4.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite H5 in H4.
  apply disj_distr_left with Ms'; auto.
Qed.

Lemma resplit_AB_disjUnionBCA :
  forall (Ms M_1 M_2 Ms' M_1' M: Mem), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> merge Ms' M_1' = merge Ms M_1 ->
      disjUnion (merge M_1 M_2) Ms M ->
        disjUnion (merge M_1' M_2) Ms' M.
Proof.
  intros.
  assert (disjUnion M_1' M_2 (merge M_1' M_2)).
  apply resplit_AB_disjBC with Ms M_1 Ms'; auto.
  unfold disjUnion in H5.
  inversion_clear H5.
  clear H7.
  assert (disjUnion Ms' M_2 (merge Ms' M_2)).
  apply resplit_AB_disjBC with Ms M_1 M_1'; auto.
  apply disjoint_commut; auto.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite <- H5.
  auto.
  unfold disjUnion in H5.
  inversion_clear H5.
  clear H8.
  unfold disjUnion in H4.
  inversion_clear H4.
  assert (merge (merge M_1 M_2) Ms = merge (merge Ms M_1) M_2).
  apply merge_ABC_CAB; auto.
  unfold disjUnion.
  split.
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  rewrite H8.
  rewrite H4.
  rewrite <- H3.
  assert (merge (merge M_1' M_2) Ms' = merge (merge Ms' M_1') M_2).
  apply merge_ABC_CAB; auto.
  auto.
Qed.


Lemma resplit_AB_disjBC_cons_aux:
  forall (Ms M_1 M_2 Ms' M_1' : Mem) (l : Address) (w w' : Word), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> not_in_dom l (merge (merge Ms M_1) M_2) ->
      not_in_dom (S l) (merge (merge Ms M_1) M_2) ->
      merge Ms' M_1' = ext_M (ext_M (merge Ms M_1) l w) (S l) w' ->
        disjoint M_1' M_2.
Proof.
  intros.
  assert (disjoint (merge Ms M_1) M_2).
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  apply disjoint_commut; auto.
  apply disjoint_commut; auto.
  assert (disjoint (ext_M (ext_M (merge Ms M_1) l w) (S l) w') M_2).
  unfold ext_M.
  apply extend_presv_disj; auto.
  apply extend_presv_disj; auto.
  apply notin_merge_distr_right with (merge Ms M_1); auto.
  apply notin_merge_distr_right with (merge Ms M_1); auto.
  rewrite <- H5 in H7.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite H8 in H7.
  apply disj_distr_left with Ms'; auto.
Qed.

Lemma resplit_AB_disjBC_cons :
  forall (Ms M_1 M_2 Ms' M_1' : Mem) (l : Address) (w w' : Word), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> not_in_dom l (merge (merge Ms M_1) M_2) ->
      not_in_dom (S l) (merge (merge Ms M_1) M_2) ->
      merge Ms' M_1' = ext_M (ext_M (merge Ms M_1) l w) (S l) w' ->
        disjUnion M_1' M_2 (merge M_1' M_2).
Proof.
  intros.
  unfold disjUnion.
  split; auto.
  apply resplit_AB_disjBC_cons_aux with Ms M_1 Ms' l w w'; auto.
Qed.

Lemma resplit_AB_disjUnionBCA_cons:
  forall (Ms M_1 M_2 Ms' M_1' M: Mem) (l : Address) (w w' : Word), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> not_in_dom l (merge (merge Ms M_1) M_2) ->
      not_in_dom (S l) (merge (merge Ms M_1) M_2) ->
      merge Ms' M_1' = ext_M (ext_M (merge Ms M_1) l w) (S l) w' -> 
      disjUnion (merge M_1 M_2) Ms M ->
        disjUnion (merge M_1' M_2) Ms' (ext_M (ext_M M l w) (S l) w').
Proof.
  intros.
  assert (disjUnion M_1' M_2 (merge M_1' M_2)).
  apply resplit_AB_disjBC_cons with Ms M_1 Ms' l w w'; auto.
  unfold disjUnion in H7.
  inversion_clear H7.
  clear H9.
  assert (disjUnion Ms' M_2 (merge Ms' M_2)).
  apply resplit_AB_disjBC_cons with Ms M_1 M_1' l w w'; auto.
  apply disjoint_commut; auto.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite <- H7.
  auto.
  unfold disjUnion in H7.
  inversion_clear H7.
  clear H10.
  unfold disjUnion in H6.
  inversion_clear H6.
  assert (merge (merge M_1 M_2) Ms = merge (merge Ms M_1) M_2).
  apply merge_ABC_CAB; auto.
  rewrite H6 in H10.
  assert (ext_M M l w = merge (ext_M (merge Ms M_1) l w) M_2).
  unfold ext_M.
  rewrite H10.
  apply put_merge_merge_put.
  rewrite H11.
  assert (ext_M (merge (ext_M (merge Ms M_1) l w) M_2) (S l) w'
              = merge (ext_M (ext_M (merge Ms M_1) l w) (S l) w') M_2).
  unfold ext_M.
  apply put_merge_merge_put.
  rewrite H12.
  rewrite <- H5.
  unfold disjUnion.
  split.
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  assert (merge (merge M_1' M_2) Ms' = merge (merge Ms' M_1') M_2).
  apply merge_ABC_CAB; auto.
  auto.
Qed.

Lemma resplit_AB_disjBC_free_aux :
  forall (Ms M_1 M_2 Ms' M_1' : Mem) (l : Address), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> in_dom l (merge Ms M_1) ->
      merge Ms' M_1' = remove_M (merge Ms M_1) l ->
        disjoint M_1' M_2.
Proof.
  intros.
  assert (disjoint (merge Ms M_1) M_2).
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  apply disjoint_commut; auto.
  apply disjoint_commut; auto.
  assert (disjoint (remove_M (merge Ms M_1) l) M_2).
  unfold remove_M.
  apply remove_presv_disj; auto.
  rewrite <- H4 in H6.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite H7 in H6.
  apply disj_distr_left with Ms'; auto.
Qed.

Lemma resplit_AB_disjBC_free :
  forall (Ms M_1 M_2 Ms' M_1' : Mem) (l : Address), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> in_dom l (merge Ms M_1) ->
      merge Ms' M_1' = remove_M (merge Ms M_1) l ->
        disjUnion M_1' M_2 (merge M_1' M_2).
Proof.
  intros.
  unfold disjUnion.
  split; auto.
  apply resplit_AB_disjBC_free_aux with Ms M_1 Ms' l; auto.
Qed.

Lemma resplit_AB_disjUnionBCA_free:
  forall (Ms M_1 M_2 Ms' M_1' M: Mem) (l : Address), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> in_dom l (merge Ms M_1) ->
      merge Ms' M_1' = remove_M (merge Ms M_1) l ->
      disjUnion (merge M_1 M_2) Ms M ->
        disjUnion (merge M_1' M_2) Ms' (remove_M M l).
Proof.
  intros.
  assert (disjUnion M_1' M_2 (merge M_1' M_2)).
  apply resplit_AB_disjBC_free with Ms M_1 Ms' l; auto.
  unfold disjUnion in H6.
  inversion_clear H6.
  clear H8.
  assert (disjUnion Ms' M_2 (merge Ms' M_2)).
  apply resplit_AB_disjBC_free with Ms M_1 M_1' l; auto.
  apply disjoint_commut; auto.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite <- H6.
  auto.
  unfold disjUnion in H6.
  inversion_clear H6.
  clear H9.
  unfold disjUnion in H5.
  inversion_clear H5.
  assert (merge (merge M_1 M_2) Ms = merge (merge Ms M_1) M_2).
  apply merge_ABC_CAB; auto.
  rewrite H5 in H9.
  assert (remove_M M l = merge (remove_M (merge Ms M_1) l) M_2).
  unfold remove_M.
  rewrite H9.
  apply disj_remove_merge_merge_remove.
  apply disjoint_commut.
  apply disj_merge_assoc.
  apply disjoint_commut; auto.
  apply disjoint_commut; auto. auto.
  unfold disjUnion.
  split.
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  rewrite H10.
  rewrite <- H4.
  assert (merge (merge M_1' M_2) Ms' = merge (merge Ms' M_1') M_2).
  apply merge_ABC_CAB; auto.
  auto.
Qed.


Lemma resplit_AB_disjBC_sw_aux :
  forall (Ms M_1 M_2 Ms' M_1' : Mem) (l : Address) (w : Word), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> in_dom l (merge Ms M_1) ->
      merge Ms' M_1' = ext_M (merge Ms M_1) l w ->
        disjoint M_1' M_2.
Proof.
  intros.
  assert (disjoint (merge Ms M_1) M_2).
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  apply disjoint_commut; auto.
  apply disjoint_commut; auto.
  assert (disjoint (ext_M (merge Ms M_1) l w) M_2).
  unfold ext_M.
  apply update_presv_disj; auto.
  rewrite <- H4 in H6.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite H7 in H6.
  apply disj_distr_left with Ms'; auto.
Qed.

Lemma resplit_AB_disjBC_sw :
  forall (Ms M_1 M_2 Ms' M_1' : Mem) (l : Address) (w : Word), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> in_dom l (merge Ms M_1) ->
      merge Ms' M_1' = ext_M (merge Ms M_1) l w ->
        disjUnion M_1' M_2 (merge M_1' M_2).
Proof.
  intros.
  unfold disjUnion.
  split; auto.
  apply resplit_AB_disjBC_sw_aux with Ms M_1 Ms' l w; auto.
Qed.

Lemma resplit_AB_disjUnionBCA_sw:
  forall (Ms M_1 M_2 Ms' M_1' M: Mem) (l : Address) (w : Word), 
    disjoint Ms M_1 -> disjoint M_1 M_2 -> disjoint Ms M_2  ->
      disjoint Ms' M_1' -> in_dom l (merge Ms M_1) ->
      merge Ms' M_1' = ext_M (merge Ms M_1) l w ->
      disjUnion (merge M_1 M_2) Ms M ->
        disjUnion (merge M_1' M_2) Ms' (ext_M M l w).
Proof.
  intros.
  assert (disjUnion M_1' M_2 (merge M_1' M_2)).
  apply resplit_AB_disjBC_sw with Ms M_1 Ms' l w; auto.
  unfold disjUnion in H6.
  inversion_clear H6.
  clear H8.
  assert (disjUnion Ms' M_2 (merge Ms' M_2)).
  apply resplit_AB_disjBC_sw with Ms M_1 M_1' l w; auto.
  apply disjoint_commut; auto.
  assert (merge Ms' M_1' = merge M_1' Ms').
  apply disj_merge_commut; auto.
  rewrite <- H6.
  auto.
  unfold disjUnion in H6.
  inversion_clear H6.
  clear H9.
  unfold disjUnion in H5.
  inversion_clear H5.
  assert (merge (merge M_1 M_2) Ms = merge (merge Ms M_1) M_2).
  apply merge_ABC_CAB; auto.
  rewrite H5 in H9.
  assert (ext_M M l w = merge (ext_M (merge Ms M_1) l w) M_2).
  unfold ext_M.
  rewrite H9.
  apply put_merge_merge_put.
  rewrite H10.
  rewrite <- H4.
  unfold disjUnion.
  split.
  apply disjoint_commut.
  apply disj_merge_assoc; auto.
  assert (merge (merge M_1' M_2) Ms' = merge (merge Ms' M_1') M_2).
  apply merge_ABC_CAB; auto.
  auto.
Qed.

Lemma intertA_presv_filterB:
  forall L l i j,
    i <> j -> not_in_dom l L -> filter_L L j = filter_L (ins_L L l i) j.
Proof.
  intros.
  apply (extensional_eqMap (filter_L L j) (filter_L (ins_L L l i) j)).
  intro.
  unfold filter_L.
  unfold ins_L.
  unfold put.
  induction (beq_l x l).
  rewrite <- a in H0.
  unfold not_in_dom in H0.
  rewrite H0.
  induction (beq_t j i).
  assert (i = j). auto. contradiction.
  auto.
  auto.
Qed.

Lemma delA_presv_filterB:
  forall L l i j,
    i <> j -> maps_to L l i -> filter_L L j = filter_L (del_L L l) j.
Proof.
  intros.
  apply (extensional_eqMap (filter_L L j) (filter_L (del_L L l) j)).
  intro.
  unfold filter_L.
  unfold del_L.
  unfold remove.
  induction (beq_l x l).
  rewrite a.
  unfold maps_to in H0.
  rewrite H0.
  induction (beq_t j i).
  assert (i = j). auto.
  contradiction. 
  auto. auto.
Qed.

Lemma dis_ext : forall M1 M2 l w, dis_M M1 M2 -> not_in_dom l (merge M1 M2) -> dis_M M1 (ext_M M2 l w).
Proof.
intros M1 M2 l w Hdis Hni.
unfold dis_M.
apply disjoint_commut.
unfold ext_M.
apply extend_presv_disj.
apply disjoint_commut; auto.
apply notin_merge_distr_left with M2; auto.
Qed.

Lemma uni_ext : forall M1 M2 l w, dis_M M1 M2 -> not_in_dom l (merge M1 M2) -> ext_M (merge M1 M2) l w = merge M1 (ext_M M2 l w).
Proof.
intros M1 M2 l w Hdis Hni.
unfold ext_M.
rewrite disj_merge_commut.
rewrite put_merge_merge_put.
rewrite disj_merge_commut.
auto.
apply disjoint_commut.
apply dis_ext.
auto.
auto.
auto.
Qed.

Lemma dom_ext : forall M l l' w, l' <> l -> not_in_dom l M -> not_in_dom l' M -> not_in_dom l' (ext_M M l w).
Proof.
intros M l l' w Hne Hni Hni'.
unfold ext_M.
unfold not_in_dom.
rewrite <- Hni'.
unfold put.
induction (beq_w l' l).
elim Hne; auto.
auto.
Qed.

Lemma dom_uni : forall M1 M2 l, dis_M M1 M2 -> not_in_dom l (merge M1 M2) -> not_in_dom l M2.
Proof.
intros M1 M2 l Hdis Hni.
apply notin_merge_distr_right with M1.
auto.
Qed.

Lemma dom_dom : forall M1 M2 f, dis_M M1 M2 -> in_dom f M2 -> in_dom f (merge M1 M2).
Proof.
intros M1 M2 f Hdis Hin.
inversion_clear Hin as [w Hw].
exists w.
compute.
assert (M1 f = None).
elim (Hdis f); intro Ht.
auto.
unfold not_in_dom in Ht.
rewrite Hw in Ht.
discriminate.
rewrite H.
auto.
Qed.

Lemma dis_rem : forall M1 M2 f, dis_M M1 M2 -> in_dom f M2 -> dis_M M1 (remove_M M2 f).
Proof.
intros M1 M2 f Hdis Hin.
intros f'.
elim (Hdis f'); intro Ht.
left; auto.
right.
unfold remove_M.
apply remove_shrinks_dom; auto.
Qed.

Lemma uni_rem : forall M1 M2 f, dis_M M1 M2 -> in_dom f M2 -> remove_M (merge M1 M2) f = merge M1 (remove_M M2 f).
Proof.
intros M1 M2 f Hdis Hin.
unfold remove_M.
rewrite disj_merge_commut.
rewrite disj_remove_merge_merge_remove.
rewrite disj_merge_commut.
auto.
apply disjoint_commut.
apply dis_rem.
auto.
auto.
apply disjoint_commut; auto.
auto.
auto.
Qed.

Lemma dis_ext2 : forall M1 M2 f w, dis_M M1 M2 -> in_dom f M2 -> dis_M M1 (ext_M M2 f w).
Proof.
intros M1 M2 f w Hdis Hin.
unfold dis_M.
apply disjoint_commut.
unfold ext_M.
apply update_presv_disj.
apply disjoint_commut; auto.
auto.
Qed.

Lemma uni_ext2 : forall M1 M2 f w, dis_M M1 M2 -> in_dom f M2 -> ext_M (merge M1 M2) f w = merge M1 (ext_M M2 f w).
Proof.
intros M1 M2 f w Hdis Hin.
unfold ext_M.
rewrite disj_merge_commut.
rewrite put_merge_merge_put.
rewrite disj_merge_commut.
auto.
apply disjoint_commut.
apply dis_ext2.
auto.
auto.
auto.
Qed.

Lemma ins_fil: forall L l k, filter_L (ins_L L l k) k = ins_L (filter_L L k) l k.
Proof.
intros L l k.
apply (extensional_eqMap (filter_L (ins_L L l k) k) (ins_L (filter_L L k) l k)).
intros l'.
unfold ins_L.
unfold put.
unfold filter_L.
induction (beq_l l' l).
induction (beq_t k k).
auto.
elim b.
auto.
auto.
Qed.

Lemma map_fil : forall L l k, maps_to (filter_L L k) l k -> maps_to L l k.
Proof.
intros L l k Hmaps.
unfold filter_L in Hmaps.
unfold maps_to in Hmaps.
unfold maps_to.
induction (L l).
induction (beq_t k a).
rewrite a0; auto.
discriminate.
auto.
Qed.

Lemma del_fil : forall L l k, filter_L (del_L L l) k = del_L (filter_L L k) l.
Proof.
intros L l k.
apply (extensional_eqMap (filter_L (del_L L l) k) (del_L (filter_L L k) l)).
intros l'.
unfold del_L.
unfold remove.
unfold filter_L.
induction (beq_l l' l).
auto.
auto.
Qed.

Lemma mer_empl : forall M, merge (emp (A:=Address) Word) M = M.
Proof.
intros M.
apply (extensional_eqMap (merge (emp (A:=Address) Word) M) M).
compute; auto.
Qed.

Lemma mer_empr : forall M, merge M (emp (A:=Address) Word) = M.
Proof.
intros M.
apply (extensional_eqMap (merge M (emp (A:=Address) Word)) M).
intro f.
compute.
induction (M f); auto.
Qed.

Lemma equ_mer : forall M1 M2 M1' M2', M1 = M1' -> M2 = M2' -> uni_M M1 M2 = uni_M M1' M2'.
Proof.
intros M1 M2 M1' M2' Heq1 Heq2.
rewrite Heq1.
rewrite Heq2.
auto.
Qed.

Lemma sub_trans : forall M1 M2 M3, sub_M M1 M2 -> sub_M M2 M3 -> sub_M M1 M3.
Proof.
intros M1 M2 M3 Hsub12 Hsub23.
intros f Hin.
rewrite (Hsub12 f Hin).
assert (Hin' : in_dom f M2).
inversion Hin as [w Hw].
exists w.
rewrite (Hsub12 f Hin) in Hw.
auto.
rewrite (Hsub23 f Hin').
auto.
Qed.

Lemma uni_dis : forall M1 M2 M3, dis_M M1 M3 -> dis_M M2 M3 -> dis_M (uni_M M1 M2) M3.
Proof.
intros M1 M2 M3 Hdis Hdis'.
unfold dis_M.
unfold uni_M.
apply disjoint_commut.
apply disj_merge_assoc.
apply disjoint_commut; auto.
apply disjoint_commut; auto.
Qed.

Lemma uni_disll : forall M1 M2 M3 M4, dis_M (uni_M M1 M2) (uni_M M3 M4) -> dis_M M1 M3.
Proof.
intros M1 M2 M3 M4 Hdis.
unfold dis_M.
apply disj_distr_left with M2.
apply disjoint_commut.
apply disj_distr_left with M4.
apply disjoint_commut.
auto.
Qed.

Lemma uni_dislr : forall M1 M2 M3 M4, dis_M (uni_M M1 M2) (uni_M M3 M4) -> dis_M M1 M4.
Proof.
intros M1 M2 M3 M4 Hdis.
unfold dis_M.
apply disj_distr_left with M2.
apply disjoint_commut.
apply disj_distr_right with M3.
apply disjoint_commut.
auto.
Qed.

Lemma uni_emp : forall M1 M2 M3 M4, emp_m M1 -> dis_M M3 M4 -> uni_M (uni_M M1 M2) (uni_M M3 M4) = uni_M (uni_M M2 M4) M3.
Proof.
intros M1 M2 M3 M4 Ht Hdis.
assert (Hemp : M1 = emp (A:=Address) Word).
apply (extensional_eqMap M1 (emp (A:=Address) Word)).
auto.
rewrite Hemp.
clear M1 Ht Hemp.
unfold uni_M.
rewrite (mer_empl M2).
rewrite (disj_merge_commut Hdis).
symmetry.
apply (merge_assoc M2 M4 M3).
Qed.